#define OPT_SPEED 1
#define SIZEOF_INT 4
#define SIZEOF_LONG 4
#define SIZEOF_LONG_LONG 8

# define MAD_VERSION		"0.11.4 (beta)"
# define MAD_PUBLISHYEAR	"2000"
# define MAD_AUTHOR		"Robert Leslie"
# define MAD_EMAIL		"rob@mars.org"

extern char const mad_version[];
extern char const mad_copyright[];
extern char const mad_author[];
extern char const mad_license[];
extern char const mad_build[];


# ifndef MAD_FIXED_H
# define MAD_FIXED_H

# if SIZEOF_INT >= 4
typedef   signed int mad_fixed_t;

typedef   signed int mad_fixed64hi_t;
typedef unsigned int mad_fixed64lo_t;
# else
typedef   signed long mad_fixed_t;

typedef   signed long mad_fixed64hi_t;
typedef unsigned long mad_fixed64lo_t;
# endif

/*
 * Fixed-point format: 0xABBBBBBB
 * A == whole part      (sign + 3 bits)
 * B == fractional part (28 bits)
 *
 * Values are signed two's complement, so the effective range is:
 * 0x80000000 to 0x7fffffff
 *       -8.0 to +7.9999999962747097015380859375
 *
 * The smallest representable value is:
 * 0x00000001 == 0.0000000037252902984619140625 (i.e. about 3.725e-9)
 *
 * 28 bits of fractional accuracy represent about
 * 8.6 digits of decimal accuracy.
 *
 * Fixed-point numbers can be added or subtracted as normal
 * integers, but multiplication requires shifting the 64-bit result
 * from 56 fractional bits back to 28 (and rounding.)
 */

# define MAD_F_FRACBITS		28

# define MAD_F_MIN		(-0x80000000L)
# define MAD_F_MAX		  0x7fffffffL

# define MAD_F_ONE		  0x10000000L

# define mad_f_tofixed(x)	((mad_fixed_t) ((x) * 268435456.0 + 0.5))
# define mad_f_todouble(x)	((double)      ((x) / 268435456.0))

# define mad_f_intpart(x)	((x) >> MAD_F_FRACBITS)
# define mad_f_fracpart(x)	((x) & ((1L << MAD_F_FRACBITS) - 1))
				/* (x should be positive) */

# define mad_f_fromint(x)	((x) << MAD_F_FRACBITS)

# define mad_f_add(x, y)	((x) + (y))
# define mad_f_sub(x, y)	((x) - (y))

# if defined(OPT_ACCURACY)
#  define mad_f_scale64(hi, lo)  \
    ((((mad_fixed_t)  \
       (((mad_fixed64hi_t) (hi) << (32 - (MAD_F_SCALEBITS - 1))) |  \
	((mad_fixed64lo_t) (lo) >> (MAD_F_SCALEBITS - 1)))) + 1) >> 1)
# else
#  define mad_f_scale64(hi, lo)  \
    ((mad_fixed_t)  \
     (((mad_fixed64hi_t) (hi) << (32 - MAD_F_SCALEBITS)) |  \
      ((mad_fixed64lo_t) (lo) >> MAD_F_SCALEBITS)))
# endif

# if defined(FPM_APPROX)

/*
 * This version is the most portable but it loses significant accuracy.
 * Furthermore, accuracy is biased against the second argument, so care
 * should be taken when ordering operands.
 *
 * The scale factors are constant as this is not used with SSO.
 *
 * Pre-rounding is required to stay within the limits of compliance.
 */
#  define mad_f_mul(x, y)	((((x) + (1L << 11)) >> 12) *  \
				 (((y) + (1L << 15)) >> 16))

# elif defined(FPM_64BIT)

/*
 * This version should be the most accurate if 64-bit (long long) types are
 * supported by the compiler, although it may not be the most efficient.
 */
#  if defined(OPT_ACCURACY)
#   define mad_f_mul(x, y)  \
    ((mad_fixed_t)  \
     ((((signed long long) (x) * (y)) +  \
       (1L << (MAD_F_SCALEBITS - 1))) >> MAD_F_SCALEBITS))
#  else
#   define mad_f_mul(x, y)  \
    ((mad_fixed_t) (((signed long long) (x) * (y)) >> MAD_F_SCALEBITS))
#  endif

#  define MAD_F_SCALEBITS  MAD_F_FRACBITS

# elif defined(FPM_INTEL)

/*
 * This Intel version is fast and accurate; the disposition of the least
 * significant bit depends on OPT_ACCURACY via mad_f_scale64().
 */
#  define mad_f_mul(x, y)  \
    ({ mad_fixed64hi_t __hi;  \
       mad_fixed64lo_t __lo;  \
       asm ("imull %3"  \
	    : "=a" (__lo), "=d" (__hi)  \
	    : "%a" (x), "rm" (y)  \
	    : "cc");  \
       mad_f_scale64(__hi, __lo);  \
    })

#  if defined(OPT_ACCURACY)
/*
 * This gives best accuracy but is not very fast.
 */
#   define MAD_F_HAVEMLA
#   define mad_f_mla(hi, lo, x, y)  \
    ({ mad_fixed64hi_t __hi;  \
       mad_fixed64lo_t __lo;  \
       asm ("imull %3"  \
	    : "=a" (__lo), "=d" (__hi)  \
	    : "%a" (x), "rm" (y)  \
	    : "cc");  \
       asm ("addl %3, %0\n\t"  \
	    "adcl %5, %1\n\t"  \
	    : "=r,rm" (*(lo)), "=r,rm" (*(hi))  \
	    : "%0,0" (*(lo)), "rm,r" (__lo), "%1,1" (*(hi)), "rm,r" (__hi)  \
	    : "cc");  \
    })
#  endif  /* OPT_ACCURACY */

#  if defined(OPT_ACCURACY)
/*
 * Surprisingly, this is faster than SHRD followed by ADC.
 */
#   undef  mad_f_scale64
#   define mad_f_scale64(hi, lo)  \
    ({ mad_fixed64hi_t __hi_;  \
       mad_fixed64lo_t __lo_;  \
       mad_fixed_t __result;  \
       asm ("addl %4, %2\n\t"  \
	    "adcl %5, %3"  \
	    : "=rm" (__lo_), "=rm" (__hi_)  \
	    : "0" (lo), "1" (hi),  \
	      "ir" (1L << (MAD_F_SCALEBITS - 1)), "ir" (0)  \
	    : "cc");  \
       asm ("shrdl %3, %2, %1"  \
	    : "=rm" (__result)  \
	    : "0" (__lo_), "r" (__hi_), "I" (MAD_F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })
#  else
#   undef mad_f_scale64
#   define mad_f_scale64(hi, lo)  \
    ({ mad_fixed_t __result;  \
       asm ("shrdl %3, %2, %1"  \
	    : "=rm" (__result)  \
	    : "0" (lo), "r" (hi), "I" (MAD_F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })
#  endif  /* OPT_ACCURACY */

#  define MAD_F_SCALEBITS  MAD_F_FRACBITS

# elif defined(FPM_ARM)

/* 
 * This ARM V4 version is as accurate as FPM_64BIT but much faster. The
 * least significant bit is properly rounded at no CPU cycle cost!
 */
#  define mad_f_mul(x, y)  \
    ({ mad_fixed64hi_t __hi;  \
       mad_fixed64lo_t __lo;  \
       mad_fixed_t __result;  \
       asm ("smull	%0, %1, %3, %4\n\t"  \
	    "movs	%0, %0, lsr %5\n\t"  \
	    "adc	%2, %0, %1, lsl %6"  \
	    : "=&r" (__lo), "=&r" (__hi), "=r" (__result)  \
	    : "%r" (x), "r" (y),  \
	      "M" (MAD_F_SCALEBITS), "M" (32 - MAD_F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })

#  define MAD_F_HAVEMLA
#  define mad_f_mla(hi, lo, x, y)  \
    asm ("smlal	%0, %1, %2, %3"  \
	 : "+r" (*(lo)), "+r" (*(hi))  \
	 : "%r" (x), "r" (y))

#  undef  mad_f_scale64
#  define mad_f_scale64(hi, lo)  \
    ({ mad_fixed_t __result;  \
       asm ("movs	%0, %1, lsr %3\n\t"  \
	    "adc	%0, %0, %2, lsl %4"  \
	    : "=r" (__result)  \
	    : "r" (lo), "r" (hi),  \
	      "M" (MAD_F_SCALEBITS), "M" (32 - MAD_F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })

#  define MAD_F_SCALEBITS  MAD_F_FRACBITS

# elif defined(FPM_MIPS)

/*
 * This MIPS version is fast and accurate; the disposition of the least
 * significant bit depends on OPT_ACCURACY via mad_f_scale64().
 */
#  define mad_f_mul(x, y)  \
    ({ mad_fixed64hi_t __hi;  \
       mad_fixed64lo_t __lo;  \
       asm ("mult %2, %3"  \
	    : "=l" (__lo), "=h" (__hi)  \
	    : "%r" (x), "r" (y));  \
       mad_f_scale64(__hi, __lo);  \
    })

# if 0  /* assembler doesn't recognize this instruction? */
#  define MAD_F_HAVEMLA
#  define mad_f_mla(hi, lo, x, y)  \
    asm ("macc r0, %2, %3"  \
	 : "+l" (*(lo)), "+h" (*(hi))  \
	 : "%r" (x), "r" (y));
# endif

#  define MAD_F_SCALEBITS  MAD_F_FRACBITS

# elif defined(FPM_SPARC)

/*
 * This SPARC V8 version is fast and accurate; the disposition of the least
 * significant bit depends on OPT_ACCURACY via mad_f_scale64().
 */
#  define mad_f_mul(x, y)  \
    ({ mad_fixed64hi_t __hi;  \
       mad_fixed64lo_t __lo;  \
       asm ("smul %2, %3, %0\n\t"  \
	    "rd %%y, %1"  \
	    : "=r" (__lo), "=r" (__hi)  \
	    : "%r" (x), "rI" (y));  \
       mad_f_scale64(__hi, __lo);  \
    })

#  define MAD_F_SCALEBITS  MAD_F_FRACBITS

# else

mad_fixed_t mad_f_mul(mad_fixed_t, mad_fixed_t);

# endif

mad_fixed_t mad_f_abs(mad_fixed_t);

# endif


# ifndef MAD_BIT_H
# define MAD_BIT_H

struct mad_bitptr {
  unsigned char const *byte;
  unsigned short cache;
  unsigned short left;
};

void mad_bit_init(struct mad_bitptr *, unsigned char const *);

# define mad_bit_finish(bitptr)  /* nothing */

unsigned int mad_bit_length(struct mad_bitptr const *,
			    struct mad_bitptr const *);

# define mad_bit_bitsleft(bitptr)  ((bitptr)->left)
unsigned char const *mad_bit_nextbyte(struct mad_bitptr const *);

void mad_bit_skip(struct mad_bitptr *, unsigned int);
unsigned long mad_bit_read(struct mad_bitptr *, unsigned int);
void mad_bit_write(struct mad_bitptr *, unsigned int, unsigned long);

unsigned short mad_bit_crc(struct mad_bitptr, unsigned int, unsigned short);

# endif


# ifndef MAD_TIMER_H
# define MAD_TIMER_H

struct mad_timer {
  unsigned long seconds;		/* whole seconds */
  unsigned long fraction;		/* 1/14112000 seconds */
};

# define MAD_TIMER_FRACPARTS	14112000L

# define MAD_TIMER_HOURS	0x0002
# define MAD_TIMER_MINUTES	0x0001
# define MAD_TIMER_SECONDS	0x0000

# define MAD_TIMER_DECISECONDS	0x0003
# define MAD_TIMER_CENTISECONDS	0x0004
# define MAD_TIMER_MILLISECONDS	0x0005

void mad_timer_init(struct mad_timer *);

# define mad_timer_finish(timer)  /* nothing */

void mad_timer_set(struct mad_timer *, unsigned long,
		   unsigned long, unsigned long);
void mad_timer_add(struct mad_timer *, struct mad_timer const *);
int mad_timer_compare(struct mad_timer const *, struct mad_timer const *);

void mad_timer_str(struct mad_timer const *, char *, char const *, int);
unsigned long mad_timer_count(struct mad_timer const *, int);
unsigned long mad_timer_fraction(struct mad_timer const *, unsigned long);

# endif


# ifndef MAD_STREAM_H
# define MAD_STREAM_H


struct mad_stream {
  unsigned char const *buffer;		/* input bitstream buffer */
  unsigned char const *bufend;		/* end of buffer */
  unsigned long skiplen;		/* bytes to skip before next frame */

  int sync;				/* stream sync found */
  unsigned int freerate;		/* free bitrate (fixed) */

  unsigned char const *this_frame;	/* start of current frame */
  unsigned char const *next_frame;	/* start of next frame */
  struct mad_bitptr ptr;		/* current processing bit pointer */

  struct mad_bitptr anc_ptr;		/* ancillary bits pointer */
  unsigned int anc_bitlen;		/* number of ancillary bits */

  unsigned char (*main_data)[1935];	/* Layer III main_data */
  unsigned int md_len;			/* bytes in main_data */

  int error;				/* error code (see below) */
};

# define MAD_ERR_BUFLEN		0x0001	/* input buffer too small (or EOF) */
# define MAD_ERR_BUFPTR		0x0002	/* invalid (null) buffer pointer */

# define MAD_ERR_NOMEM		0x0031	/* not enough memory */

# define MAD_ERR_LOSTSYNC	0x0101	/* lost synchronization */
# define MAD_ERR_BADLAYER	0x0102	/* reserved header layer value */
# define MAD_ERR_BADBITRATE	0x0103	/* forbidden bitrate value */
# define MAD_ERR_BADSAMPLEFREQ	0x0104	/* reserved sample frequency value */
# define MAD_ERR_BADEMPHASIS	0x0105	/* reserved emphasis value */

# define MAD_ERR_BADCRC		0x0201	/* CRC check failed */
# define MAD_ERR_BADBITALLOC	0x0211	/* forbidden bit allocation value */
# define MAD_ERR_BADSCALEFACTOR	0x0221	/* bad scalefactor index */
# define MAD_ERR_BADFRAMELEN	0x0231	/* bad frame length */
# define MAD_ERR_BADBIGVALUES	0x0232	/* bad big_values count */
# define MAD_ERR_BADBLOCKTYPE	0x0233	/* reserved block_type */
# define MAD_ERR_BADDATAPTR	0x0234	/* bad main_data_begin pointer */
# define MAD_ERR_BADDATALEN	0x0235	/* bad main data length */
# define MAD_ERR_BADPART3LEN	0x0236	/* bad audio data length */
# define MAD_ERR_BADHUFFTABLE	0x0237	/* bad Huffman table select */
# define MAD_ERR_BADSTEREO	0x0238	/* incompatible block_type for MS */

# define MAD_RECOVERABLE(error)	((error) & 0xff00)

void mad_stream_init(struct mad_stream *);
void mad_stream_finish(struct mad_stream *);

void mad_stream_buffer(struct mad_stream *,
		       unsigned char const *, unsigned long);
void mad_stream_skip(struct mad_stream *, unsigned long);

int mad_stream_sync(struct mad_stream *);

# endif


# ifndef MAD_FRAME_H
# define MAD_FRAME_H


struct mad_frame {
  int layer;				/* audio layer (1, 2, or 3) */
  int mode;				/* channel mode (see below) */
  int mode_ext;				/* additional mode info */
  int emphasis;				/* de-emphasis to use (see below) */

  unsigned long bitrate;		/* stream bitrate (bps) */
  unsigned int sfreq;			/* sampling frequency (Hz) */

  int flags;				/* flags (below) */
  int private;				/* private bits (below) */

  struct mad_timer duration;		/* audio playing time of frame */

  mad_fixed_t sbsample[2][36][32];	/* synthesis subband filter samples */
  mad_fixed_t (*overlap)[2][32][18];	/* Layer III block overlap data */
};

# define MAD_NCHANNELS(frame)		((frame)->mode ? 2 : 1)
# define MAD_NSBSAMPLES(frame)  \
  ((frame)->layer == 1 ? 12 :  \
   (((frame)->layer == 3 && ((frame)->flags & MAD_FLAG_LSF_EXT)) ? 18 : 36))

# define MAD_MODE_SINGLE_CHANNEL	0
# define MAD_MODE_DUAL_CHANNEL		1
# define MAD_MODE_JOINT_STEREO		2
# define MAD_MODE_STEREO		3

# define MAD_EMPH_NONE		0	/* no emphasis */
# define MAD_EMPH_50_15_MS	1	/* 50/15 microseconds */
# define MAD_EMPH_CCITT_J_17	3	/* CCITT J.17 */

# define MAD_FLAG_III_NPRIVATE	0x0007	/* number of Layer III private bits */

# define MAD_FLAG_PROTECTION	0x0010	/* frame has CRC protection */
# define MAD_FLAG_COPYRIGHT	0x0020	/* frame is copyright */
# define MAD_FLAG_ORIGINAL	0x0040	/* frame is original (else copy) */
# define MAD_FLAG_PADDING	0x0080	/* frame has additional slot */

# define MAD_FLAG_I_STEREO	0x0100	/* uses intensity joint stereo */
# define MAD_FLAG_MS_STEREO	0x0200	/* uses middle/side joint stereo */

# define MAD_FLAG_LSF_EXT	0x1000	/* lower sampling freq. extension */
# define MAD_FLAG_MC_EXT	0x2000	/* multichannel audio extension */

# define MAD_PRIV_HEADER	0x0100	/* header private bit */
# define MAD_PRIV_III		0x001f	/* Layer III private bits (up to 5) */

void mad_frame_init(struct mad_frame *);
void mad_frame_finish(struct mad_frame *);

void mad_frame_mute(struct mad_frame *);

int mad_frame_header(struct mad_frame *, struct mad_stream *,
		     unsigned short *);
int mad_frame_decode(struct mad_frame *, struct mad_stream *);

# endif


# ifndef MAD_SYNTH_H
# define MAD_SYNTH_H


struct mad_synth {
  mad_fixed_t filterout[2][2][256];	/* polyphase filterbank outputs */
  unsigned short slot;			/* current processing slot */

  unsigned short pcmlen;		/* number of PCM samples */
  mad_fixed_t pcmout[2][1152];		/* PCM sample outputs */
};

void mad_synth_init(struct mad_synth *);

# define mad_synth_finish(synth)  /* nothing */

void mad_synth_frame(struct mad_synth *, struct mad_frame const *);

# endif


# ifndef MAD_DECODER_H
# define MAD_DECODER_H


struct mad_decoder {
  int mode;

  struct {
    int pid;
    int in;
    int out;
  } async;

  struct {
    struct mad_stream stream;
    struct mad_frame frame;
    struct mad_synth synth;
  } *sync;

  void *cb_data;

  int (*input_func)(void *, struct mad_stream *);
  int (*filter_func)(void *, struct mad_frame *);
  int (*output_func)(void *, struct mad_frame const *,
		     struct mad_synth const *);
  int (*error_func)(void *, struct mad_stream *, struct mad_frame *);
};

enum {
  mad_cmd_play,
  mad_cmd_pause,
  mad_cmd_rewind,
  mad_cmd_skip,
  mad_cmd_stop
};

union mad_control {
  short command;

  struct mad_cmd_play {
    short command;
  } play;

  struct mad_cmd_pause {
    short command;
  } pause;

  struct mad_cmd_rewind {
    short command;
  } rewind;

  struct mad_cmd_skip {
    short command;
  } skip;

  struct mad_cmd_stop {
    short command;
  } stop;
};

void mad_decoder_init(struct mad_decoder *, void *,
		      int (*)(void *, struct mad_stream *),
		      int (*)(void *, struct mad_frame *),
		      int (*)(void *, struct mad_frame const *,
			      struct mad_synth const *),
		      int (*)(void *, struct mad_stream *,
			      struct mad_frame *));
int mad_decoder_finish(struct mad_decoder *);

int mad_decoder_run(struct mad_decoder *, int);
int mad_decoder_command(struct mad_decoder *, union mad_control *);

# define MAD_DECODER_SYNC	0x0000
# define MAD_DECODER_ASYNC	0x0001

# define MAD_DECODER_CONTINUE	0x0000
# define MAD_DECODER_STOP	0x0010
# define MAD_DECODER_BREAK	0x0011
# define MAD_DECODER_IGNORE	0x0020

# endif

