/*
 * mad - MPEG audio decoder
 * Copyright (C) 2000 Robert Leslie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: layer3.c,v 1.10 2000/09/24 17:49:38 rob Exp $
 */

# ifdef HAVE_CONFIG_H
#  include "config.h"
# endif

# include <stdlib.h>
# include <string.h>

# ifdef HAVE_LIMITS_H
#  include <limits.h>
# else
#  define CHAR_BIT  8
# endif

# include "fixed.h"
# include "bit.h"
# include "stream.h"
# include "frame.h"
# include "huffman.h"
# include "layer3.h"

/* --- Layer III ----------------------------------------------------------- */

enum {
  count1table_select	= 0x01,
  scalefac_scale	= 0x02,
  preflag		= 0x04,
  window_switching_flag	= 0x08,
  mixed_block_flag	= 0x10
};

struct sideinfo {
  unsigned int main_data_begin;
  unsigned int private_bits;

  unsigned char scfsi[2];

  struct granule {
    struct channel {
      /* from side info */
      unsigned short part2_3_length;
      unsigned short big_values;
      unsigned short global_gain;
      unsigned short scalefac_compress;

      unsigned char flags;
      unsigned char block_type;
      unsigned char table_select[3];
      unsigned char subblock_gain[3];
      unsigned char region0_count;
      unsigned char region1_count;

      /* from main_data */
      unsigned char scalefac_l[22];
      unsigned char scalefac_s[13][3];

      /* implied from decoding */
      unsigned short count1;
    } ch[2];
  } gr[2];
};

/*
 * scalefactor bit lengths
 * derived from section 2.4.2.7 of ISO/IEC 11172-3
 */
static
struct {
  unsigned short slen1;
  unsigned short slen2;
} const sflen[16] = {
  { 0, 0 }, { 0, 1 }, { 0, 2 }, { 0, 3 },
  { 3, 0 }, { 1, 1 }, { 1, 2 }, { 1, 3 },
  { 2, 1 }, { 2, 2 }, { 2, 3 }, { 3, 1 },
  { 3, 2 }, { 3, 3 }, { 4, 2 }, { 4, 3 }
};

/*
 * number of LSF scalefactor band values
 * derived from section 2.4.3.2 of ISO/IEC 13818-3
 */
static
unsigned short const nsfb_table[6][3][4] = {
  { {  6,  5,  5, 5 },
    {  9,  9,  9, 9 },
    {  6,  9,  9, 9 } },
  { {  6,  5,  7, 3 },
    {  9,  9, 12, 6 },
    {  6,  9, 12, 6 } },
  { { 11, 10,  0, 0 },
    { 18, 18,  0, 0 },
    { 15, 18,  0, 0 } },
  { {  7,  7,  7, 0 },
    { 12, 12, 12, 0 },
    {  6, 15, 12, 0 } },
  { {  6,  6,  6, 3 },
    { 12,  9,  9, 6 },
    {  6, 12,  9, 6 } },
  { {  8,  8,  5, 0 },
    { 15, 12,  9, 0 },
    {  6, 18,  9, 0 } }
};

/*
 * scalefactor band widths for long blocks
 * derived from Table B.8 of ISO/IEC 11172-3 and Table B.2 of ISO/IEC 13838-3
 */
static
unsigned short const sfwidth_l[6][22] = {
  /*    48 kHz */ {  4,  4,  4,  4,  4,  4,  6,  6,  6,   8,  10,
		    12, 16, 18, 22, 28, 34, 40, 46, 54,  54, 192 },
  /*  44.1 kHz */ {  4,  4,  4,  4,  4,  4,  6,  6,  8,   8,  10,
		    12, 16, 20, 24, 28, 34, 42, 50, 54,  76, 158 },
  /*    32 kHz */ {  4,  4,  4,  4,  4,  4,  6,  6,  8,  10,  12,
		    16, 20, 24, 30, 38, 46, 56, 68, 84, 102,  26 },

  /*    24 kHz */ {  6,  6,  6,  6,  6,  6,  8, 10, 12,  14,  16,
		    18, 22, 26, 32, 38, 46, 54, 62, 70,  76,  36 },
  /* 22.05 kHz */ {  6,  6,  6,  6,  6,  6,  8, 10, 12,  14,  16,
		    20, 24, 28, 32, 38, 46, 52, 60, 68,  58,  54 },
  /*    16 kHz */ {  6,  6,  6,  6,  6,  6,  8, 10, 12,  14,  16,
		    20, 24, 28, 32, 38, 46, 52, 60, 68,  58,  54 }
};

/*
 * scalefactor band widths for short blocks
 * derived from Table B.8 of ISO/IEC 11172-3 and Table B.2 of ISO/IEC 13838-3
 */
static
unsigned short const sfwidth_s[6][13] = {
  /*    48 kHz */ { 4, 4, 4, 4, 6,  6, 10, 12, 14, 16, 20, 26, 66 },
  /*  44.1 kHz */ { 4, 4, 4, 4, 6,  8, 10, 12, 14, 18, 22, 30, 56 },
  /*    32 kHz */ { 4, 4, 4, 4, 6,  8, 12, 16, 20, 26, 34, 42, 12 },

  /*    24 kHz */ { 4, 4, 4, 6, 8, 10, 12, 14, 18, 24, 32, 44, 12 },
  /* 22.05 kHz */ { 4, 4, 4, 6, 6,  8, 10, 14, 18, 26, 32, 42, 18 },
  /*    16 kHz */ { 4, 4, 4, 6, 8, 10, 12, 14, 18, 24, 30, 40, 18 }
};

/*
 * scalefactor band preemphasis (used only when preflag is set)
 * derived from Table B.6 of ISO/IEC 11172-3
 */
static
unsigned short const pretab[22] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 3, 3, 3, 2, 0
};

/*
 * table for requantization
 * rq_table[x].mantissa * 2^(rq_table[x].exponent) == x^(4/3)
 */
static
struct fixedfloat {
  unsigned long mantissa  /*: 27 */;
  unsigned short exponent /*:  5 */;
} const rq_table[8207] = {
 { 0x00000000L /* 0.000000000 */,  0 }, { 0x04000000L /* 0.250000000 */,  2 },
 { 0x050a28beL /* 0.314980262 */,  3 }, { 0x0453a5cdL /* 0.270421794 */,  4 },
 { 0x06597fa9L /* 0.396850263 */,  4 }, { 0x04466275L /* 0.267183742 */,  5 },
 { 0x05738c72L /* 0.340710111 */,  5 }, { 0x06b1fc81L /* 0.418453696 */,  5 },
 { 0x04000000L /* 0.250000000 */,  6 }, { 0x04ae20d7L /* 0.292511788 */,  6 },
 { 0x0562d694L /* 0.336630420 */,  6 }, { 0x061dae96L /* 0.382246578 */,  6 },
 { 0x06de47f4L /* 0.429267841 */,  6 }, { 0x07a44f7aL /* 0.477614858 */,  6 },
 { 0x0437be65L /* 0.263609310 */,  7 }, { 0x049fc824L /* 0.289009227 */,  7 },

 { 0x050a28beL /* 0.314980262 */,  7 }, { 0x0576c6f5L /* 0.341498336 */,  7 },
 { 0x05e58c0bL /* 0.368541759 */,  7 }, { 0x06566361L /* 0.396090870 */,  7 },
 { 0x06c93a2eL /* 0.424127753 */,  7 }, { 0x073dff3eL /* 0.452635998 */,  7 },
 { 0x07b4a2bcL /* 0.481600510 */,  7 }, { 0x04168b05L /* 0.255503674 */,  8 },
 { 0x0453a5cdL /* 0.270421794 */,  8 }, { 0x04919b6aL /* 0.285548607 */,  8 },
 { 0x04d065fbL /* 0.300878507 */,  8 }, { 0x05100000L /* 0.316406250 */,  8 },
 { 0x05506451L /* 0.332126919 */,  8 }, { 0x05918e15L /* 0.348035890 */,  8 },
 { 0x05d378bbL /* 0.364128809 */,  8 }, { 0x06161ff3L /* 0.380401563 */,  8 },

 { 0x06597fa9L /* 0.396850263 */,  8 }, { 0x069d9400L /* 0.413471222 */,  8 },
 { 0x06e2594cL /* 0.430260942 */,  8 }, { 0x0727cc11L /* 0.447216097 */,  8 },
 { 0x076de8fcL /* 0.464333519 */,  8 }, { 0x07b4ace3L /* 0.481610189 */,  8 },
 { 0x07fc14bfL /* 0.499043224 */,  8 }, { 0x04220ed7L /* 0.258314934 */,  9 },
 { 0x04466275L /* 0.267183742 */,  9 }, { 0x046b03e7L /* 0.276126771 */,  9 },
 { 0x048ff1e8L /* 0.285142811 */,  9 }, { 0x04b52b3fL /* 0.294230696 */,  9 },
 { 0x04daaec0L /* 0.303389310 */,  9 }, { 0x05007b49L /* 0.312617576 */,  9 },
 { 0x05268fc6L /* 0.321914457 */,  9 }, { 0x054ceb2aL /* 0.331278957 */,  9 },

 { 0x05738c72L /* 0.340710111 */,  9 }, { 0x059a72a5L /* 0.350206992 */,  9 },
 { 0x05c19cd3L /* 0.359768701 */,  9 }, { 0x05e90a12L /* 0.369394372 */,  9 },
 { 0x0610b982L /* 0.379083164 */,  9 }, { 0x0638aa48L /* 0.388834268 */,  9 },
 { 0x0660db91L /* 0.398646895 */,  9 }, { 0x06894c90L /* 0.408520284 */,  9 },
 { 0x06b1fc81L /* 0.418453696 */,  9 }, { 0x06daeaa1L /* 0.428446415 */,  9 },
 { 0x07041636L /* 0.438497744 */,  9 }, { 0x072d7e8bL /* 0.448607009 */,  9 },
 { 0x075722efL /* 0.458773552 */,  9 }, { 0x078102b8L /* 0.468996735 */,  9 },
 { 0x07ab1d3eL /* 0.479275937 */,  9 }, { 0x07d571e0L /* 0.489610555 */,  9 },

 { 0x04000000L /* 0.250000000 */, 10 }, { 0x04156381L /* 0.255221850 */, 10 },
 { 0x042ae32aL /* 0.260470548 */, 10 }, { 0x04407eb1L /* 0.265745823 */, 10 },
 { 0x045635cfL /* 0.271047409 */, 10 }, { 0x046c083eL /* 0.276375048 */, 10 },
 { 0x0481f5bbL /* 0.281728487 */, 10 }, { 0x0497fe03L /* 0.287107481 */, 10 },
 { 0x04ae20d7L /* 0.292511788 */, 10 }, { 0x04c45df6L /* 0.297941173 */, 10 },
 { 0x04dab524L /* 0.303395408 */, 10 }, { 0x04f12624L /* 0.308874267 */, 10 },
 { 0x0507b0bcL /* 0.314377532 */, 10 }, { 0x051e54b1L /* 0.319904987 */, 10 },
 { 0x053511cbL /* 0.325456423 */, 10 }, { 0x054be7d4L /* 0.331031635 */, 10 },

 { 0x0562d694L /* 0.336630420 */, 10 }, { 0x0579ddd8L /* 0.342252584 */, 10 },
 { 0x0590fd6cL /* 0.347897931 */, 10 }, { 0x05a8351cL /* 0.353566275 */, 10 },
 { 0x05bf84b8L /* 0.359257429 */, 10 }, { 0x05d6ec0eL /* 0.364971213 */, 10 },
 { 0x05ee6aefL /* 0.370707448 */, 10 }, { 0x0606012bL /* 0.376465960 */, 10 },
 { 0x061dae96L /* 0.382246578 */, 10 }, { 0x06357302L /* 0.388049134 */, 10 },
 { 0x064d4e43L /* 0.393873464 */, 10 }, { 0x0665402dL /* 0.399719406 */, 10 },
 { 0x067d4896L /* 0.405586801 */, 10 }, { 0x06956753L /* 0.411475493 */, 10 },
 { 0x06ad9c3dL /* 0.417385331 */, 10 }, { 0x06c5e72bL /* 0.423316162 */, 10 },

 { 0x06de47f4L /* 0.429267841 */, 10 }, { 0x06f6be73L /* 0.435240221 */, 10 },
 { 0x070f4a80L /* 0.441233161 */, 10 }, { 0x0727ebf7L /* 0.447246519 */, 10 },
 { 0x0740a2b2L /* 0.453280160 */, 10 }, { 0x07596e8dL /* 0.459333946 */, 10 },
 { 0x07724f64L /* 0.465407744 */, 10 }, { 0x078b4514L /* 0.471501425 */, 10 },
 { 0x07a44f7aL /* 0.477614858 */, 10 }, { 0x07bd6e75L /* 0.483747918 */, 10 },
 { 0x07d6a1e2L /* 0.489900479 */, 10 }, { 0x07efe9a1L /* 0.496072418 */, 10 },
 { 0x0404a2c9L /* 0.251131807 */, 11 }, { 0x04115acaL /* 0.254236974 */, 11 },
 { 0x041e1cc4L /* 0.257351652 */, 11 }, { 0x042ae8a7L /* 0.260475783 */, 11 },

 { 0x0437be65L /* 0.263609310 */, 11 }, { 0x04449deeL /* 0.266752177 */, 11 },
 { 0x04518733L /* 0.269904329 */, 11 }, { 0x045e7a26L /* 0.273065710 */, 11 },
 { 0x046b76b9L /* 0.276236269 */, 11 }, { 0x04787cdcL /* 0.279415952 */, 11 },
 { 0x04858c83L /* 0.282604707 */, 11 }, { 0x0492a59fL /* 0.285802482 */, 11 },
 { 0x049fc824L /* 0.289009227 */, 11 }, { 0x04acf402L /* 0.292224893 */, 11 },
 { 0x04ba292eL /* 0.295449429 */, 11 }, { 0x04c7679aL /* 0.298682788 */, 11 },
 { 0x04d4af3aL /* 0.301924921 */, 11 }, { 0x04e20000L /* 0.305175781 */, 11 },
 { 0x04ef59e0L /* 0.308435322 */, 11 }, { 0x04fcbcceL /* 0.311703498 */, 11 },

 { 0x050a28beL /* 0.314980262 */, 11 }, { 0x05179da4L /* 0.318265572 */, 11 },
 { 0x05251b73L /* 0.321559381 */, 11 }, { 0x0532a220L /* 0.324861647 */, 11 },
 { 0x054031a0L /* 0.328172327 */, 11 }, { 0x054dc9e7L /* 0.331491377 */, 11 },
 { 0x055b6ae9L /* 0.334818756 */, 11 }, { 0x0569149cL /* 0.338154423 */, 11 },
 { 0x0576c6f5L /* 0.341498336 */, 11 }, { 0x058481e9L /* 0.344850455 */, 11 },
 { 0x0592456dL /* 0.348210741 */, 11 }, { 0x05a01176L /* 0.351579152 */, 11 },
 { 0x05ade5faL /* 0.354955651 */, 11 }, { 0x05bbc2efL /* 0.358340200 */, 11 },
 { 0x05c9a84aL /* 0.361732758 */, 11 }, { 0x05d79601L /* 0.365133291 */, 11 },

 { 0x05e58c0bL /* 0.368541759 */, 11 }, { 0x05f38a5dL /* 0.371958126 */, 11 },
 { 0x060190eeL /* 0.375382356 */, 11 }, { 0x060f9fb3L /* 0.378814413 */, 11 },
 { 0x061db6a5L /* 0.382254261 */, 11 }, { 0x062bd5b8L /* 0.385701865 */, 11 },
 { 0x0639fce4L /* 0.389157191 */, 11 }, { 0x06482c1fL /* 0.392620204 */, 11 },
 { 0x06566361L /* 0.396090870 */, 11 }, { 0x0664a2a0L /* 0.399569155 */, 11 },
 { 0x0672e9d4L /* 0.403055027 */, 11 }, { 0x068138f3L /* 0.406548452 */, 11 },
 { 0x068f8ff5L /* 0.410049398 */, 11 }, { 0x069deed1L /* 0.413557833 */, 11 },
 { 0x06ac557fL /* 0.417073724 */, 11 }, { 0x06bac3f6L /* 0.420597041 */, 11 },

 { 0x06c93a2eL /* 0.424127753 */, 11 }, { 0x06d7b81fL /* 0.427665827 */, 11 },
 { 0x06e63dc0L /* 0.431211234 */, 11 }, { 0x06f4cb09L /* 0.434763944 */, 11 },
 { 0x07035ff3L /* 0.438323927 */, 11 }, { 0x0711fc75L /* 0.441891153 */, 11 },
 { 0x0720a087L /* 0.445465593 */, 11 }, { 0x072f4c22L /* 0.449047217 */, 11 },
 { 0x073dff3eL /* 0.452635998 */, 11 }, { 0x074cb9d3L /* 0.456231906 */, 11 },
 { 0x075b7bdbL /* 0.459834914 */, 11 }, { 0x076a454cL /* 0.463444993 */, 11 },
 { 0x07791620L /* 0.467062117 */, 11 }, { 0x0787ee50L /* 0.470686258 */, 11 },
 { 0x0796cdd4L /* 0.474317388 */, 11 }, { 0x07a5b4a5L /* 0.477955481 */, 11 },

 { 0x07b4a2bcL /* 0.481600510 */, 11 }, { 0x07c39812L /* 0.485252449 */, 11 },
 { 0x07d294a0L /* 0.488911273 */, 11 }, { 0x07e1985fL /* 0.492576954 */, 11 },
 { 0x07f0a348L /* 0.496249468 */, 11 }, { 0x07ffb554L /* 0.499928790 */, 11 },
 { 0x0407673fL /* 0.251807447 */, 12 }, { 0x040ef75eL /* 0.253653877 */, 12 },
 { 0x04168b05L /* 0.255503674 */, 12 }, { 0x041e2230L /* 0.257356825 */, 12 },
 { 0x0425bcddL /* 0.259213318 */, 12 }, { 0x042d5b07L /* 0.261073141 */, 12 },
 { 0x0434fcadL /* 0.262936282 */, 12 }, { 0x043ca1c9L /* 0.264802730 */, 12 },
 { 0x04444a5aL /* 0.266672472 */, 12 }, { 0x044bf65dL /* 0.268545497 */, 12 },

 { 0x0453a5cdL /* 0.270421794 */, 12 }, { 0x045b58a9L /* 0.272301352 */, 12 },
 { 0x04630eedL /* 0.274184158 */, 12 }, { 0x046ac896L /* 0.276070203 */, 12 },
 { 0x047285a2L /* 0.277959474 */, 12 }, { 0x047a460cL /* 0.279851960 */, 12 },
 { 0x048209d3L /* 0.281747652 */, 12 }, { 0x0489d0f4L /* 0.283646538 */, 12 },
 { 0x04919b6aL /* 0.285548607 */, 12 }, { 0x04996935L /* 0.287453849 */, 12 },
 { 0x04a13a50L /* 0.289362253 */, 12 }, { 0x04a90ebaL /* 0.291273810 */, 12 },
 { 0x04b0e66eL /* 0.293188507 */, 12 }, { 0x04b8c16cL /* 0.295106336 */, 12 },
 { 0x04c09fafL /* 0.297027285 */, 12 }, { 0x04c88135L /* 0.298951346 */, 12 },

 { 0x04d065fbL /* 0.300878507 */, 12 }, { 0x04d84dffL /* 0.302808759 */, 12 },
 { 0x04e0393eL /* 0.304742092 */, 12 }, { 0x04e827b6L /* 0.306678497 */, 12 },
 { 0x04f01963L /* 0.308617963 */, 12 }, { 0x04f80e44L /* 0.310560480 */, 12 },
 { 0x05000655L /* 0.312506041 */, 12 }, { 0x05080195L /* 0.314454634 */, 12 },
 { 0x05100000L /* 0.316406250 */, 12 }, { 0x05180194L /* 0.318360880 */, 12 },
 { 0x0520064fL /* 0.320318516 */, 12 }, { 0x05280e2dL /* 0.322279147 */, 12 },
 { 0x0530192eL /* 0.324242764 */, 12 }, { 0x0538274eL /* 0.326209359 */, 12 },
 { 0x0540388aL /* 0.328178922 */, 12 }, { 0x05484ce2L /* 0.330151445 */, 12 },

 { 0x05506451L /* 0.332126919 */, 12 }, { 0x05587ed5L /* 0.334105334 */, 12 },
 { 0x05609c6eL /* 0.336086683 */, 12 }, { 0x0568bd17L /* 0.338070956 */, 12 },
 { 0x0570e0cfL /* 0.340058145 */, 12 }, { 0x05790793L /* 0.342048241 */, 12 },
 { 0x05813162L /* 0.344041237 */, 12 }, { 0x05895e39L /* 0.346037122 */, 12 },
 { 0x05918e15L /* 0.348035890 */, 12 }, { 0x0599c0f4L /* 0.350037532 */, 12 },
 { 0x05a1f6d5L /* 0.352042040 */, 12 }, { 0x05aa2fb5L /* 0.354049405 */, 12 },
 { 0x05b26b92L /* 0.356059619 */, 12 }, { 0x05baaa69L /* 0.358072674 */, 12 },
 { 0x05c2ec39L /* 0.360088563 */, 12 }, { 0x05cb3100L /* 0.362107278 */, 12 },

 { 0x05d378bbL /* 0.364128809 */, 12 }, { 0x05dbc368L /* 0.366153151 */, 12 },
 { 0x05e41105L /* 0.368180294 */, 12 }, { 0x05ec6190L /* 0.370210231 */, 12 },
 { 0x05f4b507L /* 0.372242955 */, 12 }, { 0x05fd0b68L /* 0.374278458 */, 12 },
 { 0x060564b1L /* 0.376316732 */, 12 }, { 0x060dc0e0L /* 0.378357769 */, 12 },
 { 0x06161ff3L /* 0.380401563 */, 12 }, { 0x061e81e8L /* 0.382448106 */, 12 },
 { 0x0626e6bcL /* 0.384497391 */, 12 }, { 0x062f4e6fL /* 0.386549409 */, 12 },
 { 0x0637b8fdL /* 0.388604155 */, 12 }, { 0x06402666L /* 0.390661620 */, 12 },
 { 0x064896a7L /* 0.392721798 */, 12 }, { 0x065109beL /* 0.394784681 */, 12 },

 { 0x06597fa9L /* 0.396850263 */, 12 }, { 0x0661f867L /* 0.398918536 */, 12 },
 { 0x066a73f5L /* 0.400989493 */, 12 }, { 0x0672f252L /* 0.403063128 */, 12 },
 { 0x067b737cL /* 0.405139433 */, 12 }, { 0x0683f771L /* 0.407218402 */, 12 },
 { 0x068c7e2fL /* 0.409300027 */, 12 }, { 0x069507b5L /* 0.411384303 */, 12 },
 { 0x069d9400L /* 0.413471222 */, 12 }, { 0x06a6230fL /* 0.415560778 */, 12 },
 { 0x06aeb4e0L /* 0.417652964 */, 12 }, { 0x06b74971L /* 0.419747773 */, 12 },
 { 0x06bfe0c0L /* 0.421845199 */, 12 }, { 0x06c87accL /* 0.423945235 */, 12 },
 { 0x06d11794L /* 0.426047876 */, 12 }, { 0x06d9b714L /* 0.428153114 */, 12 },

 { 0x06e2594cL /* 0.430260942 */, 12 }, { 0x06eafe3aL /* 0.432371356 */, 12 },
 { 0x06f3a5dcL /* 0.434484348 */, 12 }, { 0x06fc5030L /* 0.436599912 */, 12 },
 { 0x0704fd35L /* 0.438718042 */, 12 }, { 0x070daceaL /* 0.440838732 */, 12 },
 { 0x07165f4bL /* 0.442961975 */, 12 }, { 0x071f1459L /* 0.445087765 */, 12 },
 { 0x0727cc11L /* 0.447216097 */, 12 }, { 0x07308671L /* 0.449346964 */, 12 },
 { 0x07394378L /* 0.451480360 */, 12 }, { 0x07420325L /* 0.453616280 */, 12 },
 { 0x074ac575L /* 0.455754717 */, 12 }, { 0x07538a67L /* 0.457895665 */, 12 },
 { 0x075c51faL /* 0.460039119 */, 12 }, { 0x07651c2cL /* 0.462185072 */, 12 },

 { 0x076de8fcL /* 0.464333519 */, 12 }, { 0x0776b867L /* 0.466484455 */, 12 },
 { 0x077f8a6dL /* 0.468637872 */, 12 }, { 0x07885f0bL /* 0.470793767 */, 12 },
 { 0x07913641L /* 0.472952132 */, 12 }, { 0x079a100cL /* 0.475112962 */, 12 },
 { 0x07a2ec6cL /* 0.477276252 */, 12 }, { 0x07abcb5fL /* 0.479441997 */, 12 },
 { 0x07b4ace3L /* 0.481610189 */, 12 }, { 0x07bd90f6L /* 0.483780825 */, 12 },
 { 0x07c67798L /* 0.485953899 */, 12 }, { 0x07cf60c7L /* 0.488129404 */, 12 },
 { 0x07d84c81L /* 0.490307336 */, 12 }, { 0x07e13ac5L /* 0.492487690 */, 12 },
 { 0x07ea2b92L /* 0.494670459 */, 12 }, { 0x07f31ee6L /* 0.496855639 */, 12 },

 { 0x07fc14bfL /* 0.499043224 */, 12 }, { 0x0402868eL /* 0.250616605 */, 13 },
 { 0x040703ffL /* 0.251712795 */, 13 }, { 0x040b82b0L /* 0.252810180 */, 13 },
 { 0x041002a1L /* 0.253908756 */, 13 }, { 0x041483d1L /* 0.255008523 */, 13 },
 { 0x04190640L /* 0.256109476 */, 13 }, { 0x041d89edL /* 0.257211614 */, 13 },
 { 0x04220ed7L /* 0.258314934 */, 13 }, { 0x042694feL /* 0.259419433 */, 13 },
 { 0x042b1c60L /* 0.260525110 */, 13 }, { 0x042fa4feL /* 0.261631960 */, 13 },
 { 0x04342ed7L /* 0.262739982 */, 13 }, { 0x0438b9e9L /* 0.263849174 */, 13 },
 { 0x043d4635L /* 0.264959533 */, 13 }, { 0x0441d3b9L /* 0.266071056 */, 13 },

 { 0x04466275L /* 0.267183742 */, 13 }, { 0x044af269L /* 0.268297587 */, 13 },
 { 0x044f8393L /* 0.269412589 */, 13 }, { 0x045415f3L /* 0.270528746 */, 13 },
 { 0x0458a989L /* 0.271646056 */, 13 }, { 0x045d3e53L /* 0.272764515 */, 13 },
 { 0x0461d451L /* 0.273884123 */, 13 }, { 0x04666b83L /* 0.275004875 */, 13 },
 { 0x046b03e7L /* 0.276126771 */, 13 }, { 0x046f9d7eL /* 0.277249808 */, 13 },
 { 0x04743847L /* 0.278373983 */, 13 }, { 0x0478d440L /* 0.279499294 */, 13 },
 { 0x047d716aL /* 0.280625739 */, 13 }, { 0x04820fc3L /* 0.281753315 */, 13 },
 { 0x0486af4cL /* 0.282882021 */, 13 }, { 0x048b5003L /* 0.284011853 */, 13 },

 { 0x048ff1e8L /* 0.285142811 */, 13 }, { 0x049494fbL /* 0.286274891 */, 13 },
 { 0x0499393aL /* 0.287408091 */, 13 }, { 0x049ddea5L /* 0.288542409 */, 13 },
 { 0x04a2853cL /* 0.289677844 */, 13 }, { 0x04a72cfeL /* 0.290814392 */, 13 },
 { 0x04abd5eaL /* 0.291952051 */, 13 }, { 0x04b08000L /* 0.293090820 */, 13 },
 { 0x04b52b3fL /* 0.294230696 */, 13 }, { 0x04b9d7a7L /* 0.295371678 */, 13 },
 { 0x04be8537L /* 0.296513762 */, 13 }, { 0x04c333eeL /* 0.297656947 */, 13 },
 { 0x04c7e3ccL /* 0.298801231 */, 13 }, { 0x04cc94d1L /* 0.299946611 */, 13 },
 { 0x04d146fbL /* 0.301093085 */, 13 }, { 0x04d5fa4bL /* 0.302240653 */, 13 },

 { 0x04daaec0L /* 0.303389310 */, 13 }, { 0x04df6458L /* 0.304539056 */, 13 },
 { 0x04e41b14L /* 0.305689888 */, 13 }, { 0x04e8d2f3L /* 0.306841804 */, 13 },
 { 0x04ed8bf5L /* 0.307994802 */, 13 }, { 0x04f24618L /* 0.309148880 */, 13 },
 { 0x04f7015dL /* 0.310304037 */, 13 }, { 0x04fbbdc3L /* 0.311460269 */, 13 },
 { 0x05007b49L /* 0.312617576 */, 13 }, { 0x050539efL /* 0.313775954 */, 13 },
 { 0x0509f9b4L /* 0.314935403 */, 13 }, { 0x050eba98L /* 0.316095920 */, 13 },
 { 0x05137c9aL /* 0.317257503 */, 13 }, { 0x05183fbaL /* 0.318420150 */, 13 },
 { 0x051d03f7L /* 0.319583859 */, 13 }, { 0x0521c950L /* 0.320748629 */, 13 },

 { 0x05268fc6L /* 0.321914457 */, 13 }, { 0x052b5757L /* 0.323081342 */, 13 },
 { 0x05302003L /* 0.324249281 */, 13 }, { 0x0534e9caL /* 0.325418273 */, 13 },
 { 0x0539b4abL /* 0.326588316 */, 13 }, { 0x053e80a6L /* 0.327759407 */, 13 },
 { 0x05434db9L /* 0.328931546 */, 13 }, { 0x05481be5L /* 0.330104730 */, 13 },
 { 0x054ceb2aL /* 0.331278957 */, 13 }, { 0x0551bb85L /* 0.332454225 */, 13 },
 { 0x05568cf8L /* 0.333630533 */, 13 }, { 0x055b5f81L /* 0.334807879 */, 13 },
 { 0x05603321L /* 0.335986261 */, 13 }, { 0x056507d6L /* 0.337165677 */, 13 },
 { 0x0569dda0L /* 0.338346125 */, 13 }, { 0x056eb47fL /* 0.339527604 */, 13 },

 { 0x05738c72L /* 0.340710111 */, 13 }, { 0x05786578L /* 0.341893646 */, 13 },
 { 0x057d3f92L /* 0.343078205 */, 13 }, { 0x05821abfL /* 0.344263788 */, 13 },
 { 0x0586f6fdL /* 0.345450393 */, 13 }, { 0x058bd44eL /* 0.346638017 */, 13 },
 { 0x0590b2b0L /* 0.347826659 */, 13 }, { 0x05959222L /* 0.349016318 */, 13 },
 { 0x059a72a5L /* 0.350206992 */, 13 }, { 0x059f5438L /* 0.351398678 */, 13 },
 { 0x05a436daL /* 0.352591376 */, 13 }, { 0x05a91a8cL /* 0.353785083 */, 13 },
 { 0x05adff4cL /* 0.354979798 */, 13 }, { 0x05b2e51aL /* 0.356175519 */, 13 },
 { 0x05b7cbf5L /* 0.357372244 */, 13 }, { 0x05bcb3deL /* 0.358569972 */, 13 },

 { 0x05c19cd3L /* 0.359768701 */, 13 }, { 0x05c686d5L /* 0.360968429 */, 13 },
 { 0x05cb71e2L /* 0.362169156 */, 13 }, { 0x05d05dfbL /* 0.363370878 */, 13 },
 { 0x05d54b1fL /* 0.364573594 */, 13 }, { 0x05da394dL /* 0.365777304 */, 13 },
 { 0x05df2885L /* 0.366982004 */, 13 }, { 0x05e418c7L /* 0.368187694 */, 13 },
 { 0x05e90a12L /* 0.369394372 */, 13 }, { 0x05edfc66L /* 0.370602036 */, 13 },
 { 0x05f2efc2L /* 0.371810684 */, 13 }, { 0x05f7e426L /* 0.373020316 */, 13 },
 { 0x05fcd992L /* 0.374230929 */, 13 }, { 0x0601d004L /* 0.375442522 */, 13 },
 { 0x0606c77dL /* 0.376655093 */, 13 }, { 0x060bbffdL /* 0.377868641 */, 13 },

 { 0x0610b982L /* 0.379083164 */, 13 }, { 0x0615b40cL /* 0.380298661 */, 13 },
 { 0x061aaf9cL /* 0.381515130 */, 13 }, { 0x061fac2fL /* 0.382732569 */, 13 },
 { 0x0624a9c7L /* 0.383950977 */, 13 }, { 0x0629a863L /* 0.385170352 */, 13 },
 { 0x062ea802L /* 0.386390694 */, 13 }, { 0x0633a8a3L /* 0.387611999 */, 13 },
 { 0x0638aa48L /* 0.388834268 */, 13 }, { 0x063daceeL /* 0.390057497 */, 13 },
 { 0x0642b096L /* 0.391281687 */, 13 }, { 0x0647b53fL /* 0.392506834 */, 13 },
 { 0x064cbae9L /* 0.393732939 */, 13 }, { 0x0651c193L /* 0.394959999 */, 13 },
 { 0x0656c93dL /* 0.396188012 */, 13 }, { 0x065bd1e7L /* 0.397416978 */, 13 },

 { 0x0660db91L /* 0.398646895 */, 13 }, { 0x0665e639L /* 0.399877761 */, 13 },
 { 0x066af1dfL /* 0.401109575 */, 13 }, { 0x066ffe84L /* 0.402342335 */, 13 },
 { 0x06750c26L /* 0.403576041 */, 13 }, { 0x067a1ac6L /* 0.404810690 */, 13 },
 { 0x067f2a62L /* 0.406046281 */, 13 }, { 0x06843afbL /* 0.407282813 */, 13 },
 { 0x06894c90L /* 0.408520284 */, 13 }, { 0x068e5f21L /* 0.409758693 */, 13 },
 { 0x069372aeL /* 0.410998038 */, 13 }, { 0x06988735L /* 0.412238319 */, 13 },
 { 0x069d9cb7L /* 0.413479532 */, 13 }, { 0x06a2b333L /* 0.414721679 */, 13 },
 { 0x06a7caa9L /* 0.415964756 */, 13 }, { 0x06ace318L /* 0.417208762 */, 13 },

 { 0x06b1fc81L /* 0.418453696 */, 13 }, { 0x06b716e2L /* 0.419699557 */, 13 },
 { 0x06bc323bL /* 0.420946343 */, 13 }, { 0x06c14e8dL /* 0.422194054 */, 13 },
 { 0x06c66bd6L /* 0.423442686 */, 13 }, { 0x06cb8a17L /* 0.424692240 */, 13 },
 { 0x06d0a94eL /* 0.425942714 */, 13 }, { 0x06d5c97cL /* 0.427194106 */, 13 },
 { 0x06daeaa1L /* 0.428446415 */, 13 }, { 0x06e00cbbL /* 0.429699640 */, 13 },
 { 0x06e52fcaL /* 0.430953779 */, 13 }, { 0x06ea53cfL /* 0.432208832 */, 13 },
 { 0x06ef78c8L /* 0.433464796 */, 13 }, { 0x06f49eb6L /* 0.434721671 */, 13 },
 { 0x06f9c597L /* 0.435979455 */, 13 }, { 0x06feed6dL /* 0.437238146 */, 13 },

 { 0x07041636L /* 0.438497744 */, 13 }, { 0x07093ff2L /* 0.439758248 */, 13 },
 { 0x070e6aa0L /* 0.441019655 */, 13 }, { 0x07139641L /* 0.442281965 */, 13 },
 { 0x0718c2d3L /* 0.443545176 */, 13 }, { 0x071df058L /* 0.444809288 */, 13 },
 { 0x07231ecdL /* 0.446074298 */, 13 }, { 0x07284e34L /* 0.447340205 */, 13 },
 { 0x072d7e8bL /* 0.448607009 */, 13 }, { 0x0732afd2L /* 0.449874708 */, 13 },
 { 0x0737e209L /* 0.451143300 */, 13 }, { 0x073d1530L /* 0.452412785 */, 13 },
 { 0x07424946L /* 0.453683161 */, 13 }, { 0x07477e4bL /* 0.454954427 */, 13 },
 { 0x074cb43eL /* 0.456226581 */, 13 }, { 0x0751eb20L /* 0.457499623 */, 13 },

 { 0x075722efL /* 0.458773552 */, 13 }, { 0x075c5bacL /* 0.460048365 */, 13 },
 { 0x07619557L /* 0.461324062 */, 13 }, { 0x0766cfeeL /* 0.462600642 */, 13 },
 { 0x076c0b72L /* 0.463878102 */, 13 }, { 0x077147e2L /* 0.465156443 */, 13 },
 { 0x0776853eL /* 0.466435663 */, 13 }, { 0x077bc385L /* 0.467715761 */, 13 },
 { 0x078102b8L /* 0.468996735 */, 13 }, { 0x078642d6L /* 0.470278584 */, 13 },
 { 0x078b83deL /* 0.471561307 */, 13 }, { 0x0790c5d1L /* 0.472844904 */, 13 },
 { 0x079608aeL /* 0.474129372 */, 13 }, { 0x079b4c74L /* 0.475414710 */, 13 },
 { 0x07a09124L /* 0.476700918 */, 13 }, { 0x07a5d6bdL /* 0.477987994 */, 13 },

 { 0x07ab1d3eL /* 0.479275937 */, 13 }, { 0x07b064a8L /* 0.480564746 */, 13 },
 { 0x07b5acfbL /* 0.481854420 */, 13 }, { 0x07baf635L /* 0.483144957 */, 13 },
 { 0x07c04056L /* 0.484436356 */, 13 }, { 0x07c58b5fL /* 0.485728617 */, 13 },
 { 0x07cad74eL /* 0.487021738 */, 13 }, { 0x07d02424L /* 0.488315717 */, 13 },
 { 0x07d571e0L /* 0.489610555 */, 13 }, { 0x07dac083L /* 0.490906249 */, 13 },
 { 0x07e0100aL /* 0.492202799 */, 13 }, { 0x07e56078L /* 0.493500203 */, 13 },
 { 0x07eab1caL /* 0.494798460 */, 13 }, { 0x07f00401L /* 0.496097570 */, 13 },
 { 0x07f5571dL /* 0.497397530 */, 13 }, { 0x07faab1cL /* 0.498698341 */, 13 },

 { 0x04000000L /* 0.250000000 */, 14 }, { 0x0402aae3L /* 0.250651254 */, 14 },
 { 0x04055638L /* 0.251302930 */, 14 }, { 0x040801ffL /* 0.251955030 */, 14 },
 { 0x040aae37L /* 0.252607552 */, 14 }, { 0x040d5ae0L /* 0.253260495 */, 14 },
 { 0x041007faL /* 0.253913860 */, 14 }, { 0x0412b586L /* 0.254567645 */, 14 },
 { 0x04156381L /* 0.255221850 */, 14 }, { 0x041811eeL /* 0.255876475 */, 14 },
 { 0x041ac0cbL /* 0.256531518 */, 14 }, { 0x041d7018L /* 0.257186980 */, 14 },
 { 0x04201fd5L /* 0.257842860 */, 14 }, { 0x0422d003L /* 0.258499157 */, 14 },
 { 0x042580a0L /* 0.259155872 */, 14 }, { 0x042831adL /* 0.259813002 */, 14 },

 { 0x042ae32aL /* 0.260470548 */, 14 }, { 0x042d9516L /* 0.261128510 */, 14 },
 { 0x04304772L /* 0.261786886 */, 14 }, { 0x0432fa3dL /* 0.262445676 */, 14 },
 { 0x0435ad76L /* 0.263104880 */, 14 }, { 0x0438611fL /* 0.263764497 */, 14 },
 { 0x043b1536L /* 0.264424527 */, 14 }, { 0x043dc9bcL /* 0.265084969 */, 14 },
 { 0x04407eb1L /* 0.265745823 */, 14 }, { 0x04433414L /* 0.266407088 */, 14 },
 { 0x0445e9e5L /* 0.267068763 */, 14 }, { 0x0448a024L /* 0.267730848 */, 14 },
 { 0x044b56d1L /* 0.268393343 */, 14 }, { 0x044e0decL /* 0.269056248 */, 14 },
 { 0x0450c575L /* 0.269719560 */, 14 }, { 0x04537d6bL /* 0.270383281 */, 14 },

 { 0x045635cfL /* 0.271047409 */, 14 }, { 0x0458ee9fL /* 0.271711944 */, 14 },
 { 0x045ba7ddL /* 0.272376886 */, 14 }, { 0x045e6188L /* 0.273042234 */, 14 },
 { 0x04611ba0L /* 0.273707988 */, 14 }, { 0x0463d625L /* 0.274374147 */, 14 },
 { 0x04669116L /* 0.275040710 */, 14 }, { 0x04694c74L /* 0.275707677 */, 14 },
 { 0x046c083eL /* 0.276375048 */, 14 }, { 0x046ec474L /* 0.277042822 */, 14 },
 { 0x04718116L /* 0.277710999 */, 14 }, { 0x04743e25L /* 0.278379578 */, 14 },
 { 0x0476fb9fL /* 0.279048558 */, 14 }, { 0x0479b984L /* 0.279717940 */, 14 },
 { 0x047c77d6L /* 0.280387722 */, 14 }, { 0x047f3693L /* 0.281057905 */, 14 },

 { 0x0481f5bbL /* 0.281728487 */, 14 }, { 0x0484b54eL /* 0.282399469 */, 14 },
 { 0x0487754cL /* 0.283070849 */, 14 }, { 0x048a35b6L /* 0.283742628 */, 14 },
 { 0x048cf68aL /* 0.284414805 */, 14 }, { 0x048fb7c8L /* 0.285087379 */, 14 },
 { 0x04927972L /* 0.285760350 */, 14 }, { 0x04953b85L /* 0.286433717 */, 14 },
 { 0x0497fe03L /* 0.287107481 */, 14 }, { 0x049ac0ebL /* 0.287781640 */, 14 },
 { 0x049d843eL /* 0.288456194 */, 14 }, { 0x04a047faL /* 0.289131142 */, 14 },
 { 0x04a30c20L /* 0.289806485 */, 14 }, { 0x04a5d0afL /* 0.290482221 */, 14 },
 { 0x04a895a8L /* 0.291158351 */, 14 }, { 0x04ab5b0bL /* 0.291834873 */, 14 },

 { 0x04ae20d7L /* 0.292511788 */, 14 }, { 0x04b0e70cL /* 0.293189094 */, 14 },
 { 0x04b3adaaL /* 0.293866792 */, 14 }, { 0x04b674b1L /* 0.294544881 */, 14 },
 { 0x04b93c21L /* 0.295223360 */, 14 }, { 0x04bc03faL /* 0.295902229 */, 14 },
 { 0x04becc3bL /* 0.296581488 */, 14 }, { 0x04c194e4L /* 0.297261136 */, 14 },
 { 0x04c45df6L /* 0.297941173 */, 14 }, { 0x04c72771L /* 0.298621598 */, 14 },
 { 0x04c9f153L /* 0.299302411 */, 14 }, { 0x04ccbb9dL /* 0.299983611 */, 14 },
 { 0x04cf864fL /* 0.300665198 */, 14 }, { 0x04d25169L /* 0.301347172 */, 14 },
 { 0x04d51cebL /* 0.302029532 */, 14 }, { 0x04d7e8d4L /* 0.302712277 */, 14 },

 { 0x04dab524L /* 0.303395408 */, 14 }, { 0x04dd81dcL /* 0.304078923 */, 14 },
 { 0x04e04efbL /* 0.304762823 */, 14 }, { 0x04e31c81L /* 0.305447106 */, 14 },
 { 0x04e5ea6eL /* 0.306131773 */, 14 }, { 0x04e8b8c2L /* 0.306816823 */, 14 },
 { 0x04eb877cL /* 0.307502256 */, 14 }, { 0x04ee569dL /* 0.308188071 */, 14 },
 { 0x04f12624L /* 0.308874267 */, 14 }, { 0x04f3f612L /* 0.309560845 */, 14 },
 { 0x04f6c666L /* 0.310247804 */, 14 }, { 0x04f99721L /* 0.310935143 */, 14 },
 { 0x04fc6841L /* 0.311622862 */, 14 }, { 0x04ff39c7L /* 0.312310961 */, 14 },
 { 0x05020bb3L /* 0.312999439 */, 14 }, { 0x0504de05L /* 0.313688296 */, 14 },

 { 0x0507b0bcL /* 0.314377532 */, 14 }, { 0x050a83d8L /* 0.315067145 */, 14 },
 { 0x050d575bL /* 0.315757136 */, 14 }, { 0x05102b42L /* 0.316447504 */, 14 },
 { 0x0512ff8eL /* 0.317138249 */, 14 }, { 0x0515d440L /* 0.317829370 */, 14 },
 { 0x0518a956L /* 0.318520867 */, 14 }, { 0x051b7ed1L /* 0.319212739 */, 14 },
 { 0x051e54b1L /* 0.319904987 */, 14 }, { 0x05212af5L /* 0.320597609 */, 14 },
 { 0x0524019eL /* 0.321290606 */, 14 }, { 0x0526d8abL /* 0.321983976 */, 14 },
 { 0x0529b01dL /* 0.322677720 */, 14 }, { 0x052c87f2L /* 0.323371837 */, 14 },
 { 0x052f602cL /* 0.324066327 */, 14 }, { 0x053238caL /* 0.324761189 */, 14 },

 { 0x053511cbL /* 0.325456423 */, 14 }, { 0x0537eb30L /* 0.326152028 */, 14 },
 { 0x053ac4f9L /* 0.326848005 */, 14 }, { 0x053d9f25L /* 0.327544352 */, 14 },
 { 0x054079b5L /* 0.328241070 */, 14 }, { 0x054354a8L /* 0.328938157 */, 14 },
 { 0x05462ffeL /* 0.329635614 */, 14 }, { 0x05490bb7L /* 0.330333440 */, 14 },
 { 0x054be7d4L /* 0.331031635 */, 14 }, { 0x054ec453L /* 0.331730198 */, 14 },
 { 0x0551a134L /* 0.332429129 */, 14 }, { 0x05547e79L /* 0.333128427 */, 14 },
 { 0x05575c20L /* 0.333828093 */, 14 }, { 0x055a3a2aL /* 0.334528126 */, 14 },
 { 0x055d1896L /* 0.335228525 */, 14 }, { 0x055ff764L /* 0.335929290 */, 14 },

 { 0x0562d694L /* 0.336630420 */, 14 }, { 0x0565b627L /* 0.337331916 */, 14 },
 { 0x0568961bL /* 0.338033777 */, 14 }, { 0x056b7671L /* 0.338736002 */, 14 },
 { 0x056e5729L /* 0.339438592 */, 14 }, { 0x05713843L /* 0.340141545 */, 14 },
 { 0x057419beL /* 0.340844862 */, 14 }, { 0x0576fb9aL /* 0.341548541 */, 14 },
 { 0x0579ddd8L /* 0.342252584 */, 14 }, { 0x057cc077L /* 0.342956988 */, 14 },
 { 0x057fa378L /* 0.343661754 */, 14 }, { 0x058286d9L /* 0.344366882 */, 14 },
 { 0x05856a9bL /* 0.345072371 */, 14 }, { 0x05884ebeL /* 0.345778221 */, 14 },
 { 0x058b3342L /* 0.346484431 */, 14 }, { 0x058e1827L /* 0.347191002 */, 14 },

 { 0x0590fd6cL /* 0.347897931 */, 14 }, { 0x0593e311L /* 0.348605221 */, 14 },
 { 0x0596c917L /* 0.349312869 */, 14 }, { 0x0599af7dL /* 0.350020876 */, 14 },
 { 0x059c9643L /* 0.350729240 */, 14 }, { 0x059f7d6aL /* 0.351437963 */, 14 },
 { 0x05a264f0L /* 0.352147044 */, 14 }, { 0x05a54cd6L /* 0.352856481 */, 14 },
 { 0x05a8351cL /* 0.353566275 */, 14 }, { 0x05ab1dc2L /* 0.354276426 */, 14 },
 { 0x05ae06c7L /* 0.354986932 */, 14 }, { 0x05b0f02bL /* 0.355697795 */, 14 },
 { 0x05b3d9f0L /* 0.356409012 */, 14 }, { 0x05b6c413L /* 0.357120585 */, 14 },
 { 0x05b9ae95L /* 0.357832512 */, 14 }, { 0x05bc9977L /* 0.358544794 */, 14 },

 { 0x05bf84b8L /* 0.359257429 */, 14 }, { 0x05c27057L /* 0.359970419 */, 14 },
 { 0x05c55c56L /* 0.360683761 */, 14 }, { 0x05c848b3L /* 0.361397456 */, 14 },
 { 0x05cb356eL /* 0.362111504 */, 14 }, { 0x05ce2289L /* 0.362825904 */, 14 },
 { 0x05d11001L /* 0.363540655 */, 14 }, { 0x05d3fdd8L /* 0.364255759 */, 14 },
 { 0x05d6ec0eL /* 0.364971213 */, 14 }, { 0x05d9daa1L /* 0.365687018 */, 14 },
 { 0x05dcc993L /* 0.366403174 */, 14 }, { 0x05dfb8e2L /* 0.367119680 */, 14 },
 { 0x05e2a890L /* 0.367836535 */, 14 }, { 0x05e5989bL /* 0.368553740 */, 14 },
 { 0x05e88904L /* 0.369271294 */, 14 }, { 0x05eb79cbL /* 0.369989197 */, 14 },

 { 0x05ee6aefL /* 0.370707448 */, 14 }, { 0x05f15c70L /* 0.371426047 */, 14 },
 { 0x05f44e4fL /* 0.372144994 */, 14 }, { 0x05f7408bL /* 0.372864289 */, 14 },
 { 0x05fa3324L /* 0.373583930 */, 14 }, { 0x05fd261bL /* 0.374303918 */, 14 },
 { 0x0600196eL /* 0.375024253 */, 14 }, { 0x06030d1eL /* 0.375744934 */, 14 },
 { 0x0606012bL /* 0.376465960 */, 14 }, { 0x0608f595L /* 0.377187332 */, 14 },
 { 0x060bea5cL /* 0.377909049 */, 14 }, { 0x060edf7fL /* 0.378631110 */, 14 },
 { 0x0611d4feL /* 0.379353516 */, 14 }, { 0x0614cadaL /* 0.380076266 */, 14 },
 { 0x0617c112L /* 0.380799360 */, 14 }, { 0x061ab7a6L /* 0.381522798 */, 14 },

 { 0x061dae96L /* 0.382246578 */, 14 }, { 0x0620a5e3L /* 0.382970701 */, 14 },
 { 0x06239d8bL /* 0.383695167 */, 14 }, { 0x0626958fL /* 0.384419975 */, 14 },
 { 0x06298defL /* 0.385145124 */, 14 }, { 0x062c86aaL /* 0.385870615 */, 14 },
 { 0x062f7fc1L /* 0.386596448 */, 14 }, { 0x06327934L /* 0.387322621 */, 14 },
 { 0x06357302L /* 0.388049134 */, 14 }, { 0x06386d2bL /* 0.388775988 */, 14 },
 { 0x063b67b0L /* 0.389503182 */, 14 }, { 0x063e6290L /* 0.390230715 */, 14 },
 { 0x06415dcbL /* 0.390958588 */, 14 }, { 0x06445960L /* 0.391686799 */, 14 },
 { 0x06475551L /* 0.392415349 */, 14 }, { 0x064a519cL /* 0.393144238 */, 14 },

 { 0x064d4e43L /* 0.393873464 */, 14 }, { 0x06504b44L /* 0.394603028 */, 14 },
 { 0x0653489fL /* 0.395332930 */, 14 }, { 0x06564655L /* 0.396063168 */, 14 },
 { 0x06594465L /* 0.396793743 */, 14 }, { 0x065c42d0L /* 0.397524655 */, 14 },
 { 0x065f4195L /* 0.398255903 */, 14 }, { 0x066240b4L /* 0.398987487 */, 14 },
 { 0x0665402dL /* 0.399719406 */, 14 }, { 0x06684000L /* 0.400451660 */, 14 },
 { 0x066b402dL /* 0.401184249 */, 14 }, { 0x066e40b3L /* 0.401917173 */, 14 },
 { 0x06714194L /* 0.402650431 */, 14 }, { 0x067442ceL /* 0.403384024 */, 14 },
 { 0x06774462L /* 0.404117949 */, 14 }, { 0x067a464fL /* 0.404852209 */, 14 },

 { 0x067d4896L /* 0.405586801 */, 14 }, { 0x06804b36L /* 0.406321726 */, 14 },
 { 0x06834e2fL /* 0.407056983 */, 14 }, { 0x06865181L /* 0.407792573 */, 14 },
 { 0x0689552cL /* 0.408528495 */, 14 }, { 0x068c5931L /* 0.409264748 */, 14 },
 { 0x068f5d8eL /* 0.410001332 */, 14 }, { 0x06926245L /* 0.410738247 */, 14 },
 { 0x06956753L /* 0.411475493 */, 14 }, { 0x06986cbbL /* 0.412213070 */, 14 },
 { 0x069b727bL /* 0.412950976 */, 14 }, { 0x069e7894L /* 0.413689213 */, 14 },
 { 0x06a17f05L /* 0.414427779 */, 14 }, { 0x06a485cfL /* 0.415166674 */, 14 },
 { 0x06a78cf1L /* 0.415905897 */, 14 }, { 0x06aa946bL /* 0.416645450 */, 14 },

 { 0x06ad9c3dL /* 0.417385331 */, 14 }, { 0x06b0a468L /* 0.418125540 */, 14 },
 { 0x06b3aceaL /* 0.418866076 */, 14 }, { 0x06b6b5c4L /* 0.419606940 */, 14 },
 { 0x06b9bef6L /* 0.420348132 */, 14 }, { 0x06bcc880L /* 0.421089650 */, 14 },
 { 0x06bfd261L /* 0.421831494 */, 14 }, { 0x06c2dc9aL /* 0.422573665 */, 14 },
 { 0x06c5e72bL /* 0.423316162 */, 14 }, { 0x06c8f213L /* 0.424058985 */, 14 },
 { 0x06cbfd52L /* 0.424802133 */, 14 }, { 0x06cf08e9L /* 0.425545607 */, 14 },
 { 0x06d214d7L /* 0.426289405 */, 14 }, { 0x06d5211cL /* 0.427033528 */, 14 },
 { 0x06d82db8L /* 0.427777975 */, 14 }, { 0x06db3aaaL /* 0.428522746 */, 14 },

 { 0x06de47f4L /* 0.429267841 */, 14 }, { 0x06e15595L /* 0.430013259 */, 14 },
 { 0x06e4638dL /* 0.430759001 */, 14 }, { 0x06e771dbL /* 0.431505065 */, 14 },
 { 0x06ea807fL /* 0.432251452 */, 14 }, { 0x06ed8f7bL /* 0.432998162 */, 14 },
 { 0x06f09eccL /* 0.433745193 */, 14 }, { 0x06f3ae75L /* 0.434492546 */, 14 },
 { 0x06f6be73L /* 0.435240221 */, 14 }, { 0x06f9cec8L /* 0.435988217 */, 14 },
 { 0x06fcdf72L /* 0.436736534 */, 14 }, { 0x06fff073L /* 0.437485172 */, 14 },
 { 0x070301caL /* 0.438234130 */, 14 }, { 0x07061377L /* 0.438983408 */, 14 },
 { 0x0709257aL /* 0.439733006 */, 14 }, { 0x070c37d2L /* 0.440482924 */, 14 },

 { 0x070f4a80L /* 0.441233161 */, 14 }, { 0x07125d84L /* 0.441983717 */, 14 },
 { 0x071570deL /* 0.442734592 */, 14 }, { 0x0718848dL /* 0.443485785 */, 14 },
 { 0x071b9891L /* 0.444237296 */, 14 }, { 0x071eacebL /* 0.444989126 */, 14 },
 { 0x0721c19aL /* 0.445741273 */, 14 }, { 0x0724d69eL /* 0.446493738 */, 14 },
 { 0x0727ebf7L /* 0.447246519 */, 14 }, { 0x072b01a6L /* 0.447999618 */, 14 },
 { 0x072e17a9L /* 0.448753033 */, 14 }, { 0x07312e01L /* 0.449506765 */, 14 },
 { 0x073444aeL /* 0.450260813 */, 14 }, { 0x07375bb0L /* 0.451015176 */, 14 },
 { 0x073a7307L /* 0.451769856 */, 14 }, { 0x073d8ab2L /* 0.452524850 */, 14 },

 { 0x0740a2b2L /* 0.453280160 */, 14 }, { 0x0743bb06L /* 0.454035784 */, 14 },
 { 0x0746d3afL /* 0.454791723 */, 14 }, { 0x0749ecacL /* 0.455547976 */, 14 },
 { 0x074d05feL /* 0.456304543 */, 14 }, { 0x07501fa3L /* 0.457061423 */, 14 },
 { 0x0753399dL /* 0.457818618 */, 14 }, { 0x075653ebL /* 0.458576125 */, 14 },
 { 0x07596e8dL /* 0.459333946 */, 14 }, { 0x075c8983L /* 0.460092079 */, 14 },
 { 0x075fa4ccL /* 0.460850524 */, 14 }, { 0x0762c06aL /* 0.461609282 */, 14 },
 { 0x0765dc5bL /* 0.462368352 */, 14 }, { 0x0768f8a0L /* 0.463127733 */, 14 },
 { 0x076c1538L /* 0.463887426 */, 14 }, { 0x076f3224L /* 0.464647430 */, 14 },

 { 0x07724f64L /* 0.465407744 */, 14 }, { 0x07756cf7L /* 0.466168370 */, 14 },
 { 0x07788addL /* 0.466929306 */, 14 }, { 0x077ba916L /* 0.467690552 */, 14 },
 { 0x077ec7a3L /* 0.468452108 */, 14 }, { 0x0781e683L /* 0.469213973 */, 14 },
 { 0x078505b5L /* 0.469976148 */, 14 }, { 0x0788253bL /* 0.470738632 */, 14 },
 { 0x078b4514L /* 0.471501425 */, 14 }, { 0x078e653fL /* 0.472264527 */, 14 },
 { 0x079185beL /* 0.473027937 */, 14 }, { 0x0794a68fL /* 0.473791655 */, 14 },
 { 0x0797c7b2L /* 0.474555681 */, 14 }, { 0x079ae929L /* 0.475320014 */, 14 },
 { 0x079e0af1L /* 0.476084655 */, 14 }, { 0x07a12d0cL /* 0.476849603 */, 14 },

 { 0x07a44f7aL /* 0.477614858 */, 14 }, { 0x07a7723aL /* 0.478380420 */, 14 },
 { 0x07aa954cL /* 0.479146288 */, 14 }, { 0x07adb8b0L /* 0.479912463 */, 14 },
 { 0x07b0dc67L /* 0.480678943 */, 14 }, { 0x07b4006fL /* 0.481445729 */, 14 },
 { 0x07b724caL /* 0.482212820 */, 14 }, { 0x07ba4976L /* 0.482980216 */, 14 },
 { 0x07bd6e75L /* 0.483747918 */, 14 }, { 0x07c093c5L /* 0.484515924 */, 14 },
 { 0x07c3b967L /* 0.485284235 */, 14 }, { 0x07c6df5aL /* 0.486052849 */, 14 },
 { 0x07ca059fL /* 0.486821768 */, 14 }, { 0x07cd2c36L /* 0.487590991 */, 14 },
 { 0x07d0531eL /* 0.488360517 */, 14 }, { 0x07d37a57L /* 0.489130346 */, 14 },

 { 0x07d6a1e2L /* 0.489900479 */, 14 }, { 0x07d9c9beL /* 0.490670914 */, 14 },
 { 0x07dcf1ecL /* 0.491441651 */, 14 }, { 0x07e01a6aL /* 0.492212691 */, 14 },
 { 0x07e3433aL /* 0.492984033 */, 14 }, { 0x07e66c5aL /* 0.493755677 */, 14 },
 { 0x07e995ccL /* 0.494527623 */, 14 }, { 0x07ecbf8eL /* 0.495299870 */, 14 },
 { 0x07efe9a1L /* 0.496072418 */, 14 }, { 0x07f31405L /* 0.496845266 */, 14 },
 { 0x07f63ebaL /* 0.497618416 */, 14 }, { 0x07f969c0L /* 0.498391866 */, 14 },
 { 0x07fc9516L /* 0.499165616 */, 14 }, { 0x07ffc0bcL /* 0.499939666 */, 14 },
 { 0x04017659L /* 0.250357008 */, 15 }, { 0x04030c7dL /* 0.250744333 */, 15 },

 { 0x0404a2c9L /* 0.251131807 */, 15 }, { 0x0406393dL /* 0.251519431 */, 15 },
 { 0x0407cfd9L /* 0.251907204 */, 15 }, { 0x0409669dL /* 0.252295127 */, 15 },
 { 0x040afd89L /* 0.252683198 */, 15 }, { 0x040c949eL /* 0.253071419 */, 15 },
 { 0x040e2bdaL /* 0.253459789 */, 15 }, { 0x040fc33eL /* 0.253848307 */, 15 },
 { 0x04115acaL /* 0.254236974 */, 15 }, { 0x0412f27eL /* 0.254625790 */, 15 },
 { 0x04148a5aL /* 0.255014755 */, 15 }, { 0x0416225dL /* 0.255403867 */, 15 },
 { 0x0417ba89L /* 0.255793128 */, 15 }, { 0x041952dcL /* 0.256182537 */, 15 },
 { 0x041aeb57L /* 0.256572095 */, 15 }, { 0x041c83faL /* 0.256961800 */, 15 },

 { 0x041e1cc4L /* 0.257351652 */, 15 }, { 0x041fb5b6L /* 0.257741653 */, 15 },
 { 0x04214ed0L /* 0.258131801 */, 15 }, { 0x0422e811L /* 0.258522097 */, 15 },
 { 0x04248179L /* 0.258912540 */, 15 }, { 0x04261b0aL /* 0.259303130 */, 15 },
 { 0x0427b4c2L /* 0.259693868 */, 15 }, { 0x04294ea1L /* 0.260084752 */, 15 },
 { 0x042ae8a7L /* 0.260475783 */, 15 }, { 0x042c82d6L /* 0.260866961 */, 15 },
 { 0x042e1d2bL /* 0.261258286 */, 15 }, { 0x042fb7a8L /* 0.261649758 */, 15 },
 { 0x0431524cL /* 0.262041376 */, 15 }, { 0x0432ed17L /* 0.262433140 */, 15 },
 { 0x0434880aL /* 0.262825051 */, 15 }, { 0x04362324L /* 0.263217107 */, 15 },

 { 0x0437be65L /* 0.263609310 */, 15 }, { 0x043959cdL /* 0.264001659 */, 15 },
 { 0x043af55dL /* 0.264394153 */, 15 }, { 0x043c9113L /* 0.264786794 */, 15 },
 { 0x043e2cf1L /* 0.265179580 */, 15 }, { 0x043fc8f6L /* 0.265572511 */, 15 },
 { 0x04416522L /* 0.265965588 */, 15 }, { 0x04430174L /* 0.266358810 */, 15 },
 { 0x04449deeL /* 0.266752177 */, 15 }, { 0x04463a8fL /* 0.267145689 */, 15 },
 { 0x0447d756L /* 0.267539347 */, 15 }, { 0x04497445L /* 0.267933149 */, 15 },
 { 0x044b115aL /* 0.268327096 */, 15 }, { 0x044cae96L /* 0.268721187 */, 15 },
 { 0x044e4bf9L /* 0.269115423 */, 15 }, { 0x044fe983L /* 0.269509804 */, 15 },

 { 0x04518733L /* 0.269904329 */, 15 }, { 0x0453250aL /* 0.270298998 */, 15 },
 { 0x0454c308L /* 0.270693811 */, 15 }, { 0x0456612dL /* 0.271088768 */, 15 },
 { 0x0457ff78L /* 0.271483869 */, 15 }, { 0x04599deaL /* 0.271879114 */, 15 },
 { 0x045b3c82L /* 0.272274503 */, 15 }, { 0x045cdb41L /* 0.272670035 */, 15 },
 { 0x045e7a26L /* 0.273065710 */, 15 }, { 0x04601932L /* 0.273461530 */, 15 },
 { 0x0461b864L /* 0.273857492 */, 15 }, { 0x046357bdL /* 0.274253597 */, 15 },
 { 0x0464f73cL /* 0.274649846 */, 15 }, { 0x046696e2L /* 0.275046238 */, 15 },
 { 0x046836aeL /* 0.275442772 */, 15 }, { 0x0469d6a0L /* 0.275839449 */, 15 },

 { 0x046b76b9L /* 0.276236269 */, 15 }, { 0x046d16f7L /* 0.276633232 */, 15 },
 { 0x046eb75cL /* 0.277030337 */, 15 }, { 0x047057e8L /* 0.277427584 */, 15 },
 { 0x0471f899L /* 0.277824973 */, 15 }, { 0x04739971L /* 0.278222505 */, 15 },
 { 0x04753a6fL /* 0.278620179 */, 15 }, { 0x0476db92L /* 0.279017995 */, 15 },
 { 0x04787cdcL /* 0.279415952 */, 15 }, { 0x047a1e4cL /* 0.279814051 */, 15 },
 { 0x047bbfe2L /* 0.280212292 */, 15 }, { 0x047d619eL /* 0.280610675 */, 15 },
 { 0x047f0380L /* 0.281009199 */, 15 }, { 0x0480a588L /* 0.281407864 */, 15 },
 { 0x048247b6L /* 0.281806670 */, 15 }, { 0x0483ea0aL /* 0.282205618 */, 15 },

 { 0x04858c83L /* 0.282604707 */, 15 }, { 0x04872f22L /* 0.283003936 */, 15 },
 { 0x0488d1e8L /* 0.283403307 */, 15 }, { 0x048a74d3L /* 0.283802818 */, 15 },
 { 0x048c17e3L /* 0.284202470 */, 15 }, { 0x048dbb1aL /* 0.284602263 */, 15 },
 { 0x048f5e76L /* 0.285002195 */, 15 }, { 0x049101f8L /* 0.285402269 */, 15 },
 { 0x0492a59fL /* 0.285802482 */, 15 }, { 0x0494496cL /* 0.286202836 */, 15 },
 { 0x0495ed5fL /* 0.286603329 */, 15 }, { 0x04979177L /* 0.287003963 */, 15 },
 { 0x049935b5L /* 0.287404737 */, 15 }, { 0x049ada19L /* 0.287805650 */, 15 },
 { 0x049c7ea1L /* 0.288206703 */, 15 }, { 0x049e2350L /* 0.288607895 */, 15 },

 { 0x049fc824L /* 0.289009227 */, 15 }, { 0x04a16d1dL /* 0.289410699 */, 15 },
 { 0x04a3123bL /* 0.289812309 */, 15 }, { 0x04a4b77fL /* 0.290214059 */, 15 },
 { 0x04a65ce8L /* 0.290615948 */, 15 }, { 0x04a80277L /* 0.291017976 */, 15 },
 { 0x04a9a82bL /* 0.291420143 */, 15 }, { 0x04ab4e04L /* 0.291822449 */, 15 },
 { 0x04acf402L /* 0.292224893 */, 15 }, { 0x04ae9a26L /* 0.292627476 */, 15 },
 { 0x04b0406eL /* 0.293030197 */, 15 }, { 0x04b1e6dcL /* 0.293433057 */, 15 },
 { 0x04b38d6fL /* 0.293836055 */, 15 }, { 0x04b53427L /* 0.294239192 */, 15 },
 { 0x04b6db05L /* 0.294642466 */, 15 }, { 0x04b88207L /* 0.295045879 */, 15 },

 { 0x04ba292eL /* 0.295449429 */, 15 }, { 0x04bbd07aL /* 0.295853118 */, 15 },
 { 0x04bd77ecL /* 0.296256944 */, 15 }, { 0x04bf1f82L /* 0.296660907 */, 15 },
 { 0x04c0c73dL /* 0.297065009 */, 15 }, { 0x04c26f1dL /* 0.297469248 */, 15 },
 { 0x04c41722L /* 0.297873624 */, 15 }, { 0x04c5bf4cL /* 0.298278137 */, 15 },
 { 0x04c7679aL /* 0.298682788 */, 15 }, { 0x04c9100dL /* 0.299087576 */, 15 },
 { 0x04cab8a6L /* 0.299492500 */, 15 }, { 0x04cc6163L /* 0.299897562 */, 15 },
 { 0x04ce0a44L /* 0.300302761 */, 15 }, { 0x04cfb34bL /* 0.300708096 */, 15 },
 { 0x04d15c76L /* 0.301113568 */, 15 }, { 0x04d305c5L /* 0.301519176 */, 15 },

 { 0x04d4af3aL /* 0.301924921 */, 15 }, { 0x04d658d2L /* 0.302330802 */, 15 },
 { 0x04d80290L /* 0.302736820 */, 15 }, { 0x04d9ac72L /* 0.303142973 */, 15 },
 { 0x04db5679L /* 0.303549263 */, 15 }, { 0x04dd00a4L /* 0.303955689 */, 15 },
 { 0x04deaaf3L /* 0.304362251 */, 15 }, { 0x04e05567L /* 0.304768948 */, 15 },
 { 0x04e20000L /* 0.305175781 */, 15 }, { 0x04e3aabdL /* 0.305582750 */, 15 },
 { 0x04e5559eL /* 0.305989854 */, 15 }, { 0x04e700a3L /* 0.306397094 */, 15 },
 { 0x04e8abcdL /* 0.306804470 */, 15 }, { 0x04ea571cL /* 0.307211980 */, 15 },
 { 0x04ec028eL /* 0.307619626 */, 15 }, { 0x04edae25L /* 0.308027406 */, 15 },

 { 0x04ef59e0L /* 0.308435322 */, 15 }, { 0x04f105bfL /* 0.308843373 */, 15 },
 { 0x04f2b1c3L /* 0.309251558 */, 15 }, { 0x04f45deaL /* 0.309659879 */, 15 },
 { 0x04f60a36L /* 0.310068333 */, 15 }, { 0x04f7b6a6L /* 0.310476923 */, 15 },
 { 0x04f9633aL /* 0.310885647 */, 15 }, { 0x04fb0ff2L /* 0.311294505 */, 15 },
 { 0x04fcbcceL /* 0.311703498 */, 15 }, { 0x04fe69ceL /* 0.312112625 */, 15 },
 { 0x050016f3L /* 0.312521885 */, 15 }, { 0x0501c43bL /* 0.312931280 */, 15 },
 { 0x050371a7L /* 0.313340809 */, 15 }, { 0x05051f37L /* 0.313750472 */, 15 },
 { 0x0506ccebL /* 0.314160269 */, 15 }, { 0x05087ac2L /* 0.314570199 */, 15 },

 { 0x050a28beL /* 0.314980262 */, 15 }, { 0x050bd6deL /* 0.315390460 */, 15 },
 { 0x050d8521L /* 0.315800790 */, 15 }, { 0x050f3388L /* 0.316211255 */, 15 },
 { 0x0510e213L /* 0.316621852 */, 15 }, { 0x051290c2L /* 0.317032582 */, 15 },
 { 0x05143f94L /* 0.317443446 */, 15 }, { 0x0515ee8aL /* 0.317854442 */, 15 },
 { 0x05179da4L /* 0.318265572 */, 15 }, { 0x05194ce1L /* 0.318676834 */, 15 },
 { 0x051afc42L /* 0.319088229 */, 15 }, { 0x051cabc7L /* 0.319499756 */, 15 },
 { 0x051e5b6fL /* 0.319911417 */, 15 }, { 0x05200b3aL /* 0.320323209 */, 15 },
 { 0x0521bb2aL /* 0.320735134 */, 15 }, { 0x05236b3dL /* 0.321147192 */, 15 },

 { 0x05251b73L /* 0.321559381 */, 15 }, { 0x0526cbcdL /* 0.321971703 */, 15 },
 { 0x05287c4aL /* 0.322384156 */, 15 }, { 0x052a2ceaL /* 0.322796742 */, 15 },
 { 0x052bddaeL /* 0.323209460 */, 15 }, { 0x052d8e96L /* 0.323622309 */, 15 },
 { 0x052f3fa1L /* 0.324035290 */, 15 }, { 0x0530f0cfL /* 0.324448403 */, 15 },
 { 0x0532a220L /* 0.324861647 */, 15 }, { 0x05345395L /* 0.325275023 */, 15 },
 { 0x0536052dL /* 0.325688530 */, 15 }, { 0x0537b6e8L /* 0.326102168 */, 15 },
 { 0x053968c6L /* 0.326515938 */, 15 }, { 0x053b1ac8L /* 0.326929839 */, 15 },
 { 0x053cccedL /* 0.327343870 */, 15 }, { 0x053e7f35L /* 0.327758033 */, 15 },

 { 0x054031a0L /* 0.328172327 */, 15 }, { 0x0541e42eL /* 0.328586751 */, 15 },
 { 0x054396dfL /* 0.329001306 */, 15 }, { 0x054549b4L /* 0.329415992 */, 15 },
 { 0x0546fcabL /* 0.329830808 */, 15 }, { 0x0548afc6L /* 0.330245755 */, 15 },
 { 0x054a6303L /* 0.330660832 */, 15 }, { 0x054c1663L /* 0.331076039 */, 15 },
 { 0x054dc9e7L /* 0.331491377 */, 15 }, { 0x054f7d8dL /* 0.331906845 */, 15 },
 { 0x05513156L /* 0.332322443 */, 15 }, { 0x0552e542L /* 0.332738170 */, 15 },
 { 0x05549951L /* 0.333154028 */, 15 }, { 0x05564d83L /* 0.333570016 */, 15 },
 { 0x055801d8L /* 0.333986133 */, 15 }, { 0x0559b64fL /* 0.334402380 */, 15 },

 { 0x055b6ae9L /* 0.334818756 */, 15 }, { 0x055d1fa6L /* 0.335235262 */, 15 },
 { 0x055ed486L /* 0.335651898 */, 15 }, { 0x05608988L /* 0.336068662 */, 15 },
 { 0x05623eadL /* 0.336485556 */, 15 }, { 0x0563f3f5L /* 0.336902579 */, 15 },
 { 0x0565a960L /* 0.337319732 */, 15 }, { 0x05675eedL /* 0.337737013 */, 15 },
 { 0x0569149cL /* 0.338154423 */, 15 }, { 0x056aca6fL /* 0.338571962 */, 15 },
 { 0x056c8064L /* 0.338989630 */, 15 }, { 0x056e367bL /* 0.339407426 */, 15 },
 { 0x056fecb5L /* 0.339825351 */, 15 }, { 0x0571a311L /* 0.340243405 */, 15 },
 { 0x05735990L /* 0.340661587 */, 15 }, { 0x05751032L /* 0.341079898 */, 15 },

 { 0x0576c6f5L /* 0.341498336 */, 15 }, { 0x05787ddcL /* 0.341916903 */, 15 },
 { 0x057a34e4L /* 0.342335598 */, 15 }, { 0x057bec0fL /* 0.342754421 */, 15 },
 { 0x057da35dL /* 0.343173373 */, 15 }, { 0x057f5accL /* 0.343592452 */, 15 },
 { 0x0581125eL /* 0.344011659 */, 15 }, { 0x0582ca12L /* 0.344430993 */, 15 },
 { 0x058481e9L /* 0.344850455 */, 15 }, { 0x058639e2L /* 0.345270045 */, 15 },
 { 0x0587f1fdL /* 0.345689763 */, 15 }, { 0x0589aa3aL /* 0.346109608 */, 15 },
 { 0x058b629aL /* 0.346529580 */, 15 }, { 0x058d1b1bL /* 0.346949679 */, 15 },
 { 0x058ed3bfL /* 0.347369906 */, 15 }, { 0x05908c85L /* 0.347790260 */, 15 },

 { 0x0592456dL /* 0.348210741 */, 15 }, { 0x0593fe77L /* 0.348631348 */, 15 },
 { 0x0595b7a3L /* 0.349052083 */, 15 }, { 0x059770f1L /* 0.349472945 */, 15 },
 { 0x05992a61L /* 0.349893933 */, 15 }, { 0x059ae3f3L /* 0.350315048 */, 15 },
 { 0x059c9da8L /* 0.350736290 */, 15 }, { 0x059e577eL /* 0.351157658 */, 15 },
 { 0x05a01176L /* 0.351579152 */, 15 }, { 0x05a1cb90L /* 0.352000773 */, 15 },
 { 0x05a385ccL /* 0.352422520 */, 15 }, { 0x05a5402aL /* 0.352844394 */, 15 },
 { 0x05a6faa9L /* 0.353266393 */, 15 }, { 0x05a8b54bL /* 0.353688519 */, 15 },
 { 0x05aa700eL /* 0.354110771 */, 15 }, { 0x05ac2af3L /* 0.354533148 */, 15 },

 { 0x05ade5faL /* 0.354955651 */, 15 }, { 0x05afa123L /* 0.355378281 */, 15 },
 { 0x05b15c6dL /* 0.355801035 */, 15 }, { 0x05b317d9L /* 0.356223916 */, 15 },
 { 0x05b4d367L /* 0.356646922 */, 15 }, { 0x05b68f16L /* 0.357070053 */, 15 },
 { 0x05b84ae7L /* 0.357493310 */, 15 }, { 0x05ba06daL /* 0.357916692 */, 15 },
 { 0x05bbc2efL /* 0.358340200 */, 15 }, { 0x05bd7f25L /* 0.358763832 */, 15 },
 { 0x05bf3b7cL /* 0.359187590 */, 15 }, { 0x05c0f7f5L /* 0.359611472 */, 15 },
 { 0x05c2b490L /* 0.360035480 */, 15 }, { 0x05c4714cL /* 0.360459613 */, 15 },
 { 0x05c62e2aL /* 0.360883870 */, 15 }, { 0x05c7eb29L /* 0.361308252 */, 15 },

 { 0x05c9a84aL /* 0.361732758 */, 15 }, { 0x05cb658cL /* 0.362157390 */, 15 },
 { 0x05cd22efL /* 0.362582145 */, 15 }, { 0x05cee074L /* 0.363007026 */, 15 },
 { 0x05d09e1bL /* 0.363432030 */, 15 }, { 0x05d25be2L /* 0.363857159 */, 15 },
 { 0x05d419cbL /* 0.364282412 */, 15 }, { 0x05d5d7d5L /* 0.364707789 */, 15 },
 { 0x05d79601L /* 0.365133291 */, 15 }, { 0x05d9544eL /* 0.365558916 */, 15 },
 { 0x05db12bcL /* 0.365984665 */, 15 }, { 0x05dcd14cL /* 0.366410538 */, 15 },
 { 0x05de8ffcL /* 0.366836535 */, 15 }, { 0x05e04eceL /* 0.367262655 */, 15 },
 { 0x05e20dc1L /* 0.367688900 */, 15 }, { 0x05e3ccd5L /* 0.368115267 */, 15 },

 { 0x05e58c0bL /* 0.368541759 */, 15 }, { 0x05e74b61L /* 0.368968373 */, 15 },
 { 0x05e90ad9L /* 0.369395111 */, 15 }, { 0x05eaca72L /* 0.369821973 */, 15 },
 { 0x05ec8a2bL /* 0.370248957 */, 15 }, { 0x05ee4a06L /* 0.370676065 */, 15 },
 { 0x05f00a02L /* 0.371103295 */, 15 }, { 0x05f1ca1fL /* 0.371530649 */, 15 },
 { 0x05f38a5dL /* 0.371958126 */, 15 }, { 0x05f54abcL /* 0.372385725 */, 15 },
 { 0x05f70b3cL /* 0.372813448 */, 15 }, { 0x05f8cbdcL /* 0.373241292 */, 15 },
 { 0x05fa8c9eL /* 0.373669260 */, 15 }, { 0x05fc4d81L /* 0.374097350 */, 15 },
 { 0x05fe0e84L /* 0.374525563 */, 15 }, { 0x05ffcfa8L /* 0.374953898 */, 15 },

 { 0x060190eeL /* 0.375382356 */, 15 }, { 0x06035254L /* 0.375810936 */, 15 },
 { 0x060513daL /* 0.376239638 */, 15 }, { 0x0606d582L /* 0.376668462 */, 15 },
 { 0x0608974aL /* 0.377097408 */, 15 }, { 0x060a5934L /* 0.377526476 */, 15 },
 { 0x060c1b3dL /* 0.377955667 */, 15 }, { 0x060ddd68L /* 0.378384979 */, 15 },
 { 0x060f9fb3L /* 0.378814413 */, 15 }, { 0x0611621fL /* 0.379243968 */, 15 },
 { 0x061324acL /* 0.379673646 */, 15 }, { 0x0614e759L /* 0.380103444 */, 15 },
 { 0x0616aa27L /* 0.380533365 */, 15 }, { 0x06186d16L /* 0.380963407 */, 15 },
 { 0x061a3025L /* 0.381393570 */, 15 }, { 0x061bf354L /* 0.381823855 */, 15 },

 { 0x061db6a5L /* 0.382254261 */, 15 }, { 0x061f7a15L /* 0.382684788 */, 15 },
 { 0x06213da7L /* 0.383115436 */, 15 }, { 0x06230158L /* 0.383546205 */, 15 },
 { 0x0624c52aL /* 0.383977096 */, 15 }, { 0x0626891dL /* 0.384408107 */, 15 },
 { 0x06284d30L /* 0.384839239 */, 15 }, { 0x062a1164L /* 0.385270492 */, 15 },
 { 0x062bd5b8L /* 0.385701865 */, 15 }, { 0x062d9a2cL /* 0.386133359 */, 15 },
 { 0x062f5ec1L /* 0.386564974 */, 15 }, { 0x06312376L /* 0.386996709 */, 15 },
 { 0x0632e84bL /* 0.387428565 */, 15 }, { 0x0634ad41L /* 0.387860541 */, 15 },
 { 0x06367257L /* 0.388292637 */, 15 }, { 0x0638378dL /* 0.388724854 */, 15 },

 { 0x0639fce4L /* 0.389157191 */, 15 }, { 0x063bc25bL /* 0.389589648 */, 15 },
 { 0x063d87f2L /* 0.390022225 */, 15 }, { 0x063f4da9L /* 0.390454922 */, 15 },
 { 0x06411380L /* 0.390887739 */, 15 }, { 0x0642d978L /* 0.391320675 */, 15 },
 { 0x06449f8fL /* 0.391753732 */, 15 }, { 0x064665c7L /* 0.392186908 */, 15 },
 { 0x06482c1fL /* 0.392620204 */, 15 }, { 0x0649f297L /* 0.393053619 */, 15 },
 { 0x064bb92fL /* 0.393487154 */, 15 }, { 0x064d7fe8L /* 0.393920808 */, 15 },
 { 0x064f46c0L /* 0.394354582 */, 15 }, { 0x06510db8L /* 0.394788475 */, 15 },
 { 0x0652d4d0L /* 0.395222488 */, 15 }, { 0x06549c09L /* 0.395656619 */, 15 },

 { 0x06566361L /* 0.396090870 */, 15 }, { 0x06582ad9L /* 0.396525239 */, 15 },
 { 0x0659f271L /* 0.396959728 */, 15 }, { 0x065bba29L /* 0.397394336 */, 15 },
 { 0x065d8201L /* 0.397829062 */, 15 }, { 0x065f49f9L /* 0.398263907 */, 15 },
 { 0x06611211L /* 0.398698871 */, 15 }, { 0x0662da49L /* 0.399133954 */, 15 },
 { 0x0664a2a0L /* 0.399569155 */, 15 }, { 0x06666b17L /* 0.400004475 */, 15 },
 { 0x066833aeL /* 0.400439913 */, 15 }, { 0x0669fc65L /* 0.400875470 */, 15 },
 { 0x066bc53cL /* 0.401311145 */, 15 }, { 0x066d8e32L /* 0.401746938 */, 15 },
 { 0x066f5748L /* 0.402182850 */, 15 }, { 0x0671207eL /* 0.402618879 */, 15 },

 { 0x0672e9d4L /* 0.403055027 */, 15 }, { 0x0674b349L /* 0.403491293 */, 15 },
 { 0x06767cdeL /* 0.403927676 */, 15 }, { 0x06784692L /* 0.404364178 */, 15 },
 { 0x067a1066L /* 0.404800797 */, 15 }, { 0x067bda5aL /* 0.405237535 */, 15 },
 { 0x067da46dL /* 0.405674390 */, 15 }, { 0x067f6ea0L /* 0.406111362 */, 15 },
 { 0x068138f3L /* 0.406548452 */, 15 }, { 0x06830365L /* 0.406985660 */, 15 },
 { 0x0684cdf6L /* 0.407422985 */, 15 }, { 0x068698a8L /* 0.407860427 */, 15 },
 { 0x06886378L /* 0.408297987 */, 15 }, { 0x068a2e68L /* 0.408735664 */, 15 },
 { 0x068bf978L /* 0.409173458 */, 15 }, { 0x068dc4a7L /* 0.409611370 */, 15 },

 { 0x068f8ff5L /* 0.410049398 */, 15 }, { 0x06915b63L /* 0.410487544 */, 15 },
 { 0x069326f0L /* 0.410925806 */, 15 }, { 0x0694f29cL /* 0.411364185 */, 15 },
 { 0x0696be68L /* 0.411802681 */, 15 }, { 0x06988a54L /* 0.412241294 */, 15 },
 { 0x069a565eL /* 0.412680024 */, 15 }, { 0x069c2288L /* 0.413118870 */, 15 },
 { 0x069deed1L /* 0.413557833 */, 15 }, { 0x069fbb3aL /* 0.413996912 */, 15 },
 { 0x06a187c1L /* 0.414436108 */, 15 }, { 0x06a35468L /* 0.414875420 */, 15 },
 { 0x06a5212fL /* 0.415314849 */, 15 }, { 0x06a6ee14L /* 0.415754393 */, 15 },
 { 0x06a8bb18L /* 0.416194054 */, 15 }, { 0x06aa883cL /* 0.416633831 */, 15 },

 { 0x06ac557fL /* 0.417073724 */, 15 }, { 0x06ae22e1L /* 0.417513734 */, 15 },
 { 0x06aff062L /* 0.417953859 */, 15 }, { 0x06b1be03L /* 0.418394100 */, 15 },
 { 0x06b38bc2L /* 0.418834457 */, 15 }, { 0x06b559a1L /* 0.419274929 */, 15 },
 { 0x06b7279eL /* 0.419715518 */, 15 }, { 0x06b8f5bbL /* 0.420156222 */, 15 },
 { 0x06bac3f6L /* 0.420597041 */, 15 }, { 0x06bc9251L /* 0.421037977 */, 15 },
 { 0x06be60cbL /* 0.421479027 */, 15 }, { 0x06c02f63L /* 0.421920193 */, 15 },
 { 0x06c1fe1bL /* 0.422361475 */, 15 }, { 0x06c3ccf1L /* 0.422802871 */, 15 },
 { 0x06c59be7L /* 0.423244383 */, 15 }, { 0x06c76afbL /* 0.423686010 */, 15 },

 { 0x06c93a2eL /* 0.424127753 */, 15 }, { 0x06cb0981L /* 0.424569610 */, 15 },
 { 0x06ccd8f2L /* 0.425011582 */, 15 }, { 0x06cea881L /* 0.425453669 */, 15 },
 { 0x06d07830L /* 0.425895871 */, 15 }, { 0x06d247feL /* 0.426338188 */, 15 },
 { 0x06d417eaL /* 0.426780620 */, 15 }, { 0x06d5e7f5L /* 0.427223166 */, 15 },
 { 0x06d7b81fL /* 0.427665827 */, 15 }, { 0x06d98868L /* 0.428108603 */, 15 },
 { 0x06db58cfL /* 0.428551493 */, 15 }, { 0x06dd2955L /* 0.428994497 */, 15 },
 { 0x06def9faL /* 0.429437616 */, 15 }, { 0x06e0cabeL /* 0.429880849 */, 15 },
 { 0x06e29ba0L /* 0.430324197 */, 15 }, { 0x06e46ca1L /* 0.430767659 */, 15 },

 { 0x06e63dc0L /* 0.431211234 */, 15 }, { 0x06e80efeL /* 0.431654924 */, 15 },
 { 0x06e9e05bL /* 0.432098728 */, 15 }, { 0x06ebb1d6L /* 0.432542647 */, 15 },
 { 0x06ed8370L /* 0.432986678 */, 15 }, { 0x06ef5529L /* 0.433430824 */, 15 },
 { 0x06f12700L /* 0.433875084 */, 15 }, { 0x06f2f8f5L /* 0.434319457 */, 15 },
 { 0x06f4cb09L /* 0.434763944 */, 15 }, { 0x06f69d3cL /* 0.435208545 */, 15 },
 { 0x06f86f8dL /* 0.435653259 */, 15 }, { 0x06fa41fdL /* 0.436098087 */, 15 },
 { 0x06fc148bL /* 0.436543029 */, 15 }, { 0x06fde737L /* 0.436988083 */, 15 },
 { 0x06ffba02L /* 0.437433251 */, 15 }, { 0x07018cebL /* 0.437878533 */, 15 },

 { 0x07035ff3L /* 0.438323927 */, 15 }, { 0x07053319L /* 0.438769435 */, 15 },
 { 0x0707065dL /* 0.439215056 */, 15 }, { 0x0708d9c0L /* 0.439660790 */, 15 },
 { 0x070aad41L /* 0.440106636 */, 15 }, { 0x070c80e1L /* 0.440552596 */, 15 },
 { 0x070e549fL /* 0.440998669 */, 15 }, { 0x0710287bL /* 0.441444855 */, 15 },
 { 0x0711fc75L /* 0.441891153 */, 15 }, { 0x0713d08dL /* 0.442337564 */, 15 },
 { 0x0715a4c4L /* 0.442784088 */, 15 }, { 0x07177919L /* 0.443230724 */, 15 },
 { 0x07194d8cL /* 0.443677473 */, 15 }, { 0x071b221eL /* 0.444124334 */, 15 },
 { 0x071cf6ceL /* 0.444571308 */, 15 }, { 0x071ecb9bL /* 0.445018394 */, 15 },

 { 0x0720a087L /* 0.445465593 */, 15 }, { 0x07227591L /* 0.445912903 */, 15 },
 { 0x07244ab9L /* 0.446360326 */, 15 }, { 0x07262000L /* 0.446807861 */, 15 },
 { 0x0727f564L /* 0.447255509 */, 15 }, { 0x0729cae7L /* 0.447703268 */, 15 },
 { 0x072ba087L /* 0.448151139 */, 15 }, { 0x072d7646L /* 0.448599122 */, 15 },
 { 0x072f4c22L /* 0.449047217 */, 15 }, { 0x0731221dL /* 0.449495424 */, 15 },
 { 0x0732f835L /* 0.449943742 */, 15 }, { 0x0734ce6cL /* 0.450392173 */, 15 },
 { 0x0736a4c1L /* 0.450840715 */, 15 }, { 0x07387b33L /* 0.451289368 */, 15 },
 { 0x073a51c4L /* 0.451738133 */, 15 }, { 0x073c2872L /* 0.452187010 */, 15 },

 { 0x073dff3eL /* 0.452635998 */, 15 }, { 0x073fd628L /* 0.453085097 */, 15 },
 { 0x0741ad30L /* 0.453534308 */, 15 }, { 0x07438456L /* 0.453983630 */, 15 },
 { 0x07455b9aL /* 0.454433063 */, 15 }, { 0x074732fcL /* 0.454882607 */, 15 },
 { 0x07490a7bL /* 0.455332262 */, 15 }, { 0x074ae218L /* 0.455782029 */, 15 },
 { 0x074cb9d3L /* 0.456231906 */, 15 }, { 0x074e91acL /* 0.456681894 */, 15 },
 { 0x075069a3L /* 0.457131993 */, 15 }, { 0x075241b7L /* 0.457582203 */, 15 },
 { 0x075419e9L /* 0.458032524 */, 15 }, { 0x0755f239L /* 0.458482956 */, 15 },
 { 0x0757caa7L /* 0.458933498 */, 15 }, { 0x0759a332L /* 0.459384151 */, 15 },

 { 0x075b7bdbL /* 0.459834914 */, 15 }, { 0x075d54a1L /* 0.460285788 */, 15 },
 { 0x075f2d85L /* 0.460736772 */, 15 }, { 0x07610687L /* 0.461187867 */, 15 },
 { 0x0762dfa6L /* 0.461639071 */, 15 }, { 0x0764b8e3L /* 0.462090387 */, 15 },
 { 0x0766923eL /* 0.462541812 */, 15 }, { 0x07686bb6L /* 0.462993348 */, 15 },
 { 0x076a454cL /* 0.463444993 */, 15 }, { 0x076c1effL /* 0.463896749 */, 15 },
 { 0x076df8d0L /* 0.464348615 */, 15 }, { 0x076fd2beL /* 0.464800591 */, 15 },
 { 0x0771accaL /* 0.465252676 */, 15 }, { 0x077386f3L /* 0.465704872 */, 15 },
 { 0x0775613aL /* 0.466157177 */, 15 }, { 0x07773b9eL /* 0.466609592 */, 15 },

 { 0x07791620L /* 0.467062117 */, 15 }, { 0x077af0bfL /* 0.467514751 */, 15 },
 { 0x077ccb7cL /* 0.467967495 */, 15 }, { 0x077ea656L /* 0.468420349 */, 15 },
 { 0x0780814dL /* 0.468873312 */, 15 }, { 0x07825c62L /* 0.469326384 */, 15 },
 { 0x07843794L /* 0.469779566 */, 15 }, { 0x078612e3L /* 0.470232857 */, 15 },
 { 0x0787ee50L /* 0.470686258 */, 15 }, { 0x0789c9daL /* 0.471139767 */, 15 },
 { 0x078ba581L /* 0.471593386 */, 15 }, { 0x078d8146L /* 0.472047114 */, 15 },
 { 0x078f5d28L /* 0.472500951 */, 15 }, { 0x07913927L /* 0.472954896 */, 15 },
 { 0x07931543L /* 0.473408951 */, 15 }, { 0x0794f17dL /* 0.473863115 */, 15 },

 { 0x0796cdd4L /* 0.474317388 */, 15 }, { 0x0798aa48L /* 0.474771769 */, 15 },
 { 0x079a86d9L /* 0.475226259 */, 15 }, { 0x079c6388L /* 0.475680858 */, 15 },
 { 0x079e4053L /* 0.476135565 */, 15 }, { 0x07a01d3cL /* 0.476590381 */, 15 },
 { 0x07a1fa42L /* 0.477045306 */, 15 }, { 0x07a3d765L /* 0.477500339 */, 15 },
 { 0x07a5b4a5L /* 0.477955481 */, 15 }, { 0x07a79202L /* 0.478410731 */, 15 },
 { 0x07a96f7dL /* 0.478866089 */, 15 }, { 0x07ab4d14L /* 0.479321555 */, 15 },
 { 0x07ad2ac8L /* 0.479777130 */, 15 }, { 0x07af089aL /* 0.480232813 */, 15 },
 { 0x07b0e688L /* 0.480688604 */, 15 }, { 0x07b2c494L /* 0.481144503 */, 15 },

 { 0x07b4a2bcL /* 0.481600510 */, 15 }, { 0x07b68102L /* 0.482056625 */, 15 },
 { 0x07b85f64L /* 0.482512848 */, 15 }, { 0x07ba3de4L /* 0.482969179 */, 15 },
 { 0x07bc1c80L /* 0.483425618 */, 15 }, { 0x07bdfb39L /* 0.483882164 */, 15 },
 { 0x07bfda0fL /* 0.484338818 */, 15 }, { 0x07c1b902L /* 0.484795580 */, 15 },
 { 0x07c39812L /* 0.485252449 */, 15 }, { 0x07c5773fL /* 0.485709426 */, 15 },
 { 0x07c75689L /* 0.486166511 */, 15 }, { 0x07c935efL /* 0.486623703 */, 15 },
 { 0x07cb1573L /* 0.487081002 */, 15 }, { 0x07ccf513L /* 0.487538409 */, 15 },
 { 0x07ced4d0L /* 0.487995923 */, 15 }, { 0x07d0b4aaL /* 0.488453544 */, 15 },

 { 0x07d294a0L /* 0.488911273 */, 15 }, { 0x07d474b3L /* 0.489369108 */, 15 },
 { 0x07d654e4L /* 0.489827051 */, 15 }, { 0x07d83530L /* 0.490285101 */, 15 },
 { 0x07da159aL /* 0.490743258 */, 15 }, { 0x07dbf620L /* 0.491201522 */, 15 },
 { 0x07ddd6c3L /* 0.491659892 */, 15 }, { 0x07dfb783L /* 0.492118370 */, 15 },
 { 0x07e1985fL /* 0.492576954 */, 15 }, { 0x07e37958L /* 0.493035645 */, 15 },
 { 0x07e55a6eL /* 0.493494443 */, 15 }, { 0x07e73ba0L /* 0.493953348 */, 15 },
 { 0x07e91cefL /* 0.494412359 */, 15 }, { 0x07eafe5aL /* 0.494871476 */, 15 },
 { 0x07ecdfe2L /* 0.495330701 */, 15 }, { 0x07eec187L /* 0.495790031 */, 15 },

 { 0x07f0a348L /* 0.496249468 */, 15 }, { 0x07f28526L /* 0.496709012 */, 15 },
 { 0x07f46720L /* 0.497168662 */, 15 }, { 0x07f64937L /* 0.497628418 */, 15 },
 { 0x07f82b6aL /* 0.498088280 */, 15 }, { 0x07fa0dbaL /* 0.498548248 */, 15 },
 { 0x07fbf026L /* 0.499008323 */, 15 }, { 0x07fdd2afL /* 0.499468503 */, 15 },
 { 0x07ffb554L /* 0.499928790 */, 15 }, { 0x0400cc0bL /* 0.250194591 */, 16 },
 { 0x0401bd7aL /* 0.250424840 */, 16 }, { 0x0402aef7L /* 0.250655143 */, 16 },
 { 0x0403a083L /* 0.250885498 */, 16 }, { 0x0404921cL /* 0.251115906 */, 16 },
 { 0x040583c4L /* 0.251346367 */, 16 }, { 0x0406757aL /* 0.251576880 */, 16 },

 { 0x0407673fL /* 0.251807447 */, 16 }, { 0x04085911L /* 0.252038066 */, 16 },
 { 0x04094af1L /* 0.252268738 */, 16 }, { 0x040a3ce0L /* 0.252499463 */, 16 },
 { 0x040b2eddL /* 0.252730240 */, 16 }, { 0x040c20e8L /* 0.252961071 */, 16 },
 { 0x040d1301L /* 0.253191953 */, 16 }, { 0x040e0529L /* 0.253422889 */, 16 },
 { 0x040ef75eL /* 0.253653877 */, 16 }, { 0x040fe9a1L /* 0.253884918 */, 16 },
 { 0x0410dbf3L /* 0.254116011 */, 16 }, { 0x0411ce53L /* 0.254347157 */, 16 },
 { 0x0412c0c1L /* 0.254578356 */, 16 }, { 0x0413b33dL /* 0.254809606 */, 16 },
 { 0x0414a5c7L /* 0.255040910 */, 16 }, { 0x0415985fL /* 0.255272266 */, 16 },

 { 0x04168b05L /* 0.255503674 */, 16 }, { 0x04177db9L /* 0.255735135 */, 16 },
 { 0x0418707cL /* 0.255966648 */, 16 }, { 0x0419634cL /* 0.256198213 */, 16 },
 { 0x041a562aL /* 0.256429831 */, 16 }, { 0x041b4917L /* 0.256661501 */, 16 },
 { 0x041c3c11L /* 0.256893223 */, 16 }, { 0x041d2f1aL /* 0.257124998 */, 16 },
 { 0x041e2230L /* 0.257356825 */, 16 }, { 0x041f1555L /* 0.257588704 */, 16 },
 { 0x04200888L /* 0.257820635 */, 16 }, { 0x0420fbc8L /* 0.258052619 */, 16 },
 { 0x0421ef17L /* 0.258284654 */, 16 }, { 0x0422e273L /* 0.258516742 */, 16 },
 { 0x0423d5deL /* 0.258748882 */, 16 }, { 0x0424c956L /* 0.258981074 */, 16 },

 { 0x0425bcddL /* 0.259213318 */, 16 }, { 0x0426b071L /* 0.259445614 */, 16 },
 { 0x0427a414L /* 0.259677962 */, 16 }, { 0x042897c4L /* 0.259910362 */, 16 },
 { 0x04298b83L /* 0.260142814 */, 16 }, { 0x042a7f4fL /* 0.260375318 */, 16 },
 { 0x042b7329L /* 0.260607874 */, 16 }, { 0x042c6711L /* 0.260840481 */, 16 },
 { 0x042d5b07L /* 0.261073141 */, 16 }, { 0x042e4f0bL /* 0.261305852 */, 16 },
 { 0x042f431dL /* 0.261538616 */, 16 }, { 0x0430373dL /* 0.261771431 */, 16 },
 { 0x04312b6bL /* 0.262004297 */, 16 }, { 0x04321fa6L /* 0.262237216 */, 16 },
 { 0x043313f0L /* 0.262470186 */, 16 }, { 0x04340847L /* 0.262703208 */, 16 },

 { 0x0434fcadL /* 0.262936282 */, 16 }, { 0x0435f120L /* 0.263169407 */, 16 },
 { 0x0436e5a1L /* 0.263402584 */, 16 }, { 0x0437da2fL /* 0.263635813 */, 16 },
 { 0x0438ceccL /* 0.263869093 */, 16 }, { 0x0439c377L /* 0.264102425 */, 16 },
 { 0x043ab82fL /* 0.264335808 */, 16 }, { 0x043bacf5L /* 0.264569243 */, 16 },
 { 0x043ca1c9L /* 0.264802730 */, 16 }, { 0x043d96abL /* 0.265036267 */, 16 },
 { 0x043e8b9bL /* 0.265269857 */, 16 }, { 0x043f8098L /* 0.265503498 */, 16 },
 { 0x044075a3L /* 0.265737190 */, 16 }, { 0x04416abcL /* 0.265970933 */, 16 },
 { 0x04425fe3L /* 0.266204728 */, 16 }, { 0x04435518L /* 0.266438574 */, 16 },

 { 0x04444a5aL /* 0.266672472 */, 16 }, { 0x04453fabL /* 0.266906421 */, 16 },
 { 0x04463508L /* 0.267140421 */, 16 }, { 0x04472a74L /* 0.267374472 */, 16 },
 { 0x04481feeL /* 0.267608575 */, 16 }, { 0x04491575L /* 0.267842729 */, 16 },
 { 0x044a0b0aL /* 0.268076934 */, 16 }, { 0x044b00acL /* 0.268311190 */, 16 },
 { 0x044bf65dL /* 0.268545497 */, 16 }, { 0x044cec1bL /* 0.268779856 */, 16 },
 { 0x044de1e7L /* 0.269014265 */, 16 }, { 0x044ed7c0L /* 0.269248726 */, 16 },
 { 0x044fcda8L /* 0.269483238 */, 16 }, { 0x0450c39cL /* 0.269717800 */, 16 },
 { 0x0451b99fL /* 0.269952414 */, 16 }, { 0x0452afafL /* 0.270187079 */, 16 },

 { 0x0453a5cdL /* 0.270421794 */, 16 }, { 0x04549bf9L /* 0.270656561 */, 16 },
 { 0x04559232L /* 0.270891379 */, 16 }, { 0x04568879L /* 0.271126247 */, 16 },
 { 0x04577eceL /* 0.271361166 */, 16 }, { 0x04587530L /* 0.271596136 */, 16 },
 { 0x04596ba0L /* 0.271831157 */, 16 }, { 0x045a621eL /* 0.272066229 */, 16 },
 { 0x045b58a9L /* 0.272301352 */, 16 }, { 0x045c4f42L /* 0.272536525 */, 16 },
 { 0x045d45e9L /* 0.272771749 */, 16 }, { 0x045e3c9dL /* 0.273007024 */, 16 },
 { 0x045f335eL /* 0.273242350 */, 16 }, { 0x04602a2eL /* 0.273477726 */, 16 },
 { 0x0461210bL /* 0.273713153 */, 16 }, { 0x046217f5L /* 0.273948630 */, 16 },

 { 0x04630eedL /* 0.274184158 */, 16 }, { 0x046405f3L /* 0.274419737 */, 16 },
 { 0x0464fd06L /* 0.274655366 */, 16 }, { 0x0465f427L /* 0.274891046 */, 16 },
 { 0x0466eb55L /* 0.275126776 */, 16 }, { 0x0467e291L /* 0.275362557 */, 16 },
 { 0x0468d9dbL /* 0.275598389 */, 16 }, { 0x0469d132L /* 0.275834270 */, 16 },
 { 0x046ac896L /* 0.276070203 */, 16 }, { 0x046bc009L /* 0.276306185 */, 16 },
 { 0x046cb788L /* 0.276542218 */, 16 }, { 0x046daf15L /* 0.276778302 */, 16 },
 { 0x046ea6b0L /* 0.277014435 */, 16 }, { 0x046f9e58L /* 0.277250619 */, 16 },
 { 0x0470960eL /* 0.277486854 */, 16 }, { 0x04718dd1L /* 0.277723139 */, 16 },

 { 0x047285a2L /* 0.277959474 */, 16 }, { 0x04737d80L /* 0.278195859 */, 16 },
 { 0x0474756cL /* 0.278432294 */, 16 }, { 0x04756d65L /* 0.278668780 */, 16 },
 { 0x0476656bL /* 0.278905316 */, 16 }, { 0x04775d7fL /* 0.279141902 */, 16 },
 { 0x047855a1L /* 0.279378538 */, 16 }, { 0x04794dd0L /* 0.279615224 */, 16 },
 { 0x047a460cL /* 0.279851960 */, 16 }, { 0x047b3e56L /* 0.280088747 */, 16 },
 { 0x047c36aeL /* 0.280325583 */, 16 }, { 0x047d2f12L /* 0.280562470 */, 16 },
 { 0x047e2784L /* 0.280799406 */, 16 }, { 0x047f2004L /* 0.281036393 */, 16 },
 { 0x04801891L /* 0.281273429 */, 16 }, { 0x0481112bL /* 0.281510516 */, 16 },

 { 0x048209d3L /* 0.281747652 */, 16 }, { 0x04830288L /* 0.281984838 */, 16 },
 { 0x0483fb4bL /* 0.282222075 */, 16 }, { 0x0484f41bL /* 0.282459361 */, 16 },
 { 0x0485ecf8L /* 0.282696697 */, 16 }, { 0x0486e5e3L /* 0.282934082 */, 16 },
 { 0x0487dedbL /* 0.283171518 */, 16 }, { 0x0488d7e1L /* 0.283409003 */, 16 },
 { 0x0489d0f4L /* 0.283646538 */, 16 }, { 0x048aca14L /* 0.283884123 */, 16 },
 { 0x048bc341L /* 0.284121757 */, 16 }, { 0x048cbc7cL /* 0.284359441 */, 16 },
 { 0x048db5c4L /* 0.284597175 */, 16 }, { 0x048eaf1aL /* 0.284834959 */, 16 },
 { 0x048fa87dL /* 0.285072792 */, 16 }, { 0x0490a1edL /* 0.285310675 */, 16 },

 { 0x04919b6aL /* 0.285548607 */, 16 }, { 0x049294f5L /* 0.285786589 */, 16 },
 { 0x04938e8dL /* 0.286024621 */, 16 }, { 0x04948833L /* 0.286262702 */, 16 },
 { 0x049581e5L /* 0.286500832 */, 16 }, { 0x04967ba5L /* 0.286739012 */, 16 },
 { 0x04977573L /* 0.286977242 */, 16 }, { 0x04986f4dL /* 0.287215521 */, 16 },
 { 0x04996935L /* 0.287453849 */, 16 }, { 0x049a632aL /* 0.287692227 */, 16 },
 { 0x049b5d2cL /* 0.287930654 */, 16 }, { 0x049c573cL /* 0.288169131 */, 16 },
 { 0x049d5159L /* 0.288407657 */, 16 }, { 0x049e4b83L /* 0.288646232 */, 16 },
 { 0x049f45baL /* 0.288884857 */, 16 }, { 0x04a03ffeL /* 0.289123530 */, 16 },

 { 0x04a13a50L /* 0.289362253 */, 16 }, { 0x04a234afL /* 0.289601026 */, 16 },
 { 0x04a32f1bL /* 0.289839847 */, 16 }, { 0x04a42995L /* 0.290078718 */, 16 },
 { 0x04a5241bL /* 0.290317638 */, 16 }, { 0x04a61eafL /* 0.290556607 */, 16 },
 { 0x04a71950L /* 0.290795626 */, 16 }, { 0x04a813feL /* 0.291034693 */, 16 },
 { 0x04a90ebaL /* 0.291273810 */, 16 }, { 0x04aa0982L /* 0.291512975 */, 16 },
 { 0x04ab0458L /* 0.291752190 */, 16 }, { 0x04abff3bL /* 0.291991453 */, 16 },
 { 0x04acfa2bL /* 0.292230766 */, 16 }, { 0x04adf528L /* 0.292470128 */, 16 },
 { 0x04aef032L /* 0.292709539 */, 16 }, { 0x04afeb4aL /* 0.292948998 */, 16 },

 { 0x04b0e66eL /* 0.293188507 */, 16 }, { 0x04b1e1a0L /* 0.293428065 */, 16 },
 { 0x04b2dcdfL /* 0.293667671 */, 16 }, { 0x04b3d82bL /* 0.293907326 */, 16 },
 { 0x04b4d384L /* 0.294147031 */, 16 }, { 0x04b5ceeaL /* 0.294386784 */, 16 },
 { 0x04b6ca5eL /* 0.294626585 */, 16 }, { 0x04b7c5deL /* 0.294866436 */, 16 },
 { 0x04b8c16cL /* 0.295106336 */, 16 }, { 0x04b9bd06L /* 0.295346284 */, 16 },
 { 0x04bab8aeL /* 0.295586281 */, 16 }, { 0x04bbb463L /* 0.295826327 */, 16 },
 { 0x04bcb024L /* 0.296066421 */, 16 }, { 0x04bdabf3L /* 0.296306564 */, 16 },
 { 0x04bea7cfL /* 0.296546756 */, 16 }, { 0x04bfa3b8L /* 0.296786996 */, 16 },

 { 0x04c09fafL /* 0.297027285 */, 16 }, { 0x04c19bb2L /* 0.297267623 */, 16 },
 { 0x04c297c2L /* 0.297508009 */, 16 }, { 0x04c393dfL /* 0.297748444 */, 16 },
 { 0x04c49009L /* 0.297988927 */, 16 }, { 0x04c58c41L /* 0.298229459 */, 16 },
 { 0x04c68885L /* 0.298470039 */, 16 }, { 0x04c784d6L /* 0.298710668 */, 16 },
 { 0x04c88135L /* 0.298951346 */, 16 }, { 0x04c97da0L /* 0.299192071 */, 16 },
 { 0x04ca7a18L /* 0.299432846 */, 16 }, { 0x04cb769eL /* 0.299673668 */, 16 },
 { 0x04cc7330L /* 0.299914539 */, 16 }, { 0x04cd6fcfL /* 0.300155459 */, 16 },
 { 0x04ce6c7bL /* 0.300396426 */, 16 }, { 0x04cf6935L /* 0.300637443 */, 16 },

 { 0x04d065fbL /* 0.300878507 */, 16 }, { 0x04d162ceL /* 0.301119620 */, 16 },
 { 0x04d25faeL /* 0.301360781 */, 16 }, { 0x04d35c9bL /* 0.301601990 */, 16 },
 { 0x04d45995L /* 0.301843247 */, 16 }, { 0x04d5569cL /* 0.302084553 */, 16 },
 { 0x04d653b0L /* 0.302325907 */, 16 }, { 0x04d750d1L /* 0.302567309 */, 16 },
 { 0x04d84dffL /* 0.302808759 */, 16 }, { 0x04d94b3aL /* 0.303050257 */, 16 },
 { 0x04da4881L /* 0.303291804 */, 16 }, { 0x04db45d6L /* 0.303533399 */, 16 },
 { 0x04dc4337L /* 0.303775041 */, 16 }, { 0x04dd40a6L /* 0.304016732 */, 16 },
 { 0x04de3e21L /* 0.304258471 */, 16 }, { 0x04df3ba9L /* 0.304500257 */, 16 },

 { 0x04e0393eL /* 0.304742092 */, 16 }, { 0x04e136e0L /* 0.304983975 */, 16 },
 { 0x04e2348fL /* 0.305225906 */, 16 }, { 0x04e3324bL /* 0.305467885 */, 16 },
 { 0x04e43013L /* 0.305709911 */, 16 }, { 0x04e52de9L /* 0.305951986 */, 16 },
 { 0x04e62bcbL /* 0.306194108 */, 16 }, { 0x04e729baL /* 0.306436279 */, 16 },
 { 0x04e827b6L /* 0.306678497 */, 16 }, { 0x04e925bfL /* 0.306920763 */, 16 },
 { 0x04ea23d4L /* 0.307163077 */, 16 }, { 0x04eb21f7L /* 0.307405438 */, 16 },
 { 0x04ec2026L /* 0.307647848 */, 16 }, { 0x04ed1e62L /* 0.307890305 */, 16 },
 { 0x04ee1cabL /* 0.308132810 */, 16 }, { 0x04ef1b01L /* 0.308375362 */, 16 },

 { 0x04f01963L /* 0.308617963 */, 16 }, { 0x04f117d3L /* 0.308860611 */, 16 },
 { 0x04f2164fL /* 0.309103306 */, 16 }, { 0x04f314d8L /* 0.309346050 */, 16 },
 { 0x04f4136dL /* 0.309588841 */, 16 }, { 0x04f51210L /* 0.309831679 */, 16 },
 { 0x04f610bfL /* 0.310074565 */, 16 }, { 0x04f70f7bL /* 0.310317499 */, 16 },
 { 0x04f80e44L /* 0.310560480 */, 16 }, { 0x04f90d19L /* 0.310803509 */, 16 },
 { 0x04fa0bfcL /* 0.311046586 */, 16 }, { 0x04fb0aebL /* 0.311289710 */, 16 },
 { 0x04fc09e7L /* 0.311532881 */, 16 }, { 0x04fd08efL /* 0.311776100 */, 16 },
 { 0x04fe0805L /* 0.312019366 */, 16 }, { 0x04ff0727L /* 0.312262680 */, 16 },

 { 0x05000655L /* 0.312506041 */, 16 }, { 0x05010591L /* 0.312749449 */, 16 },
 { 0x050204d9L /* 0.312992905 */, 16 }, { 0x0503042eL /* 0.313236408 */, 16 },
 { 0x0504038fL /* 0.313479959 */, 16 }, { 0x050502feL /* 0.313723556 */, 16 },
 { 0x05060279L /* 0.313967202 */, 16 }, { 0x05070200L /* 0.314210894 */, 16 },
 { 0x05080195L /* 0.314454634 */, 16 }, { 0x05090136L /* 0.314698420 */, 16 },
 { 0x050a00e3L /* 0.314942255 */, 16 }, { 0x050b009eL /* 0.315186136 */, 16 },
 { 0x050c0065L /* 0.315430064 */, 16 }, { 0x050d0039L /* 0.315674040 */, 16 },
 { 0x050e0019L /* 0.315918063 */, 16 }, { 0x050f0006L /* 0.316162133 */, 16 },

 { 0x05100000L /* 0.316406250 */, 16 }, { 0x05110006L /* 0.316650414 */, 16 },
 { 0x05120019L /* 0.316894625 */, 16 }, { 0x05130039L /* 0.317138884 */, 16 },
 { 0x05140065L /* 0.317383189 */, 16 }, { 0x0515009eL /* 0.317627541 */, 16 },
 { 0x051600e3L /* 0.317871941 */, 16 }, { 0x05170135L /* 0.318116387 */, 16 },
 { 0x05180194L /* 0.318360880 */, 16 }, { 0x051901ffL /* 0.318605421 */, 16 },
 { 0x051a0277L /* 0.318850008 */, 16 }, { 0x051b02fcL /* 0.319094642 */, 16 },
 { 0x051c038dL /* 0.319339323 */, 16 }, { 0x051d042aL /* 0.319584051 */, 16 },
 { 0x051e04d4L /* 0.319828826 */, 16 }, { 0x051f058bL /* 0.320073647 */, 16 },

 { 0x0520064fL /* 0.320318516 */, 16 }, { 0x0521071fL /* 0.320563431 */, 16 },
 { 0x052207fbL /* 0.320808393 */, 16 }, { 0x052308e4L /* 0.321053402 */, 16 },
 { 0x052409daL /* 0.321298457 */, 16 }, { 0x05250adcL /* 0.321543560 */, 16 },
 { 0x05260beaL /* 0.321788709 */, 16 }, { 0x05270d06L /* 0.322033904 */, 16 },
 { 0x05280e2dL /* 0.322279147 */, 16 }, { 0x05290f62L /* 0.322524436 */, 16 },
 { 0x052a10a3L /* 0.322769771 */, 16 }, { 0x052b11f0L /* 0.323015154 */, 16 },
 { 0x052c134aL /* 0.323260583 */, 16 }, { 0x052d14b0L /* 0.323506058 */, 16 },
 { 0x052e1623L /* 0.323751580 */, 16 }, { 0x052f17a2L /* 0.323997149 */, 16 },

 { 0x0530192eL /* 0.324242764 */, 16 }, { 0x05311ac6L /* 0.324488426 */, 16 },
 { 0x05321c6bL /* 0.324734134 */, 16 }, { 0x05331e1cL /* 0.324979889 */, 16 },
 { 0x05341fdaL /* 0.325225690 */, 16 }, { 0x053521a4L /* 0.325471538 */, 16 },
 { 0x0536237bL /* 0.325717432 */, 16 }, { 0x0537255eL /* 0.325963372 */, 16 },
 { 0x0538274eL /* 0.326209359 */, 16 }, { 0x0539294aL /* 0.326455392 */, 16 },
 { 0x053a2b52L /* 0.326701472 */, 16 }, { 0x053b2d67L /* 0.326947598 */, 16 },
 { 0x053c2f89L /* 0.327193770 */, 16 }, { 0x053d31b6L /* 0.327439989 */, 16 },
 { 0x053e33f1L /* 0.327686254 */, 16 }, { 0x053f3637L /* 0.327932565 */, 16 },

 { 0x0540388aL /* 0.328178922 */, 16 }, { 0x05413aeaL /* 0.328425326 */, 16 },
 { 0x05423d56L /* 0.328671776 */, 16 }, { 0x05433fceL /* 0.328918272 */, 16 },
 { 0x05444253L /* 0.329164814 */, 16 }, { 0x054544e4L /* 0.329411403 */, 16 },
 { 0x05464781L /* 0.329658038 */, 16 }, { 0x05474a2bL /* 0.329904718 */, 16 },
 { 0x05484ce2L /* 0.330151445 */, 16 }, { 0x05494fa4L /* 0.330398218 */, 16 },
 { 0x054a5273L /* 0.330645037 */, 16 }, { 0x054b554eL /* 0.330891903 */, 16 },
 { 0x054c5836L /* 0.331138814 */, 16 }, { 0x054d5b2aL /* 0.331385771 */, 16 },
 { 0x054e5e2bL /* 0.331632774 */, 16 }, { 0x054f6138L /* 0.331879824 */, 16 },

 { 0x05506451L /* 0.332126919 */, 16 }, { 0x05516776L /* 0.332374060 */, 16 },
 { 0x05526aa8L /* 0.332621247 */, 16 }, { 0x05536de6L /* 0.332868480 */, 16 },
 { 0x05547131L /* 0.333115759 */, 16 }, { 0x05557487L /* 0.333363084 */, 16 },
 { 0x055677eaL /* 0.333610455 */, 16 }, { 0x05577b5aL /* 0.333857872 */, 16 },
 { 0x05587ed5L /* 0.334105334 */, 16 }, { 0x0559825eL /* 0.334352843 */, 16 },
 { 0x055a85f2L /* 0.334600397 */, 16 }, { 0x055b8992L /* 0.334847997 */, 16 },
 { 0x055c8d3fL /* 0.335095642 */, 16 }, { 0x055d90f9L /* 0.335343334 */, 16 },
 { 0x055e94beL /* 0.335591071 */, 16 }, { 0x055f9890L /* 0.335838854 */, 16 },

 { 0x05609c6eL /* 0.336086683 */, 16 }, { 0x0561a058L /* 0.336334557 */, 16 },
 { 0x0562a44fL /* 0.336582477 */, 16 }, { 0x0563a851L /* 0.336830443 */, 16 },
 { 0x0564ac60L /* 0.337078454 */, 16 }, { 0x0565b07cL /* 0.337326511 */, 16 },
 { 0x0566b4a3L /* 0.337574614 */, 16 }, { 0x0567b8d7L /* 0.337822762 */, 16 },
 { 0x0568bd17L /* 0.338070956 */, 16 }, { 0x0569c163L /* 0.338319195 */, 16 },
 { 0x056ac5bcL /* 0.338567480 */, 16 }, { 0x056bca20L /* 0.338815811 */, 16 },
 { 0x056cce91L /* 0.339064186 */, 16 }, { 0x056dd30eL /* 0.339312608 */, 16 },
 { 0x056ed798L /* 0.339561075 */, 16 }, { 0x056fdc2dL /* 0.339809587 */, 16 },

 { 0x0570e0cfL /* 0.340058145 */, 16 }, { 0x0571e57dL /* 0.340306748 */, 16 },
 { 0x0572ea37L /* 0.340555397 */, 16 }, { 0x0573eefdL /* 0.340804091 */, 16 },
 { 0x0574f3d0L /* 0.341052830 */, 16 }, { 0x0575f8aeL /* 0.341301615 */, 16 },
 { 0x0576fd99L /* 0.341550445 */, 16 }, { 0x05780290L /* 0.341799321 */, 16 },
 { 0x05790793L /* 0.342048241 */, 16 }, { 0x057a0ca3L /* 0.342297207 */, 16 },
 { 0x057b11beL /* 0.342546219 */, 16 }, { 0x057c16e6L /* 0.342795275 */, 16 },
 { 0x057d1c1aL /* 0.343044377 */, 16 }, { 0x057e2159L /* 0.343293524 */, 16 },
 { 0x057f26a6L /* 0.343542717 */, 16 }, { 0x05802bfeL /* 0.343791954 */, 16 },

 { 0x05813162L /* 0.344041237 */, 16 }, { 0x058236d2L /* 0.344290564 */, 16 },
 { 0x05833c4fL /* 0.344539937 */, 16 }, { 0x058441d8L /* 0.344789356 */, 16 },
 { 0x0585476cL /* 0.345038819 */, 16 }, { 0x05864d0dL /* 0.345288327 */, 16 },
 { 0x058752baL /* 0.345537880 */, 16 }, { 0x05885873L /* 0.345787479 */, 16 },
 { 0x05895e39L /* 0.346037122 */, 16 }, { 0x058a640aL /* 0.346286811 */, 16 },
 { 0x058b69e7L /* 0.346536545 */, 16 }, { 0x058c6fd1L /* 0.346786323 */, 16 },
 { 0x058d75c6L /* 0.347036147 */, 16 }, { 0x058e7bc8L /* 0.347286015 */, 16 },
 { 0x058f81d5L /* 0.347535929 */, 16 }, { 0x059087efL /* 0.347785887 */, 16 },

 { 0x05918e15L /* 0.348035890 */, 16 }, { 0x05929447L /* 0.348285939 */, 16 },
 { 0x05939a84L /* 0.348536032 */, 16 }, { 0x0594a0ceL /* 0.348786170 */, 16 },
 { 0x0595a724L /* 0.349036353 */, 16 }, { 0x0596ad86L /* 0.349286580 */, 16 },
 { 0x0597b3f4L /* 0.349536853 */, 16 }, { 0x0598ba6eL /* 0.349787170 */, 16 },
 { 0x0599c0f4L /* 0.350037532 */, 16 }, { 0x059ac786L /* 0.350287939 */, 16 },
 { 0x059bce25L /* 0.350538391 */, 16 }, { 0x059cd4cfL /* 0.350788887 */, 16 },
 { 0x059ddb85L /* 0.351039428 */, 16 }, { 0x059ee247L /* 0.351290014 */, 16 },
 { 0x059fe915L /* 0.351540645 */, 16 }, { 0x05a0efefL /* 0.351791320 */, 16 },

 { 0x05a1f6d5L /* 0.352042040 */, 16 }, { 0x05a2fdc7L /* 0.352292804 */, 16 },
 { 0x05a404c5L /* 0.352543613 */, 16 }, { 0x05a50bcfL /* 0.352794467 */, 16 },
 { 0x05a612e5L /* 0.353045365 */, 16 }, { 0x05a71a07L /* 0.353296308 */, 16 },
 { 0x05a82135L /* 0.353547296 */, 16 }, { 0x05a9286fL /* 0.353798328 */, 16 },
 { 0x05aa2fb5L /* 0.354049405 */, 16 }, { 0x05ab3707L /* 0.354300526 */, 16 },
 { 0x05ac3e65L /* 0.354551691 */, 16 }, { 0x05ad45ceL /* 0.354802901 */, 16 },
 { 0x05ae4d44L /* 0.355054156 */, 16 }, { 0x05af54c6L /* 0.355305455 */, 16 },
 { 0x05b05c53L /* 0.355556799 */, 16 }, { 0x05b163edL /* 0.355808187 */, 16 },

 { 0x05b26b92L /* 0.356059619 */, 16 }, { 0x05b37343L /* 0.356311096 */, 16 },
 { 0x05b47b00L /* 0.356562617 */, 16 }, { 0x05b582c9L /* 0.356814182 */, 16 },
 { 0x05b68a9eL /* 0.357065792 */, 16 }, { 0x05b7927fL /* 0.357317446 */, 16 },
 { 0x05b89a6cL /* 0.357569145 */, 16 }, { 0x05b9a265L /* 0.357820887 */, 16 },
 { 0x05baaa69L /* 0.358072674 */, 16 }, { 0x05bbb27aL /* 0.358324506 */, 16 },
 { 0x05bcba96L /* 0.358576381 */, 16 }, { 0x05bdc2beL /* 0.358828301 */, 16 },
 { 0x05becaf2L /* 0.359080265 */, 16 }, { 0x05bfd332L /* 0.359332273 */, 16 },
 { 0x05c0db7eL /* 0.359584326 */, 16 }, { 0x05c1e3d6L /* 0.359836423 */, 16 },

 { 0x05c2ec39L /* 0.360088563 */, 16 }, { 0x05c3f4a9L /* 0.360340748 */, 16 },
 { 0x05c4fd24L /* 0.360592977 */, 16 }, { 0x05c605abL /* 0.360845251 */, 16 },
 { 0x05c70e3eL /* 0.361097568 */, 16 }, { 0x05c816ddL /* 0.361349929 */, 16 },
 { 0x05c91f87L /* 0.361602335 */, 16 }, { 0x05ca283eL /* 0.361854784 */, 16 },
 { 0x05cb3100L /* 0.362107278 */, 16 }, { 0x05cc39ceL /* 0.362359815 */, 16 },
 { 0x05cd42a8L /* 0.362612397 */, 16 }, { 0x05ce4b8dL /* 0.362865022 */, 16 },
 { 0x05cf547fL /* 0.363117692 */, 16 }, { 0x05d05d7cL /* 0.363370405 */, 16 },
 { 0x05d16685L /* 0.363623163 */, 16 }, { 0x05d26f9aL /* 0.363875964 */, 16 },

 { 0x05d378bbL /* 0.364128809 */, 16 }, { 0x05d481e7L /* 0.364381698 */, 16 },
 { 0x05d58b1fL /* 0.364634632 */, 16 }, { 0x05d69463L /* 0.364887608 */, 16 },
 { 0x05d79db3L /* 0.365140629 */, 16 }, { 0x05d8a70fL /* 0.365393694 */, 16 },
 { 0x05d9b076L /* 0.365646802 */, 16 }, { 0x05dab9e9L /* 0.365899955 */, 16 },
 { 0x05dbc368L /* 0.366153151 */, 16 }, { 0x05dcccf2L /* 0.366406390 */, 16 },
 { 0x05ddd689L /* 0.366659674 */, 16 }, { 0x05dee02bL /* 0.366913001 */, 16 },
 { 0x05dfe9d8L /* 0.367166372 */, 16 }, { 0x05e0f392L /* 0.367419787 */, 16 },
 { 0x05e1fd57L /* 0.367673246 */, 16 }, { 0x05e30728L /* 0.367926748 */, 16 },

 { 0x05e41105L /* 0.368180294 */, 16 }, { 0x05e51aedL /* 0.368433883 */, 16 },
 { 0x05e624e1L /* 0.368687517 */, 16 }, { 0x05e72ee1L /* 0.368941193 */, 16 },
 { 0x05e838edL /* 0.369194914 */, 16 }, { 0x05e94304L /* 0.369448678 */, 16 },
 { 0x05ea4d27L /* 0.369702485 */, 16 }, { 0x05eb5756L /* 0.369956336 */, 16 },
 { 0x05ec6190L /* 0.370210231 */, 16 }, { 0x05ed6bd6L /* 0.370464169 */, 16 },
 { 0x05ee7628L /* 0.370718151 */, 16 }, { 0x05ef8085L /* 0.370972177 */, 16 },
 { 0x05f08aeeL /* 0.371226245 */, 16 }, { 0x05f19563L /* 0.371480358 */, 16 },
 { 0x05f29fe3L /* 0.371734513 */, 16 }, { 0x05f3aa6fL /* 0.371988712 */, 16 },

 { 0x05f4b507L /* 0.372242955 */, 16 }, { 0x05f5bfabL /* 0.372497241 */, 16 },
 { 0x05f6ca5aL /* 0.372751570 */, 16 }, { 0x05f7d514L /* 0.373005943 */, 16 },
 { 0x05f8dfdbL /* 0.373260359 */, 16 }, { 0x05f9eaadL /* 0.373514819 */, 16 },
 { 0x05faf58aL /* 0.373769322 */, 16 }, { 0x05fc0073L /* 0.374023868 */, 16 },
 { 0x05fd0b68L /* 0.374278458 */, 16 }, { 0x05fe1669L /* 0.374533091 */, 16 },
 { 0x05ff2175L /* 0.374787767 */, 16 }, { 0x06002c8dL /* 0.375042486 */, 16 },
 { 0x060137b0L /* 0.375297249 */, 16 }, { 0x060242dfL /* 0.375552055 */, 16 },
 { 0x06034e19L /* 0.375806904 */, 16 }, { 0x0604595fL /* 0.376061796 */, 16 },

 { 0x060564b1L /* 0.376316732 */, 16 }, { 0x0606700fL /* 0.376571710 */, 16 },
 { 0x06077b77L /* 0.376826732 */, 16 }, { 0x060886ecL /* 0.377081797 */, 16 },
 { 0x0609926cL /* 0.377336905 */, 16 }, { 0x060a9df8L /* 0.377592057 */, 16 },
 { 0x060ba98fL /* 0.377847251 */, 16 }, { 0x060cb532L /* 0.378102489 */, 16 },
 { 0x060dc0e0L /* 0.378357769 */, 16 }, { 0x060ecc9aL /* 0.378613093 */, 16 },
 { 0x060fd860L /* 0.378868460 */, 16 }, { 0x0610e431L /* 0.379123870 */, 16 },
 { 0x0611f00dL /* 0.379379322 */, 16 }, { 0x0612fbf5L /* 0.379634818 */, 16 },
 { 0x061407e9L /* 0.379890357 */, 16 }, { 0x061513e8L /* 0.380145939 */, 16 },

 { 0x06161ff3L /* 0.380401563 */, 16 }, { 0x06172c09L /* 0.380657231 */, 16 },
 { 0x0618382bL /* 0.380912942 */, 16 }, { 0x06194458L /* 0.381168695 */, 16 },
 { 0x061a5091L /* 0.381424492 */, 16 }, { 0x061b5cd5L /* 0.381680331 */, 16 },
 { 0x061c6925L /* 0.381936213 */, 16 }, { 0x061d7581L /* 0.382192138 */, 16 },
 { 0x061e81e8L /* 0.382448106 */, 16 }, { 0x061f8e5aL /* 0.382704117 */, 16 },
 { 0x06209ad8L /* 0.382960171 */, 16 }, { 0x0621a761L /* 0.383216267 */, 16 },
 { 0x0622b3f6L /* 0.383472406 */, 16 }, { 0x0623c096L /* 0.383728588 */, 16 },
 { 0x0624cd42L /* 0.383984813 */, 16 }, { 0x0625d9f9L /* 0.384241080 */, 16 },

 { 0x0626e6bcL /* 0.384497391 */, 16 }, { 0x0627f38aL /* 0.384753744 */, 16 },
 { 0x06290064L /* 0.385010139 */, 16 }, { 0x062a0d49L /* 0.385266578 */, 16 },
 { 0x062b1a3aL /* 0.385523059 */, 16 }, { 0x062c2736L /* 0.385779582 */, 16 },
 { 0x062d343dL /* 0.386036149 */, 16 }, { 0x062e4150L /* 0.386292758 */, 16 },
 { 0x062f4e6fL /* 0.386549409 */, 16 }, { 0x06305b99L /* 0.386806104 */, 16 },
 { 0x063168ceL /* 0.387062840 */, 16 }, { 0x0632760fL /* 0.387319620 */, 16 },
 { 0x0633835bL /* 0.387576442 */, 16 }, { 0x063490b2L /* 0.387833306 */, 16 },
 { 0x06359e15L /* 0.388090213 */, 16 }, { 0x0636ab83L /* 0.388347163 */, 16 },

 { 0x0637b8fdL /* 0.388604155 */, 16 }, { 0x0638c682L /* 0.388861190 */, 16 },
 { 0x0639d413L /* 0.389118267 */, 16 }, { 0x063ae1afL /* 0.389375386 */, 16 },
 { 0x063bef56L /* 0.389632548 */, 16 }, { 0x063cfd09L /* 0.389889752 */, 16 },
 { 0x063e0ac7L /* 0.390146999 */, 16 }, { 0x063f1891L /* 0.390404289 */, 16 },
 { 0x06402666L /* 0.390661620 */, 16 }, { 0x06413446L /* 0.390918994 */, 16 },
 { 0x06424232L /* 0.391176411 */, 16 }, { 0x06435029L /* 0.391433869 */, 16 },
 { 0x06445e2bL /* 0.391691371 */, 16 }, { 0x06456c39L /* 0.391948914 */, 16 },
 { 0x06467a52L /* 0.392206500 */, 16 }, { 0x06478877L /* 0.392464128 */, 16 },

 { 0x064896a7L /* 0.392721798 */, 16 }, { 0x0649a4e2L /* 0.392979511 */, 16 },
 { 0x064ab328L /* 0.393237266 */, 16 }, { 0x064bc17aL /* 0.393495063 */, 16 },
 { 0x064ccfd8L /* 0.393752902 */, 16 }, { 0x064dde40L /* 0.394010784 */, 16 },
 { 0x064eecb4L /* 0.394268707 */, 16 }, { 0x064ffb33L /* 0.394526673 */, 16 },
 { 0x065109beL /* 0.394784681 */, 16 }, { 0x06521854L /* 0.395042732 */, 16 },
 { 0x065326f5L /* 0.395300824 */, 16 }, { 0x065435a1L /* 0.395558959 */, 16 },
 { 0x06554459L /* 0.395817135 */, 16 }, { 0x0656531cL /* 0.396075354 */, 16 },
 { 0x065761eaL /* 0.396333615 */, 16 }, { 0x065870c4L /* 0.396591918 */, 16 },

 { 0x06597fa9L /* 0.396850263 */, 16 }, { 0x065a8e99L /* 0.397108650 */, 16 },
 { 0x065b9d95L /* 0.397367079 */, 16 }, { 0x065cac9cL /* 0.397625550 */, 16 },
 { 0x065dbbaeL /* 0.397884063 */, 16 }, { 0x065ecacbL /* 0.398142619 */, 16 },
 { 0x065fd9f4L /* 0.398401216 */, 16 }, { 0x0660e928L /* 0.398659855 */, 16 },
 { 0x0661f867L /* 0.398918536 */, 16 }, { 0x066307b1L /* 0.399177259 */, 16 },
 { 0x06641707L /* 0.399436024 */, 16 }, { 0x06652668L /* 0.399694831 */, 16 },
 { 0x066635d4L /* 0.399953679 */, 16 }, { 0x0667454cL /* 0.400212570 */, 16 },
 { 0x066854ceL /* 0.400471503 */, 16 }, { 0x0669645cL /* 0.400730477 */, 16 },

 { 0x066a73f5L /* 0.400989493 */, 16 }, { 0x066b839aL /* 0.401248551 */, 16 },
 { 0x066c9349L /* 0.401507651 */, 16 }, { 0x066da304L /* 0.401766793 */, 16 },
 { 0x066eb2caL /* 0.402025976 */, 16 }, { 0x066fc29bL /* 0.402285202 */, 16 },
 { 0x0670d278L /* 0.402544469 */, 16 }, { 0x0671e25fL /* 0.402803777 */, 16 },
 { 0x0672f252L /* 0.403063128 */, 16 }, { 0x06740250L /* 0.403322520 */, 16 },
 { 0x0675125aL /* 0.403581954 */, 16 }, { 0x0676226eL /* 0.403841430 */, 16 },
 { 0x0677328eL /* 0.404100947 */, 16 }, { 0x067842b9L /* 0.404360506 */, 16 },
 { 0x067952efL /* 0.404620107 */, 16 }, { 0x067a6330L /* 0.404879749 */, 16 },

 { 0x067b737cL /* 0.405139433 */, 16 }, { 0x067c83d4L /* 0.405399159 */, 16 },
 { 0x067d9436L /* 0.405658926 */, 16 }, { 0x067ea4a4L /* 0.405918735 */, 16 },
 { 0x067fb51dL /* 0.406178585 */, 16 }, { 0x0680c5a2L /* 0.406438477 */, 16 },
 { 0x0681d631L /* 0.406698410 */, 16 }, { 0x0682e6cbL /* 0.406958385 */, 16 },
 { 0x0683f771L /* 0.407218402 */, 16 }, { 0x06850822L /* 0.407478460 */, 16 },
 { 0x068618deL /* 0.407738559 */, 16 }, { 0x068729a5L /* 0.407998700 */, 16 },
 { 0x06883a77L /* 0.408258883 */, 16 }, { 0x06894b55L /* 0.408519107 */, 16 },
 { 0x068a5c3dL /* 0.408779372 */, 16 }, { 0x068b6d31L /* 0.409039679 */, 16 },

 { 0x068c7e2fL /* 0.409300027 */, 16 }, { 0x068d8f39L /* 0.409560417 */, 16 },
 { 0x068ea04eL /* 0.409820848 */, 16 }, { 0x068fb16eL /* 0.410081321 */, 16 },
 { 0x0690c299L /* 0.410341834 */, 16 }, { 0x0691d3cfL /* 0.410602390 */, 16 },
 { 0x0692e511L /* 0.410862986 */, 16 }, { 0x0693f65dL /* 0.411123624 */, 16 },
 { 0x069507b5L /* 0.411384303 */, 16 }, { 0x06961917L /* 0.411645024 */, 16 },
 { 0x06972a85L /* 0.411905785 */, 16 }, { 0x06983bfeL /* 0.412166588 */, 16 },
 { 0x06994d82L /* 0.412427433 */, 16 }, { 0x069a5f11L /* 0.412688318 */, 16 },
 { 0x069b70abL /* 0.412949245 */, 16 }, { 0x069c8250L /* 0.413210213 */, 16 },

 { 0x069d9400L /* 0.413471222 */, 16 }, { 0x069ea5bbL /* 0.413732273 */, 16 },
 { 0x069fb781L /* 0.413993364 */, 16 }, { 0x06a0c953L /* 0.414254497 */, 16 },
 { 0x06a1db2fL /* 0.414515671 */, 16 }, { 0x06a2ed16L /* 0.414776886 */, 16 },
 { 0x06a3ff09L /* 0.415038142 */, 16 }, { 0x06a51106L /* 0.415299440 */, 16 },
 { 0x06a6230fL /* 0.415560778 */, 16 }, { 0x06a73522L /* 0.415822157 */, 16 },
 { 0x06a84741L /* 0.416083578 */, 16 }, { 0x06a9596aL /* 0.416345040 */, 16 },
 { 0x06aa6b9fL /* 0.416606542 */, 16 }, { 0x06ab7ddfL /* 0.416868086 */, 16 },
 { 0x06ac9029L /* 0.417129671 */, 16 }, { 0x06ada27fL /* 0.417391297 */, 16 },

 { 0x06aeb4e0L /* 0.417652964 */, 16 }, { 0x06afc74bL /* 0.417914672 */, 16 },
 { 0x06b0d9c2L /* 0.418176420 */, 16 }, { 0x06b1ec43L /* 0.418438210 */, 16 },
 { 0x06b2fed0L /* 0.418700041 */, 16 }, { 0x06b41168L /* 0.418961912 */, 16 },
 { 0x06b5240aL /* 0.419223825 */, 16 }, { 0x06b636b8L /* 0.419485778 */, 16 },
 { 0x06b74971L /* 0.419747773 */, 16 }, { 0x06b85c34L /* 0.420009808 */, 16 },
 { 0x06b96f03L /* 0.420271884 */, 16 }, { 0x06ba81dcL /* 0.420534001 */, 16 },
 { 0x06bb94c1L /* 0.420796159 */, 16 }, { 0x06bca7b0L /* 0.421058358 */, 16 },
 { 0x06bdbaaaL /* 0.421320597 */, 16 }, { 0x06becdb0L /* 0.421582878 */, 16 },

 { 0x06bfe0c0L /* 0.421845199 */, 16 }, { 0x06c0f3dbL /* 0.422107561 */, 16 },
 { 0x06c20702L /* 0.422369964 */, 16 }, { 0x06c31a33L /* 0.422632407 */, 16 },
 { 0x06c42d6fL /* 0.422894891 */, 16 }, { 0x06c540b6L /* 0.423157416 */, 16 },
 { 0x06c65408L /* 0.423419982 */, 16 }, { 0x06c76765L /* 0.423682588 */, 16 },
 { 0x06c87accL /* 0.423945235 */, 16 }, { 0x06c98e3fL /* 0.424207923 */, 16 },
 { 0x06caa1bdL /* 0.424470652 */, 16 }, { 0x06cbb545L /* 0.424733421 */, 16 },
 { 0x06ccc8d9L /* 0.424996230 */, 16 }, { 0x06cddc77L /* 0.425259081 */, 16 },
 { 0x06cef020L /* 0.425521972 */, 16 }, { 0x06d003d4L /* 0.425784903 */, 16 },

 { 0x06d11794L /* 0.426047876 */, 16 }, { 0x06d22b5eL /* 0.426310889 */, 16 },
 { 0x06d33f32L /* 0.426573942 */, 16 }, { 0x06d45312L /* 0.426837036 */, 16 },
 { 0x06d566fdL /* 0.427100170 */, 16 }, { 0x06d67af2L /* 0.427363345 */, 16 },
 { 0x06d78ef3L /* 0.427626561 */, 16 }, { 0x06d8a2feL /* 0.427889817 */, 16 },
 { 0x06d9b714L /* 0.428153114 */, 16 }, { 0x06dacb35L /* 0.428416451 */, 16 },
 { 0x06dbdf61L /* 0.428679828 */, 16 }, { 0x06dcf398L /* 0.428943246 */, 16 },
 { 0x06de07d9L /* 0.429206704 */, 16 }, { 0x06df1c26L /* 0.429470203 */, 16 },
 { 0x06e0307dL /* 0.429733743 */, 16 }, { 0x06e144dfL /* 0.429997322 */, 16 },

 { 0x06e2594cL /* 0.430260942 */, 16 }, { 0x06e36dc4L /* 0.430524603 */, 16 },
 { 0x06e48246L /* 0.430788304 */, 16 }, { 0x06e596d4L /* 0.431052045 */, 16 },
 { 0x06e6ab6cL /* 0.431315826 */, 16 }, { 0x06e7c00fL /* 0.431579648 */, 16 },
 { 0x06e8d4bdL /* 0.431843511 */, 16 }, { 0x06e9e976L /* 0.432107413 */, 16 },
 { 0x06eafe3aL /* 0.432371356 */, 16 }, { 0x06ec1308L /* 0.432635339 */, 16 },
 { 0x06ed27e2L /* 0.432899362 */, 16 }, { 0x06ee3cc6L /* 0.433163426 */, 16 },
 { 0x06ef51b4L /* 0.433427530 */, 16 }, { 0x06f066aeL /* 0.433691674 */, 16 },
 { 0x06f17bb3L /* 0.433955859 */, 16 }, { 0x06f290c2L /* 0.434220083 */, 16 },

 { 0x06f3a5dcL /* 0.434484348 */, 16 }, { 0x06f4bb01L /* 0.434748653 */, 16 },
 { 0x06f5d030L /* 0.435012998 */, 16 }, { 0x06f6e56bL /* 0.435277383 */, 16 },
 { 0x06f7fab0L /* 0.435541809 */, 16 }, { 0x06f91000L /* 0.435806274 */, 16 },
 { 0x06fa255aL /* 0.436070780 */, 16 }, { 0x06fb3ac0L /* 0.436335326 */, 16 },
 { 0x06fc5030L /* 0.436599912 */, 16 }, { 0x06fd65abL /* 0.436864538 */, 16 },
 { 0x06fe7b31L /* 0.437129204 */, 16 }, { 0x06ff90c2L /* 0.437393910 */, 16 },
 { 0x0700a65dL /* 0.437658657 */, 16 }, { 0x0701bc03L /* 0.437923443 */, 16 },
 { 0x0702d1b4L /* 0.438188269 */, 16 }, { 0x0703e76fL /* 0.438453136 */, 16 },

 { 0x0704fd35L /* 0.438718042 */, 16 }, { 0x07061306L /* 0.438982988 */, 16 },
 { 0x070728e2L /* 0.439247975 */, 16 }, { 0x07083ec9L /* 0.439513001 */, 16 },
 { 0x070954baL /* 0.439778067 */, 16 }, { 0x070a6ab6L /* 0.440043173 */, 16 },
 { 0x070b80bcL /* 0.440308320 */, 16 }, { 0x070c96ceL /* 0.440573506 */, 16 },
 { 0x070daceaL /* 0.440838732 */, 16 }, { 0x070ec310L /* 0.441103997 */, 16 },
 { 0x070fd942L /* 0.441369303 */, 16 }, { 0x0710ef7eL /* 0.441634649 */, 16 },
 { 0x071205c5L /* 0.441900034 */, 16 }, { 0x07131c17L /* 0.442165460 */, 16 },
 { 0x07143273L /* 0.442430925 */, 16 }, { 0x071548daL /* 0.442696430 */, 16 },

 { 0x07165f4bL /* 0.442961975 */, 16 }, { 0x071775c8L /* 0.443227559 */, 16 },
 { 0x07188c4fL /* 0.443493184 */, 16 }, { 0x0719a2e0L /* 0.443758848 */, 16 },
 { 0x071ab97dL /* 0.444024552 */, 16 }, { 0x071bd024L /* 0.444290296 */, 16 },
 { 0x071ce6d6L /* 0.444556079 */, 16 }, { 0x071dfd92L /* 0.444821902 */, 16 },
 { 0x071f1459L /* 0.445087765 */, 16 }, { 0x07202b2bL /* 0.445353668 */, 16 },
 { 0x07214207L /* 0.445619610 */, 16 }, { 0x072258eeL /* 0.445885592 */, 16 },
 { 0x07236fe0L /* 0.446151614 */, 16 }, { 0x072486dcL /* 0.446417675 */, 16 },
 { 0x07259de3L /* 0.446683776 */, 16 }, { 0x0726b4f4L /* 0.446949917 */, 16 },

 { 0x0727cc11L /* 0.447216097 */, 16 }, { 0x0728e338L /* 0.447482317 */, 16 },
 { 0x0729fa69L /* 0.447748576 */, 16 }, { 0x072b11a5L /* 0.448014875 */, 16 },
 { 0x072c28ecL /* 0.448281214 */, 16 }, { 0x072d403dL /* 0.448547592 */, 16 },
 { 0x072e5799L /* 0.448814010 */, 16 }, { 0x072f6f00L /* 0.449080467 */, 16 },
 { 0x07308671L /* 0.449346964 */, 16 }, { 0x07319dedL /* 0.449613501 */, 16 },
 { 0x0732b573L /* 0.449880076 */, 16 }, { 0x0733cd04L /* 0.450146692 */, 16 },
 { 0x0734e4a0L /* 0.450413347 */, 16 }, { 0x0735fc46L /* 0.450680041 */, 16 },
 { 0x073713f7L /* 0.450946775 */, 16 }, { 0x07382bb2L /* 0.451213548 */, 16 },

 { 0x07394378L /* 0.451480360 */, 16 }, { 0x073a5b49L /* 0.451747213 */, 16 },
 { 0x073b7324L /* 0.452014104 */, 16 }, { 0x073c8b0aL /* 0.452281035 */, 16 },
 { 0x073da2faL /* 0.452548005 */, 16 }, { 0x073ebaf5L /* 0.452815015 */, 16 },
 { 0x073fd2faL /* 0.453082064 */, 16 }, { 0x0740eb0aL /* 0.453349152 */, 16 },
 { 0x07420325L /* 0.453616280 */, 16 }, { 0x07431b4aL /* 0.453883447 */, 16 },
 { 0x0744337aL /* 0.454150653 */, 16 }, { 0x07454bb4L /* 0.454417899 */, 16 },
 { 0x074663f8L /* 0.454685184 */, 16 }, { 0x07477c48L /* 0.454952508 */, 16 },
 { 0x074894a2L /* 0.455219872 */, 16 }, { 0x0749ad06L /* 0.455487275 */, 16 },

 { 0x074ac575L /* 0.455754717 */, 16 }, { 0x074bddeeL /* 0.456022198 */, 16 },
 { 0x074cf672L /* 0.456289719 */, 16 }, { 0x074e0f01L /* 0.456557278 */, 16 },
 { 0x074f279aL /* 0.456824877 */, 16 }, { 0x0750403eL /* 0.457092516 */, 16 },
 { 0x075158ecL /* 0.457360193 */, 16 }, { 0x075271a4L /* 0.457627909 */, 16 },
 { 0x07538a67L /* 0.457895665 */, 16 }, { 0x0754a335L /* 0.458163460 */, 16 },
 { 0x0755bc0dL /* 0.458431294 */, 16 }, { 0x0756d4f0L /* 0.458699167 */, 16 },
 { 0x0757edddL /* 0.458967079 */, 16 }, { 0x075906d5L /* 0.459235030 */, 16 },
 { 0x075a1fd7L /* 0.459503021 */, 16 }, { 0x075b38e3L /* 0.459771050 */, 16 },

 { 0x075c51faL /* 0.460039119 */, 16 }, { 0x075d6b1cL /* 0.460307226 */, 16 },
 { 0x075e8448L /* 0.460575373 */, 16 }, { 0x075f9d7fL /* 0.460843559 */, 16 },
 { 0x0760b6c0L /* 0.461111783 */, 16 }, { 0x0761d00bL /* 0.461380047 */, 16 },
 { 0x0762e961L /* 0.461648350 */, 16 }, { 0x076402c1L /* 0.461916691 */, 16 },
 { 0x07651c2cL /* 0.462185072 */, 16 }, { 0x076635a2L /* 0.462453492 */, 16 },
 { 0x07674f22L /* 0.462721950 */, 16 }, { 0x076868acL /* 0.462990448 */, 16 },
 { 0x07698240L /* 0.463258984 */, 16 }, { 0x076a9be0L /* 0.463527560 */, 16 },
 { 0x076bb589L /* 0.463796174 */, 16 }, { 0x076ccf3dL /* 0.464064827 */, 16 },

 { 0x076de8fcL /* 0.464333519 */, 16 }, { 0x076f02c5L /* 0.464602250 */, 16 },
 { 0x07701c98L /* 0.464871020 */, 16 }, { 0x07713676L /* 0.465139829 */, 16 },
 { 0x0772505eL /* 0.465408676 */, 16 }, { 0x07736a51L /* 0.465677563 */, 16 },
 { 0x0774844eL /* 0.465946488 */, 16 }, { 0x07759e55L /* 0.466215452 */, 16 },
 { 0x0776b867L /* 0.466484455 */, 16 }, { 0x0777d283L /* 0.466753496 */, 16 },
 { 0x0778ecaaL /* 0.467022577 */, 16 }, { 0x077a06dbL /* 0.467291696 */, 16 },
 { 0x077b2117L /* 0.467560854 */, 16 }, { 0x077c3b5dL /* 0.467830050 */, 16 },
 { 0x077d55adL /* 0.468099285 */, 16 }, { 0x077e7008L /* 0.468368560 */, 16 },

 { 0x077f8a6dL /* 0.468637872 */, 16 }, { 0x0780a4dcL /* 0.468907224 */, 16 },
 { 0x0781bf56L /* 0.469176614 */, 16 }, { 0x0782d9daL /* 0.469446043 */, 16 },
 { 0x0783f469L /* 0.469715510 */, 16 }, { 0x07850f02L /* 0.469985016 */, 16 },
 { 0x078629a5L /* 0.470254561 */, 16 }, { 0x07874453L /* 0.470524145 */, 16 },
 { 0x07885f0bL /* 0.470793767 */, 16 }, { 0x078979ceL /* 0.471063427 */, 16 },
 { 0x078a949aL /* 0.471333126 */, 16 }, { 0x078baf72L /* 0.471602864 */, 16 },
 { 0x078cca53L /* 0.471872641 */, 16 }, { 0x078de53fL /* 0.472142456 */, 16 },
 { 0x078f0035L /* 0.472412309 */, 16 }, { 0x07901b36L /* 0.472682201 */, 16 },

 { 0x07913641L /* 0.472952132 */, 16 }, { 0x07925156L /* 0.473222101 */, 16 },
 { 0x07936c76L /* 0.473492108 */, 16 }, { 0x079487a0L /* 0.473762155 */, 16 },
 { 0x0795a2d4L /* 0.474032239 */, 16 }, { 0x0796be13L /* 0.474302362 */, 16 },
 { 0x0797d95cL /* 0.474572524 */, 16 }, { 0x0798f4afL /* 0.474842724 */, 16 },
 { 0x079a100cL /* 0.475112962 */, 16 }, { 0x079b2b74L /* 0.475383239 */, 16 },
 { 0x079c46e7L /* 0.475653554 */, 16 }, { 0x079d6263L /* 0.475923908 */, 16 },
 { 0x079e7deaL /* 0.476194300 */, 16 }, { 0x079f997bL /* 0.476464731 */, 16 },
 { 0x07a0b516L /* 0.476735200 */, 16 }, { 0x07a1d0bcL /* 0.477005707 */, 16 },

 { 0x07a2ec6cL /* 0.477276252 */, 16 }, { 0x07a40827L /* 0.477546836 */, 16 },
 { 0x07a523ebL /* 0.477817459 */, 16 }, { 0x07a63fbaL /* 0.478088119 */, 16 },
 { 0x07a75b93L /* 0.478358818 */, 16 }, { 0x07a87777L /* 0.478629555 */, 16 },
 { 0x07a99364L /* 0.478900331 */, 16 }, { 0x07aaaf5cL /* 0.479171145 */, 16 },
 { 0x07abcb5fL /* 0.479441997 */, 16 }, { 0x07ace76bL /* 0.479712887 */, 16 },
 { 0x07ae0382L /* 0.479983816 */, 16 }, { 0x07af1fa3L /* 0.480254782 */, 16 },
 { 0x07b03bcfL /* 0.480525787 */, 16 }, { 0x07b15804L /* 0.480796831 */, 16 },
 { 0x07b27444L /* 0.481067912 */, 16 }, { 0x07b3908eL /* 0.481339032 */, 16 },

 { 0x07b4ace3L /* 0.481610189 */, 16 }, { 0x07b5c941L /* 0.481881385 */, 16 },
 { 0x07b6e5aaL /* 0.482152620 */, 16 }, { 0x07b8021dL /* 0.482423892 */, 16 },
 { 0x07b91e9bL /* 0.482695202 */, 16 }, { 0x07ba3b22L /* 0.482966551 */, 16 },
 { 0x07bb57b4L /* 0.483237938 */, 16 }, { 0x07bc7450L /* 0.483509362 */, 16 },
 { 0x07bd90f6L /* 0.483780825 */, 16 }, { 0x07beada7L /* 0.484052326 */, 16 },
 { 0x07bfca61L /* 0.484323865 */, 16 }, { 0x07c0e726L /* 0.484595443 */, 16 },
 { 0x07c203f5L /* 0.484867058 */, 16 }, { 0x07c320cfL /* 0.485138711 */, 16 },
 { 0x07c43db2L /* 0.485410402 */, 16 }, { 0x07c55aa0L /* 0.485682131 */, 16 },

 { 0x07c67798L /* 0.485953899 */, 16 }, { 0x07c7949aL /* 0.486225704 */, 16 },
 { 0x07c8b1a7L /* 0.486497547 */, 16 }, { 0x07c9cebdL /* 0.486769429 */, 16 },
 { 0x07caebdeL /* 0.487041348 */, 16 }, { 0x07cc0909L /* 0.487313305 */, 16 },
 { 0x07cd263eL /* 0.487585300 */, 16 }, { 0x07ce437dL /* 0.487857333 */, 16 },
 { 0x07cf60c7L /* 0.488129404 */, 16 }, { 0x07d07e1bL /* 0.488401513 */, 16 },
 { 0x07d19b79L /* 0.488673660 */, 16 }, { 0x07d2b8e1L /* 0.488945845 */, 16 },
 { 0x07d3d653L /* 0.489218067 */, 16 }, { 0x07d4f3cfL /* 0.489490328 */, 16 },
 { 0x07d61156L /* 0.489762626 */, 16 }, { 0x07d72ee6L /* 0.490034962 */, 16 },

 { 0x07d84c81L /* 0.490307336 */, 16 }, { 0x07d96a26L /* 0.490579748 */, 16 },
 { 0x07da87d5L /* 0.490852198 */, 16 }, { 0x07dba58fL /* 0.491124686 */, 16 },
 { 0x07dcc352L /* 0.491397211 */, 16 }, { 0x07dde120L /* 0.491669774 */, 16 },
 { 0x07defef7L /* 0.491942375 */, 16 }, { 0x07e01cd9L /* 0.492215014 */, 16 },
 { 0x07e13ac5L /* 0.492487690 */, 16 }, { 0x07e258bcL /* 0.492760404 */, 16 },
 { 0x07e376bcL /* 0.493033156 */, 16 }, { 0x07e494c6L /* 0.493305946 */, 16 },
 { 0x07e5b2dbL /* 0.493578773 */, 16 }, { 0x07e6d0f9L /* 0.493851638 */, 16 },
 { 0x07e7ef22L /* 0.494124541 */, 16 }, { 0x07e90d55L /* 0.494397481 */, 16 },

 { 0x07ea2b92L /* 0.494670459 */, 16 }, { 0x07eb49d9L /* 0.494943475 */, 16 },
 { 0x07ec682aL /* 0.495216529 */, 16 }, { 0x07ed8686L /* 0.495489620 */, 16 },
 { 0x07eea4ebL /* 0.495762748 */, 16 }, { 0x07efc35bL /* 0.496035915 */, 16 },
 { 0x07f0e1d4L /* 0.496309119 */, 16 }, { 0x07f20058L /* 0.496582360 */, 16 },
 { 0x07f31ee6L /* 0.496855639 */, 16 }, { 0x07f43d7eL /* 0.497128956 */, 16 },
 { 0x07f55c20L /* 0.497402310 */, 16 }, { 0x07f67accL /* 0.497675702 */, 16 },
 { 0x07f79982L /* 0.497949132 */, 16 }, { 0x07f8b842L /* 0.498222598 */, 16 },
 { 0x07f9d70cL /* 0.498496103 */, 16 }, { 0x07faf5e1L /* 0.498769645 */, 16 },

 { 0x07fc14bfL /* 0.499043224 */, 16 }, { 0x07fd33a8L /* 0.499316841 */, 16 },
 { 0x07fe529aL /* 0.499590496 */, 16 }, { 0x07ff7197L /* 0.499864188 */, 16 },
 { 0x0400484fL /* 0.250068959 */, 17 }, { 0x0400d7d7L /* 0.250205842 */, 17 },
 { 0x04016764L /* 0.250342744 */, 17 }, { 0x0401f6f7L /* 0.250479665 */, 17 },
 { 0x0402868eL /* 0.250616605 */, 17 }, { 0x0403162bL /* 0.250753563 */, 17 },
 { 0x0403a5ccL /* 0.250890540 */, 17 }, { 0x04043573L /* 0.251027536 */, 17 },
 { 0x0404c51eL /* 0.251164550 */, 17 }, { 0x040554cfL /* 0.251301583 */, 17 },
 { 0x0405e484L /* 0.251438635 */, 17 }, { 0x0406743fL /* 0.251575706 */, 17 },

 { 0x040703ffL /* 0.251712795 */, 17 }, { 0x040793c3L /* 0.251849903 */, 17 },
 { 0x0408238dL /* 0.251987029 */, 17 }, { 0x0408b35bL /* 0.252124174 */, 17 },
 { 0x0409432fL /* 0.252261338 */, 17 }, { 0x0409d308L /* 0.252398520 */, 17 },
 { 0x040a62e5L /* 0.252535721 */, 17 }, { 0x040af2c8L /* 0.252672941 */, 17 },
 { 0x040b82b0L /* 0.252810180 */, 17 }, { 0x040c129cL /* 0.252947436 */, 17 },
 { 0x040ca28eL /* 0.253084712 */, 17 }, { 0x040d3284L /* 0.253222006 */, 17 },
 { 0x040dc280L /* 0.253359319 */, 17 }, { 0x040e5281L /* 0.253496651 */, 17 },
 { 0x040ee286L /* 0.253634001 */, 17 }, { 0x040f7291L /* 0.253771369 */, 17 },

 { 0x041002a1L /* 0.253908756 */, 17 }, { 0x041092b5L /* 0.254046162 */, 17 },
 { 0x041122cfL /* 0.254183587 */, 17 }, { 0x0411b2edL /* 0.254321030 */, 17 },
 { 0x04124311L /* 0.254458491 */, 17 }, { 0x0412d339L /* 0.254595971 */, 17 },
 { 0x04136367L /* 0.254733470 */, 17 }, { 0x0413f399L /* 0.254870987 */, 17 },
 { 0x041483d1L /* 0.255008523 */, 17 }, { 0x0415140dL /* 0.255146077 */, 17 },
 { 0x0415a44fL /* 0.255283650 */, 17 }, { 0x04163495L /* 0.255421241 */, 17 },
 { 0x0416c4e1L /* 0.255558851 */, 17 }, { 0x04175531L /* 0.255696480 */, 17 },
 { 0x0417e586L /* 0.255834127 */, 17 }, { 0x041875e1L /* 0.255971792 */, 17 },

 { 0x04190640L /* 0.256109476 */, 17 }, { 0x041996a4L /* 0.256247179 */, 17 },
 { 0x041a270dL /* 0.256384900 */, 17 }, { 0x041ab77bL /* 0.256522639 */, 17 },
 { 0x041b47efL /* 0.256660397 */, 17 }, { 0x041bd867L /* 0.256798174 */, 17 },
 { 0x041c68e4L /* 0.256935969 */, 17 }, { 0x041cf966L /* 0.257073782 */, 17 },
 { 0x041d89edL /* 0.257211614 */, 17 }, { 0x041e1a79L /* 0.257349465 */, 17 },
 { 0x041eab0aL /* 0.257487334 */, 17 }, { 0x041f3b9fL /* 0.257625221 */, 17 },
 { 0x041fcc3aL /* 0.257763127 */, 17 }, { 0x04205cdaL /* 0.257901051 */, 17 },
 { 0x0420ed7fL /* 0.258038994 */, 17 }, { 0x04217e28L /* 0.258176955 */, 17 },

 { 0x04220ed7L /* 0.258314934 */, 17 }, { 0x04229f8aL /* 0.258452932 */, 17 },
 { 0x04233043L /* 0.258590948 */, 17 }, { 0x0423c100L /* 0.258728983 */, 17 },
 { 0x042451c3L /* 0.258867036 */, 17 }, { 0x0424e28aL /* 0.259005108 */, 17 },
 { 0x04257356L /* 0.259143198 */, 17 }, { 0x04260428L /* 0.259281307 */, 17 },
 { 0x042694feL /* 0.259419433 */, 17 }, { 0x042725d9L /* 0.259557579 */, 17 },
 { 0x0427b6b9L /* 0.259695742 */, 17 }, { 0x0428479eL /* 0.259833924 */, 17 },
 { 0x0428d888L /* 0.259972124 */, 17 }, { 0x04296976L /* 0.260110343 */, 17 },
 { 0x0429fa6aL /* 0.260248580 */, 17 }, { 0x042a8b63L /* 0.260386836 */, 17 },

 { 0x042b1c60L /* 0.260525110 */, 17 }, { 0x042bad63L /* 0.260663402 */, 17 },
 { 0x042c3e6aL /* 0.260801712 */, 17 }, { 0x042ccf77L /* 0.260940041 */, 17 },
 { 0x042d6088L /* 0.261078388 */, 17 }, { 0x042df19eL /* 0.261216754 */, 17 },
 { 0x042e82b9L /* 0.261355137 */, 17 }, { 0x042f13d9L /* 0.261493540 */, 17 },
 { 0x042fa4feL /* 0.261631960 */, 17 }, { 0x04303628L /* 0.261770399 */, 17 },
 { 0x0430c757L /* 0.261908856 */, 17 }, { 0x0431588bL /* 0.262047331 */, 17 },
 { 0x0431e9c3L /* 0.262185825 */, 17 }, { 0x04327b01L /* 0.262324337 */, 17 },
 { 0x04330c43L /* 0.262462867 */, 17 }, { 0x04339d8aL /* 0.262601416 */, 17 },

 { 0x04342ed7L /* 0.262739982 */, 17 }, { 0x0434c028L /* 0.262878568 */, 17 },
 { 0x0435517eL /* 0.263017171 */, 17 }, { 0x0435e2d9L /* 0.263155792 */, 17 },
 { 0x04367439L /* 0.263294432 */, 17 }, { 0x0437059eL /* 0.263433090 */, 17 },
 { 0x04379707L /* 0.263571767 */, 17 }, { 0x04382876L /* 0.263710461 */, 17 },
 { 0x0438b9e9L /* 0.263849174 */, 17 }, { 0x04394b61L /* 0.263987905 */, 17 },
 { 0x0439dcdfL /* 0.264126655 */, 17 }, { 0x043a6e61L /* 0.264265422 */, 17 },
 { 0x043affe8L /* 0.264404208 */, 17 }, { 0x043b9174L /* 0.264543012 */, 17 },
 { 0x043c2305L /* 0.264681834 */, 17 }, { 0x043cb49aL /* 0.264820674 */, 17 },

 { 0x043d4635L /* 0.264959533 */, 17 }, { 0x043dd7d4L /* 0.265098410 */, 17 },
 { 0x043e6979L /* 0.265237305 */, 17 }, { 0x043efb22L /* 0.265376218 */, 17 },
 { 0x043f8cd0L /* 0.265515149 */, 17 }, { 0x04401e83L /* 0.265654099 */, 17 },
 { 0x0440b03bL /* 0.265793066 */, 17 }, { 0x044141f7L /* 0.265932052 */, 17 },
 { 0x0441d3b9L /* 0.266071056 */, 17 }, { 0x04426580L /* 0.266210078 */, 17 },
 { 0x0442f74bL /* 0.266349119 */, 17 }, { 0x0443891bL /* 0.266488177 */, 17 },
 { 0x04441af0L /* 0.266627254 */, 17 }, { 0x0444accaL /* 0.266766349 */, 17 },
 { 0x04453ea9L /* 0.266905462 */, 17 }, { 0x0445d08dL /* 0.267044593 */, 17 },

 { 0x04466275L /* 0.267183742 */, 17 }, { 0x0446f463L /* 0.267322909 */, 17 },
 { 0x04478655L /* 0.267462094 */, 17 }, { 0x0448184cL /* 0.267601298 */, 17 },
 { 0x0448aa48L /* 0.267740519 */, 17 }, { 0x04493c49L /* 0.267879759 */, 17 },
 { 0x0449ce4fL /* 0.268019017 */, 17 }, { 0x044a6059L /* 0.268158293 */, 17 },
 { 0x044af269L /* 0.268297587 */, 17 }, { 0x044b847dL /* 0.268436899 */, 17 },
 { 0x044c1696L /* 0.268576229 */, 17 }, { 0x044ca8b4L /* 0.268715577 */, 17 },
 { 0x044d3ad7L /* 0.268854943 */, 17 }, { 0x044dccffL /* 0.268994328 */, 17 },
 { 0x044e5f2bL /* 0.269133730 */, 17 }, { 0x044ef15dL /* 0.269273150 */, 17 },

 { 0x044f8393L /* 0.269412589 */, 17 }, { 0x045015ceL /* 0.269552045 */, 17 },
 { 0x0450a80eL /* 0.269691520 */, 17 }, { 0x04513a53L /* 0.269831013 */, 17 },
 { 0x0451cc9cL /* 0.269970523 */, 17 }, { 0x04525eebL /* 0.270110052 */, 17 },
 { 0x0452f13eL /* 0.270249599 */, 17 }, { 0x04538396L /* 0.270389163 */, 17 },
 { 0x045415f3L /* 0.270528746 */, 17 }, { 0x0454a855L /* 0.270668347 */, 17 },
 { 0x04553abbL /* 0.270807965 */, 17 }, { 0x0455cd27L /* 0.270947602 */, 17 },
 { 0x04565f97L /* 0.271087257 */, 17 }, { 0x0456f20cL /* 0.271226930 */, 17 },
 { 0x04578486L /* 0.271366620 */, 17 }, { 0x04581705L /* 0.271506329 */, 17 },

 { 0x0458a989L /* 0.271646056 */, 17 }, { 0x04593c11L /* 0.271785800 */, 17 },
 { 0x0459ce9eL /* 0.271925563 */, 17 }, { 0x045a6130L /* 0.272065343 */, 17 },
 { 0x045af3c7L /* 0.272205142 */, 17 }, { 0x045b8663L /* 0.272344958 */, 17 },
 { 0x045c1903L /* 0.272484793 */, 17 }, { 0x045caba9L /* 0.272624645 */, 17 },
 { 0x045d3e53L /* 0.272764515 */, 17 }, { 0x045dd102L /* 0.272904403 */, 17 },
 { 0x045e63b6L /* 0.273044310 */, 17 }, { 0x045ef66eL /* 0.273184234 */, 17 },
 { 0x045f892bL /* 0.273324176 */, 17 }, { 0x04601beeL /* 0.273464136 */, 17 },
 { 0x0460aeb5L /* 0.273604113 */, 17 }, { 0x04614180L /* 0.273744109 */, 17 },

 { 0x0461d451L /* 0.273884123 */, 17 }, { 0x04626727L /* 0.274024154 */, 17 },
 { 0x0462fa01L /* 0.274164204 */, 17 }, { 0x04638ce0L /* 0.274304271 */, 17 },
 { 0x04641fc4L /* 0.274444356 */, 17 }, { 0x0464b2acL /* 0.274584459 */, 17 },
 { 0x0465459aL /* 0.274724580 */, 17 }, { 0x0465d88cL /* 0.274864719 */, 17 },
 { 0x04666b83L /* 0.275004875 */, 17 }, { 0x0466fe7fL /* 0.275145050 */, 17 },
 { 0x0467917fL /* 0.275285242 */, 17 }, { 0x04682485L /* 0.275425452 */, 17 },
 { 0x0468b78fL /* 0.275565681 */, 17 }, { 0x04694a9eL /* 0.275705926 */, 17 },
 { 0x0469ddb2L /* 0.275846190 */, 17 }, { 0x046a70caL /* 0.275986472 */, 17 },

 { 0x046b03e7L /* 0.276126771 */, 17 }, { 0x046b970aL /* 0.276267088 */, 17 },
 { 0x046c2a31L /* 0.276407423 */, 17 }, { 0x046cbd5cL /* 0.276547776 */, 17 },
 { 0x046d508dL /* 0.276688147 */, 17 }, { 0x046de3c2L /* 0.276828535 */, 17 },
 { 0x046e76fcL /* 0.276968942 */, 17 }, { 0x046f0a3bL /* 0.277109366 */, 17 },
 { 0x046f9d7eL /* 0.277249808 */, 17 }, { 0x047030c7L /* 0.277390267 */, 17 },
 { 0x0470c414L /* 0.277530745 */, 17 }, { 0x04715766L /* 0.277671240 */, 17 },
 { 0x0471eabcL /* 0.277811753 */, 17 }, { 0x04727e18L /* 0.277952284 */, 17 },
 { 0x04731178L /* 0.278092832 */, 17 }, { 0x0473a4ddL /* 0.278233399 */, 17 },

 { 0x04743847L /* 0.278373983 */, 17 }, { 0x0474cbb5L /* 0.278514584 */, 17 },
 { 0x04755f29L /* 0.278655204 */, 17 }, { 0x0475f2a1L /* 0.278795841 */, 17 },
 { 0x0476861dL /* 0.278936496 */, 17 }, { 0x0477199fL /* 0.279077169 */, 17 },
 { 0x0477ad25L /* 0.279217860 */, 17 }, { 0x047840b0L /* 0.279358568 */, 17 },
 { 0x0478d440L /* 0.279499294 */, 17 }, { 0x047967d5L /* 0.279640037 */, 17 },
 { 0x0479fb6eL /* 0.279780799 */, 17 }, { 0x047a8f0cL /* 0.279921578 */, 17 },
 { 0x047b22afL /* 0.280062375 */, 17 }, { 0x047bb657L /* 0.280203189 */, 17 },
 { 0x047c4a03L /* 0.280344021 */, 17 }, { 0x047cddb4L /* 0.280484871 */, 17 },

 { 0x047d716aL /* 0.280625739 */, 17 }, { 0x047e0524L /* 0.280766624 */, 17 },
 { 0x047e98e4L /* 0.280907527 */, 17 }, { 0x047f2ca8L /* 0.281048447 */, 17 },
 { 0x047fc071L /* 0.281189385 */, 17 }, { 0x0480543eL /* 0.281330341 */, 17 },
 { 0x0480e811L /* 0.281471315 */, 17 }, { 0x04817be8L /* 0.281612306 */, 17 },
 { 0x04820fc3L /* 0.281753315 */, 17 }, { 0x0482a3a4L /* 0.281894341 */, 17 },
 { 0x04833789L /* 0.282035386 */, 17 }, { 0x0483cb73L /* 0.282176447 */, 17 },
 { 0x04845f62L /* 0.282317527 */, 17 }, { 0x0484f355L /* 0.282458624 */, 17 },
 { 0x0485874dL /* 0.282599738 */, 17 }, { 0x04861b4aL /* 0.282740871 */, 17 },

 { 0x0486af4cL /* 0.282882021 */, 17 }, { 0x04874352L /* 0.283023188 */, 17 },
 { 0x0487d75dL /* 0.283164373 */, 17 }, { 0x04886b6dL /* 0.283305576 */, 17 },
 { 0x0488ff82L /* 0.283446796 */, 17 }, { 0x0489939bL /* 0.283588034 */, 17 },
 { 0x048a27b9L /* 0.283729290 */, 17 }, { 0x048abbdcL /* 0.283870563 */, 17 },
 { 0x048b5003L /* 0.284011853 */, 17 }, { 0x048be42fL /* 0.284153161 */, 17 },
 { 0x048c7860L /* 0.284294487 */, 17 }, { 0x048d0c96L /* 0.284435831 */, 17 },
 { 0x048da0d0L /* 0.284577192 */, 17 }, { 0x048e350fL /* 0.284718570 */, 17 },
 { 0x048ec953L /* 0.284859966 */, 17 }, { 0x048f5d9bL /* 0.285001380 */, 17 },

 { 0x048ff1e8L /* 0.285142811 */, 17 }, { 0x0490863aL /* 0.285284259 */, 17 },
 { 0x04911a91L /* 0.285425726 */, 17 }, { 0x0491aeecL /* 0.285567209 */, 17 },
 { 0x0492434cL /* 0.285708711 */, 17 }, { 0x0492d7b0L /* 0.285850229 */, 17 },
 { 0x04936c1aL /* 0.285991766 */, 17 }, { 0x04940088L /* 0.286133319 */, 17 },
 { 0x049494fbL /* 0.286274891 */, 17 }, { 0x04952972L /* 0.286416480 */, 17 },
 { 0x0495bdeeL /* 0.286558086 */, 17 }, { 0x0496526fL /* 0.286699710 */, 17 },
 { 0x0496e6f5L /* 0.286841351 */, 17 }, { 0x04977b7fL /* 0.286983010 */, 17 },
 { 0x0498100eL /* 0.287124686 */, 17 }, { 0x0498a4a1L /* 0.287266380 */, 17 },

 { 0x0499393aL /* 0.287408091 */, 17 }, { 0x0499cdd7L /* 0.287549820 */, 17 },
 { 0x049a6278L /* 0.287691566 */, 17 }, { 0x049af71fL /* 0.287833330 */, 17 },
 { 0x049b8bcaL /* 0.287975111 */, 17 }, { 0x049c207aL /* 0.288116909 */, 17 },
 { 0x049cb52eL /* 0.288258725 */, 17 }, { 0x049d49e7L /* 0.288400559 */, 17 },
 { 0x049ddea5L /* 0.288542409 */, 17 }, { 0x049e7367L /* 0.288684278 */, 17 },
 { 0x049f082fL /* 0.288826163 */, 17 }, { 0x049f9cfaL /* 0.288968067 */, 17 },
 { 0x04a031cbL /* 0.289109987 */, 17 }, { 0x04a0c6a0L /* 0.289251925 */, 17 },
 { 0x04a15b7aL /* 0.289393881 */, 17 }, { 0x04a1f059L /* 0.289535854 */, 17 },

 { 0x04a2853cL /* 0.289677844 */, 17 }, { 0x04a31a24L /* 0.289819851 */, 17 },
 { 0x04a3af10L /* 0.289961876 */, 17 }, { 0x04a44401L /* 0.290103919 */, 17 },
 { 0x04a4d8f7L /* 0.290245979 */, 17 }, { 0x04a56df2L /* 0.290388056 */, 17 },
 { 0x04a602f1L /* 0.290530150 */, 17 }, { 0x04a697f5L /* 0.290672262 */, 17 },
 { 0x04a72cfeL /* 0.290814392 */, 17 }, { 0x04a7c20bL /* 0.290956538 */, 17 },
 { 0x04a8571dL /* 0.291098703 */, 17 }, { 0x04a8ec33L /* 0.291240884 */, 17 },
 { 0x04a9814eL /* 0.291383083 */, 17 }, { 0x04aa166eL /* 0.291525299 */, 17 },
 { 0x04aaab93L /* 0.291667532 */, 17 }, { 0x04ab40bcL /* 0.291809783 */, 17 },

 { 0x04abd5eaL /* 0.291952051 */, 17 }, { 0x04ac6b1cL /* 0.292094337 */, 17 },
 { 0x04ad0053L /* 0.292236640 */, 17 }, { 0x04ad958fL /* 0.292378960 */, 17 },
 { 0x04ae2ad0L /* 0.292521297 */, 17 }, { 0x04aec015L /* 0.292663652 */, 17 },
 { 0x04af555eL /* 0.292806024 */, 17 }, { 0x04afeaadL /* 0.292948414 */, 17 },
 { 0x04b08000L /* 0.293090820 */, 17 }, { 0x04b11557L /* 0.293233244 */, 17 },
 { 0x04b1aab4L /* 0.293375686 */, 17 }, { 0x04b24015L /* 0.293518144 */, 17 },
 { 0x04b2d57aL /* 0.293660620 */, 17 }, { 0x04b36ae4L /* 0.293803113 */, 17 },
 { 0x04b40053L /* 0.293945624 */, 17 }, { 0x04b495c7L /* 0.294088151 */, 17 },

 { 0x04b52b3fL /* 0.294230696 */, 17 }, { 0x04b5c0bcL /* 0.294373259 */, 17 },
 { 0x04b6563dL /* 0.294515838 */, 17 }, { 0x04b6ebc3L /* 0.294658435 */, 17 },
 { 0x04b7814eL /* 0.294801049 */, 17 }, { 0x04b816ddL /* 0.294943680 */, 17 },
 { 0x04b8ac71L /* 0.295086329 */, 17 }, { 0x04b9420aL /* 0.295228995 */, 17 },
 { 0x04b9d7a7L /* 0.295371678 */, 17 }, { 0x04ba6d49L /* 0.295514378 */, 17 },
 { 0x04bb02efL /* 0.295657095 */, 17 }, { 0x04bb989aL /* 0.295799830 */, 17 },
 { 0x04bc2e4aL /* 0.295942582 */, 17 }, { 0x04bcc3feL /* 0.296085351 */, 17 },
 { 0x04bd59b7L /* 0.296228138 */, 17 }, { 0x04bdef74L /* 0.296370941 */, 17 },

 { 0x04be8537L /* 0.296513762 */, 17 }, { 0x04bf1afdL /* 0.296656600 */, 17 },
 { 0x04bfb0c9L /* 0.296799455 */, 17 }, { 0x04c04699L /* 0.296942327 */, 17 },
 { 0x04c0dc6dL /* 0.297085217 */, 17 }, { 0x04c17247L /* 0.297228124 */, 17 },
 { 0x04c20824L /* 0.297371048 */, 17 }, { 0x04c29e07L /* 0.297513989 */, 17 },
 { 0x04c333eeL /* 0.297656947 */, 17 }, { 0x04c3c9daL /* 0.297799922 */, 17 },
 { 0x04c45fcaL /* 0.297942915 */, 17 }, { 0x04c4f5bfL /* 0.298085925 */, 17 },
 { 0x04c58bb8L /* 0.298228951 */, 17 }, { 0x04c621b6L /* 0.298371996 */, 17 },
 { 0x04c6b7b9L /* 0.298515057 */, 17 }, { 0x04c74dc0L /* 0.298658135 */, 17 },

 { 0x04c7e3ccL /* 0.298801231 */, 17 }, { 0x04c879ddL /* 0.298944343 */, 17 },
 { 0x04c90ff2L /* 0.299087473 */, 17 }, { 0x04c9a60cL /* 0.299230620 */, 17 },
 { 0x04ca3c2aL /* 0.299373784 */, 17 }, { 0x04cad24dL /* 0.299516965 */, 17 },
 { 0x04cb6874L /* 0.299660163 */, 17 }, { 0x04cbfea0L /* 0.299803378 */, 17 },
 { 0x04cc94d1L /* 0.299946611 */, 17 }, { 0x04cd2b06L /* 0.300089860 */, 17 },
 { 0x04cdc140L /* 0.300233127 */, 17 }, { 0x04ce577fL /* 0.300376411 */, 17 },
 { 0x04ceedc2L /* 0.300519711 */, 17 }, { 0x04cf8409L /* 0.300663029 */, 17 },
 { 0x04d01a55L /* 0.300806364 */, 17 }, { 0x04d0b0a6L /* 0.300949716 */, 17 },

 { 0x04d146fbL /* 0.301093085 */, 17 }, { 0x04d1dd55L /* 0.301236472 */, 17 },
 { 0x04d273b4L /* 0.301379875 */, 17 }, { 0x04d30a17L /* 0.301523295 */, 17 },
 { 0x04d3a07fL /* 0.301666733 */, 17 }, { 0x04d436ebL /* 0.301810187 */, 17 },
 { 0x04d4cd5cL /* 0.301953659 */, 17 }, { 0x04d563d1L /* 0.302097147 */, 17 },
 { 0x04d5fa4bL /* 0.302240653 */, 17 }, { 0x04d690caL /* 0.302384175 */, 17 },
 { 0x04d7274dL /* 0.302527715 */, 17 }, { 0x04d7bdd5L /* 0.302671271 */, 17 },
 { 0x04d85461L /* 0.302814845 */, 17 }, { 0x04d8eaf2L /* 0.302958436 */, 17 },
 { 0x04d98187L /* 0.303102044 */, 17 }, { 0x04da1821L /* 0.303245668 */, 17 },

 { 0x04daaec0L /* 0.303389310 */, 17 }, { 0x04db4563L /* 0.303532969 */, 17 },
 { 0x04dbdc0aL /* 0.303676645 */, 17 }, { 0x04dc72b7L /* 0.303820337 */, 17 },
 { 0x04dd0967L /* 0.303964047 */, 17 }, { 0x04dda01dL /* 0.304107774 */, 17 },
 { 0x04de36d7L /* 0.304251517 */, 17 }, { 0x04decd95L /* 0.304395278 */, 17 },
 { 0x04df6458L /* 0.304539056 */, 17 }, { 0x04dffb20L /* 0.304682850 */, 17 },
 { 0x04e091ecL /* 0.304826662 */, 17 }, { 0x04e128bcL /* 0.304970491 */, 17 },
 { 0x04e1bf92L /* 0.305114336 */, 17 }, { 0x04e2566bL /* 0.305258199 */, 17 },
 { 0x04e2ed4aL /* 0.305402078 */, 17 }, { 0x04e3842dL /* 0.305545974 */, 17 },

 { 0x04e41b14L /* 0.305689888 */, 17 }, { 0x04e4b200L /* 0.305833818 */, 17 },
 { 0x04e548f1L /* 0.305977765 */, 17 }, { 0x04e5dfe6L /* 0.306121729 */, 17 },
 { 0x04e676dfL /* 0.306265710 */, 17 }, { 0x04e70ddeL /* 0.306409708 */, 17 },
 { 0x04e7a4e0L /* 0.306553723 */, 17 }, { 0x04e83be7L /* 0.306697755 */, 17 },
 { 0x04e8d2f3L /* 0.306841804 */, 17 }, { 0x04e96a04L /* 0.306985869 */, 17 },
 { 0x04ea0118L /* 0.307129952 */, 17 }, { 0x04ea9832L /* 0.307274051 */, 17 },
 { 0x04eb2f50L /* 0.307418168 */, 17 }, { 0x04ebc672L /* 0.307562301 */, 17 },
 { 0x04ec5d99L /* 0.307706451 */, 17 }, { 0x04ecf4c5L /* 0.307850618 */, 17 },

 { 0x04ed8bf5L /* 0.307994802 */, 17 }, { 0x04ee2329L /* 0.308139003 */, 17 },
 { 0x04eeba63L /* 0.308283220 */, 17 }, { 0x04ef51a0L /* 0.308427455 */, 17 },
 { 0x04efe8e2L /* 0.308571706 */, 17 }, { 0x04f08029L /* 0.308715974 */, 17 },
 { 0x04f11774L /* 0.308860260 */, 17 }, { 0x04f1aec4L /* 0.309004561 */, 17 },
 { 0x04f24618L /* 0.309148880 */, 17 }, { 0x04f2dd71L /* 0.309293216 */, 17 },
 { 0x04f374cfL /* 0.309437568 */, 17 }, { 0x04f40c30L /* 0.309581938 */, 17 },
 { 0x04f4a397L /* 0.309726324 */, 17 }, { 0x04f53b02L /* 0.309870727 */, 17 },
 { 0x04f5d271L /* 0.310015147 */, 17 }, { 0x04f669e5L /* 0.310159583 */, 17 },

 { 0x04f7015dL /* 0.310304037 */, 17 }, { 0x04f798daL /* 0.310448507 */, 17 },
 { 0x04f8305cL /* 0.310592994 */, 17 }, { 0x04f8c7e2L /* 0.310737498 */, 17 },
 { 0x04f95f6cL /* 0.310882018 */, 17 }, { 0x04f9f6fbL /* 0.311026556 */, 17 },
 { 0x04fa8e8fL /* 0.311171110 */, 17 }, { 0x04fb2627L /* 0.311315681 */, 17 },
 { 0x04fbbdc3L /* 0.311460269 */, 17 }, { 0x04fc5564L /* 0.311604874 */, 17 },
 { 0x04fced0aL /* 0.311749495 */, 17 }, { 0x04fd84b4L /* 0.311894133 */, 17 },
 { 0x04fe1c62L /* 0.312038788 */, 17 }, { 0x04feb415L /* 0.312183460 */, 17 },
 { 0x04ff4bcdL /* 0.312328148 */, 17 }, { 0x04ffe389L /* 0.312472854 */, 17 },

 { 0x05007b49L /* 0.312617576 */, 17 }, { 0x0501130eL /* 0.312762314 */, 17 },
 { 0x0501aad8L /* 0.312907070 */, 17 }, { 0x050242a6L /* 0.313051842 */, 17 },
 { 0x0502da78L /* 0.313196631 */, 17 }, { 0x0503724fL /* 0.313341437 */, 17 },
 { 0x05040a2bL /* 0.313486259 */, 17 }, { 0x0504a20bL /* 0.313631098 */, 17 },
 { 0x050539efL /* 0.313775954 */, 17 }, { 0x0505d1d8L /* 0.313920827 */, 17 },
 { 0x050669c5L /* 0.314065716 */, 17 }, { 0x050701b7L /* 0.314210622 */, 17 },
 { 0x050799aeL /* 0.314355545 */, 17 }, { 0x050831a9L /* 0.314500484 */, 17 },
 { 0x0508c9a8L /* 0.314645440 */, 17 }, { 0x050961acL /* 0.314790413 */, 17 },

 { 0x0509f9b4L /* 0.314935403 */, 17 }, { 0x050a91c1L /* 0.315080409 */, 17 },
 { 0x050b29d2L /* 0.315225432 */, 17 }, { 0x050bc1e8L /* 0.315370472 */, 17 },
 { 0x050c5a02L /* 0.315515528 */, 17 }, { 0x050cf221L /* 0.315660601 */, 17 },
 { 0x050d8a44L /* 0.315805690 */, 17 }, { 0x050e226cL /* 0.315950797 */, 17 },
 { 0x050eba98L /* 0.316095920 */, 17 }, { 0x050f52c9L /* 0.316241059 */, 17 },
 { 0x050feafeL /* 0.316386216 */, 17 }, { 0x05108337L /* 0.316531388 */, 17 },
 { 0x05111b75L /* 0.316676578 */, 17 }, { 0x0511b3b8L /* 0.316821784 */, 17 },
 { 0x05124bffL /* 0.316967007 */, 17 }, { 0x0512e44aL /* 0.317112247 */, 17 },

 { 0x05137c9aL /* 0.317257503 */, 17 }, { 0x051414eeL /* 0.317402775 */, 17 },
 { 0x0514ad47L /* 0.317548065 */, 17 }, { 0x051545a5L /* 0.317693371 */, 17 },
 { 0x0515de06L /* 0.317838693 */, 17 }, { 0x0516766dL /* 0.317984033 */, 17 },
 { 0x05170ed7L /* 0.318129388 */, 17 }, { 0x0517a746L /* 0.318274761 */, 17 },
 { 0x05183fbaL /* 0.318420150 */, 17 }, { 0x0518d832L /* 0.318565555 */, 17 },
 { 0x051970aeL /* 0.318710978 */, 17 }, { 0x051a092fL /* 0.318856416 */, 17 },
 { 0x051aa1b5L /* 0.319001872 */, 17 }, { 0x051b3a3fL /* 0.319147344 */, 17 },
 { 0x051bd2cdL /* 0.319292832 */, 17 }, { 0x051c6b60L /* 0.319438338 */, 17 },

 { 0x051d03f7L /* 0.319583859 */, 17 }, { 0x051d9c92L /* 0.319729398 */, 17 },
 { 0x051e3532L /* 0.319874952 */, 17 }, { 0x051ecdd7L /* 0.320020524 */, 17 },
 { 0x051f6680L /* 0.320166112 */, 17 }, { 0x051fff2dL /* 0.320311716 */, 17 },
 { 0x052097dfL /* 0.320457337 */, 17 }, { 0x05213095L /* 0.320602975 */, 17 },
 { 0x0521c950L /* 0.320748629 */, 17 }, { 0x0522620fL /* 0.320894300 */, 17 },
 { 0x0522fad3L /* 0.321039987 */, 17 }, { 0x0523939bL /* 0.321185691 */, 17 },
 { 0x05242c68L /* 0.321331411 */, 17 }, { 0x0524c538L /* 0.321477148 */, 17 },
 { 0x05255e0eL /* 0.321622901 */, 17 }, { 0x0525f6e8L /* 0.321768671 */, 17 },

 { 0x05268fc6L /* 0.321914457 */, 17 }, { 0x052728a9L /* 0.322060260 */, 17 },
 { 0x0527c190L /* 0.322206079 */, 17 }, { 0x05285a7bL /* 0.322351915 */, 17 },
 { 0x0528f36bL /* 0.322497768 */, 17 }, { 0x05298c5fL /* 0.322643636 */, 17 },
 { 0x052a2558L /* 0.322789522 */, 17 }, { 0x052abe55L /* 0.322935424 */, 17 },
 { 0x052b5757L /* 0.323081342 */, 17 }, { 0x052bf05dL /* 0.323227277 */, 17 },
 { 0x052c8968L /* 0.323373228 */, 17 }, { 0x052d2277L /* 0.323519196 */, 17 },
 { 0x052dbb8aL /* 0.323665180 */, 17 }, { 0x052e54a2L /* 0.323811180 */, 17 },
 { 0x052eedbeL /* 0.323957197 */, 17 }, { 0x052f86deL /* 0.324103231 */, 17 },

 { 0x05302003L /* 0.324249281 */, 17 }, { 0x0530b92dL /* 0.324395347 */, 17 },
 { 0x0531525bL /* 0.324541430 */, 17 }, { 0x0531eb8dL /* 0.324687530 */, 17 },
 { 0x053284c4L /* 0.324833646 */, 17 }, { 0x05331dffL /* 0.324979778 */, 17 },
 { 0x0533b73eL /* 0.325125926 */, 17 }, { 0x05345082L /* 0.325272091 */, 17 },
 { 0x0534e9caL /* 0.325418273 */, 17 }, { 0x05358317L /* 0.325564471 */, 17 },
 { 0x05361c68L /* 0.325710685 */, 17 }, { 0x0536b5beL /* 0.325856916 */, 17 },
 { 0x05374f17L /* 0.326003163 */, 17 }, { 0x0537e876L /* 0.326149427 */, 17 },
 { 0x053881d9L /* 0.326295707 */, 17 }, { 0x05391b40L /* 0.326442003 */, 17 },

 { 0x0539b4abL /* 0.326588316 */, 17 }, { 0x053a4e1bL /* 0.326734645 */, 17 },
 { 0x053ae78fL /* 0.326880990 */, 17 }, { 0x053b8108L /* 0.327027352 */, 17 },
 { 0x053c1a85L /* 0.327173730 */, 17 }, { 0x053cb407L /* 0.327320125 */, 17 },
 { 0x053d4d8dL /* 0.327466536 */, 17 }, { 0x053de717L /* 0.327612963 */, 17 },
 { 0x053e80a6L /* 0.327759407 */, 17 }, { 0x053f1a39L /* 0.327905867 */, 17 },
 { 0x053fb3d0L /* 0.328052344 */, 17 }, { 0x05404d6cL /* 0.328198837 */, 17 },
 { 0x0540e70cL /* 0.328345346 */, 17 }, { 0x054180b1L /* 0.328491871 */, 17 },
 { 0x05421a5aL /* 0.328638413 */, 17 }, { 0x0542b407L /* 0.328784971 */, 17 },

 { 0x05434db9L /* 0.328931546 */, 17 }, { 0x0543e76fL /* 0.329078137 */, 17 },
 { 0x0544812aL /* 0.329224744 */, 17 }, { 0x05451ae9L /* 0.329371367 */, 17 },
 { 0x0545b4acL /* 0.329518007 */, 17 }, { 0x05464e74L /* 0.329664663 */, 17 },
 { 0x0546e840L /* 0.329811336 */, 17 }, { 0x05478211L /* 0.329958024 */, 17 },
 { 0x05481be5L /* 0.330104730 */, 17 }, { 0x0548b5bfL /* 0.330251451 */, 17 },
 { 0x05494f9cL /* 0.330398189 */, 17 }, { 0x0549e97eL /* 0.330544943 */, 17 },
 { 0x054a8364L /* 0.330691713 */, 17 }, { 0x054b1d4fL /* 0.330838499 */, 17 },
 { 0x054bb73eL /* 0.330985302 */, 17 }, { 0x054c5132L /* 0.331132121 */, 17 },

 { 0x054ceb2aL /* 0.331278957 */, 17 }, { 0x054d8526L /* 0.331425808 */, 17 },
 { 0x054e1f26L /* 0.331572676 */, 17 }, { 0x054eb92bL /* 0.331719560 */, 17 },
 { 0x054f5334L /* 0.331866461 */, 17 }, { 0x054fed42L /* 0.332013377 */, 17 },
 { 0x05508754L /* 0.332160310 */, 17 }, { 0x0551216bL /* 0.332307260 */, 17 },
 { 0x0551bb85L /* 0.332454225 */, 17 }, { 0x055255a4L /* 0.332601207 */, 17 },
 { 0x0552efc8L /* 0.332748205 */, 17 }, { 0x055389f0L /* 0.332895219 */, 17 },
 { 0x0554241cL /* 0.333042249 */, 17 }, { 0x0554be4cL /* 0.333189296 */, 17 },
 { 0x05555881L /* 0.333336359 */, 17 }, { 0x0555f2baL /* 0.333483438 */, 17 },

 { 0x05568cf8L /* 0.333630533 */, 17 }, { 0x0557273aL /* 0.333777645 */, 17 },
 { 0x0557c180L /* 0.333924772 */, 17 }, { 0x05585bcbL /* 0.334071916 */, 17 },
 { 0x0558f61aL /* 0.334219076 */, 17 }, { 0x0559906dL /* 0.334366253 */, 17 },
 { 0x055a2ac5L /* 0.334513445 */, 17 }, { 0x055ac521L /* 0.334660654 */, 17 },
 { 0x055b5f81L /* 0.334807879 */, 17 }, { 0x055bf9e6L /* 0.334955120 */, 17 },
 { 0x055c944fL /* 0.335102377 */, 17 }, { 0x055d2ebdL /* 0.335249651 */, 17 },
 { 0x055dc92eL /* 0.335396941 */, 17 }, { 0x055e63a5L /* 0.335544246 */, 17 },
 { 0x055efe1fL /* 0.335691568 */, 17 }, { 0x055f989eL /* 0.335838906 */, 17 },

 { 0x05603321L /* 0.335986261 */, 17 }, { 0x0560cda8L /* 0.336133631 */, 17 },
 { 0x05616834L /* 0.336281018 */, 17 }, { 0x056202c4L /* 0.336428421 */, 17 },
 { 0x05629d59L /* 0.336575840 */, 17 }, { 0x056337f2L /* 0.336723275 */, 17 },
 { 0x0563d28fL /* 0.336870726 */, 17 }, { 0x05646d30L /* 0.337018193 */, 17 },
 { 0x056507d6L /* 0.337165677 */, 17 }, { 0x0565a280L /* 0.337313176 */, 17 },
 { 0x05663d2fL /* 0.337460692 */, 17 }, { 0x0566d7e1L /* 0.337608224 */, 17 },
 { 0x05677298L /* 0.337755772 */, 17 }, { 0x05680d54L /* 0.337903336 */, 17 },
 { 0x0568a814L /* 0.338050916 */, 17 }, { 0x056942d8L /* 0.338198513 */, 17 },

 { 0x0569dda0L /* 0.338346125 */, 17 }, { 0x056a786dL /* 0.338493753 */, 17 },
 { 0x056b133eL /* 0.338641398 */, 17 }, { 0x056bae13L /* 0.338789059 */, 17 },
 { 0x056c48edL /* 0.338936736 */, 17 }, { 0x056ce3cbL /* 0.339084429 */, 17 },
 { 0x056d7eadL /* 0.339232138 */, 17 }, { 0x056e1994L /* 0.339379863 */, 17 },
 { 0x056eb47fL /* 0.339527604 */, 17 }, { 0x056f4f6eL /* 0.339675361 */, 17 },
 { 0x056fea62L /* 0.339823134 */, 17 }, { 0x0570855aL /* 0.339970924 */, 17 },
 { 0x05712056L /* 0.340118729 */, 17 }, { 0x0571bb56L /* 0.340266550 */, 17 },
 { 0x0572565bL /* 0.340414388 */, 17 }, { 0x0572f164L /* 0.340562242 */, 17 },

 { 0x05738c72L /* 0.340710111 */, 17 }, { 0x05742784L /* 0.340857997 */, 17 },
 { 0x0574c29aL /* 0.341005899 */, 17 }, { 0x05755db4L /* 0.341153816 */, 17 },
 { 0x0575f8d3L /* 0.341301750 */, 17 }, { 0x057693f6L /* 0.341449700 */, 17 },
 { 0x05772f1dL /* 0.341597666 */, 17 }, { 0x0577ca49L /* 0.341745648 */, 17 },
 { 0x05786578L /* 0.341893646 */, 17 }, { 0x057900adL /* 0.342041659 */, 17 },
 { 0x05799be5L /* 0.342189689 */, 17 }, { 0x057a3722L /* 0.342337735 */, 17 },
 { 0x057ad263L /* 0.342485797 */, 17 }, { 0x057b6da8L /* 0.342633875 */, 17 },
 { 0x057c08f2L /* 0.342781969 */, 17 }, { 0x057ca440L /* 0.342930079 */, 17 },

 { 0x057d3f92L /* 0.343078205 */, 17 }, { 0x057ddae9L /* 0.343226347 */, 17 },
 { 0x057e7644L /* 0.343374505 */, 17 }, { 0x057f11a3L /* 0.343522679 */, 17 },
 { 0x057fad06L /* 0.343670869 */, 17 }, { 0x0580486eL /* 0.343819075 */, 17 },
 { 0x0580e3daL /* 0.343967296 */, 17 }, { 0x05817f4aL /* 0.344115534 */, 17 },
 { 0x05821abfL /* 0.344263788 */, 17 }, { 0x0582b638L /* 0.344412058 */, 17 },
 { 0x058351b5L /* 0.344560343 */, 17 }, { 0x0583ed36L /* 0.344708645 */, 17 },
 { 0x058488bcL /* 0.344856963 */, 17 }, { 0x05852446L /* 0.345005296 */, 17 },
 { 0x0585bfd4L /* 0.345153646 */, 17 }, { 0x05865b67L /* 0.345302011 */, 17 },

 { 0x0586f6fdL /* 0.345450393 */, 17 }, { 0x05879298L /* 0.345598790 */, 17 },
 { 0x05882e38L /* 0.345747203 */, 17 }, { 0x0588c9dcL /* 0.345895632 */, 17 },
 { 0x05896583L /* 0.346044077 */, 17 }, { 0x058a0130L /* 0.346192538 */, 17 },
 { 0x058a9ce0L /* 0.346341015 */, 17 }, { 0x058b3895L /* 0.346489508 */, 17 },
 { 0x058bd44eL /* 0.346638017 */, 17 }, { 0x058c700bL /* 0.346786542 */, 17 },
 { 0x058d0bcdL /* 0.346935082 */, 17 }, { 0x058da793L /* 0.347083639 */, 17 },
 { 0x058e435dL /* 0.347232211 */, 17 }, { 0x058edf2bL /* 0.347380799 */, 17 },
 { 0x058f7afeL /* 0.347529403 */, 17 }, { 0x059016d5L /* 0.347678023 */, 17 },

 { 0x0590b2b0L /* 0.347826659 */, 17 }, { 0x05914e8fL /* 0.347975311 */, 17 },
 { 0x0591ea73L /* 0.348123979 */, 17 }, { 0x0592865bL /* 0.348272662 */, 17 },
 { 0x05932247L /* 0.348421362 */, 17 }, { 0x0593be37L /* 0.348570077 */, 17 },
 { 0x05945a2cL /* 0.348718808 */, 17 }, { 0x0594f625L /* 0.348867555 */, 17 },
 { 0x05959222L /* 0.349016318 */, 17 }, { 0x05962e24L /* 0.349165097 */, 17 },
 { 0x0596ca2aL /* 0.349313892 */, 17 }, { 0x05976634L /* 0.349462702 */, 17 },
 { 0x05980242L /* 0.349611528 */, 17 }, { 0x05989e54L /* 0.349760370 */, 17 },
 { 0x05993a6bL /* 0.349909228 */, 17 }, { 0x0599d686L /* 0.350058102 */, 17 },

 { 0x059a72a5L /* 0.350206992 */, 17 }, { 0x059b0ec9L /* 0.350355897 */, 17 },
 { 0x059baaf1L /* 0.350504818 */, 17 }, { 0x059c471dL /* 0.350653756 */, 17 },
 { 0x059ce34dL /* 0.350802708 */, 17 }, { 0x059d7f81L /* 0.350951677 */, 17 },
 { 0x059e1bbaL /* 0.351100662 */, 17 }, { 0x059eb7f7L /* 0.351249662 */, 17 },
 { 0x059f5438L /* 0.351398678 */, 17 }, { 0x059ff07eL /* 0.351547710 */, 17 },
 { 0x05a08cc7L /* 0.351696758 */, 17 }, { 0x05a12915L /* 0.351845821 */, 17 },
 { 0x05a1c567L /* 0.351994901 */, 17 }, { 0x05a261beL /* 0.352143996 */, 17 },
 { 0x05a2fe18L /* 0.352293107 */, 17 }, { 0x05a39a77L /* 0.352442233 */, 17 },

 { 0x05a436daL /* 0.352591376 */, 17 }, { 0x05a4d342L /* 0.352740534 */, 17 },
 { 0x05a56fadL /* 0.352889708 */, 17 }, { 0x05a60c1dL /* 0.353038898 */, 17 },
 { 0x05a6a891L /* 0.353188103 */, 17 }, { 0x05a7450aL /* 0.353337325 */, 17 },
 { 0x05a7e186L /* 0.353486562 */, 17 }, { 0x05a87e07L /* 0.353635814 */, 17 },
 { 0x05a91a8cL /* 0.353785083 */, 17 }, { 0x05a9b715L /* 0.353934367 */, 17 },
 { 0x05aa53a2L /* 0.354083667 */, 17 }, { 0x05aaf034L /* 0.354232983 */, 17 },
 { 0x05ab8ccaL /* 0.354382314 */, 17 }, { 0x05ac2964L /* 0.354531662 */, 17 },
 { 0x05acc602L /* 0.354681025 */, 17 }, { 0x05ad62a5L /* 0.354830403 */, 17 },

 { 0x05adff4cL /* 0.354979798 */, 17 }, { 0x05ae9bf7L /* 0.355129208 */, 17 },
 { 0x05af38a6L /* 0.355278634 */, 17 }, { 0x05afd559L /* 0.355428075 */, 17 },
 { 0x05b07211L /* 0.355577533 */, 17 }, { 0x05b10ecdL /* 0.355727006 */, 17 },
 { 0x05b1ab8dL /* 0.355876494 */, 17 }, { 0x05b24851L /* 0.356025999 */, 17 },
 { 0x05b2e51aL /* 0.356175519 */, 17 }, { 0x05b381e6L /* 0.356325054 */, 17 },
 { 0x05b41eb7L /* 0.356474606 */, 17 }, { 0x05b4bb8cL /* 0.356624173 */, 17 },
 { 0x05b55866L /* 0.356773756 */, 17 }, { 0x05b5f543L /* 0.356923354 */, 17 },
 { 0x05b69225L /* 0.357072969 */, 17 }, { 0x05b72f0bL /* 0.357222598 */, 17 },

 { 0x05b7cbf5L /* 0.357372244 */, 17 }, { 0x05b868e3L /* 0.357521905 */, 17 },
 { 0x05b905d6L /* 0.357671582 */, 17 }, { 0x05b9a2cdL /* 0.357821275 */, 17 },
 { 0x05ba3fc8L /* 0.357970983 */, 17 }, { 0x05badcc7L /* 0.358120707 */, 17 },
 { 0x05bb79caL /* 0.358270446 */, 17 }, { 0x05bc16d2L /* 0.358420201 */, 17 },
 { 0x05bcb3deL /* 0.358569972 */, 17 }, { 0x05bd50eeL /* 0.358719758 */, 17 },
 { 0x05bdee02L /* 0.358869560 */, 17 }, { 0x05be8b1aL /* 0.359019378 */, 17 },
 { 0x05bf2837L /* 0.359169211 */, 17 }, { 0x05bfc558L /* 0.359319060 */, 17 },
 { 0x05c0627dL /* 0.359468925 */, 17 }, { 0x05c0ffa6L /* 0.359618805 */, 17 },

 { 0x05c19cd3L /* 0.359768701 */, 17 }, { 0x05c23a05L /* 0.359918612 */, 17 },
 { 0x05c2d73aL /* 0.360068540 */, 17 }, { 0x05c37474L /* 0.360218482 */, 17 },
 { 0x05c411b2L /* 0.360368440 */, 17 }, { 0x05c4aef5L /* 0.360518414 */, 17 },
 { 0x05c54c3bL /* 0.360668404 */, 17 }, { 0x05c5e986L /* 0.360818409 */, 17 },
 { 0x05c686d5L /* 0.360968429 */, 17 }, { 0x05c72428L /* 0.361118466 */, 17 },
 { 0x05c7c17fL /* 0.361268517 */, 17 }, { 0x05c85edaL /* 0.361418585 */, 17 },
 { 0x05c8fc3aL /* 0.361568668 */, 17 }, { 0x05c9999eL /* 0.361718766 */, 17 },
 { 0x05ca3706L /* 0.361868881 */, 17 }, { 0x05cad472L /* 0.362019010 */, 17 },

 { 0x05cb71e2L /* 0.362169156 */, 17 }, { 0x05cc0f57L /* 0.362319316 */, 17 },
 { 0x05ccaccfL /* 0.362469493 */, 17 }, { 0x05cd4a4cL /* 0.362619685 */, 17 },
 { 0x05cde7cdL /* 0.362769892 */, 17 }, { 0x05ce8552L /* 0.362920115 */, 17 },
 { 0x05cf22dcL /* 0.363070354 */, 17 }, { 0x05cfc069L /* 0.363220608 */, 17 },
 { 0x05d05dfbL /* 0.363370878 */, 17 }, { 0x05d0fb91L /* 0.363521163 */, 17 },
 { 0x05d1992bL /* 0.363671464 */, 17 }, { 0x05d236c9L /* 0.363821780 */, 17 },
 { 0x05d2d46cL /* 0.363972112 */, 17 }, { 0x05d37212L /* 0.364122459 */, 17 },
 { 0x05d40fbdL /* 0.364272822 */, 17 }, { 0x05d4ad6cL /* 0.364423200 */, 17 },

 { 0x05d54b1fL /* 0.364573594 */, 17 }, { 0x05d5e8d6L /* 0.364724004 */, 17 },
 { 0x05d68691L /* 0.364874429 */, 17 }, { 0x05d72451L /* 0.365024869 */, 17 },
 { 0x05d7c215L /* 0.365175325 */, 17 }, { 0x05d85fdcL /* 0.365325796 */, 17 },
 { 0x05d8fda8L /* 0.365476283 */, 17 }, { 0x05d99b79L /* 0.365626786 */, 17 },
 { 0x05da394dL /* 0.365777304 */, 17 }, { 0x05dad726L /* 0.365927837 */, 17 },
 { 0x05db7502L /* 0.366078386 */, 17 }, { 0x05dc12e3L /* 0.366228950 */, 17 },
 { 0x05dcb0c8L /* 0.366379530 */, 17 }, { 0x05dd4eb1L /* 0.366530125 */, 17 },
 { 0x05ddec9eL /* 0.366680736 */, 17 }, { 0x05de8a90L /* 0.366831362 */, 17 },

 { 0x05df2885L /* 0.366982004 */, 17 }, { 0x05dfc67fL /* 0.367132661 */, 17 },
 { 0x05e0647dL /* 0.367283334 */, 17 }, { 0x05e1027fL /* 0.367434022 */, 17 },
 { 0x05e1a085L /* 0.367584725 */, 17 }, { 0x05e23e8fL /* 0.367735444 */, 17 },
 { 0x05e2dc9eL /* 0.367886179 */, 17 }, { 0x05e37ab0L /* 0.368036929 */, 17 },
 { 0x05e418c7L /* 0.368187694 */, 17 }, { 0x05e4b6e2L /* 0.368338475 */, 17 },
 { 0x05e55501L /* 0.368489271 */, 17 }, { 0x05e5f324L /* 0.368640082 */, 17 },
 { 0x05e6914cL /* 0.368790909 */, 17 }, { 0x05e72f77L /* 0.368941752 */, 17 },
 { 0x05e7cda7L /* 0.369092610 */, 17 }, { 0x05e86bdaL /* 0.369243483 */, 17 },

 { 0x05e90a12L /* 0.369394372 */, 17 }, { 0x05e9a84eL /* 0.369545276 */, 17 },
 { 0x05ea468eL /* 0.369696195 */, 17 }, { 0x05eae4d3L /* 0.369847130 */, 17 },
 { 0x05eb831bL /* 0.369998080 */, 17 }, { 0x05ec2168L /* 0.370149046 */, 17 },
 { 0x05ecbfb8L /* 0.370300027 */, 17 }, { 0x05ed5e0dL /* 0.370451024 */, 17 },
 { 0x05edfc66L /* 0.370602036 */, 17 }, { 0x05ee9ac3L /* 0.370753063 */, 17 },
 { 0x05ef3924L /* 0.370904105 */, 17 }, { 0x05efd78aL /* 0.371055163 */, 17 },
 { 0x05f075f3L /* 0.371206237 */, 17 }, { 0x05f11461L /* 0.371357326 */, 17 },
 { 0x05f1b2d3L /* 0.371508430 */, 17 }, { 0x05f25148L /* 0.371659549 */, 17 },

 { 0x05f2efc2L /* 0.371810684 */, 17 }, { 0x05f38e40L /* 0.371961834 */, 17 },
 { 0x05f42cc3L /* 0.372113000 */, 17 }, { 0x05f4cb49L /* 0.372264181 */, 17 },
 { 0x05f569d3L /* 0.372415377 */, 17 }, { 0x05f60862L /* 0.372566589 */, 17 },
 { 0x05f6a6f5L /* 0.372717816 */, 17 }, { 0x05f7458bL /* 0.372869058 */, 17 },
 { 0x05f7e426L /* 0.373020316 */, 17 }, { 0x05f882c5L /* 0.373171589 */, 17 },
 { 0x05f92169L /* 0.373322877 */, 17 }, { 0x05f9c010L /* 0.373474181 */, 17 },
 { 0x05fa5ebbL /* 0.373625500 */, 17 }, { 0x05fafd6bL /* 0.373776834 */, 17 },
 { 0x05fb9c1eL /* 0.373928184 */, 17 }, { 0x05fc3ad6L /* 0.374079549 */, 17 },

 { 0x05fcd992L /* 0.374230929 */, 17 }, { 0x05fd7852L /* 0.374382325 */, 17 },
 { 0x05fe1716L /* 0.374533735 */, 17 }, { 0x05feb5deL /* 0.374685162 */, 17 },
 { 0x05ff54aaL /* 0.374836603 */, 17 }, { 0x05fff37bL /* 0.374988060 */, 17 },
 { 0x0600924fL /* 0.375139532 */, 17 }, { 0x06013128L /* 0.375291019 */, 17 },
 { 0x0601d004L /* 0.375442522 */, 17 }, { 0x06026ee5L /* 0.375594040 */, 17 },
 { 0x06030dcaL /* 0.375745573 */, 17 }, { 0x0603acb3L /* 0.375897122 */, 17 },
 { 0x06044ba0L /* 0.376048685 */, 17 }, { 0x0604ea91L /* 0.376200265 */, 17 },
 { 0x06058987L /* 0.376351859 */, 17 }, { 0x06062880L /* 0.376503468 */, 17 },

 { 0x0606c77dL /* 0.376655093 */, 17 }, { 0x0607667fL /* 0.376806733 */, 17 },
 { 0x06080585L /* 0.376958389 */, 17 }, { 0x0608a48fL /* 0.377110059 */, 17 },
 { 0x0609439cL /* 0.377261745 */, 17 }, { 0x0609e2aeL /* 0.377413446 */, 17 },
 { 0x060a81c4L /* 0.377565163 */, 17 }, { 0x060b20dfL /* 0.377716894 */, 17 },
 { 0x060bbffdL /* 0.377868641 */, 17 }, { 0x060c5f1fL /* 0.378020403 */, 17 },
 { 0x060cfe46L /* 0.378172181 */, 17 }, { 0x060d9d70L /* 0.378323973 */, 17 },
 { 0x060e3c9fL /* 0.378475781 */, 17 }, { 0x060edbd1L /* 0.378627604 */, 17 },
 { 0x060f7b08L /* 0.378779442 */, 17 }, { 0x06101a43L /* 0.378931296 */, 17 },

 { 0x0610b982L /* 0.379083164 */, 17 }, { 0x061158c5L /* 0.379235048 */, 17 },
 { 0x0611f80cL /* 0.379386947 */, 17 }, { 0x06129757L /* 0.379538862 */, 17 },
 { 0x061336a6L /* 0.379690791 */, 17 }, { 0x0613d5faL /* 0.379842736 */, 17 },
 { 0x06147551L /* 0.379994696 */, 17 }, { 0x061514adL /* 0.380146671 */, 17 },
 { 0x0615b40cL /* 0.380298661 */, 17 }, { 0x06165370L /* 0.380450666 */, 17 },
 { 0x0616f2d8L /* 0.380602687 */, 17 }, { 0x06179243L /* 0.380754723 */, 17 },
 { 0x061831b3L /* 0.380906774 */, 17 }, { 0x0618d127L /* 0.381058840 */, 17 },
 { 0x0619709fL /* 0.381210921 */, 17 }, { 0x061a101bL /* 0.381363018 */, 17 },

 { 0x061aaf9cL /* 0.381515130 */, 17 }, { 0x061b4f20L /* 0.381667257 */, 17 },
 { 0x061beea8L /* 0.381819399 */, 17 }, { 0x061c8e34L /* 0.381971556 */, 17 },
 { 0x061d2dc5L /* 0.382123728 */, 17 }, { 0x061dcd59L /* 0.382275916 */, 17 },
 { 0x061e6cf2L /* 0.382428118 */, 17 }, { 0x061f0c8fL /* 0.382580336 */, 17 },
 { 0x061fac2fL /* 0.382732569 */, 17 }, { 0x06204bd4L /* 0.382884817 */, 17 },
 { 0x0620eb7dL /* 0.383037080 */, 17 }, { 0x06218b2aL /* 0.383189358 */, 17 },
 { 0x06222adbL /* 0.383341652 */, 17 }, { 0x0622ca90L /* 0.383493960 */, 17 },
 { 0x06236a49L /* 0.383646284 */, 17 }, { 0x06240a06L /* 0.383798623 */, 17 },

 { 0x0624a9c7L /* 0.383950977 */, 17 }, { 0x0625498dL /* 0.384103346 */, 17 },
 { 0x0625e956L /* 0.384255730 */, 17 }, { 0x06268923L /* 0.384408129 */, 17 },
 { 0x062728f5L /* 0.384560544 */, 17 }, { 0x0627c8caL /* 0.384712973 */, 17 },
 { 0x062868a4L /* 0.384865418 */, 17 }, { 0x06290881L /* 0.385017878 */, 17 },
 { 0x0629a863L /* 0.385170352 */, 17 }, { 0x062a4849L /* 0.385322842 */, 17 },
 { 0x062ae832L /* 0.385475347 */, 17 }, { 0x062b8820L /* 0.385627867 */, 17 },
 { 0x062c2812L /* 0.385780402 */, 17 }, { 0x062cc808L /* 0.385932953 */, 17 },
 { 0x062d6802L /* 0.386085518 */, 17 }, { 0x062e0800L /* 0.386238098 */, 17 },

 { 0x062ea802L /* 0.386390694 */, 17 }, { 0x062f4808L /* 0.386543304 */, 17 },
 { 0x062fe812L /* 0.386695930 */, 17 }, { 0x06308820L /* 0.386848570 */, 17 },
 { 0x06312832L /* 0.387001226 */, 17 }, { 0x0631c849L /* 0.387153897 */, 17 },
 { 0x06326863L /* 0.387306582 */, 17 }, { 0x06330881L /* 0.387459283 */, 17 },
 { 0x0633a8a3L /* 0.387611999 */, 17 }, { 0x063448caL /* 0.387764730 */, 17 },
 { 0x0634e8f4L /* 0.387917476 */, 17 }, { 0x06358923L /* 0.388070237 */, 17 },
 { 0x06362955L /* 0.388223013 */, 17 }, { 0x0636c98cL /* 0.388375804 */, 17 },
 { 0x063769c6L /* 0.388528610 */, 17 }, { 0x06380a05L /* 0.388681431 */, 17 },

 { 0x0638aa48L /* 0.388834268 */, 17 }, { 0x06394a8eL /* 0.388987119 */, 17 },
 { 0x0639ead9L /* 0.389139985 */, 17 }, { 0x063a8b28L /* 0.389292866 */, 17 },
 { 0x063b2b7bL /* 0.389445762 */, 17 }, { 0x063bcbd1L /* 0.389598674 */, 17 },
 { 0x063c6c2cL /* 0.389751600 */, 17 }, { 0x063d0c8bL /* 0.389904541 */, 17 },
 { 0x063daceeL /* 0.390057497 */, 17 }, { 0x063e4d55L /* 0.390210468 */, 17 },
 { 0x063eedc0L /* 0.390363455 */, 17 }, { 0x063f8e2fL /* 0.390516456 */, 17 },
 { 0x06402ea2L /* 0.390669472 */, 17 }, { 0x0640cf19L /* 0.390822503 */, 17 },
 { 0x06416f94L /* 0.390975549 */, 17 }, { 0x06421013L /* 0.391128611 */, 17 },

 { 0x0642b096L /* 0.391281687 */, 17 }, { 0x0643511dL /* 0.391434778 */, 17 },
 { 0x0643f1a8L /* 0.391587884 */, 17 }, { 0x06449237L /* 0.391741005 */, 17 },
 { 0x064532caL /* 0.391894141 */, 17 }, { 0x0645d361L /* 0.392047292 */, 17 },
 { 0x064673fcL /* 0.392200458 */, 17 }, { 0x0647149cL /* 0.392353638 */, 17 },
 { 0x0647b53fL /* 0.392506834 */, 17 }, { 0x064855e6L /* 0.392660045 */, 17 },
 { 0x0648f691L /* 0.392813271 */, 17 }, { 0x06499740L /* 0.392966511 */, 17 },
 { 0x064a37f4L /* 0.393119767 */, 17 }, { 0x064ad8abL /* 0.393273038 */, 17 },
 { 0x064b7966L /* 0.393426323 */, 17 }, { 0x064c1a25L /* 0.393579623 */, 17 },

 { 0x064cbae9L /* 0.393732939 */, 17 }, { 0x064d5bb0L /* 0.393886269 */, 17 },
 { 0x064dfc7bL /* 0.394039614 */, 17 }, { 0x064e9d4bL /* 0.394192974 */, 17 },
 { 0x064f3e1eL /* 0.394346349 */, 17 }, { 0x064fdef5L /* 0.394499739 */, 17 },
 { 0x06507fd0L /* 0.394653144 */, 17 }, { 0x065120b0L /* 0.394806564 */, 17 },
 { 0x0651c193L /* 0.394959999 */, 17 }, { 0x0652627aL /* 0.395113448 */, 17 },
 { 0x06530366L /* 0.395266913 */, 17 }, { 0x0653a455L /* 0.395420392 */, 17 },
 { 0x06544548L /* 0.395573886 */, 17 }, { 0x0654e640L /* 0.395727395 */, 17 },
 { 0x0655873bL /* 0.395880919 */, 17 }, { 0x0656283aL /* 0.396034458 */, 17 },

 { 0x0656c93dL /* 0.396188012 */, 17 }, { 0x06576a45L /* 0.396341581 */, 17 },
 { 0x06580b50L /* 0.396495164 */, 17 }, { 0x0658ac5fL /* 0.396648763 */, 17 },
 { 0x06594d73L /* 0.396802376 */, 17 }, { 0x0659ee8aL /* 0.396956004 */, 17 },
 { 0x065a8fa5L /* 0.397109647 */, 17 }, { 0x065b30c4L /* 0.397263305 */, 17 },
 { 0x065bd1e7L /* 0.397416978 */, 17 }, { 0x065c730fL /* 0.397570666 */, 17 },
 { 0x065d143aL /* 0.397724368 */, 17 }, { 0x065db569L /* 0.397878085 */, 17 },
 { 0x065e569cL /* 0.398031818 */, 17 }, { 0x065ef7d3L /* 0.398185565 */, 17 },
 { 0x065f990eL /* 0.398339326 */, 17 }, { 0x06603a4eL /* 0.398493103 */, 17 },

 { 0x0660db91L /* 0.398646895 */, 17 }, { 0x06617cd8L /* 0.398800701 */, 17 },
 { 0x06621e23L /* 0.398954522 */, 17 }, { 0x0662bf72L /* 0.399108358 */, 17 },
 { 0x066360c5L /* 0.399262209 */, 17 }, { 0x0664021cL /* 0.399416075 */, 17 },
 { 0x0664a377L /* 0.399569955 */, 17 }, { 0x066544d6L /* 0.399723851 */, 17 },
 { 0x0665e639L /* 0.399877761 */, 17 }, { 0x066687a0L /* 0.400031686 */, 17 },
 { 0x0667290bL /* 0.400185625 */, 17 }, { 0x0667ca79L /* 0.400339580 */, 17 },
 { 0x06686becL /* 0.400493549 */, 17 }, { 0x06690d63L /* 0.400647534 */, 17 },
 { 0x0669aedeL /* 0.400801533 */, 17 }, { 0x066a505dL /* 0.400955546 */, 17 },

 { 0x066af1dfL /* 0.401109575 */, 17 }, { 0x066b9366L /* 0.401263618 */, 17 },
 { 0x066c34f1L /* 0.401417676 */, 17 }, { 0x066cd67fL /* 0.401571749 */, 17 },
 { 0x066d7812L /* 0.401725837 */, 17 }, { 0x066e19a9L /* 0.401879939 */, 17 },
 { 0x066ebb43L /* 0.402034056 */, 17 }, { 0x066f5ce2L /* 0.402188188 */, 17 },
 { 0x066ffe84L /* 0.402342335 */, 17 }, { 0x0670a02aL /* 0.402496497 */, 17 },
 { 0x067141d5L /* 0.402650673 */, 17 }, { 0x0671e383L /* 0.402804864 */, 17 },
 { 0x06728535L /* 0.402959070 */, 17 }, { 0x067326ecL /* 0.403113291 */, 17 },
 { 0x0673c8a6L /* 0.403267526 */, 17 }, { 0x06746a64L /* 0.403421776 */, 17 },

 { 0x06750c26L /* 0.403576041 */, 17 }, { 0x0675adecL /* 0.403730320 */, 17 },
 { 0x06764fb6L /* 0.403884615 */, 17 }, { 0x0676f184L /* 0.404038924 */, 17 },
 { 0x06779356L /* 0.404193247 */, 17 }, { 0x0678352cL /* 0.404347586 */, 17 },
 { 0x0678d706L /* 0.404501939 */, 17 }, { 0x067978e4L /* 0.404656307 */, 17 },
 { 0x067a1ac6L /* 0.404810690 */, 17 }, { 0x067abcacL /* 0.404965087 */, 17 },
 { 0x067b5e95L /* 0.405119499 */, 17 }, { 0x067c0083L /* 0.405273926 */, 17 },
 { 0x067ca275L /* 0.405428368 */, 17 }, { 0x067d446aL /* 0.405582824 */, 17 },
 { 0x067de664L /* 0.405737295 */, 17 }, { 0x067e8861L /* 0.405891781 */, 17 },

 { 0x067f2a62L /* 0.406046281 */, 17 }, { 0x067fcc68L /* 0.406200796 */, 17 },
 { 0x06806e71L /* 0.406355326 */, 17 }, { 0x0681107eL /* 0.406509870 */, 17 },
 { 0x0681b28fL /* 0.406664429 */, 17 }, { 0x068254a4L /* 0.406819003 */, 17 },
 { 0x0682f6bdL /* 0.406973592 */, 17 }, { 0x068398daL /* 0.407128195 */, 17 },
 { 0x06843afbL /* 0.407282813 */, 17 }, { 0x0684dd20L /* 0.407437445 */, 17 },
 { 0x06857f49L /* 0.407592093 */, 17 }, { 0x06862176L /* 0.407746754 */, 17 },
 { 0x0686c3a6L /* 0.407901431 */, 17 }, { 0x068765dbL /* 0.408056122 */, 17 },
 { 0x06880814L /* 0.408210828 */, 17 }, { 0x0688aa50L /* 0.408365549 */, 17 },

 { 0x06894c90L /* 0.408520284 */, 17 }, { 0x0689eed5L /* 0.408675034 */, 17 },
 { 0x068a911dL /* 0.408829798 */, 17 }, { 0x068b3369L /* 0.408984577 */, 17 },
 { 0x068bd5b9L /* 0.409139371 */, 17 }, { 0x068c780eL /* 0.409294180 */, 17 },
 { 0x068d1a66L /* 0.409449003 */, 17 }, { 0x068dbcc1L /* 0.409603840 */, 17 },
 { 0x068e5f21L /* 0.409758693 */, 17 }, { 0x068f0185L /* 0.409913560 */, 17 },
 { 0x068fa3edL /* 0.410068441 */, 17 }, { 0x06904658L /* 0.410223338 */, 17 },
 { 0x0690e8c8L /* 0.410378249 */, 17 }, { 0x06918b3cL /* 0.410533174 */, 17 },
 { 0x06922db3L /* 0.410688114 */, 17 }, { 0x0692d02eL /* 0.410843069 */, 17 },

 { 0x069372aeL /* 0.410998038 */, 17 }, { 0x06941531L /* 0.411153022 */, 17 },
 { 0x0694b7b8L /* 0.411308021 */, 17 }, { 0x06955a43L /* 0.411463034 */, 17 },
 { 0x0695fcd2L /* 0.411618062 */, 17 }, { 0x06969f65L /* 0.411773104 */, 17 },
 { 0x069741fbL /* 0.411928161 */, 17 }, { 0x0697e496L /* 0.412083232 */, 17 },
 { 0x06988735L /* 0.412238319 */, 17 }, { 0x069929d7L /* 0.412393419 */, 17 },
 { 0x0699cc7eL /* 0.412548535 */, 17 }, { 0x069a6f28L /* 0.412703664 */, 17 },
 { 0x069b11d6L /* 0.412858809 */, 17 }, { 0x069bb489L /* 0.413013968 */, 17 },
 { 0x069c573fL /* 0.413169142 */, 17 }, { 0x069cf9f9L /* 0.413324330 */, 17 },

 { 0x069d9cb7L /* 0.413479532 */, 17 }, { 0x069e3f78L /* 0.413634750 */, 17 },
 { 0x069ee23eL /* 0.413789982 */, 17 }, { 0x069f8508L /* 0.413945228 */, 17 },
 { 0x06a027d5L /* 0.414100489 */, 17 }, { 0x06a0caa7L /* 0.414255765 */, 17 },
 { 0x06a16d7cL /* 0.414411055 */, 17 }, { 0x06a21055L /* 0.414566359 */, 17 },
 { 0x06a2b333L /* 0.414721679 */, 17 }, { 0x06a35614L /* 0.414877012 */, 17 },
 { 0x06a3f8f9L /* 0.415032361 */, 17 }, { 0x06a49be2L /* 0.415187723 */, 17 },
 { 0x06a53eceL /* 0.415343101 */, 17 }, { 0x06a5e1bfL /* 0.415498493 */, 17 },
 { 0x06a684b4L /* 0.415653899 */, 17 }, { 0x06a727acL /* 0.415809320 */, 17 },

 { 0x06a7caa9L /* 0.415964756 */, 17 }, { 0x06a86da9L /* 0.416120206 */, 17 },
 { 0x06a910adL /* 0.416275670 */, 17 }, { 0x06a9b3b5L /* 0.416431149 */, 17 },
 { 0x06aa56c1L /* 0.416586643 */, 17 }, { 0x06aaf9d1L /* 0.416742151 */, 17 },
 { 0x06ab9ce5L /* 0.416897673 */, 17 }, { 0x06ac3ffcL /* 0.417053210 */, 17 },
 { 0x06ace318L /* 0.417208762 */, 17 }, { 0x06ad8637L /* 0.417364328 */, 17 },
 { 0x06ae295bL /* 0.417519909 */, 17 }, { 0x06aecc82L /* 0.417675504 */, 17 },
 { 0x06af6fadL /* 0.417831113 */, 17 }, { 0x06b012dcL /* 0.417986737 */, 17 },
 { 0x06b0b60fL /* 0.418142376 */, 17 }, { 0x06b15946L /* 0.418298029 */, 17 },

 { 0x06b1fc81L /* 0.418453696 */, 17 }, { 0x06b29fbfL /* 0.418609378 */, 17 },
 { 0x06b34302L /* 0.418765075 */, 17 }, { 0x06b3e648L /* 0.418920786 */, 17 },
 { 0x06b48992L /* 0.419076511 */, 17 }, { 0x06b52ce0L /* 0.419232251 */, 17 },
 { 0x06b5d032L /* 0.419388005 */, 17 }, { 0x06b67388L /* 0.419543774 */, 17 },
 { 0x06b716e2L /* 0.419699557 */, 17 }, { 0x06b7ba3fL /* 0.419855355 */, 17 },
 { 0x06b85da1L /* 0.420011167 */, 17 }, { 0x06b90106L /* 0.420166994 */, 17 },
 { 0x06b9a470L /* 0.420322835 */, 17 }, { 0x06ba47ddL /* 0.420478690 */, 17 },
 { 0x06baeb4eL /* 0.420634560 */, 17 }, { 0x06bb8ec3L /* 0.420790445 */, 17 },

 { 0x06bc323bL /* 0.420946343 */, 17 }, { 0x06bcd5b8L /* 0.421102257 */, 17 },
 { 0x06bd7939L /* 0.421258184 */, 17 }, { 0x06be1cbdL /* 0.421414127 */, 17 },
 { 0x06bec045L /* 0.421570083 */, 17 }, { 0x06bf63d1L /* 0.421726054 */, 17 },
 { 0x06c00761L /* 0.421882040 */, 17 }, { 0x06c0aaf5L /* 0.422038039 */, 17 },
 { 0x06c14e8dL /* 0.422194054 */, 17 }, { 0x06c1f229L /* 0.422350082 */, 17 },
 { 0x06c295c8L /* 0.422506125 */, 17 }, { 0x06c3396cL /* 0.422662183 */, 17 },
 { 0x06c3dd13L /* 0.422818255 */, 17 }, { 0x06c480beL /* 0.422974341 */, 17 },
 { 0x06c5246dL /* 0.423130442 */, 17 }, { 0x06c5c820L /* 0.423286557 */, 17 },

 { 0x06c66bd6L /* 0.423442686 */, 17 }, { 0x06c70f91L /* 0.423598830 */, 17 },
 { 0x06c7b34fL /* 0.423754988 */, 17 }, { 0x06c85712L /* 0.423911161 */, 17 },
 { 0x06c8fad8L /* 0.424067348 */, 17 }, { 0x06c99ea2L /* 0.424223550 */, 17 },
 { 0x06ca4270L /* 0.424379765 */, 17 }, { 0x06cae641L /* 0.424535996 */, 17 },
 { 0x06cb8a17L /* 0.424692240 */, 17 }, { 0x06cc2df0L /* 0.424848499 */, 17 },
 { 0x06ccd1ceL /* 0.425004772 */, 17 }, { 0x06cd75afL /* 0.425161060 */, 17 },
 { 0x06ce1994L /* 0.425317362 */, 17 }, { 0x06cebd7dL /* 0.425473678 */, 17 },
 { 0x06cf6169L /* 0.425630009 */, 17 }, { 0x06d0055aL /* 0.425786354 */, 17 },

 { 0x06d0a94eL /* 0.425942714 */, 17 }, { 0x06d14d47L /* 0.426099088 */, 17 },
 { 0x06d1f143L /* 0.426255476 */, 17 }, { 0x06d29543L /* 0.426411878 */, 17 },
 { 0x06d33947L /* 0.426568295 */, 17 }, { 0x06d3dd4eL /* 0.426724726 */, 17 },
 { 0x06d4815aL /* 0.426881172 */, 17 }, { 0x06d52569L /* 0.427037632 */, 17 },
 { 0x06d5c97cL /* 0.427194106 */, 17 }, { 0x06d66d93L /* 0.427350594 */, 17 },
 { 0x06d711aeL /* 0.427507097 */, 17 }, { 0x06d7b5cdL /* 0.427663614 */, 17 },
 { 0x06d859f0L /* 0.427820146 */, 17 }, { 0x06d8fe16L /* 0.427976692 */, 17 },
 { 0x06d9a240L /* 0.428133252 */, 17 }, { 0x06da466fL /* 0.428289826 */, 17 },

 { 0x06daeaa1L /* 0.428446415 */, 17 }, { 0x06db8ed6L /* 0.428603018 */, 17 },
 { 0x06dc3310L /* 0.428759635 */, 17 }, { 0x06dcd74dL /* 0.428916267 */, 17 },
 { 0x06dd7b8fL /* 0.429072913 */, 17 }, { 0x06de1fd4L /* 0.429229573 */, 17 },
 { 0x06dec41dL /* 0.429386248 */, 17 }, { 0x06df686aL /* 0.429542937 */, 17 },
 { 0x06e00cbbL /* 0.429699640 */, 17 }, { 0x06e0b10fL /* 0.429856357 */, 17 },
 { 0x06e15567L /* 0.430013089 */, 17 }, { 0x06e1f9c4L /* 0.430169835 */, 17 },
 { 0x06e29e24L /* 0.430326595 */, 17 }, { 0x06e34287L /* 0.430483370 */, 17 },
 { 0x06e3e6efL /* 0.430640159 */, 17 }, { 0x06e48b5bL /* 0.430796962 */, 17 },

 { 0x06e52fcaL /* 0.430953779 */, 17 }, { 0x06e5d43dL /* 0.431110611 */, 17 },
 { 0x06e678b4L /* 0.431267457 */, 17 }, { 0x06e71d2fL /* 0.431424317 */, 17 },
 { 0x06e7c1aeL /* 0.431581192 */, 17 }, { 0x06e86630L /* 0.431738080 */, 17 },
 { 0x06e90ab7L /* 0.431894983 */, 17 }, { 0x06e9af41L /* 0.432051900 */, 17 },
 { 0x06ea53cfL /* 0.432208832 */, 17 }, { 0x06eaf860L /* 0.432365778 */, 17 },
 { 0x06eb9cf6L /* 0.432522737 */, 17 }, { 0x06ec418fL /* 0.432679712 */, 17 },
 { 0x06ece62dL /* 0.432836700 */, 17 }, { 0x06ed8aceL /* 0.432993703 */, 17 },
 { 0x06ee2f73L /* 0.433150720 */, 17 }, { 0x06eed41bL /* 0.433307751 */, 17 },

 { 0x06ef78c8L /* 0.433464796 */, 17 }, { 0x06f01d78L /* 0.433621856 */, 17 },
 { 0x06f0c22cL /* 0.433778929 */, 17 }, { 0x06f166e4L /* 0.433936017 */, 17 },
 { 0x06f20ba0L /* 0.434093120 */, 17 }, { 0x06f2b060L /* 0.434250236 */, 17 },
 { 0x06f35523L /* 0.434407367 */, 17 }, { 0x06f3f9ebL /* 0.434564512 */, 17 },
 { 0x06f49eb6L /* 0.434721671 */, 17 }, { 0x06f54385L /* 0.434878844 */, 17 },
 { 0x06f5e857L /* 0.435036032 */, 17 }, { 0x06f68d2eL /* 0.435193233 */, 17 },
 { 0x06f73208L /* 0.435350449 */, 17 }, { 0x06f7d6e6L /* 0.435507679 */, 17 },
 { 0x06f87bc8L /* 0.435664924 */, 17 }, { 0x06f920aeL /* 0.435822182 */, 17 },

 { 0x06f9c597L /* 0.435979455 */, 17 }, { 0x06fa6a85L /* 0.436136741 */, 17 },
 { 0x06fb0f76L /* 0.436294042 */, 17 }, { 0x06fbb46bL /* 0.436451358 */, 17 },
 { 0x06fc5964L /* 0.436608687 */, 17 }, { 0x06fcfe60L /* 0.436766031 */, 17 },
 { 0x06fda361L /* 0.436923388 */, 17 }, { 0x06fe4865L /* 0.437080760 */, 17 },
 { 0x06feed6dL /* 0.437238146 */, 17 }, { 0x06ff9279L /* 0.437395547 */, 17 },
 { 0x07003788L /* 0.437552961 */, 17 }, { 0x0700dc9cL /* 0.437710389 */, 17 },
 { 0x070181b3L /* 0.437867832 */, 17 }, { 0x070226ceL /* 0.438025289 */, 17 },
 { 0x0702cbedL /* 0.438182760 */, 17 }, { 0x0703710fL /* 0.438340245 */, 17 },

 { 0x07041636L /* 0.438497744 */, 17 }, { 0x0704bb60L /* 0.438655258 */, 17 },
 { 0x0705608eL /* 0.438812785 */, 17 }, { 0x070605c0L /* 0.438970327 */, 17 },
 { 0x0706aaf5L /* 0.439127883 */, 17 }, { 0x0707502fL /* 0.439285453 */, 17 },
 { 0x0707f56cL /* 0.439443037 */, 17 }, { 0x07089aadL /* 0.439600635 */, 17 },
 { 0x07093ff2L /* 0.439758248 */, 17 }, { 0x0709e53aL /* 0.439915874 */, 17 },
 { 0x070a8a86L /* 0.440073515 */, 17 }, { 0x070b2fd7L /* 0.440231170 */, 17 },
 { 0x070bd52aL /* 0.440388839 */, 17 }, { 0x070c7a82L /* 0.440546522 */, 17 },
 { 0x070d1fdeL /* 0.440704219 */, 17 }, { 0x070dc53dL /* 0.440861930 */, 17 },

 { 0x070e6aa0L /* 0.441019655 */, 17 }, { 0x070f1007L /* 0.441177395 */, 17 },
 { 0x070fb571L /* 0.441335148 */, 17 }, { 0x07105ae0L /* 0.441492916 */, 17 },
 { 0x07110052L /* 0.441650697 */, 17 }, { 0x0711a5c8L /* 0.441808493 */, 17 },
 { 0x07124b42L /* 0.441966303 */, 17 }, { 0x0712f0bfL /* 0.442124127 */, 17 },
 { 0x07139641L /* 0.442281965 */, 17 }, { 0x07143bc6L /* 0.442439817 */, 17 },
 { 0x0714e14fL /* 0.442597683 */, 17 }, { 0x071586dbL /* 0.442755564 */, 17 },
 { 0x07162c6cL /* 0.442913458 */, 17 }, { 0x0716d200L /* 0.443071366 */, 17 },
 { 0x07177798L /* 0.443229289 */, 17 }, { 0x07181d34L /* 0.443387226 */, 17 },

 { 0x0718c2d3L /* 0.443545176 */, 17 }, { 0x07196877L /* 0.443703141 */, 17 },
 { 0x071a0e1eL /* 0.443861120 */, 17 }, { 0x071ab3c9L /* 0.444019113 */, 17 },
 { 0x071b5977L /* 0.444177119 */, 17 }, { 0x071bff2aL /* 0.444335140 */, 17 },
 { 0x071ca4e0L /* 0.444493175 */, 17 }, { 0x071d4a9aL /* 0.444651224 */, 17 },
 { 0x071df058L /* 0.444809288 */, 17 }, { 0x071e9619L /* 0.444967365 */, 17 },
 { 0x071f3bdeL /* 0.445125456 */, 17 }, { 0x071fe1a8L /* 0.445283561 */, 17 },
 { 0x07208774L /* 0.445441680 */, 17 }, { 0x07212d45L /* 0.445599814 */, 17 },
 { 0x0721d319L /* 0.445757961 */, 17 }, { 0x072278f1L /* 0.445916122 */, 17 },

 { 0x07231ecdL /* 0.446074298 */, 17 }, { 0x0723c4adL /* 0.446232487 */, 17 },
 { 0x07246a90L /* 0.446390690 */, 17 }, { 0x07251077L /* 0.446548908 */, 17 },
 { 0x0725b662L /* 0.446707139 */, 17 }, { 0x07265c51L /* 0.446865385 */, 17 },
 { 0x07270244L /* 0.447023644 */, 17 }, { 0x0727a83aL /* 0.447181918 */, 17 },
 { 0x07284e34L /* 0.447340205 */, 17 }, { 0x0728f431L /* 0.447498507 */, 17 },
 { 0x07299a33L /* 0.447656822 */, 17 }, { 0x072a4038L /* 0.447815152 */, 17 },
 { 0x072ae641L /* 0.447973495 */, 17 }, { 0x072b8c4eL /* 0.448131853 */, 17 },
 { 0x072c325eL /* 0.448290224 */, 17 }, { 0x072cd873L /* 0.448448609 */, 17 },

 { 0x072d7e8bL /* 0.448607009 */, 17 }, { 0x072e24a7L /* 0.448765422 */, 17 },
 { 0x072ecac6L /* 0.448923850 */, 17 }, { 0x072f70e9L /* 0.449082291 */, 17 },
 { 0x07301710L /* 0.449240746 */, 17 }, { 0x0730bd3bL /* 0.449399216 */, 17 },
 { 0x0731636aL /* 0.449557699 */, 17 }, { 0x0732099cL /* 0.449716196 */, 17 },
 { 0x0732afd2L /* 0.449874708 */, 17 }, { 0x0733560cL /* 0.450033233 */, 17 },
 { 0x0733fc49L /* 0.450191772 */, 17 }, { 0x0734a28bL /* 0.450350325 */, 17 },
 { 0x073548d0L /* 0.450508892 */, 17 }, { 0x0735ef18L /* 0.450667473 */, 17 },
 { 0x07369565L /* 0.450826068 */, 17 }, { 0x07373bb5L /* 0.450984677 */, 17 },

 { 0x0737e209L /* 0.451143300 */, 17 }, { 0x07388861L /* 0.451301937 */, 17 },
 { 0x07392ebcL /* 0.451460588 */, 17 }, { 0x0739d51cL /* 0.451619252 */, 17 },
 { 0x073a7b7fL /* 0.451777931 */, 17 }, { 0x073b21e5L /* 0.451936623 */, 17 },
 { 0x073bc850L /* 0.452095330 */, 17 }, { 0x073c6ebeL /* 0.452254050 */, 17 },
 { 0x073d1530L /* 0.452412785 */, 17 }, { 0x073dbba6L /* 0.452571533 */, 17 },
 { 0x073e621fL /* 0.452730295 */, 17 }, { 0x073f089cL /* 0.452889071 */, 17 },
 { 0x073faf1dL /* 0.453047861 */, 17 }, { 0x074055a2L /* 0.453206665 */, 17 },
 { 0x0740fc2aL /* 0.453365483 */, 17 }, { 0x0741a2b6L /* 0.453524315 */, 17 },

 { 0x07424946L /* 0.453683161 */, 17 }, { 0x0742efd9L /* 0.453842020 */, 17 },
 { 0x07439671L /* 0.454000894 */, 17 }, { 0x07443d0cL /* 0.454159781 */, 17 },
 { 0x0744e3aaL /* 0.454318683 */, 17 }, { 0x07458a4dL /* 0.454477598 */, 17 },
 { 0x074630f3L /* 0.454636527 */, 17 }, { 0x0746d79dL /* 0.454795470 */, 17 },
 { 0x07477e4bL /* 0.454954427 */, 17 }, { 0x074824fcL /* 0.455113397 */, 17 },
 { 0x0748cbb1L /* 0.455272382 */, 17 }, { 0x0749726aL /* 0.455431381 */, 17 },
 { 0x074a1927L /* 0.455590393 */, 17 }, { 0x074abfe7L /* 0.455749419 */, 17 },
 { 0x074b66abL /* 0.455908459 */, 17 }, { 0x074c0d73L /* 0.456067513 */, 17 },

 { 0x074cb43eL /* 0.456226581 */, 17 }, { 0x074d5b0dL /* 0.456385663 */, 17 },
 { 0x074e01e0L /* 0.456544759 */, 17 }, { 0x074ea8b7L /* 0.456703868 */, 17 },
 { 0x074f4f91L /* 0.456862992 */, 17 }, { 0x074ff66fL /* 0.457022129 */, 17 },
 { 0x07509d51L /* 0.457181280 */, 17 }, { 0x07514437L /* 0.457340445 */, 17 },
 { 0x0751eb20L /* 0.457499623 */, 17 }, { 0x0752920dL /* 0.457658816 */, 17 },
 { 0x075338fdL /* 0.457818022 */, 17 }, { 0x0753dff2L /* 0.457977243 */, 17 },
 { 0x075486eaL /* 0.458136477 */, 17 }, { 0x07552de6L /* 0.458295725 */, 17 },
 { 0x0755d4e5L /* 0.458454987 */, 17 }, { 0x07567be8L /* 0.458614262 */, 17 },

 { 0x075722efL /* 0.458773552 */, 17 }, { 0x0757c9faL /* 0.458932855 */, 17 },
 { 0x07587108L /* 0.459092172 */, 17 }, { 0x0759181aL /* 0.459251503 */, 17 },
 { 0x0759bf30L /* 0.459410848 */, 17 }, { 0x075a664aL /* 0.459570206 */, 17 },
 { 0x075b0d67L /* 0.459729579 */, 17 }, { 0x075bb488L /* 0.459888965 */, 17 },
 { 0x075c5bacL /* 0.460048365 */, 17 }, { 0x075d02d5L /* 0.460207779 */, 17 },
 { 0x075daa01L /* 0.460367206 */, 17 }, { 0x075e5130L /* 0.460526648 */, 17 },
 { 0x075ef864L /* 0.460686103 */, 17 }, { 0x075f9f9bL /* 0.460845572 */, 17 },
 { 0x076046d6L /* 0.461005055 */, 17 }, { 0x0760ee14L /* 0.461164552 */, 17 },

 { 0x07619557L /* 0.461324062 */, 17 }, { 0x07623c9dL /* 0.461483586 */, 17 },
 { 0x0762e3e6L /* 0.461643124 */, 17 }, { 0x07638b34L /* 0.461802676 */, 17 },
 { 0x07643285L /* 0.461962242 */, 17 }, { 0x0764d9d9L /* 0.462121821 */, 17 },
 { 0x07658132L /* 0.462281414 */, 17 }, { 0x0766288eL /* 0.462441021 */, 17 },
 { 0x0766cfeeL /* 0.462600642 */, 17 }, { 0x07677751L /* 0.462760276 */, 17 },
 { 0x07681eb9L /* 0.462919924 */, 17 }, { 0x0768c624L /* 0.463079586 */, 17 },
 { 0x07696d92L /* 0.463239262 */, 17 }, { 0x076a1505L /* 0.463398951 */, 17 },
 { 0x076abc7bL /* 0.463558655 */, 17 }, { 0x076b63f4L /* 0.463718372 */, 17 },

 { 0x076c0b72L /* 0.463878102 */, 17 }, { 0x076cb2f3L /* 0.464037847 */, 17 },
 { 0x076d5a78L /* 0.464197605 */, 17 }, { 0x076e0200L /* 0.464357377 */, 17 },
 { 0x076ea98cL /* 0.464517163 */, 17 }, { 0x076f511cL /* 0.464676962 */, 17 },
 { 0x076ff8b0L /* 0.464836776 */, 17 }, { 0x0770a047L /* 0.464996603 */, 17 },
 { 0x077147e2L /* 0.465156443 */, 17 }, { 0x0771ef80L /* 0.465316298 */, 17 },
 { 0x07729723L /* 0.465476166 */, 17 }, { 0x07733ec9L /* 0.465636048 */, 17 },
 { 0x0773e672L /* 0.465795943 */, 17 }, { 0x07748e20L /* 0.465955853 */, 17 },
 { 0x077535d1L /* 0.466115776 */, 17 }, { 0x0775dd85L /* 0.466275713 */, 17 },

 { 0x0776853eL /* 0.466435663 */, 17 }, { 0x07772cfaL /* 0.466595627 */, 17 },
 { 0x0777d4baL /* 0.466755605 */, 17 }, { 0x07787c7dL /* 0.466915597 */, 17 },
 { 0x07792444L /* 0.467075602 */, 17 }, { 0x0779cc0fL /* 0.467235621 */, 17 },
 { 0x077a73ddL /* 0.467395654 */, 17 }, { 0x077b1bafL /* 0.467555701 */, 17 },
 { 0x077bc385L /* 0.467715761 */, 17 }, { 0x077c6b5fL /* 0.467875835 */, 17 },
 { 0x077d133cL /* 0.468035922 */, 17 }, { 0x077dbb1dL /* 0.468196023 */, 17 },
 { 0x077e6301L /* 0.468356138 */, 17 }, { 0x077f0ae9L /* 0.468516267 */, 17 },
 { 0x077fb2d5L /* 0.468676409 */, 17 }, { 0x07805ac5L /* 0.468836565 */, 17 },

 { 0x078102b8L /* 0.468996735 */, 17 }, { 0x0781aaafL /* 0.469156918 */, 17 },
 { 0x078252aaL /* 0.469317115 */, 17 }, { 0x0782faa8L /* 0.469477326 */, 17 },
 { 0x0783a2aaL /* 0.469637550 */, 17 }, { 0x07844aafL /* 0.469797788 */, 17 },
 { 0x0784f2b8L /* 0.469958040 */, 17 }, { 0x07859ac5L /* 0.470118305 */, 17 },
 { 0x078642d6L /* 0.470278584 */, 17 }, { 0x0786eaeaL /* 0.470438877 */, 17 },
 { 0x07879302L /* 0.470599183 */, 17 }, { 0x07883b1eL /* 0.470759503 */, 17 },
 { 0x0788e33dL /* 0.470919836 */, 17 }, { 0x07898b60L /* 0.471080184 */, 17 },
 { 0x078a3386L /* 0.471240545 */, 17 }, { 0x078adbb0L /* 0.471400919 */, 17 },

 { 0x078b83deL /* 0.471561307 */, 17 }, { 0x078c2c10L /* 0.471721709 */, 17 },
 { 0x078cd445L /* 0.471882125 */, 17 }, { 0x078d7c7eL /* 0.472042554 */, 17 },
 { 0x078e24baL /* 0.472202996 */, 17 }, { 0x078eccfbL /* 0.472363453 */, 17 },
 { 0x078f753eL /* 0.472523923 */, 17 }, { 0x07901d86L /* 0.472684406 */, 17 },
 { 0x0790c5d1L /* 0.472844904 */, 17 }, { 0x07916e20L /* 0.473005414 */, 17 },
 { 0x07921672L /* 0.473165939 */, 17 }, { 0x0792bec8L /* 0.473326477 */, 17 },
 { 0x07936722L /* 0.473487029 */, 17 }, { 0x07940f80L /* 0.473647594 */, 17 },
 { 0x0794b7e1L /* 0.473808173 */, 17 }, { 0x07956045L /* 0.473968765 */, 17 },

 { 0x079608aeL /* 0.474129372 */, 17 }, { 0x0796b11aL /* 0.474289991 */, 17 },
 { 0x0797598aL /* 0.474450625 */, 17 }, { 0x079801fdL /* 0.474611272 */, 17 },
 { 0x0798aa74L /* 0.474771932 */, 17 }, { 0x079952eeL /* 0.474932606 */, 17 },
 { 0x0799fb6dL /* 0.475093294 */, 17 }, { 0x079aa3efL /* 0.475253995 */, 17 },
 { 0x079b4c74L /* 0.475414710 */, 17 }, { 0x079bf4fdL /* 0.475575439 */, 17 },
 { 0x079c9d8aL /* 0.475736181 */, 17 }, { 0x079d461bL /* 0.475896936 */, 17 },
 { 0x079deeafL /* 0.476057705 */, 17 }, { 0x079e9747L /* 0.476218488 */, 17 },
 { 0x079f3fe2L /* 0.476379285 */, 17 }, { 0x079fe881L /* 0.476540095 */, 17 },

 { 0x07a09124L /* 0.476700918 */, 17 }, { 0x07a139caL /* 0.476861755 */, 17 },
 { 0x07a1e274L /* 0.477022606 */, 17 }, { 0x07a28b22L /* 0.477183470 */, 17 },
 { 0x07a333d3L /* 0.477344348 */, 17 }, { 0x07a3dc88L /* 0.477505239 */, 17 },
 { 0x07a48541L /* 0.477666144 */, 17 }, { 0x07a52dfdL /* 0.477827062 */, 17 },
 { 0x07a5d6bdL /* 0.477987994 */, 17 }, { 0x07a67f80L /* 0.478148940 */, 17 },
 { 0x07a72847L /* 0.478309899 */, 17 }, { 0x07a7d112L /* 0.478470871 */, 17 },
 { 0x07a879e1L /* 0.478631857 */, 17 }, { 0x07a922b3L /* 0.478792857 */, 17 },
 { 0x07a9cb88L /* 0.478953870 */, 17 }, { 0x07aa7462L /* 0.479114897 */, 17 },

 { 0x07ab1d3eL /* 0.479275937 */, 17 }, { 0x07abc61fL /* 0.479436991 */, 17 },
 { 0x07ac6f03L /* 0.479598058 */, 17 }, { 0x07ad17ebL /* 0.479759139 */, 17 },
 { 0x07adc0d6L /* 0.479920233 */, 17 }, { 0x07ae69c6L /* 0.480081341 */, 17 },
 { 0x07af12b8L /* 0.480242463 */, 17 }, { 0x07afbbafL /* 0.480403598 */, 17 },
 { 0x07b064a8L /* 0.480564746 */, 17 }, { 0x07b10da6L /* 0.480725908 */, 17 },
 { 0x07b1b6a7L /* 0.480887083 */, 17 }, { 0x07b25facL /* 0.481048272 */, 17 },
 { 0x07b308b5L /* 0.481209475 */, 17 }, { 0x07b3b1c1L /* 0.481370691 */, 17 },
 { 0x07b45ad0L /* 0.481531920 */, 17 }, { 0x07b503e4L /* 0.481693163 */, 17 },

 { 0x07b5acfbL /* 0.481854420 */, 17 }, { 0x07b65615L /* 0.482015690 */, 17 },
 { 0x07b6ff33L /* 0.482176973 */, 17 }, { 0x07b7a855L /* 0.482338270 */, 17 },
 { 0x07b8517bL /* 0.482499580 */, 17 }, { 0x07b8faa4L /* 0.482660904 */, 17 },
 { 0x07b9a3d0L /* 0.482822242 */, 17 }, { 0x07ba4d01L /* 0.482983592 */, 17 },
 { 0x07baf635L /* 0.483144957 */, 17 }, { 0x07bb9f6cL /* 0.483306335 */, 17 },
 { 0x07bc48a7L /* 0.483467726 */, 17 }, { 0x07bcf1e6L /* 0.483629131 */, 17 },
 { 0x07bd9b28L /* 0.483790549 */, 17 }, { 0x07be446eL /* 0.483951980 */, 17 },
 { 0x07beedb8L /* 0.484113426 */, 17 }, { 0x07bf9705L /* 0.484274884 */, 17 },

 { 0x07c04056L /* 0.484436356 */, 17 }, { 0x07c0e9aaL /* 0.484597842 */, 17 },
 { 0x07c19302L /* 0.484759341 */, 17 }, { 0x07c23c5eL /* 0.484920853 */, 17 },
 { 0x07c2e5bdL /* 0.485082379 */, 17 }, { 0x07c38f20L /* 0.485243918 */, 17 },
 { 0x07c43887L /* 0.485405471 */, 17 }, { 0x07c4e1f1L /* 0.485567037 */, 17 },
 { 0x07c58b5fL /* 0.485728617 */, 17 }, { 0x07c634d0L /* 0.485890210 */, 17 },
 { 0x07c6de45L /* 0.486051817 */, 17 }, { 0x07c787bdL /* 0.486213436 */, 17 },
 { 0x07c83139L /* 0.486375070 */, 17 }, { 0x07c8dab9L /* 0.486536717 */, 17 },
 { 0x07c9843cL /* 0.486698377 */, 17 }, { 0x07ca2dc3L /* 0.486860051 */, 17 },

 { 0x07cad74eL /* 0.487021738 */, 17 }, { 0x07cb80dcL /* 0.487183438 */, 17 },
 { 0x07cc2a6eL /* 0.487345152 */, 17 }, { 0x07ccd403L /* 0.487506879 */, 17 },
 { 0x07cd7d9cL /* 0.487668620 */, 17 }, { 0x07ce2739L /* 0.487830374 */, 17 },
 { 0x07ced0d9L /* 0.487992142 */, 17 }, { 0x07cf7a7dL /* 0.488153923 */, 17 },
 { 0x07d02424L /* 0.488315717 */, 17 }, { 0x07d0cdcfL /* 0.488477525 */, 17 },
 { 0x07d1777eL /* 0.488639346 */, 17 }, { 0x07d22130L /* 0.488801181 */, 17 },
 { 0x07d2cae5L /* 0.488963029 */, 17 }, { 0x07d3749fL /* 0.489124890 */, 17 },
 { 0x07d41e5cL /* 0.489286765 */, 17 }, { 0x07d4c81cL /* 0.489448653 */, 17 },

 { 0x07d571e0L /* 0.489610555 */, 17 }, { 0x07d61ba8L /* 0.489772470 */, 17 },
 { 0x07d6c573L /* 0.489934398 */, 17 }, { 0x07d76f42L /* 0.490096340 */, 17 },
 { 0x07d81915L /* 0.490258295 */, 17 }, { 0x07d8c2ebL /* 0.490420263 */, 17 },
 { 0x07d96cc4L /* 0.490582245 */, 17 }, { 0x07da16a2L /* 0.490744240 */, 17 },
 { 0x07dac083L /* 0.490906249 */, 17 }, { 0x07db6a67L /* 0.491068271 */, 17 },
 { 0x07dc144fL /* 0.491230306 */, 17 }, { 0x07dcbe3bL /* 0.491392355 */, 17 },
 { 0x07dd682aL /* 0.491554417 */, 17 }, { 0x07de121dL /* 0.491716492 */, 17 },
 { 0x07debc13L /* 0.491878581 */, 17 }, { 0x07df660dL /* 0.492040683 */, 17 },

 { 0x07e0100aL /* 0.492202799 */, 17 }, { 0x07e0ba0cL /* 0.492364928 */, 17 },
 { 0x07e16410L /* 0.492527070 */, 17 }, { 0x07e20e19L /* 0.492689225 */, 17 },
 { 0x07e2b824L /* 0.492851394 */, 17 }, { 0x07e36234L /* 0.493013576 */, 17 },
 { 0x07e40c47L /* 0.493175772 */, 17 }, { 0x07e4b65eL /* 0.493337981 */, 17 },
 { 0x07e56078L /* 0.493500203 */, 17 }, { 0x07e60a95L /* 0.493662438 */, 17 },
 { 0x07e6b4b7L /* 0.493824687 */, 17 }, { 0x07e75edcL /* 0.493986949 */, 17 },
 { 0x07e80904L /* 0.494149225 */, 17 }, { 0x07e8b330L /* 0.494311514 */, 17 },
 { 0x07e95d60L /* 0.494473816 */, 17 }, { 0x07ea0793L /* 0.494636131 */, 17 },

 { 0x07eab1caL /* 0.494798460 */, 17 }, { 0x07eb5c04L /* 0.494960802 */, 17 },
 { 0x07ec0642L /* 0.495123158 */, 17 }, { 0x07ecb084L /* 0.495285526 */, 17 },
 { 0x07ed5ac9L /* 0.495447908 */, 17 }, { 0x07ee0512L /* 0.495610304 */, 17 },
 { 0x07eeaf5eL /* 0.495772712 */, 17 }, { 0x07ef59aeL /* 0.495935134 */, 17 },
 { 0x07f00401L /* 0.496097570 */, 17 }, { 0x07f0ae58L /* 0.496260018 */, 17 },
 { 0x07f158b3L /* 0.496422480 */, 17 }, { 0x07f20311L /* 0.496584955 */, 17 },
 { 0x07f2ad72L /* 0.496747444 */, 17 }, { 0x07f357d8L /* 0.496909945 */, 17 },
 { 0x07f40240L /* 0.497072460 */, 17 }, { 0x07f4acadL /* 0.497234989 */, 17 },

 { 0x07f5571dL /* 0.497397530 */, 17 }, { 0x07f60190L /* 0.497560085 */, 17 },
 { 0x07f6ac07L /* 0.497722653 */, 17 }, { 0x07f75682L /* 0.497885235 */, 17 },
 { 0x07f80100L /* 0.498047829 */, 17 }, { 0x07f8ab82L /* 0.498210437 */, 17 },
 { 0x07f95607L /* 0.498373058 */, 17 }, { 0x07fa0090L /* 0.498535693 */, 17 },
 { 0x07faab1cL /* 0.498698341 */, 17 }, { 0x07fb55acL /* 0.498861002 */, 17 },
 { 0x07fc0040L /* 0.499023676 */, 17 }, { 0x07fcaad7L /* 0.499186364 */, 17 },
 { 0x07fd5572L /* 0.499349064 */, 17 }, { 0x07fe0010L /* 0.499511778 */, 17 },
 { 0x07feaab2L /* 0.499674506 */, 17 }, { 0x07ff5557L /* 0.499837246 */, 17 },

 { 0x04000000L /* 0.250000000 */, 18 }, { 0x04005556L /* 0.250081384 */, 18 },
 { 0x0400aaaeL /* 0.250162774 */, 18 }, { 0x04010008L /* 0.250244170 */, 18 },
 { 0x04015563L /* 0.250325574 */, 18 }, { 0x0401aac1L /* 0.250406984 */, 18 },
 { 0x04020020L /* 0.250488400 */, 18 }, { 0x04025581L /* 0.250569824 */, 18 },
 { 0x0402aae3L /* 0.250651254 */, 18 }, { 0x04030048L /* 0.250732690 */, 18 },
 { 0x040355aeL /* 0.250814133 */, 18 }, { 0x0403ab16L /* 0.250895583 */, 18 },
 { 0x04040080L /* 0.250977039 */, 18 }, { 0x040455ebL /* 0.251058502 */, 18 },
 { 0x0404ab59L /* 0.251139971 */, 18 }, { 0x040500c8L /* 0.251221448 */, 18 },

 { 0x04055638L /* 0.251302930 */, 18 }, { 0x0405ababL /* 0.251384420 */, 18 },
 { 0x0406011fL /* 0.251465916 */, 18 }, { 0x04065696L /* 0.251547418 */, 18 },
 { 0x0406ac0eL /* 0.251628927 */, 18 }, { 0x04070187L /* 0.251710443 */, 18 },
 { 0x04075703L /* 0.251791965 */, 18 }, { 0x0407ac80L /* 0.251873494 */, 18 },
 { 0x040801ffL /* 0.251955030 */, 18 }, { 0x04085780L /* 0.252036572 */, 18 },
 { 0x0408ad02L /* 0.252118121 */, 18 }, { 0x04090287L /* 0.252199676 */, 18 },
 { 0x0409580dL /* 0.252281238 */, 18 }, { 0x0409ad95L /* 0.252362807 */, 18 },
 { 0x040a031eL /* 0.252444382 */, 18 }, { 0x040a58aaL /* 0.252525963 */, 18 },

 { 0x040aae37L /* 0.252607552 */, 18 }, { 0x040b03c6L /* 0.252689147 */, 18 },
 { 0x040b5957L /* 0.252770748 */, 18 }, { 0x040baee9L /* 0.252852356 */, 18 },
 { 0x040c047eL /* 0.252933971 */, 18 }, { 0x040c5a14L /* 0.253015592 */, 18 },
 { 0x040cafabL /* 0.253097220 */, 18 }, { 0x040d0545L /* 0.253178854 */, 18 },
 { 0x040d5ae0L /* 0.253260495 */, 18 }, { 0x040db07dL /* 0.253342143 */, 18 },
 { 0x040e061cL /* 0.253423797 */, 18 }, { 0x040e5bbdL /* 0.253505457 */, 18 },
 { 0x040eb15fL /* 0.253587125 */, 18 }, { 0x040f0703L /* 0.253668799 */, 18 },
 { 0x040f5ca9L /* 0.253750479 */, 18 }, { 0x040fb251L /* 0.253832166 */, 18 },

 { 0x041007faL /* 0.253913860 */, 18 }, { 0x04105da6L /* 0.253995560 */, 18 },
 { 0x0410b353L /* 0.254077266 */, 18 }, { 0x04110901L /* 0.254158980 */, 18 },
 { 0x04115eb2L /* 0.254240700 */, 18 }, { 0x0411b464L /* 0.254322426 */, 18 },
 { 0x04120a18L /* 0.254404159 */, 18 }, { 0x04125fceL /* 0.254485899 */, 18 },
 { 0x0412b586L /* 0.254567645 */, 18 }, { 0x04130b3fL /* 0.254649397 */, 18 },
 { 0x041360faL /* 0.254731157 */, 18 }, { 0x0413b6b7L /* 0.254812922 */, 18 },
 { 0x04140c75L /* 0.254894695 */, 18 }, { 0x04146236L /* 0.254976474 */, 18 },
 { 0x0414b7f8L /* 0.255058259 */, 18 }, { 0x04150dbcL /* 0.255140051 */, 18 },

 { 0x04156381L /* 0.255221850 */, 18 }, { 0x0415b949L /* 0.255303655 */, 18 },
 { 0x04160f12L /* 0.255385467 */, 18 }, { 0x041664ddL /* 0.255467285 */, 18 },
 { 0x0416baaaL /* 0.255549110 */, 18 }, { 0x04171078L /* 0.255630941 */, 18 },
 { 0x04176648L /* 0.255712779 */, 18 }, { 0x0417bc1aL /* 0.255794624 */, 18 },
 { 0x041811eeL /* 0.255876475 */, 18 }, { 0x041867c3L /* 0.255958332 */, 18 },
 { 0x0418bd9bL /* 0.256040196 */, 18 }, { 0x04191374L /* 0.256122067 */, 18 },
 { 0x0419694eL /* 0.256203944 */, 18 }, { 0x0419bf2bL /* 0.256285828 */, 18 },
 { 0x041a1509L /* 0.256367718 */, 18 }, { 0x041a6ae9L /* 0.256449615 */, 18 },

 { 0x041ac0cbL /* 0.256531518 */, 18 }, { 0x041b16aeL /* 0.256613428 */, 18 },
 { 0x041b6c94L /* 0.256695344 */, 18 }, { 0x041bc27bL /* 0.256777267 */, 18 },
 { 0x041c1863L /* 0.256859197 */, 18 }, { 0x041c6e4eL /* 0.256941133 */, 18 },
 { 0x041cc43aL /* 0.257023076 */, 18 }, { 0x041d1a28L /* 0.257105025 */, 18 },
 { 0x041d7018L /* 0.257186980 */, 18 }, { 0x041dc60aL /* 0.257268942 */, 18 },
 { 0x041e1bfdL /* 0.257350911 */, 18 }, { 0x041e71f2L /* 0.257432886 */, 18 },
 { 0x041ec7e9L /* 0.257514868 */, 18 }, { 0x041f1de1L /* 0.257596856 */, 18 },
 { 0x041f73dcL /* 0.257678851 */, 18 }, { 0x041fc9d8L /* 0.257760852 */, 18 },

 { 0x04201fd5L /* 0.257842860 */, 18 }, { 0x042075d5L /* 0.257924875 */, 18 },
 { 0x0420cbd6L /* 0.258006895 */, 18 }, { 0x042121d9L /* 0.258088923 */, 18 },
 { 0x042177deL /* 0.258170957 */, 18 }, { 0x0421cde5L /* 0.258252997 */, 18 },
 { 0x042223edL /* 0.258335044 */, 18 }, { 0x042279f7L /* 0.258417097 */, 18 },
 { 0x0422d003L /* 0.258499157 */, 18 }, { 0x04232611L /* 0.258581224 */, 18 },
 { 0x04237c20L /* 0.258663297 */, 18 }, { 0x0423d231L /* 0.258745376 */, 18 },
 { 0x04242844L /* 0.258827462 */, 18 }, { 0x04247e58L /* 0.258909555 */, 18 },
 { 0x0424d46eL /* 0.258991654 */, 18 }, { 0x04252a87L /* 0.259073760 */, 18 },

 { 0x042580a0L /* 0.259155872 */, 18 }, { 0x0425d6bcL /* 0.259237990 */, 18 },
 { 0x04262cd9L /* 0.259320115 */, 18 }, { 0x042682f8L /* 0.259402247 */, 18 },
 { 0x0426d919L /* 0.259484385 */, 18 }, { 0x04272f3bL /* 0.259566529 */, 18 },
 { 0x04278560L /* 0.259648680 */, 18 }, { 0x0427db86L /* 0.259730838 */, 18 },
 { 0x042831adL /* 0.259813002 */, 18 }, { 0x042887d7L /* 0.259895173 */, 18 },
 { 0x0428de02L /* 0.259977350 */, 18 }, { 0x0429342fL /* 0.260059533 */, 18 },
 { 0x04298a5eL /* 0.260141723 */, 18 }, { 0x0429e08eL /* 0.260223920 */, 18 },
 { 0x042a36c0L /* 0.260306123 */, 18 }, { 0x042a8cf4L /* 0.260388332 */, 18 },

 { 0x042ae32aL /* 0.260470548 */, 18 }, { 0x042b3962L /* 0.260552771 */, 18 },
 { 0x042b8f9bL /* 0.260635000 */, 18 }, { 0x042be5d6L /* 0.260717235 */, 18 },
 { 0x042c3c12L /* 0.260799477 */, 18 }, { 0x042c9251L /* 0.260881725 */, 18 },
 { 0x042ce891L /* 0.260963980 */, 18 }, { 0x042d3ed3L /* 0.261046242 */, 18 },
 { 0x042d9516L /* 0.261128510 */, 18 }, { 0x042deb5cL /* 0.261210784 */, 18 },
 { 0x042e41a3L /* 0.261293065 */, 18 }, { 0x042e97ecL /* 0.261375352 */, 18 },
 { 0x042eee36L /* 0.261457646 */, 18 }, { 0x042f4482L /* 0.261539946 */, 18 },
 { 0x042f9ad1L /* 0.261622253 */, 18 }, { 0x042ff120L /* 0.261704566 */, 18 },

 { 0x04304772L /* 0.261786886 */, 18 }, { 0x04309dc5L /* 0.261869212 */, 18 },
 { 0x0430f41aL /* 0.261951545 */, 18 }, { 0x04314a71L /* 0.262033884 */, 18 },
 { 0x0431a0c9L /* 0.262116229 */, 18 }, { 0x0431f723L /* 0.262198581 */, 18 },
 { 0x04324d7fL /* 0.262280940 */, 18 }, { 0x0432a3ddL /* 0.262363305 */, 18 },
 { 0x0432fa3dL /* 0.262445676 */, 18 }, { 0x0433509eL /* 0.262528054 */, 18 },
 { 0x0433a701L /* 0.262610438 */, 18 }, { 0x0433fd65L /* 0.262692829 */, 18 },
 { 0x043453ccL /* 0.262775227 */, 18 }, { 0x0434aa34L /* 0.262857630 */, 18 },
 { 0x0435009dL /* 0.262940040 */, 18 }, { 0x04355709L /* 0.263022457 */, 18 },

 { 0x0435ad76L /* 0.263104880 */, 18 }, { 0x043603e5L /* 0.263187310 */, 18 },
 { 0x04365a56L /* 0.263269746 */, 18 }, { 0x0436b0c9L /* 0.263352188 */, 18 },
 { 0x0437073dL /* 0.263434637 */, 18 }, { 0x04375db3L /* 0.263517093 */, 18 },
 { 0x0437b42aL /* 0.263599554 */, 18 }, { 0x04380aa4L /* 0.263682023 */, 18 },
 { 0x0438611fL /* 0.263764497 */, 18 }, { 0x0438b79cL /* 0.263846979 */, 18 },
 { 0x04390e1aL /* 0.263929466 */, 18 }, { 0x0439649bL /* 0.264011960 */, 18 },
 { 0x0439bb1dL /* 0.264094461 */, 18 }, { 0x043a11a1L /* 0.264176968 */, 18 },
 { 0x043a6826L /* 0.264259481 */, 18 }, { 0x043abeadL /* 0.264342001 */, 18 },

 { 0x043b1536L /* 0.264424527 */, 18 }, { 0x043b6bc1L /* 0.264507060 */, 18 },
 { 0x043bc24dL /* 0.264589599 */, 18 }, { 0x043c18dcL /* 0.264672145 */, 18 },
 { 0x043c6f6cL /* 0.264754697 */, 18 }, { 0x043cc5fdL /* 0.264837255 */, 18 },
 { 0x043d1c91L /* 0.264919820 */, 18 }, { 0x043d7326L /* 0.265002392 */, 18 },
 { 0x043dc9bcL /* 0.265084969 */, 18 }, { 0x043e2055L /* 0.265167554 */, 18 },
 { 0x043e76efL /* 0.265250144 */, 18 }, { 0x043ecd8bL /* 0.265332741 */, 18 },
 { 0x043f2429L /* 0.265415345 */, 18 }, { 0x043f7ac8L /* 0.265497955 */, 18 },
 { 0x043fd169L /* 0.265580571 */, 18 }, { 0x0440280cL /* 0.265663194 */, 18 },

 { 0x04407eb1L /* 0.265745823 */, 18 }, { 0x0440d557L /* 0.265828459 */, 18 },
 { 0x04412bffL /* 0.265911101 */, 18 }, { 0x044182a9L /* 0.265993749 */, 18 },
 { 0x0441d955L /* 0.266076404 */, 18 }, { 0x04423002L /* 0.266159065 */, 18 },
 { 0x044286b1L /* 0.266241733 */, 18 }, { 0x0442dd61L /* 0.266324407 */, 18 },
 { 0x04433414L /* 0.266407088 */, 18 }, { 0x04438ac8L /* 0.266489775 */, 18 },
 { 0x0443e17eL /* 0.266572468 */, 18 }, { 0x04443835L /* 0.266655168 */, 18 },
 { 0x04448eefL /* 0.266737874 */, 18 }, { 0x0444e5aaL /* 0.266820587 */, 18 },
 { 0x04453c66L /* 0.266903306 */, 18 }, { 0x04459325L /* 0.266986031 */, 18 },

 { 0x0445e9e5L /* 0.267068763 */, 18 }, { 0x044640a7L /* 0.267151501 */, 18 },
 { 0x0446976aL /* 0.267234246 */, 18 }, { 0x0446ee30L /* 0.267316997 */, 18 },
 { 0x044744f7L /* 0.267399755 */, 18 }, { 0x04479bc0L /* 0.267482518 */, 18 },
 { 0x0447f28aL /* 0.267565289 */, 18 }, { 0x04484956L /* 0.267648065 */, 18 },
 { 0x0448a024L /* 0.267730848 */, 18 }, { 0x0448f6f4L /* 0.267813638 */, 18 },
 { 0x04494dc5L /* 0.267896434 */, 18 }, { 0x0449a498L /* 0.267979236 */, 18 },
 { 0x0449fb6dL /* 0.268062045 */, 18 }, { 0x044a5243L /* 0.268144860 */, 18 },
 { 0x044aa91cL /* 0.268227681 */, 18 }, { 0x044afff6L /* 0.268310509 */, 18 },

 { 0x044b56d1L /* 0.268393343 */, 18 }, { 0x044badafL /* 0.268476184 */, 18 },
 { 0x044c048eL /* 0.268559031 */, 18 }, { 0x044c5b6fL /* 0.268641885 */, 18 },
 { 0x044cb251L /* 0.268724744 */, 18 }, { 0x044d0935L /* 0.268807611 */, 18 },
 { 0x044d601bL /* 0.268890483 */, 18 }, { 0x044db703L /* 0.268973362 */, 18 },
 { 0x044e0decL /* 0.269056248 */, 18 }, { 0x044e64d7L /* 0.269139139 */, 18 },
 { 0x044ebbc4L /* 0.269222037 */, 18 }, { 0x044f12b3L /* 0.269304942 */, 18 },
 { 0x044f69a3L /* 0.269387853 */, 18 }, { 0x044fc095L /* 0.269470770 */, 18 },
 { 0x04501788L /* 0.269553694 */, 18 }, { 0x04506e7eL /* 0.269636624 */, 18 },

 { 0x0450c575L /* 0.269719560 */, 18 }, { 0x04511c6eL /* 0.269802503 */, 18 },
 { 0x04517368L /* 0.269885452 */, 18 }, { 0x0451ca64L /* 0.269968408 */, 18 },
 { 0x04522162L /* 0.270051370 */, 18 }, { 0x04527862L /* 0.270134338 */, 18 },
 { 0x0452cf63L /* 0.270217312 */, 18 }, { 0x04532666L /* 0.270300293 */, 18 },
 { 0x04537d6bL /* 0.270383281 */, 18 }, { 0x0453d472L /* 0.270466275 */, 18 },
 { 0x04542b7aL /* 0.270549275 */, 18 }, { 0x04548284L /* 0.270632281 */, 18 },
 { 0x0454d98fL /* 0.270715294 */, 18 }, { 0x0455309cL /* 0.270798313 */, 18 },
 { 0x045587abL /* 0.270881339 */, 18 }, { 0x0455debcL /* 0.270964371 */, 18 },

 { 0x045635cfL /* 0.271047409 */, 18 }, { 0x04568ce3L /* 0.271130454 */, 18 },
 { 0x0456e3f9L /* 0.271213505 */, 18 }, { 0x04573b10L /* 0.271296562 */, 18 },
 { 0x04579229L /* 0.271379626 */, 18 }, { 0x0457e944L /* 0.271462696 */, 18 },
 { 0x04584061L /* 0.271545772 */, 18 }, { 0x0458977fL /* 0.271628855 */, 18 },
 { 0x0458ee9fL /* 0.271711944 */, 18 }, { 0x045945c1L /* 0.271795040 */, 18 },
 { 0x04599ce5L /* 0.271878142 */, 18 }, { 0x0459f40aL /* 0.271961250 */, 18 },
 { 0x045a4b31L /* 0.272044365 */, 18 }, { 0x045aa259L /* 0.272127486 */, 18 },
 { 0x045af984L /* 0.272210613 */, 18 }, { 0x045b50b0L /* 0.272293746 */, 18 },

 { 0x045ba7ddL /* 0.272376886 */, 18 }, { 0x045bff0dL /* 0.272460033 */, 18 },
 { 0x045c563eL /* 0.272543185 */, 18 }, { 0x045cad71L /* 0.272626344 */, 18 },
 { 0x045d04a5L /* 0.272709510 */, 18 }, { 0x045d5bdcL /* 0.272792681 */, 18 },
 { 0x045db313L /* 0.272875859 */, 18 }, { 0x045e0a4dL /* 0.272959044 */, 18 },
 { 0x045e6188L /* 0.273042234 */, 18 }, { 0x045eb8c5L /* 0.273125431 */, 18 },
 { 0x045f1004L /* 0.273208635 */, 18 }, { 0x045f6745L /* 0.273291844 */, 18 },
 { 0x045fbe87L /* 0.273375060 */, 18 }, { 0x046015cbL /* 0.273458283 */, 18 },
 { 0x04606d10L /* 0.273541511 */, 18 }, { 0x0460c457L /* 0.273624747 */, 18 },

 { 0x04611ba0L /* 0.273707988 */, 18 }, { 0x046172ebL /* 0.273791236 */, 18 },
 { 0x0461ca37L /* 0.273874490 */, 18 }, { 0x04622185L /* 0.273957750 */, 18 },
 { 0x046278d5L /* 0.274041017 */, 18 }, { 0x0462d026L /* 0.274124290 */, 18 },
 { 0x0463277aL /* 0.274207569 */, 18 }, { 0x04637eceL /* 0.274290855 */, 18 },
 { 0x0463d625L /* 0.274374147 */, 18 }, { 0x04642d7dL /* 0.274457445 */, 18 },
 { 0x046484d7L /* 0.274540749 */, 18 }, { 0x0464dc33L /* 0.274624060 */, 18 },
 { 0x04653390L /* 0.274707378 */, 18 }, { 0x04658aefL /* 0.274790701 */, 18 },
 { 0x0465e250L /* 0.274874031 */, 18 }, { 0x046639b2L /* 0.274957367 */, 18 },

 { 0x04669116L /* 0.275040710 */, 18 }, { 0x0466e87cL /* 0.275124059 */, 18 },
 { 0x04673fe3L /* 0.275207414 */, 18 }, { 0x0467974dL /* 0.275290775 */, 18 },
 { 0x0467eeb7L /* 0.275374143 */, 18 }, { 0x04684624L /* 0.275457517 */, 18 },
 { 0x04689d92L /* 0.275540897 */, 18 }, { 0x0468f502L /* 0.275624284 */, 18 },
 { 0x04694c74L /* 0.275707677 */, 18 }, { 0x0469a3e7L /* 0.275791076 */, 18 },
 { 0x0469fb5cL /* 0.275874482 */, 18 }, { 0x046a52d3L /* 0.275957894 */, 18 },
 { 0x046aaa4bL /* 0.276041312 */, 18 }, { 0x046b01c5L /* 0.276124737 */, 18 },
 { 0x046b5941L /* 0.276208167 */, 18 }, { 0x046bb0bfL /* 0.276291605 */, 18 },

 { 0x046c083eL /* 0.276375048 */, 18 }, { 0x046c5fbfL /* 0.276458498 */, 18 },
 { 0x046cb741L /* 0.276541954 */, 18 }, { 0x046d0ec5L /* 0.276625416 */, 18 },
 { 0x046d664bL /* 0.276708885 */, 18 }, { 0x046dbdd3L /* 0.276792360 */, 18 },
 { 0x046e155cL /* 0.276875841 */, 18 }, { 0x046e6ce7L /* 0.276959328 */, 18 },
 { 0x046ec474L /* 0.277042822 */, 18 }, { 0x046f1c02L /* 0.277126322 */, 18 },
 { 0x046f7392L /* 0.277209829 */, 18 }, { 0x046fcb24L /* 0.277293341 */, 18 },
 { 0x047022b8L /* 0.277376860 */, 18 }, { 0x04707a4dL /* 0.277460385 */, 18 },
 { 0x0470d1e4L /* 0.277543917 */, 18 }, { 0x0471297cL /* 0.277627455 */, 18 },

 { 0x04718116L /* 0.277710999 */, 18 }, { 0x0471d8b2L /* 0.277794549 */, 18 },
 { 0x04723050L /* 0.277878106 */, 18 }, { 0x047287efL /* 0.277961669 */, 18 },
 { 0x0472df90L /* 0.278045238 */, 18 }, { 0x04733733L /* 0.278128813 */, 18 },
 { 0x04738ed7L /* 0.278212395 */, 18 }, { 0x0473e67dL /* 0.278295983 */, 18 },
 { 0x04743e25L /* 0.278379578 */, 18 }, { 0x047495ceL /* 0.278463178 */, 18 },
 { 0x0474ed79L /* 0.278546785 */, 18 }, { 0x04754526L /* 0.278630398 */, 18 },
 { 0x04759cd4L /* 0.278714018 */, 18 }, { 0x0475f484L /* 0.278797643 */, 18 },
 { 0x04764c36L /* 0.278881275 */, 18 }, { 0x0476a3eaL /* 0.278964914 */, 18 },

 { 0x0476fb9fL /* 0.279048558 */, 18 }, { 0x04775356L /* 0.279132209 */, 18 },
 { 0x0477ab0eL /* 0.279215866 */, 18 }, { 0x047802c8L /* 0.279299529 */, 18 },
 { 0x04785a84L /* 0.279383199 */, 18 }, { 0x0478b242L /* 0.279466875 */, 18 },
 { 0x04790a01L /* 0.279550557 */, 18 }, { 0x047961c2L /* 0.279634245 */, 18 },
 { 0x0479b984L /* 0.279717940 */, 18 }, { 0x047a1149L /* 0.279801641 */, 18 },
 { 0x047a690fL /* 0.279885348 */, 18 }, { 0x047ac0d6L /* 0.279969061 */, 18 },
 { 0x047b18a0L /* 0.280052781 */, 18 }, { 0x047b706bL /* 0.280136507 */, 18 },
 { 0x047bc837L /* 0.280220239 */, 18 }, { 0x047c2006L /* 0.280303978 */, 18 },

 { 0x047c77d6L /* 0.280387722 */, 18 }, { 0x047ccfa8L /* 0.280471473 */, 18 },
 { 0x047d277bL /* 0.280555230 */, 18 }, { 0x047d7f50L /* 0.280638994 */, 18 },
 { 0x047dd727L /* 0.280722764 */, 18 }, { 0x047e2effL /* 0.280806540 */, 18 },
 { 0x047e86d9L /* 0.280890322 */, 18 }, { 0x047edeb5L /* 0.280974110 */, 18 },
 { 0x047f3693L /* 0.281057905 */, 18 }, { 0x047f8e72L /* 0.281141706 */, 18 },
 { 0x047fe653L /* 0.281225513 */, 18 }, { 0x04803e35L /* 0.281309326 */, 18 },
 { 0x04809619L /* 0.281393146 */, 18 }, { 0x0480edffL /* 0.281476972 */, 18 },
 { 0x048145e7L /* 0.281560804 */, 18 }, { 0x04819dd0L /* 0.281644643 */, 18 },

 { 0x0481f5bbL /* 0.281728487 */, 18 }, { 0x04824da7L /* 0.281812338 */, 18 },
 { 0x0482a595L /* 0.281896195 */, 18 }, { 0x0482fd85L /* 0.281980059 */, 18 },
 { 0x04835577L /* 0.282063928 */, 18 }, { 0x0483ad6aL /* 0.282147804 */, 18 },
 { 0x0484055fL /* 0.282231686 */, 18 }, { 0x04845d56L /* 0.282315574 */, 18 },
 { 0x0484b54eL /* 0.282399469 */, 18 }, { 0x04850d48L /* 0.282483370 */, 18 },
 { 0x04856544L /* 0.282567277 */, 18 }, { 0x0485bd41L /* 0.282651190 */, 18 },
 { 0x04861540L /* 0.282735109 */, 18 }, { 0x04866d40L /* 0.282819035 */, 18 },
 { 0x0486c543L /* 0.282902967 */, 18 }, { 0x04871d47L /* 0.282986905 */, 18 },

 { 0x0487754cL /* 0.283070849 */, 18 }, { 0x0487cd54L /* 0.283154800 */, 18 },
 { 0x0488255dL /* 0.283238757 */, 18 }, { 0x04887d67L /* 0.283322720 */, 18 },
 { 0x0488d574L /* 0.283406689 */, 18 }, { 0x04892d82L /* 0.283490665 */, 18 },
 { 0x04898591L /* 0.283574646 */, 18 }, { 0x0489dda3L /* 0.283658634 */, 18 },
 { 0x048a35b6L /* 0.283742628 */, 18 }, { 0x048a8dcaL /* 0.283826629 */, 18 },
 { 0x048ae5e1L /* 0.283910635 */, 18 }, { 0x048b3df9L /* 0.283994648 */, 18 },
 { 0x048b9612L /* 0.284078667 */, 18 }, { 0x048bee2eL /* 0.284162692 */, 18 },
 { 0x048c464bL /* 0.284246723 */, 18 }, { 0x048c9e69L /* 0.284330761 */, 18 },

 { 0x048cf68aL /* 0.284414805 */, 18 }, { 0x048d4eacL /* 0.284498855 */, 18 },
 { 0x048da6cfL /* 0.284582911 */, 18 }, { 0x048dfef5L /* 0.284666974 */, 18 },
 { 0x048e571cL /* 0.284751042 */, 18 }, { 0x048eaf44L /* 0.284835117 */, 18 },
 { 0x048f076fL /* 0.284919198 */, 18 }, { 0x048f5f9bL /* 0.285003285 */, 18 },
 { 0x048fb7c8L /* 0.285087379 */, 18 }, { 0x04900ff8L /* 0.285171479 */, 18 },
 { 0x04906829L /* 0.285255584 */, 18 }, { 0x0490c05bL /* 0.285339697 */, 18 },
 { 0x04911890L /* 0.285423815 */, 18 }, { 0x049170c6L /* 0.285507939 */, 18 },
 { 0x0491c8fdL /* 0.285592070 */, 18 }, { 0x04922137L /* 0.285676207 */, 18 },

 { 0x04927972L /* 0.285760350 */, 18 }, { 0x0492d1aeL /* 0.285844499 */, 18 },
 { 0x049329edL /* 0.285928655 */, 18 }, { 0x0493822cL /* 0.286012816 */, 18 },
 { 0x0493da6eL /* 0.286096984 */, 18 }, { 0x049432b1L /* 0.286181158 */, 18 },
 { 0x04948af6L /* 0.286265338 */, 18 }, { 0x0494e33dL /* 0.286349525 */, 18 },
 { 0x04953b85L /* 0.286433717 */, 18 }, { 0x049593cfL /* 0.286517916 */, 18 },
 { 0x0495ec1bL /* 0.286602121 */, 18 }, { 0x04964468L /* 0.286686332 */, 18 },
 { 0x04969cb7L /* 0.286770550 */, 18 }, { 0x0496f508L /* 0.286854773 */, 18 },
 { 0x04974d5aL /* 0.286939003 */, 18 }, { 0x0497a5aeL /* 0.287023239 */, 18 },

 { 0x0497fe03L /* 0.287107481 */, 18 }, { 0x0498565aL /* 0.287191729 */, 18 },
 { 0x0498aeb3L /* 0.287275983 */, 18 }, { 0x0499070eL /* 0.287360244 */, 18 },
 { 0x04995f6aL /* 0.287444511 */, 18 }, { 0x0499b7c8L /* 0.287528784 */, 18 },
 { 0x049a1027L /* 0.287613063 */, 18 }, { 0x049a6889L /* 0.287697348 */, 18 },
 { 0x049ac0ebL /* 0.287781640 */, 18 }, { 0x049b1950L /* 0.287865937 */, 18 },
 { 0x049b71b6L /* 0.287950241 */, 18 }, { 0x049bca1eL /* 0.288034551 */, 18 },
 { 0x049c2287L /* 0.288118867 */, 18 }, { 0x049c7af2L /* 0.288203190 */, 18 },
 { 0x049cd35fL /* 0.288287518 */, 18 }, { 0x049d2bceL /* 0.288371853 */, 18 },

 { 0x049d843eL /* 0.288456194 */, 18 }, { 0x049ddcafL /* 0.288540541 */, 18 },
 { 0x049e3523L /* 0.288624894 */, 18 }, { 0x049e8d98L /* 0.288709253 */, 18 },
 { 0x049ee60eL /* 0.288793619 */, 18 }, { 0x049f3e87L /* 0.288877990 */, 18 },
 { 0x049f9701L /* 0.288962368 */, 18 }, { 0x049fef7cL /* 0.289046752 */, 18 },
 { 0x04a047faL /* 0.289131142 */, 18 }, { 0x04a0a079L /* 0.289215538 */, 18 },
 { 0x04a0f8f9L /* 0.289299941 */, 18 }, { 0x04a1517cL /* 0.289384349 */, 18 },
 { 0x04a1a9ffL /* 0.289468764 */, 18 }, { 0x04a20285L /* 0.289553185 */, 18 },
 { 0x04a25b0cL /* 0.289637612 */, 18 }, { 0x04a2b395L /* 0.289722045 */, 18 },

 { 0x04a30c20L /* 0.289806485 */, 18 }, { 0x04a364acL /* 0.289890930 */, 18 },
 { 0x04a3bd3aL /* 0.289975382 */, 18 }, { 0x04a415c9L /* 0.290059840 */, 18 },
 { 0x04a46e5aL /* 0.290144304 */, 18 }, { 0x04a4c6edL /* 0.290228774 */, 18 },
 { 0x04a51f81L /* 0.290313250 */, 18 }, { 0x04a57818L /* 0.290397733 */, 18 },
 { 0x04a5d0afL /* 0.290482221 */, 18 }, { 0x04a62949L /* 0.290566716 */, 18 },
 { 0x04a681e4L /* 0.290651217 */, 18 }, { 0x04a6da80L /* 0.290735724 */, 18 },
 { 0x04a7331fL /* 0.290820237 */, 18 }, { 0x04a78bbfL /* 0.290904756 */, 18 },
 { 0x04a7e460L /* 0.290989281 */, 18 }, { 0x04a83d03L /* 0.291073813 */, 18 },

 { 0x04a895a8L /* 0.291158351 */, 18 }, { 0x04a8ee4fL /* 0.291242894 */, 18 },
 { 0x04a946f7L /* 0.291327444 */, 18 }, { 0x04a99fa1L /* 0.291412001 */, 18 },
 { 0x04a9f84cL /* 0.291496563 */, 18 }, { 0x04aa50faL /* 0.291581131 */, 18 },
 { 0x04aaa9a8L /* 0.291665706 */, 18 }, { 0x04ab0259L /* 0.291750286 */, 18 },
 { 0x04ab5b0bL /* 0.291834873 */, 18 }, { 0x04abb3bfL /* 0.291919466 */, 18 },
 { 0x04ac0c74L /* 0.292004065 */, 18 }, { 0x04ac652bL /* 0.292088670 */, 18 },
 { 0x04acbde4L /* 0.292173281 */, 18 }, { 0x04ad169eL /* 0.292257899 */, 18 },
 { 0x04ad6f5aL /* 0.292342522 */, 18 }, { 0x04adc818L /* 0.292427152 */, 18 },

 { 0x04ae20d7L /* 0.292511788 */, 18 }, { 0x04ae7998L /* 0.292596430 */, 18 },
 { 0x04aed25aL /* 0.292681078 */, 18 }, { 0x04af2b1eL /* 0.292765732 */, 18 },
 { 0x04af83e4L /* 0.292850392 */, 18 }, { 0x04afdcacL /* 0.292935058 */, 18 },
 { 0x04b03575L /* 0.293019731 */, 18 }, { 0x04b08e40L /* 0.293104409 */, 18 },
 { 0x04b0e70cL /* 0.293189094 */, 18 }, { 0x04b13fdaL /* 0.293273785 */, 18 },
 { 0x04b198aaL /* 0.293358482 */, 18 }, { 0x04b1f17bL /* 0.293443185 */, 18 },
 { 0x04b24a4eL /* 0.293527894 */, 18 }, { 0x04b2a322L /* 0.293612609 */, 18 },
 { 0x04b2fbf9L /* 0.293697331 */, 18 }, { 0x04b354d1L /* 0.293782058 */, 18 },

 { 0x04b3adaaL /* 0.293866792 */, 18 }, { 0x04b40685L /* 0.293951532 */, 18 },
 { 0x04b45f62L /* 0.294036278 */, 18 }, { 0x04b4b840L /* 0.294121029 */, 18 },
 { 0x04b51120L /* 0.294205788 */, 18 }, { 0x04b56a02L /* 0.294290552 */, 18 },
 { 0x04b5c2e6L /* 0.294375322 */, 18 }, { 0x04b61bcbL /* 0.294460098 */, 18 },
 { 0x04b674b1L /* 0.294544881 */, 18 }, { 0x04b6cd99L /* 0.294629669 */, 18 },
 { 0x04b72683L /* 0.294714464 */, 18 }, { 0x04b77f6fL /* 0.294799265 */, 18 },
 { 0x04b7d85cL /* 0.294884072 */, 18 }, { 0x04b8314bL /* 0.294968885 */, 18 },
 { 0x04b88a3bL /* 0.295053704 */, 18 }, { 0x04b8e32dL /* 0.295138529 */, 18 },

 { 0x04b93c21L /* 0.295223360 */, 18 }, { 0x04b99516L /* 0.295308197 */, 18 },
 { 0x04b9ee0dL /* 0.295393041 */, 18 }, { 0x04ba4706L /* 0.295477890 */, 18 },
 { 0x04baa000L /* 0.295562746 */, 18 }, { 0x04baf8fcL /* 0.295647608 */, 18 },
 { 0x04bb51faL /* 0.295732476 */, 18 }, { 0x04bbaaf9L /* 0.295817349 */, 18 },
 { 0x04bc03faL /* 0.295902229 */, 18 }, { 0x04bc5cfcL /* 0.295987115 */, 18 },
 { 0x04bcb600L /* 0.296072008 */, 18 }, { 0x04bd0f06L /* 0.296156906 */, 18 },
 { 0x04bd680dL /* 0.296241810 */, 18 }, { 0x04bdc116L /* 0.296326721 */, 18 },
 { 0x04be1a21L /* 0.296411637 */, 18 }, { 0x04be732dL /* 0.296496560 */, 18 },

 { 0x04becc3bL /* 0.296581488 */, 18 }, { 0x04bf254aL /* 0.296666423 */, 18 },
 { 0x04bf7e5bL /* 0.296751364 */, 18 }, { 0x04bfd76eL /* 0.296836311 */, 18 },
 { 0x04c03083L /* 0.296921264 */, 18 }, { 0x04c08999L /* 0.297006223 */, 18 },
 { 0x04c0e2b0L /* 0.297091188 */, 18 }, { 0x04c13bcaL /* 0.297176159 */, 18 },
 { 0x04c194e4L /* 0.297261136 */, 18 }, { 0x04c1ee01L /* 0.297346120 */, 18 },
 { 0x04c2471fL /* 0.297431109 */, 18 }, { 0x04c2a03fL /* 0.297516105 */, 18 },
 { 0x04c2f960L /* 0.297601106 */, 18 }, { 0x04c35283L /* 0.297686114 */, 18 },
 { 0x04c3aba8L /* 0.297771128 */, 18 }, { 0x04c404ceL /* 0.297856147 */, 18 },

 { 0x04c45df6L /* 0.297941173 */, 18 }, { 0x04c4b720L /* 0.298026205 */, 18 },
 { 0x04c5104bL /* 0.298111243 */, 18 }, { 0x04c56978L /* 0.298196287 */, 18 },
 { 0x04c5c2a7L /* 0.298281337 */, 18 }, { 0x04c61bd7L /* 0.298366393 */, 18 },
 { 0x04c67508L /* 0.298451456 */, 18 }, { 0x04c6ce3cL /* 0.298536524 */, 18 },
 { 0x04c72771L /* 0.298621598 */, 18 }, { 0x04c780a7L /* 0.298706679 */, 18 },
 { 0x04c7d9dfL /* 0.298791765 */, 18 }, { 0x04c83319L /* 0.298876858 */, 18 },
 { 0x04c88c55L /* 0.298961956 */, 18 }, { 0x04c8e592L /* 0.299047061 */, 18 },
 { 0x04c93ed1L /* 0.299132172 */, 18 }, { 0x04c99811L /* 0.299217288 */, 18 },

 { 0x04c9f153L /* 0.299302411 */, 18 }, { 0x04ca4a97L /* 0.299387540 */, 18 },
 { 0x04caa3dcL /* 0.299472675 */, 18 }, { 0x04cafd23L /* 0.299557816 */, 18 },
 { 0x04cb566bL /* 0.299642963 */, 18 }, { 0x04cbafb5L /* 0.299728116 */, 18 },
 { 0x04cc0901L /* 0.299813275 */, 18 }, { 0x04cc624eL /* 0.299898440 */, 18 },
 { 0x04ccbb9dL /* 0.299983611 */, 18 }, { 0x04cd14eeL /* 0.300068789 */, 18 },
 { 0x04cd6e40L /* 0.300153972 */, 18 }, { 0x04cdc794L /* 0.300239161 */, 18 },
 { 0x04ce20e9L /* 0.300324357 */, 18 }, { 0x04ce7a40L /* 0.300409558 */, 18 },
 { 0x04ced399L /* 0.300494765 */, 18 }, { 0x04cf2cf3L /* 0.300579979 */, 18 },

 { 0x04cf864fL /* 0.300665198 */, 18 }, { 0x04cfdfadL /* 0.300750424 */, 18 },
 { 0x04d0390cL /* 0.300835656 */, 18 }, { 0x04d0926dL /* 0.300920893 */, 18 },
 { 0x04d0ebcfL /* 0.301006137 */, 18 }, { 0x04d14533L /* 0.301091387 */, 18 },
 { 0x04d19e99L /* 0.301176643 */, 18 }, { 0x04d1f800L /* 0.301261904 */, 18 },
 { 0x04d25169L /* 0.301347172 */, 18 }, { 0x04d2aad4L /* 0.301432446 */, 18 },
 { 0x04d30440L /* 0.301517726 */, 18 }, { 0x04d35daeL /* 0.301603012 */, 18 },
 { 0x04d3b71dL /* 0.301688304 */, 18 }, { 0x04d4108eL /* 0.301773602 */, 18 },
 { 0x04d46a01L /* 0.301858906 */, 18 }, { 0x04d4c375L /* 0.301944216 */, 18 },

 { 0x04d51cebL /* 0.302029532 */, 18 }, { 0x04d57662L /* 0.302114854 */, 18 },
 { 0x04d5cfdbL /* 0.302200182 */, 18 }, { 0x04d62956L /* 0.302285516 */, 18 },
 { 0x04d682d2L /* 0.302370856 */, 18 }, { 0x04d6dc50L /* 0.302456203 */, 18 },
 { 0x04d735d0L /* 0.302541555 */, 18 }, { 0x04d78f51L /* 0.302626913 */, 18 },
 { 0x04d7e8d4L /* 0.302712277 */, 18 }, { 0x04d84258L /* 0.302797648 */, 18 },
 { 0x04d89bdeL /* 0.302883024 */, 18 }, { 0x04d8f566L /* 0.302968406 */, 18 },
 { 0x04d94eefL /* 0.303053794 */, 18 }, { 0x04d9a87aL /* 0.303139189 */, 18 },
 { 0x04da0207L /* 0.303224589 */, 18 }, { 0x04da5b95L /* 0.303309995 */, 18 },

 { 0x04dab524L /* 0.303395408 */, 18 }, { 0x04db0eb6L /* 0.303480826 */, 18 },
 { 0x04db6849L /* 0.303566251 */, 18 }, { 0x04dbc1ddL /* 0.303651681 */, 18 },
 { 0x04dc1b73L /* 0.303737117 */, 18 }, { 0x04dc750bL /* 0.303822560 */, 18 },
 { 0x04dccea5L /* 0.303908008 */, 18 }, { 0x04dd2840L /* 0.303993463 */, 18 },
 { 0x04dd81dcL /* 0.304078923 */, 18 }, { 0x04dddb7aL /* 0.304164390 */, 18 },
 { 0x04de351aL /* 0.304249862 */, 18 }, { 0x04de8ebcL /* 0.304335340 */, 18 },
 { 0x04dee85fL /* 0.304420825 */, 18 }, { 0x04df4203L /* 0.304506315 */, 18 },
 { 0x04df9baaL /* 0.304591812 */, 18 }, { 0x04dff552L /* 0.304677314 */, 18 },

 { 0x04e04efbL /* 0.304762823 */, 18 }, { 0x04e0a8a6L /* 0.304848337 */, 18 },
 { 0x04e10253L /* 0.304933858 */, 18 }, { 0x04e15c01L /* 0.305019384 */, 18 },
 { 0x04e1b5b1L /* 0.305104917 */, 18 }, { 0x04e20f63L /* 0.305190455 */, 18 },
 { 0x04e26916L /* 0.305275999 */, 18 }, { 0x04e2c2cbL /* 0.305361550 */, 18 },
 { 0x04e31c81L /* 0.305447106 */, 18 }, { 0x04e37639L /* 0.305532669 */, 18 },
 { 0x04e3cff3L /* 0.305618237 */, 18 }, { 0x04e429aeL /* 0.305703811 */, 18 },
 { 0x04e4836bL /* 0.305789392 */, 18 }, { 0x04e4dd29L /* 0.305874978 */, 18 },
 { 0x04e536e9L /* 0.305960571 */, 18 }, { 0x04e590abL /* 0.306046169 */, 18 },

 { 0x04e5ea6eL /* 0.306131773 */, 18 }, { 0x04e64433L /* 0.306217383 */, 18 },
 { 0x04e69df9L /* 0.306303000 */, 18 }, { 0x04e6f7c1L /* 0.306388622 */, 18 },
 { 0x04e7518bL /* 0.306474250 */, 18 }, { 0x04e7ab56L /* 0.306559885 */, 18 },
 { 0x04e80523L /* 0.306645525 */, 18 }, { 0x04e85ef2L /* 0.306731171 */, 18 },
 { 0x04e8b8c2L /* 0.306816823 */, 18 }, { 0x04e91293L /* 0.306902481 */, 18 },
 { 0x04e96c67L /* 0.306988145 */, 18 }, { 0x04e9c63bL /* 0.307073816 */, 18 },
 { 0x04ea2012L /* 0.307159492 */, 18 }, { 0x04ea79eaL /* 0.307245174 */, 18 },
 { 0x04ead3c4L /* 0.307330862 */, 18 }, { 0x04eb2d9fL /* 0.307416556 */, 18 },

 { 0x04eb877cL /* 0.307502256 */, 18 }, { 0x04ebe15bL /* 0.307587962 */, 18 },
 { 0x04ec3b3bL /* 0.307673674 */, 18 }, { 0x04ec951cL /* 0.307759392 */, 18 },
 { 0x04ecef00L /* 0.307845115 */, 18 }, { 0x04ed48e5L /* 0.307930845 */, 18 },
 { 0x04eda2cbL /* 0.308016581 */, 18 }, { 0x04edfcb3L /* 0.308102323 */, 18 },
 { 0x04ee569dL /* 0.308188071 */, 18 }, { 0x04eeb088L /* 0.308273824 */, 18 },
 { 0x04ef0a75L /* 0.308359584 */, 18 }, { 0x04ef6464L /* 0.308445350 */, 18 },
 { 0x04efbe54L /* 0.308531121 */, 18 }, { 0x04f01846L /* 0.308616899 */, 18 },
 { 0x04f07239L /* 0.308702682 */, 18 }, { 0x04f0cc2eL /* 0.308788472 */, 18 },

 { 0x04f12624L /* 0.308874267 */, 18 }, { 0x04f1801dL /* 0.308960068 */, 18 },
 { 0x04f1da16L /* 0.309045876 */, 18 }, { 0x04f23412L /* 0.309131689 */, 18 },
 { 0x04f28e0fL /* 0.309217508 */, 18 }, { 0x04f2e80dL /* 0.309303334 */, 18 },
 { 0x04f3420dL /* 0.309389165 */, 18 }, { 0x04f39c0fL /* 0.309475002 */, 18 },
 { 0x04f3f612L /* 0.309560845 */, 18 }, { 0x04f45017L /* 0.309646694 */, 18 },
 { 0x04f4aa1eL /* 0.309732549 */, 18 }, { 0x04f50426L /* 0.309818410 */, 18 },
 { 0x04f55e30L /* 0.309904277 */, 18 }, { 0x04f5b83bL /* 0.309990150 */, 18 },
 { 0x04f61248L /* 0.310076028 */, 18 }, { 0x04f66c56L /* 0.310161913 */, 18 },

 { 0x04f6c666L /* 0.310247804 */, 18 }, { 0x04f72078L /* 0.310333700 */, 18 },
 { 0x04f77a8bL /* 0.310419603 */, 18 }, { 0x04f7d4a0L /* 0.310505511 */, 18 },
 { 0x04f82eb7L /* 0.310591426 */, 18 }, { 0x04f888cfL /* 0.310677346 */, 18 },
 { 0x04f8e2e9L /* 0.310763272 */, 18 }, { 0x04f93d04L /* 0.310849205 */, 18 },
 { 0x04f99721L /* 0.310935143 */, 18 }, { 0x04f9f13fL /* 0.311021087 */, 18 },
 { 0x04fa4b5fL /* 0.311107037 */, 18 }, { 0x04faa581L /* 0.311192993 */, 18 },
 { 0x04faffa4L /* 0.311278955 */, 18 }, { 0x04fb59c9L /* 0.311364923 */, 18 },
 { 0x04fbb3efL /* 0.311450897 */, 18 }, { 0x04fc0e17L /* 0.311536877 */, 18 },

 { 0x04fc6841L /* 0.311622862 */, 18 }, { 0x04fcc26cL /* 0.311708854 */, 18 },
 { 0x04fd1c99L /* 0.311794851 */, 18 }, { 0x04fd76c7L /* 0.311880855 */, 18 },
 { 0x04fdd0f7L /* 0.311966864 */, 18 }, { 0x04fe2b29L /* 0.312052880 */, 18 },
 { 0x04fe855cL /* 0.312138901 */, 18 }, { 0x04fedf91L /* 0.312224928 */, 18 },
 { 0x04ff39c7L /* 0.312310961 */, 18 }, { 0x04ff93ffL /* 0.312397000 */, 18 },
 { 0x04ffee38L /* 0.312483045 */, 18 }, { 0x05004874L /* 0.312569096 */, 18 },
 { 0x0500a2b0L /* 0.312655153 */, 18 }, { 0x0500fcefL /* 0.312741216 */, 18 },
 { 0x0501572eL /* 0.312827284 */, 18 }, { 0x0501b170L /* 0.312913359 */, 18 },

 { 0x05020bb3L /* 0.312999439 */, 18 }, { 0x050265f8L /* 0.313085526 */, 18 },
 { 0x0502c03eL /* 0.313171618 */, 18 }, { 0x05031a86L /* 0.313257716 */, 18 },
 { 0x050374cfL /* 0.313343820 */, 18 }, { 0x0503cf1aL /* 0.313429931 */, 18 },
 { 0x05042967L /* 0.313516047 */, 18 }, { 0x050483b5L /* 0.313602168 */, 18 },
 { 0x0504de05L /* 0.313688296 */, 18 }, { 0x05053856L /* 0.313774430 */, 18 },
 { 0x050592a9L /* 0.313860570 */, 18 }, { 0x0505ecfdL /* 0.313946715 */, 18 },
 { 0x05064754L /* 0.314032867 */, 18 }, { 0x0506a1abL /* 0.314119024 */, 18 },
 { 0x0506fc04L /* 0.314205187 */, 18 }, { 0x0507565fL /* 0.314291357 */, 18 },

 { 0x0507b0bcL /* 0.314377532 */, 18 }, { 0x05080b1aL /* 0.314463713 */, 18 },
 { 0x05086579L /* 0.314549900 */, 18 }, { 0x0508bfdbL /* 0.314636092 */, 18 },
 { 0x05091a3dL /* 0.314722291 */, 18 }, { 0x050974a2L /* 0.314808496 */, 18 },
 { 0x0509cf08L /* 0.314894706 */, 18 }, { 0x050a296fL /* 0.314980923 */, 18 },
 { 0x050a83d8L /* 0.315067145 */, 18 }, { 0x050ade43L /* 0.315153373 */, 18 },
 { 0x050b38afL /* 0.315239607 */, 18 }, { 0x050b931dL /* 0.315325847 */, 18 },
 { 0x050bed8dL /* 0.315412093 */, 18 }, { 0x050c47feL /* 0.315498345 */, 18 },
 { 0x050ca271L /* 0.315584603 */, 18 }, { 0x050cfce5L /* 0.315670866 */, 18 },

 { 0x050d575bL /* 0.315757136 */, 18 }, { 0x050db1d2L /* 0.315843411 */, 18 },
 { 0x050e0c4bL /* 0.315929693 */, 18 }, { 0x050e66c5L /* 0.316015980 */, 18 },
 { 0x050ec141L /* 0.316102273 */, 18 }, { 0x050f1bbfL /* 0.316188572 */, 18 },
 { 0x050f763eL /* 0.316274877 */, 18 }, { 0x050fd0bfL /* 0.316361187 */, 18 },
 { 0x05102b42L /* 0.316447504 */, 18 }, { 0x051085c6L /* 0.316533826 */, 18 },
 { 0x0510e04bL /* 0.316620155 */, 18 }, { 0x05113ad3L /* 0.316706489 */, 18 },
 { 0x0511955bL /* 0.316792829 */, 18 }, { 0x0511efe6L /* 0.316879175 */, 18 },
 { 0x05124a72L /* 0.316965527 */, 18 }, { 0x0512a4ffL /* 0.317051885 */, 18 },

 { 0x0512ff8eL /* 0.317138249 */, 18 }, { 0x05135a1fL /* 0.317224618 */, 18 },
 { 0x0513b4b1L /* 0.317310994 */, 18 }, { 0x05140f45L /* 0.317397375 */, 18 },
 { 0x051469daL /* 0.317483762 */, 18 }, { 0x0514c471L /* 0.317570155 */, 18 },
 { 0x05151f0aL /* 0.317656554 */, 18 }, { 0x051579a4L /* 0.317742959 */, 18 },
 { 0x0515d440L /* 0.317829370 */, 18 }, { 0x05162eddL /* 0.317915786 */, 18 },
 { 0x0516897cL /* 0.318002209 */, 18 }, { 0x0516e41cL /* 0.318088637 */, 18 },
 { 0x05173ebeL /* 0.318175071 */, 18 }, { 0x05179962L /* 0.318261511 */, 18 },
 { 0x0517f407L /* 0.318347957 */, 18 }, { 0x05184eaeL /* 0.318434409 */, 18 },

 { 0x0518a956L /* 0.318520867 */, 18 }, { 0x05190400L /* 0.318607330 */, 18 },
 { 0x05195eabL /* 0.318693800 */, 18 }, { 0x0519b958L /* 0.318780275 */, 18 },
 { 0x051a1407L /* 0.318866756 */, 18 }, { 0x051a6eb7L /* 0.318953243 */, 18 },
 { 0x051ac969L /* 0.319039736 */, 18 }, { 0x051b241cL /* 0.319126235 */, 18 },
 { 0x051b7ed1L /* 0.319212739 */, 18 }, { 0x051bd987L /* 0.319299250 */, 18 },
 { 0x051c3440L /* 0.319385766 */, 18 }, { 0x051c8ef9L /* 0.319472288 */, 18 },
 { 0x051ce9b4L /* 0.319558816 */, 18 }, { 0x051d4471L /* 0.319645350 */, 18 },
 { 0x051d9f2fL /* 0.319731890 */, 18 }, { 0x051df9efL /* 0.319818435 */, 18 },

 { 0x051e54b1L /* 0.319904987 */, 18 }, { 0x051eaf74L /* 0.319991544 */, 18 },
 { 0x051f0a38L /* 0.320078107 */, 18 }, { 0x051f64ffL /* 0.320164676 */, 18 },
 { 0x051fbfc6L /* 0.320251251 */, 18 }, { 0x05201a90L /* 0.320337832 */, 18 },
 { 0x0520755bL /* 0.320424419 */, 18 }, { 0x0520d027L /* 0.320511011 */, 18 },
 { 0x05212af5L /* 0.320597609 */, 18 }, { 0x052185c5L /* 0.320684213 */, 18 },
 { 0x0521e096L /* 0.320770823 */, 18 }, { 0x05223b69L /* 0.320857439 */, 18 },
 { 0x0522963dL /* 0.320944061 */, 18 }, { 0x0522f113L /* 0.321030688 */, 18 },
 { 0x05234beaL /* 0.321117322 */, 18 }, { 0x0523a6c3L /* 0.321203961 */, 18 },

 { 0x0524019eL /* 0.321290606 */, 18 }, { 0x05245c7aL /* 0.321377257 */, 18 },
 { 0x0524b758L /* 0.321463913 */, 18 }, { 0x05251237L /* 0.321550576 */, 18 },
 { 0x05256d18L /* 0.321637244 */, 18 }, { 0x0525c7fbL /* 0.321723919 */, 18 },
 { 0x052622dfL /* 0.321810599 */, 18 }, { 0x05267dc4L /* 0.321897285 */, 18 },
 { 0x0526d8abL /* 0.321983976 */, 18 }, { 0x05273394L /* 0.322070674 */, 18 },
 { 0x05278e7eL /* 0.322157377 */, 18 }, { 0x0527e96aL /* 0.322244087 */, 18 },
 { 0x05284457L /* 0.322330802 */, 18 }, { 0x05289f46L /* 0.322417523 */, 18 },
 { 0x0528fa37L /* 0.322504249 */, 18 }, { 0x05295529L /* 0.322590982 */, 18 },

 { 0x0529b01dL /* 0.322677720 */, 18 }, { 0x052a0b12L /* 0.322764465 */, 18 },
 { 0x052a6609L /* 0.322851215 */, 18 }, { 0x052ac101L /* 0.322937971 */, 18 },
 { 0x052b1bfbL /* 0.323024732 */, 18 }, { 0x052b76f7L /* 0.323111500 */, 18 },
 { 0x052bd1f4L /* 0.323198273 */, 18 }, { 0x052c2cf2L /* 0.323285052 */, 18 },
 { 0x052c87f2L /* 0.323371837 */, 18 }, { 0x052ce2f4L /* 0.323458628 */, 18 },
 { 0x052d3df7L /* 0.323545425 */, 18 }, { 0x052d98fcL /* 0.323632227 */, 18 },
 { 0x052df403L /* 0.323719036 */, 18 }, { 0x052e4f0bL /* 0.323805850 */, 18 },
 { 0x052eaa14L /* 0.323892670 */, 18 }, { 0x052f051fL /* 0.323979496 */, 18 },

 { 0x052f602cL /* 0.324066327 */, 18 }, { 0x052fbb3aL /* 0.324153165 */, 18 },
 { 0x0530164aL /* 0.324240008 */, 18 }, { 0x0530715bL /* 0.324326857 */, 18 },
 { 0x0530cc6eL /* 0.324413712 */, 18 }, { 0x05312783L /* 0.324500572 */, 18 },
 { 0x05318299L /* 0.324587439 */, 18 }, { 0x0531ddb0L /* 0.324674311 */, 18 },
 { 0x053238caL /* 0.324761189 */, 18 }, { 0x053293e4L /* 0.324848073 */, 18 },
 { 0x0532ef01L /* 0.324934963 */, 18 }, { 0x05334a1eL /* 0.325021858 */, 18 },
 { 0x0533a53eL /* 0.325108760 */, 18 }, { 0x0534005fL /* 0.325195667 */, 18 },
 { 0x05345b81L /* 0.325282580 */, 18 }, { 0x0534b6a5L /* 0.325369498 */, 18 },

 { 0x053511cbL /* 0.325456423 */, 18 }, { 0x05356cf2L /* 0.325543353 */, 18 },
 { 0x0535c81bL /* 0.325630290 */, 18 }, { 0x05362345L /* 0.325717232 */, 18 },
 { 0x05367e71L /* 0.325804179 */, 18 }, { 0x0536d99fL /* 0.325891133 */, 18 },
 { 0x053734ceL /* 0.325978092 */, 18 }, { 0x05378ffeL /* 0.326065057 */, 18 },
 { 0x0537eb30L /* 0.326152028 */, 18 }, { 0x05384664L /* 0.326239005 */, 18 },
 { 0x0538a199L /* 0.326325988 */, 18 }, { 0x0538fcd0L /* 0.326412976 */, 18 },
 { 0x05395808L /* 0.326499970 */, 18 }, { 0x0539b342L /* 0.326586970 */, 18 },
 { 0x053a0e7dL /* 0.326673976 */, 18 }, { 0x053a69baL /* 0.326760988 */, 18 },

 { 0x053ac4f9L /* 0.326848005 */, 18 }, { 0x053b2039L /* 0.326935028 */, 18 },
 { 0x053b7b7bL /* 0.327022057 */, 18 }, { 0x053bd6beL /* 0.327109092 */, 18 },
 { 0x053c3203L /* 0.327196132 */, 18 }, { 0x053c8d49L /* 0.327283178 */, 18 },
 { 0x053ce891L /* 0.327370231 */, 18 }, { 0x053d43daL /* 0.327457288 */, 18 },
 { 0x053d9f25L /* 0.327544352 */, 18 }, { 0x053dfa72L /* 0.327631421 */, 18 },
 { 0x053e55c0L /* 0.327718497 */, 18 }, { 0x053eb10fL /* 0.327805578 */, 18 },
 { 0x053f0c61L /* 0.327892665 */, 18 }, { 0x053f67b3L /* 0.327979757 */, 18 },
 { 0x053fc308L /* 0.328066855 */, 18 }, { 0x05401e5eL /* 0.328153960 */, 18 },

 { 0x054079b5L /* 0.328241070 */, 18 }, { 0x0540d50eL /* 0.328328185 */, 18 },
 { 0x05413068L /* 0.328415307 */, 18 }, { 0x05418bc4L /* 0.328502434 */, 18 },
 { 0x0541e722L /* 0.328589567 */, 18 }, { 0x05424281L /* 0.328676706 */, 18 },
 { 0x05429de2L /* 0.328763850 */, 18 }, { 0x0542f944L /* 0.328851001 */, 18 },
 { 0x054354a8L /* 0.328938157 */, 18 }, { 0x0543b00dL /* 0.329025319 */, 18 },
 { 0x05440b74L /* 0.329112486 */, 18 }, { 0x054466ddL /* 0.329199660 */, 18 },
 { 0x0544c247L /* 0.329286839 */, 18 }, { 0x05451db2L /* 0.329374024 */, 18 },
 { 0x0545791fL /* 0.329461215 */, 18 }, { 0x0545d48eL /* 0.329548411 */, 18 },

 { 0x05462ffeL /* 0.329635614 */, 18 }, { 0x05468b70L /* 0.329722822 */, 18 },
 { 0x0546e6e3L /* 0.329810036 */, 18 }, { 0x05474258L /* 0.329897255 */, 18 },
 { 0x05479dceL /* 0.329984481 */, 18 }, { 0x0547f946L /* 0.330071712 */, 18 },
 { 0x054854c0L /* 0.330158949 */, 18 }, { 0x0548b03bL /* 0.330246191 */, 18 },
 { 0x05490bb7L /* 0.330333440 */, 18 }, { 0x05496735L /* 0.330420694 */, 18 },
 { 0x0549c2b5L /* 0.330507954 */, 18 }, { 0x054a1e36L /* 0.330595220 */, 18 },
 { 0x054a79b9L /* 0.330682491 */, 18 }, { 0x054ad53dL /* 0.330769768 */, 18 },
 { 0x054b30c3L /* 0.330857051 */, 18 }, { 0x054b8c4bL /* 0.330944340 */, 18 },

 { 0x054be7d4L /* 0.331031635 */, 18 }, { 0x054c435eL /* 0.331118935 */, 18 },
 { 0x054c9eeaL /* 0.331206241 */, 18 }, { 0x054cfa78L /* 0.331293553 */, 18 },
 { 0x054d5607L /* 0.331380870 */, 18 }, { 0x054db197L /* 0.331468193 */, 18 },
 { 0x054e0d2aL /* 0.331555522 */, 18 }, { 0x054e68bdL /* 0.331642857 */, 18 },
 { 0x054ec453L /* 0.331730198 */, 18 }, { 0x054f1fe9L /* 0.331817544 */, 18 },
 { 0x054f7b82L /* 0.331904896 */, 18 }, { 0x054fd71cL /* 0.331992254 */, 18 },
 { 0x055032b7L /* 0.332079617 */, 18 }, { 0x05508e54L /* 0.332166986 */, 18 },
 { 0x0550e9f3L /* 0.332254361 */, 18 }, { 0x05514593L /* 0.332341742 */, 18 },

 { 0x0551a134L /* 0.332429129 */, 18 }, { 0x0551fcd8L /* 0.332516521 */, 18 },
 { 0x0552587cL /* 0.332603919 */, 18 }, { 0x0552b423L /* 0.332691323 */, 18 },
 { 0x05530fcaL /* 0.332778732 */, 18 }, { 0x05536b74L /* 0.332866147 */, 18 },
 { 0x0553c71fL /* 0.332953568 */, 18 }, { 0x055422cbL /* 0.333040995 */, 18 },
 { 0x05547e79L /* 0.333128427 */, 18 }, { 0x0554da29L /* 0.333215865 */, 18 },
 { 0x055535daL /* 0.333303309 */, 18 }, { 0x0555918cL /* 0.333390759 */, 18 },
 { 0x0555ed40L /* 0.333478214 */, 18 }, { 0x055648f6L /* 0.333565675 */, 18 },
 { 0x0556a4adL /* 0.333653142 */, 18 }, { 0x05570066L /* 0.333740615 */, 18 },

 { 0x05575c20L /* 0.333828093 */, 18 }, { 0x0557b7dcL /* 0.333915577 */, 18 },
 { 0x05581399L /* 0.334003067 */, 18 }, { 0x05586f58L /* 0.334090562 */, 18 },
 { 0x0558cb19L /* 0.334178063 */, 18 }, { 0x055926dbL /* 0.334265570 */, 18 },
 { 0x0559829eL /* 0.334353083 */, 18 }, { 0x0559de63L /* 0.334440601 */, 18 },
 { 0x055a3a2aL /* 0.334528126 */, 18 }, { 0x055a95f2L /* 0.334615655 */, 18 },
 { 0x055af1bbL /* 0.334703191 */, 18 }, { 0x055b4d87L /* 0.334790732 */, 18 },
 { 0x055ba953L /* 0.334878279 */, 18 }, { 0x055c0522L /* 0.334965832 */, 18 },
 { 0x055c60f1L /* 0.335053391 */, 18 }, { 0x055cbcc3L /* 0.335140955 */, 18 },

 { 0x055d1896L /* 0.335228525 */, 18 }, { 0x055d746aL /* 0.335316100 */, 18 },
 { 0x055dd040L /* 0.335403682 */, 18 }, { 0x055e2c17L /* 0.335491269 */, 18 },
 { 0x055e87f0L /* 0.335578861 */, 18 }, { 0x055ee3cbL /* 0.335666460 */, 18 },
 { 0x055f3fa7L /* 0.335754064 */, 18 }, { 0x055f9b85L /* 0.335841674 */, 18 },
 { 0x055ff764L /* 0.335929290 */, 18 }, { 0x05605344L /* 0.336016911 */, 18 },
 { 0x0560af27L /* 0.336104538 */, 18 }, { 0x05610b0aL /* 0.336192171 */, 18 },
 { 0x056166f0L /* 0.336279809 */, 18 }, { 0x0561c2d7L /* 0.336367453 */, 18 },
 { 0x05621ebfL /* 0.336455103 */, 18 }, { 0x05627aa9L /* 0.336542759 */, 18 },

 { 0x0562d694L /* 0.336630420 */, 18 }, { 0x05633281L /* 0.336718087 */, 18 },
 { 0x05638e70L /* 0.336805760 */, 18 }, { 0x0563ea60L /* 0.336893439 */, 18 },
 { 0x05644651L /* 0.336981123 */, 18 }, { 0x0564a244L /* 0.337068813 */, 18 },
 { 0x0564fe39L /* 0.337156508 */, 18 }, { 0x05655a2fL /* 0.337244209 */, 18 },
 { 0x0565b627L /* 0.337331916 */, 18 }, { 0x05661220L /* 0.337419629 */, 18 },
 { 0x05666e1aL /* 0.337507347 */, 18 }, { 0x0566ca17L /* 0.337595071 */, 18 },
 { 0x05672614L /* 0.337682801 */, 18 }, { 0x05678214L /* 0.337770537 */, 18 },
 { 0x0567de15L /* 0.337858278 */, 18 }, { 0x05683a17L /* 0.337946025 */, 18 },

 { 0x0568961bL /* 0.338033777 */, 18 }, { 0x0568f220L /* 0.338121535 */, 18 },
 { 0x05694e27L /* 0.338209299 */, 18 }, { 0x0569aa30L /* 0.338297069 */, 18 },
 { 0x056a063aL /* 0.338384844 */, 18 }, { 0x056a6245L /* 0.338472625 */, 18 },
 { 0x056abe52L /* 0.338560412 */, 18 }, { 0x056b1a61L /* 0.338648204 */, 18 },
 { 0x056b7671L /* 0.338736002 */, 18 }, { 0x056bd283L /* 0.338823806 */, 18 },
 { 0x056c2e96L /* 0.338911616 */, 18 }, { 0x056c8aabL /* 0.338999431 */, 18 },
 { 0x056ce6c1L /* 0.339087252 */, 18 }, { 0x056d42d9L /* 0.339175078 */, 18 },
 { 0x056d9ef2L /* 0.339262910 */, 18 }, { 0x056dfb0dL /* 0.339350748 */, 18 },

 { 0x056e5729L /* 0.339438592 */, 18 }, { 0x056eb347L /* 0.339526441 */, 18 },
 { 0x056f0f66L /* 0.339614296 */, 18 }, { 0x056f6b87L /* 0.339702157 */, 18 },
 { 0x056fc7aaL /* 0.339790023 */, 18 }, { 0x057023cdL /* 0.339877895 */, 18 },
 { 0x05707ff3L /* 0.339965773 */, 18 }, { 0x0570dc1aL /* 0.340053656 */, 18 },
 { 0x05713843L /* 0.340141545 */, 18 }, { 0x0571946dL /* 0.340229440 */, 18 },
 { 0x0571f098L /* 0.340317340 */, 18 }, { 0x05724cc5L /* 0.340405246 */, 18 },
 { 0x0572a8f4L /* 0.340493158 */, 18 }, { 0x05730524L /* 0.340581075 */, 18 },
 { 0x05736156L /* 0.340668999 */, 18 }, { 0x0573bd89L /* 0.340756927 */, 18 },

 { 0x057419beL /* 0.340844862 */, 18 }, { 0x057475f4L /* 0.340932802 */, 18 },
 { 0x0574d22cL /* 0.341020748 */, 18 }, { 0x05752e65L /* 0.341108699 */, 18 },
 { 0x05758aa0L /* 0.341196656 */, 18 }, { 0x0575e6dcL /* 0.341284619 */, 18 },
 { 0x0576431aL /* 0.341372587 */, 18 }, { 0x05769f59L /* 0.341460562 */, 18 },
 { 0x0576fb9aL /* 0.341548541 */, 18 }, { 0x057757ddL /* 0.341636527 */, 18 },
 { 0x0577b421L /* 0.341724518 */, 18 }, { 0x05781066L /* 0.341812515 */, 18 },
 { 0x05786cadL /* 0.341900517 */, 18 }, { 0x0578c8f5L /* 0.341988525 */, 18 },
 { 0x0579253fL /* 0.342076539 */, 18 }, { 0x0579818bL /* 0.342164558 */, 18 },

 { 0x0579ddd8L /* 0.342252584 */, 18 }, { 0x057a3a27L /* 0.342340614 */, 18 },
 { 0x057a9677L /* 0.342428651 */, 18 }, { 0x057af2c8L /* 0.342516693 */, 18 },
 { 0x057b4f1cL /* 0.342604741 */, 18 }, { 0x057bab70L /* 0.342692794 */, 18 },
 { 0x057c07c6L /* 0.342780853 */, 18 }, { 0x057c641eL /* 0.342868918 */, 18 },
 { 0x057cc077L /* 0.342956988 */, 18 }, { 0x057d1cd2L /* 0.343045064 */, 18 },
 { 0x057d792eL /* 0.343133146 */, 18 }, { 0x057dd58cL /* 0.343221233 */, 18 },
 { 0x057e31ebL /* 0.343309326 */, 18 }, { 0x057e8e4cL /* 0.343397425 */, 18 },
 { 0x057eeaaeL /* 0.343485529 */, 18 }, { 0x057f4712L /* 0.343573639 */, 18 },

 { 0x057fa378L /* 0.343661754 */, 18 }, { 0x057fffdeL /* 0.343749876 */, 18 },
 { 0x05805c47L /* 0.343838003 */, 18 }, { 0x0580b8b1L /* 0.343926135 */, 18 },
 { 0x0581151cL /* 0.344014273 */, 18 }, { 0x05817189L /* 0.344102417 */, 18 },
 { 0x0581cdf7L /* 0.344190566 */, 18 }, { 0x05822a67L /* 0.344278722 */, 18 },
 { 0x058286d9L /* 0.344366882 */, 18 }, { 0x0582e34cL /* 0.344455049 */, 18 },
 { 0x05833fc0L /* 0.344543221 */, 18 }, { 0x05839c36L /* 0.344631398 */, 18 },
 { 0x0583f8aeL /* 0.344719582 */, 18 }, { 0x05845527L /* 0.344807771 */, 18 },
 { 0x0584b1a1L /* 0.344895965 */, 18 }, { 0x05850e1eL /* 0.344984165 */, 18 },

 { 0x05856a9bL /* 0.345072371 */, 18 }, { 0x0585c71aL /* 0.345160583 */, 18 },
 { 0x0586239bL /* 0.345248800 */, 18 }, { 0x0586801dL /* 0.345337023 */, 18 },
 { 0x0586dca1L /* 0.345425251 */, 18 }, { 0x05873926L /* 0.345513485 */, 18 },
 { 0x058795acL /* 0.345601725 */, 18 }, { 0x0587f235L /* 0.345689970 */, 18 },
 { 0x05884ebeL /* 0.345778221 */, 18 }, { 0x0588ab49L /* 0.345866478 */, 18 },
 { 0x058907d6L /* 0.345954740 */, 18 }, { 0x05896464L /* 0.346043008 */, 18 },
 { 0x0589c0f4L /* 0.346131281 */, 18 }, { 0x058a1d85L /* 0.346219560 */, 18 },
 { 0x058a7a18L /* 0.346307845 */, 18 }, { 0x058ad6acL /* 0.346396135 */, 18 },

 { 0x058b3342L /* 0.346484431 */, 18 }, { 0x058b8fd9L /* 0.346572733 */, 18 },
 { 0x058bec72L /* 0.346661040 */, 18 }, { 0x058c490cL /* 0.346749353 */, 18 },
 { 0x058ca5a8L /* 0.346837671 */, 18 }, { 0x058d0246L /* 0.346925996 */, 18 },
 { 0x058d5ee4L /* 0.347014325 */, 18 }, { 0x058dbb85L /* 0.347102661 */, 18 },
 { 0x058e1827L /* 0.347191002 */, 18 }, { 0x058e74caL /* 0.347279348 */, 18 },
 { 0x058ed16fL /* 0.347367700 */, 18 }, { 0x058f2e15L /* 0.347456058 */, 18 },
 { 0x058f8abdL /* 0.347544422 */, 18 }, { 0x058fe766L /* 0.347632791 */, 18 },
 { 0x05904411L /* 0.347721165 */, 18 }, { 0x0590a0beL /* 0.347809546 */, 18 },

 { 0x0590fd6cL /* 0.347897931 */, 18 }, { 0x05915a1bL /* 0.347986323 */, 18 },
 { 0x0591b6ccL /* 0.348074720 */, 18 }, { 0x0592137eL /* 0.348163123 */, 18 },
 { 0x05927032L /* 0.348251531 */, 18 }, { 0x0592cce8L /* 0.348339945 */, 18 },
 { 0x0593299fL /* 0.348428365 */, 18 }, { 0x05938657L /* 0.348516790 */, 18 },
 { 0x0593e311L /* 0.348605221 */, 18 }, { 0x05943fcdL /* 0.348693657 */, 18 },
 { 0x05949c8aL /* 0.348782099 */, 18 }, { 0x0594f948L /* 0.348870547 */, 18 },
 { 0x05955608L /* 0.348959000 */, 18 }, { 0x0595b2caL /* 0.349047459 */, 18 },
 { 0x05960f8cL /* 0.349135923 */, 18 }, { 0x05966c51L /* 0.349224393 */, 18 },

 { 0x0596c917L /* 0.349312869 */, 18 }, { 0x059725deL /* 0.349401350 */, 18 },
 { 0x059782a7L /* 0.349489837 */, 18 }, { 0x0597df72L /* 0.349578329 */, 18 },
 { 0x05983c3eL /* 0.349666827 */, 18 }, { 0x0598990cL /* 0.349755331 */, 18 },
 { 0x0598f5dbL /* 0.349843840 */, 18 }, { 0x059952abL /* 0.349932355 */, 18 },
 { 0x0599af7dL /* 0.350020876 */, 18 }, { 0x059a0c51L /* 0.350109402 */, 18 },
 { 0x059a6926L /* 0.350197933 */, 18 }, { 0x059ac5fcL /* 0.350286470 */, 18 },
 { 0x059b22d4L /* 0.350375013 */, 18 }, { 0x059b7faeL /* 0.350463562 */, 18 },
 { 0x059bdc89L /* 0.350552116 */, 18 }, { 0x059c3965L /* 0.350640675 */, 18 },

 { 0x059c9643L /* 0.350729240 */, 18 }, { 0x059cf323L /* 0.350817811 */, 18 },
 { 0x059d5004L /* 0.350906388 */, 18 }, { 0x059dace6L /* 0.350994970 */, 18 },
 { 0x059e09cbL /* 0.351083557 */, 18 }, { 0x059e66b0L /* 0.351172150 */, 18 },
 { 0x059ec397L /* 0.351260749 */, 18 }, { 0x059f2080L /* 0.351349353 */, 18 },
 { 0x059f7d6aL /* 0.351437963 */, 18 }, { 0x059fda55L /* 0.351526579 */, 18 },
 { 0x05a03742L /* 0.351615200 */, 18 }, { 0x05a09431L /* 0.351703827 */, 18 },
 { 0x05a0f121L /* 0.351792459 */, 18 }, { 0x05a14e12L /* 0.351881097 */, 18 },
 { 0x05a1ab05L /* 0.351969740 */, 18 }, { 0x05a207faL /* 0.352058389 */, 18 },

 { 0x05a264f0L /* 0.352147044 */, 18 }, { 0x05a2c1e7L /* 0.352235704 */, 18 },
 { 0x05a31ee1L /* 0.352324369 */, 18 }, { 0x05a37bdbL /* 0.352413041 */, 18 },
 { 0x05a3d8d7L /* 0.352501718 */, 18 }, { 0x05a435d5L /* 0.352590400 */, 18 },
 { 0x05a492d4L /* 0.352679088 */, 18 }, { 0x05a4efd4L /* 0.352767782 */, 18 },
 { 0x05a54cd6L /* 0.352856481 */, 18 }, { 0x05a5a9daL /* 0.352945186 */, 18 },
 { 0x05a606dfL /* 0.353033896 */, 18 }, { 0x05a663e5L /* 0.353122612 */, 18 },
 { 0x05a6c0edL /* 0.353211333 */, 18 }, { 0x05a71df7L /* 0.353300061 */, 18 },
 { 0x05a77b02L /* 0.353388793 */, 18 }, { 0x05a7d80eL /* 0.353477531 */, 18 },

 { 0x05a8351cL /* 0.353566275 */, 18 }, { 0x05a8922cL /* 0.353655024 */, 18 },
 { 0x05a8ef3cL /* 0.353743779 */, 18 }, { 0x05a94c4fL /* 0.353832540 */, 18 },
 { 0x05a9a963L /* 0.353921306 */, 18 }, { 0x05aa0678L /* 0.354010077 */, 18 },
 { 0x05aa638fL /* 0.354098855 */, 18 }, { 0x05aac0a8L /* 0.354187637 */, 18 },
 { 0x05ab1dc2L /* 0.354276426 */, 18 }, { 0x05ab7addL /* 0.354365220 */, 18 },
 { 0x05abd7faL /* 0.354454019 */, 18 }, { 0x05ac3518L /* 0.354542824 */, 18 },
 { 0x05ac9238L /* 0.354631635 */, 18 }, { 0x05acef5aL /* 0.354720451 */, 18 },
 { 0x05ad4c7dL /* 0.354809272 */, 18 }, { 0x05ada9a1L /* 0.354898100 */, 18 },

 { 0x05ae06c7L /* 0.354986932 */, 18 }, { 0x05ae63eeL /* 0.355075771 */, 18 },
 { 0x05aec117L /* 0.355164615 */, 18 }, { 0x05af1e41L /* 0.355253464 */, 18 },
 { 0x05af7b6dL /* 0.355342319 */, 18 }, { 0x05afd89bL /* 0.355431180 */, 18 },
 { 0x05b035c9L /* 0.355520046 */, 18 }, { 0x05b092faL /* 0.355608917 */, 18 },
 { 0x05b0f02bL /* 0.355697795 */, 18 }, { 0x05b14d5fL /* 0.355786677 */, 18 },
 { 0x05b1aa94L /* 0.355875566 */, 18 }, { 0x05b207caL /* 0.355964460 */, 18 },
 { 0x05b26502L /* 0.356053359 */, 18 }, { 0x05b2c23bL /* 0.356142264 */, 18 },
 { 0x05b31f76L /* 0.356231175 */, 18 }, { 0x05b37cb2L /* 0.356320091 */, 18 },

 { 0x05b3d9f0L /* 0.356409012 */, 18 }, { 0x05b4372fL /* 0.356497940 */, 18 },
 { 0x05b4946fL /* 0.356586872 */, 18 }, { 0x05b4f1b2L /* 0.356675811 */, 18 },
 { 0x05b54ef5L /* 0.356764754 */, 18 }, { 0x05b5ac3aL /* 0.356853704 */, 18 },
 { 0x05b60981L /* 0.356942659 */, 18 }, { 0x05b666c9L /* 0.357031619 */, 18 },
 { 0x05b6c413L /* 0.357120585 */, 18 }, { 0x05b7215eL /* 0.357209557 */, 18 },
 { 0x05b77eabL /* 0.357298534 */, 18 }, { 0x05b7dbf9L /* 0.357387516 */, 18 },
 { 0x05b83948L /* 0.357476504 */, 18 }, { 0x05b89699L /* 0.357565498 */, 18 },
 { 0x05b8f3ecL /* 0.357654497 */, 18 }, { 0x05b95140L /* 0.357743502 */, 18 },

 { 0x05b9ae95L /* 0.357832512 */, 18 }, { 0x05ba0becL /* 0.357921528 */, 18 },
 { 0x05ba6945L /* 0.358010550 */, 18 }, { 0x05bac69fL /* 0.358099576 */, 18 },
 { 0x05bb23faL /* 0.358188609 */, 18 }, { 0x05bb8157L /* 0.358277647 */, 18 },
 { 0x05bbdeb6L /* 0.358366690 */, 18 }, { 0x05bc3c16L /* 0.358455739 */, 18 },
 { 0x05bc9977L /* 0.358544794 */, 18 }, { 0x05bcf6daL /* 0.358633854 */, 18 },
 { 0x05bd543eL /* 0.358722920 */, 18 }, { 0x05bdb1a4L /* 0.358811991 */, 18 },
 { 0x05be0f0bL /* 0.358901067 */, 18 }, { 0x05be6c74L /* 0.358990150 */, 18 },
 { 0x05bec9dfL /* 0.359079237 */, 18 }, { 0x05bf274aL /* 0.359168331 */, 18 },

 { 0x05bf84b8L /* 0.359257429 */, 18 }, { 0x05bfe226L /* 0.359346534 */, 18 },
 { 0x05c03f97L /* 0.359435644 */, 18 }, { 0x05c09d08L /* 0.359524759 */, 18 },
 { 0x05c0fa7cL /* 0.359613880 */, 18 }, { 0x05c157f0L /* 0.359703006 */, 18 },
 { 0x05c1b566L /* 0.359792138 */, 18 }, { 0x05c212deL /* 0.359881276 */, 18 },
 { 0x05c27057L /* 0.359970419 */, 18 }, { 0x05c2cdd2L /* 0.360059567 */, 18 },
 { 0x05c32b4eL /* 0.360148721 */, 18 }, { 0x05c388cbL /* 0.360237881 */, 18 },
 { 0x05c3e64bL /* 0.360327046 */, 18 }, { 0x05c443cbL /* 0.360416216 */, 18 },
 { 0x05c4a14dL /* 0.360505392 */, 18 }, { 0x05c4fed1L /* 0.360594574 */, 18 },

 { 0x05c55c56L /* 0.360683761 */, 18 }, { 0x05c5b9dcL /* 0.360772953 */, 18 },
 { 0x05c61764L /* 0.360862152 */, 18 }, { 0x05c674edL /* 0.360951355 */, 18 },
 { 0x05c6d278L /* 0.361040564 */, 18 }, { 0x05c73005L /* 0.361129779 */, 18 },
 { 0x05c78d93L /* 0.361218999 */, 18 }, { 0x05c7eb22L /* 0.361308225 */, 18 },
 { 0x05c848b3L /* 0.361397456 */, 18 }, { 0x05c8a645L /* 0.361486693 */, 18 },
 { 0x05c903d9L /* 0.361575935 */, 18 }, { 0x05c9616eL /* 0.361665183 */, 18 },
 { 0x05c9bf05L /* 0.361754436 */, 18 }, { 0x05ca1c9dL /* 0.361843695 */, 18 },
 { 0x05ca7a37L /* 0.361932959 */, 18 }, { 0x05cad7d2L /* 0.362022229 */, 18 },

 { 0x05cb356eL /* 0.362111504 */, 18 }, { 0x05cb930dL /* 0.362200785 */, 18 },
 { 0x05cbf0acL /* 0.362290071 */, 18 }, { 0x05cc4e4dL /* 0.362379362 */, 18 },
 { 0x05ccabf0L /* 0.362468660 */, 18 }, { 0x05cd0994L /* 0.362557962 */, 18 },
 { 0x05cd6739L /* 0.362647271 */, 18 }, { 0x05cdc4e0L /* 0.362736584 */, 18 },
 { 0x05ce2289L /* 0.362825904 */, 18 }, { 0x05ce8033L /* 0.362915228 */, 18 },
 { 0x05cedddeL /* 0.363004559 */, 18 }, { 0x05cf3b8bL /* 0.363093894 */, 18 },
 { 0x05cf9939L /* 0.363183236 */, 18 }, { 0x05cff6e9L /* 0.363272582 */, 18 },
 { 0x05d0549aL /* 0.363361935 */, 18 }, { 0x05d0b24dL /* 0.363451292 */, 18 },

 { 0x05d11001L /* 0.363540655 */, 18 }, { 0x05d16db7L /* 0.363630024 */, 18 },
 { 0x05d1cb6eL /* 0.363719398 */, 18 }, { 0x05d22927L /* 0.363808778 */, 18 },
 { 0x05d286e1L /* 0.363898163 */, 18 }, { 0x05d2e49dL /* 0.363987554 */, 18 },
 { 0x05d3425aL /* 0.364076950 */, 18 }, { 0x05d3a018L /* 0.364166352 */, 18 },
 { 0x05d3fdd8L /* 0.364255759 */, 18 }, { 0x05d45b9aL /* 0.364345171 */, 18 },
 { 0x05d4b95dL /* 0.364434589 */, 18 }, { 0x05d51721L /* 0.364524013 */, 18 },
 { 0x05d574e7L /* 0.364613442 */, 18 }, { 0x05d5d2afL /* 0.364702877 */, 18 },
 { 0x05d63078L /* 0.364792317 */, 18 }, { 0x05d68e42L /* 0.364881762 */, 18 },

 { 0x05d6ec0eL /* 0.364971213 */, 18 }, { 0x05d749dbL /* 0.365060669 */, 18 },
 { 0x05d7a7aaL /* 0.365150131 */, 18 }, { 0x05d8057aL /* 0.365239599 */, 18 },
 { 0x05d8634cL /* 0.365329072 */, 18 }, { 0x05d8c11fL /* 0.365418550 */, 18 },
 { 0x05d91ef4L /* 0.365508034 */, 18 }, { 0x05d97ccaL /* 0.365597523 */, 18 },
 { 0x05d9daa1L /* 0.365687018 */, 18 }, { 0x05da387aL /* 0.365776518 */, 18 },
 { 0x05da9655L /* 0.365866024 */, 18 }, { 0x05daf431L /* 0.365955536 */, 18 },
 { 0x05db520eL /* 0.366045052 */, 18 }, { 0x05dbafedL /* 0.366134574 */, 18 },
 { 0x05dc0dceL /* 0.366224102 */, 18 }, { 0x05dc6bafL /* 0.366313635 */, 18 },

 { 0x05dcc993L /* 0.366403174 */, 18 }, { 0x05dd2778L /* 0.366492718 */, 18 },
 { 0x05dd855eL /* 0.366582267 */, 18 }, { 0x05dde346L /* 0.366671822 */, 18 },
 { 0x05de412fL /* 0.366761383 */, 18 }, { 0x05de9f1aL /* 0.366850949 */, 18 },
 { 0x05defd06L /* 0.366940520 */, 18 }, { 0x05df5af3L /* 0.367030097 */, 18 },
 { 0x05dfb8e2L /* 0.367119680 */, 18 }, { 0x05e016d3L /* 0.367209267 */, 18 },
 { 0x05e074c5L /* 0.367298861 */, 18 }, { 0x05e0d2b8L /* 0.367388459 */, 18 },
 { 0x05e130adL /* 0.367478064 */, 18 }, { 0x05e18ea4L /* 0.367567673 */, 18 },
 { 0x05e1ec9cL /* 0.367657288 */, 18 }, { 0x05e24a95L /* 0.367746909 */, 18 },

 { 0x05e2a890L /* 0.367836535 */, 18 }, { 0x05e3068cL /* 0.367926167 */, 18 },
 { 0x05e3648aL /* 0.368015804 */, 18 }, { 0x05e3c289L /* 0.368105446 */, 18 },
 { 0x05e4208aL /* 0.368195094 */, 18 }, { 0x05e47e8cL /* 0.368284747 */, 18 },
 { 0x05e4dc8fL /* 0.368374406 */, 18 }, { 0x05e53a94L /* 0.368464070 */, 18 },
 { 0x05e5989bL /* 0.368553740 */, 18 }, { 0x05e5f6a3L /* 0.368643415 */, 18 },
 { 0x05e654acL /* 0.368733096 */, 18 }, { 0x05e6b2b7L /* 0.368822782 */, 18 },
 { 0x05e710c4L /* 0.368912473 */, 18 }, { 0x05e76ed2L /* 0.369002170 */, 18 },
 { 0x05e7cce1L /* 0.369091873 */, 18 }, { 0x05e82af2L /* 0.369181581 */, 18 },

 { 0x05e88904L /* 0.369271294 */, 18 }, { 0x05e8e718L /* 0.369361013 */, 18 },
 { 0x05e9452dL /* 0.369450737 */, 18 }, { 0x05e9a343L /* 0.369540467 */, 18 },
 { 0x05ea015cL /* 0.369630202 */, 18 }, { 0x05ea5f75L /* 0.369719942 */, 18 },
 { 0x05eabd90L /* 0.369809688 */, 18 }, { 0x05eb1badL /* 0.369899440 */, 18 },
 { 0x05eb79cbL /* 0.369989197 */, 18 }, { 0x05ebd7eaL /* 0.370078959 */, 18 },
 { 0x05ec360bL /* 0.370168727 */, 18 }, { 0x05ec942dL /* 0.370258500 */, 18 },
 { 0x05ecf251L /* 0.370348279 */, 18 }, { 0x05ed5076L /* 0.370438063 */, 18 },
 { 0x05edae9dL /* 0.370527853 */, 18 }, { 0x05ee0cc5L /* 0.370617648 */, 18 },

 { 0x05ee6aefL /* 0.370707448 */, 18 }, { 0x05eec91aL /* 0.370797254 */, 18 },
 { 0x05ef2746L /* 0.370887065 */, 18 }, { 0x05ef8574L /* 0.370976882 */, 18 },
 { 0x05efe3a4L /* 0.371066704 */, 18 }, { 0x05f041d5L /* 0.371156532 */, 18 },
 { 0x05f0a007L /* 0.371246365 */, 18 }, { 0x05f0fe3bL /* 0.371336203 */, 18 },
 { 0x05f15c70L /* 0.371426047 */, 18 }, { 0x05f1baa7L /* 0.371515897 */, 18 },
 { 0x05f218dfL /* 0.371605751 */, 18 }, { 0x05f27719L /* 0.371695612 */, 18 },
 { 0x05f2d554L /* 0.371785477 */, 18 }, { 0x05f33390L /* 0.371875348 */, 18 },
 { 0x05f391cfL /* 0.371965225 */, 18 }, { 0x05f3f00eL /* 0.372055107 */, 18 },

 { 0x05f44e4fL /* 0.372144994 */, 18 }, { 0x05f4ac91L /* 0.372234887 */, 18 },
 { 0x05f50ad5L /* 0.372324785 */, 18 }, { 0x05f5691bL /* 0.372414689 */, 18 },
 { 0x05f5c761L /* 0.372504598 */, 18 }, { 0x05f625aaL /* 0.372594513 */, 18 },
 { 0x05f683f3L /* 0.372684433 */, 18 }, { 0x05f6e23fL /* 0.372774358 */, 18 },
 { 0x05f7408bL /* 0.372864289 */, 18 }, { 0x05f79ed9L /* 0.372954225 */, 18 },
 { 0x05f7fd29L /* 0.373044167 */, 18 }, { 0x05f85b7aL /* 0.373134114 */, 18 },
 { 0x05f8b9ccL /* 0.373224066 */, 18 }, { 0x05f91820L /* 0.373314024 */, 18 },
 { 0x05f97675L /* 0.373403987 */, 18 }, { 0x05f9d4ccL /* 0.373493956 */, 18 },

 { 0x05fa3324L /* 0.373583930 */, 18 }, { 0x05fa917eL /* 0.373673910 */, 18 },
 { 0x05faefd9L /* 0.373763895 */, 18 }, { 0x05fb4e36L /* 0.373853885 */, 18 },
 { 0x05fbac94L /* 0.373943881 */, 18 }, { 0x05fc0af3L /* 0.374033882 */, 18 },
 { 0x05fc6954L /* 0.374123889 */, 18 }, { 0x05fcc7b7L /* 0.374213901 */, 18 },
 { 0x05fd261bL /* 0.374303918 */, 18 }, { 0x05fd8480L /* 0.374393941 */, 18 },
 { 0x05fde2e7L /* 0.374483970 */, 18 }, { 0x05fe414fL /* 0.374574003 */, 18 },
 { 0x05fe9fb9L /* 0.374664042 */, 18 }, { 0x05fefe24L /* 0.374754087 */, 18 },
 { 0x05ff5c91L /* 0.374844137 */, 18 }, { 0x05ffbaffL /* 0.374934192 */, 18 },

 { 0x0600196eL /* 0.375024253 */, 18 }, { 0x060077dfL /* 0.375114319 */, 18 },
 { 0x0600d651L /* 0.375204391 */, 18 }, { 0x060134c5L /* 0.375294468 */, 18 },
 { 0x0601933bL /* 0.375384550 */, 18 }, { 0x0601f1b1L /* 0.375474638 */, 18 },
 { 0x0602502aL /* 0.375564731 */, 18 }, { 0x0602aea3L /* 0.375654830 */, 18 },
 { 0x06030d1eL /* 0.375744934 */, 18 }, { 0x06036b9bL /* 0.375835043 */, 18 },
 { 0x0603ca19L /* 0.375925158 */, 18 }, { 0x06042898L /* 0.376015278 */, 18 },
 { 0x06048719L /* 0.376105404 */, 18 }, { 0x0604e59cL /* 0.376195535 */, 18 },
 { 0x0605441fL /* 0.376285671 */, 18 }, { 0x0605a2a5L /* 0.376375813 */, 18 },

 { 0x0606012bL /* 0.376465960 */, 18 }, { 0x06065fb4L /* 0.376556113 */, 18 },
 { 0x0606be3dL /* 0.376646271 */, 18 }, { 0x06071cc8L /* 0.376736434 */, 18 },
 { 0x06077b55L /* 0.376826603 */, 18 }, { 0x0607d9e3L /* 0.376916777 */, 18 },
 { 0x06083872L /* 0.377006957 */, 18 }, { 0x06089703L /* 0.377097141 */, 18 },
 { 0x0608f595L /* 0.377187332 */, 18 }, { 0x06095429L /* 0.377277528 */, 18 },
 { 0x0609b2beL /* 0.377367729 */, 18 }, { 0x060a1155L /* 0.377457935 */, 18 },
 { 0x060a6fedL /* 0.377548147 */, 18 }, { 0x060ace86L /* 0.377638364 */, 18 },
 { 0x060b2d21L /* 0.377728587 */, 18 }, { 0x060b8bbeL /* 0.377818815 */, 18 },

 { 0x060bea5cL /* 0.377909049 */, 18 }, { 0x060c48fbL /* 0.377999288 */, 18 },
 { 0x060ca79cL /* 0.378089532 */, 18 }, { 0x060d063eL /* 0.378179781 */, 18 },
 { 0x060d64e1L /* 0.378270036 */, 18 }, { 0x060dc387L /* 0.378360297 */, 18 },
 { 0x060e222dL /* 0.378450563 */, 18 }, { 0x060e80d5L /* 0.378540834 */, 18 },
 { 0x060edf7fL /* 0.378631110 */, 18 }, { 0x060f3e29L /* 0.378721392 */, 18 },
 { 0x060f9cd6L /* 0.378811680 */, 18 }, { 0x060ffb83L /* 0.378901972 */, 18 },
 { 0x06105a33L /* 0.378992270 */, 18 }, { 0x0610b8e3L /* 0.379082574 */, 18 },
 { 0x06111795L /* 0.379172883 */, 18 }, { 0x06117649L /* 0.379263197 */, 18 },

 { 0x0611d4feL /* 0.379353516 */, 18 }, { 0x061233b4L /* 0.379443841 */, 18 },
 { 0x0612926cL /* 0.379534172 */, 18 }, { 0x0612f125L /* 0.379624507 */, 18 },
 { 0x06134fe0L /* 0.379714848 */, 18 }, { 0x0613ae9cL /* 0.379805195 */, 18 },
 { 0x06140d5aL /* 0.379895547 */, 18 }, { 0x06146c19L /* 0.379985904 */, 18 },
 { 0x0614cadaL /* 0.380076266 */, 18 }, { 0x0615299cL /* 0.380166634 */, 18 },
 { 0x0615885fL /* 0.380257008 */, 18 }, { 0x0615e724L /* 0.380347386 */, 18 },
 { 0x061645eaL /* 0.380437770 */, 18 }, { 0x0616a4b2L /* 0.380528160 */, 18 },
 { 0x0617037bL /* 0.380618555 */, 18 }, { 0x06176246L /* 0.380708955 */, 18 },

 { 0x0617c112L /* 0.380799360 */, 18 }, { 0x06181fdfL /* 0.380889771 */, 18 },
 { 0x06187eaeL /* 0.380980187 */, 18 }, { 0x0618dd7eL /* 0.381070609 */, 18 },
 { 0x06193c50L /* 0.381161036 */, 18 }, { 0x06199b24L /* 0.381251468 */, 18 },
 { 0x0619f9f8L /* 0.381341906 */, 18 }, { 0x061a58ceL /* 0.381432349 */, 18 },
 { 0x061ab7a6L /* 0.381522798 */, 18 }, { 0x061b167fL /* 0.381613251 */, 18 },
 { 0x061b7559L /* 0.381703711 */, 18 }, { 0x061bd435L /* 0.381794175 */, 18 },
 { 0x061c3313L /* 0.381884645 */, 18 }, { 0x061c91f1L /* 0.381975120 */, 18 },
 { 0x061cf0d2L /* 0.382065601 */, 18 }, { 0x061d4fb3L /* 0.382156087 */, 18 },

 { 0x061dae96L /* 0.382246578 */, 18 }, { 0x061e0d7bL /* 0.382337075 */, 18 },
 { 0x061e6c61L /* 0.382427577 */, 18 }, { 0x061ecb48L /* 0.382518084 */, 18 },
 { 0x061f2a31L /* 0.382608597 */, 18 }, { 0x061f891bL /* 0.382699115 */, 18 },
 { 0x061fe807L /* 0.382789638 */, 18 }, { 0x062046f4L /* 0.382880167 */, 18 },
 { 0x0620a5e3L /* 0.382970701 */, 18 }, { 0x062104d3L /* 0.383061241 */, 18 },
 { 0x062163c4L /* 0.383151786 */, 18 }, { 0x0621c2b7L /* 0.383242336 */, 18 },
 { 0x062221abL /* 0.383332891 */, 18 }, { 0x062280a1L /* 0.383423452 */, 18 },
 { 0x0622df98L /* 0.383514018 */, 18 }, { 0x06233e91L /* 0.383604590 */, 18 },

 { 0x06239d8bL /* 0.383695167 */, 18 }, { 0x0623fc86L /* 0.383785749 */, 18 },
 { 0x06245b83L /* 0.383876337 */, 18 }, { 0x0624ba82L /* 0.383966930 */, 18 },
 { 0x06251981L /* 0.384057528 */, 18 }, { 0x06257883L /* 0.384148132 */, 18 },
 { 0x0625d785L /* 0.384238741 */, 18 }, { 0x06263689L /* 0.384329355 */, 18 },
 { 0x0626958fL /* 0.384419975 */, 18 }, { 0x0626f496L /* 0.384510600 */, 18 },
 { 0x0627539eL /* 0.384601230 */, 18 }, { 0x0627b2a8L /* 0.384691866 */, 18 },
 { 0x062811b3L /* 0.384782507 */, 18 }, { 0x062870c0L /* 0.384873153 */, 18 },
 { 0x0628cfceL /* 0.384963805 */, 18 }, { 0x06292edeL /* 0.385054462 */, 18 },

 { 0x06298defL /* 0.385145124 */, 18 }, { 0x0629ed01L /* 0.385235792 */, 18 },
 { 0x062a4c15L /* 0.385326465 */, 18 }, { 0x062aab2aL /* 0.385417143 */, 18 },
 { 0x062b0a41L /* 0.385507827 */, 18 }, { 0x062b6959L /* 0.385598516 */, 18 },
 { 0x062bc873L /* 0.385689211 */, 18 }, { 0x062c278eL /* 0.385779910 */, 18 },
 { 0x062c86aaL /* 0.385870615 */, 18 }, { 0x062ce5c8L /* 0.385961326 */, 18 },
 { 0x062d44e8L /* 0.386052041 */, 18 }, { 0x062da408L /* 0.386142762 */, 18 },
 { 0x062e032aL /* 0.386233489 */, 18 }, { 0x062e624eL /* 0.386324221 */, 18 },
 { 0x062ec173L /* 0.386414958 */, 18 }, { 0x062f209aL /* 0.386505700 */, 18 },

 { 0x062f7fc1L /* 0.386596448 */, 18 }, { 0x062fdeebL /* 0.386687201 */, 18 },
 { 0x06303e16L /* 0.386777959 */, 18 }, { 0x06309d42L /* 0.386868723 */, 18 },
 { 0x0630fc6fL /* 0.386959492 */, 18 }, { 0x06315b9eL /* 0.387050266 */, 18 },
 { 0x0631bacfL /* 0.387141045 */, 18 }, { 0x06321a01L /* 0.387231830 */, 18 },
 { 0x06327934L /* 0.387322621 */, 18 }, { 0x0632d869L /* 0.387413416 */, 18 },
 { 0x0633379fL /* 0.387504217 */, 18 }, { 0x063396d7L /* 0.387595023 */, 18 },
 { 0x0633f610L /* 0.387685835 */, 18 }, { 0x0634554aL /* 0.387776652 */, 18 },
 { 0x0634b486L /* 0.387867474 */, 18 }, { 0x063513c3L /* 0.387958301 */, 18 },

 { 0x06357302L /* 0.388049134 */, 18 }, { 0x0635d242L /* 0.388139972 */, 18 },
 { 0x06363184L /* 0.388230816 */, 18 }, { 0x063690c7L /* 0.388321665 */, 18 },
 { 0x0636f00bL /* 0.388412519 */, 18 }, { 0x06374f51L /* 0.388503378 */, 18 },
 { 0x0637ae99L /* 0.388594243 */, 18 }, { 0x06380de1L /* 0.388685113 */, 18 },
 { 0x06386d2bL /* 0.388775988 */, 18 }, { 0x0638cc77L /* 0.388866869 */, 18 },
 { 0x06392bc4L /* 0.388957755 */, 18 }, { 0x06398b12L /* 0.389048646 */, 18 },
 { 0x0639ea62L /* 0.389139542 */, 18 }, { 0x063a49b4L /* 0.389230444 */, 18 },
 { 0x063aa906L /* 0.389321352 */, 18 }, { 0x063b085aL /* 0.389412264 */, 18 },

 { 0x063b67b0L /* 0.389503182 */, 18 }, { 0x063bc707L /* 0.389594105 */, 18 },
 { 0x063c265fL /* 0.389685033 */, 18 }, { 0x063c85b9L /* 0.389775967 */, 18 },
 { 0x063ce514L /* 0.389866906 */, 18 }, { 0x063d4471L /* 0.389957850 */, 18 },
 { 0x063da3cfL /* 0.390048800 */, 18 }, { 0x063e032fL /* 0.390139755 */, 18 },
 { 0x063e6290L /* 0.390230715 */, 18 }, { 0x063ec1f2L /* 0.390321681 */, 18 },
 { 0x063f2156L /* 0.390412651 */, 18 }, { 0x063f80bbL /* 0.390503628 */, 18 },
 { 0x063fe022L /* 0.390594609 */, 18 }, { 0x06403f8aL /* 0.390685596 */, 18 },
 { 0x06409ef3L /* 0.390776588 */, 18 }, { 0x0640fe5eL /* 0.390867585 */, 18 },

 { 0x06415dcbL /* 0.390958588 */, 18 }, { 0x0641bd38L /* 0.391049596 */, 18 },
 { 0x06421ca7L /* 0.391140609 */, 18 }, { 0x06427c18L /* 0.391231627 */, 18 },
 { 0x0642db8aL /* 0.391322651 */, 18 }, { 0x06433afdL /* 0.391413680 */, 18 },
 { 0x06439a72L /* 0.391504714 */, 18 }, { 0x0643f9e9L /* 0.391595754 */, 18 },
 { 0x06445960L /* 0.391686799 */, 18 }, { 0x0644b8d9L /* 0.391777849 */, 18 },
 { 0x06451854L /* 0.391868905 */, 18 }, { 0x064577d0L /* 0.391959966 */, 18 },
 { 0x0645d74dL /* 0.392051032 */, 18 }, { 0x064636ccL /* 0.392142103 */, 18 },
 { 0x0646964cL /* 0.392233180 */, 18 }, { 0x0646f5ceL /* 0.392324262 */, 18 },

 { 0x06475551L /* 0.392415349 */, 18 }, { 0x0647b4d5L /* 0.392506442 */, 18 },
 { 0x0648145bL /* 0.392597540 */, 18 }, { 0x064873e3L /* 0.392688643 */, 18 },
 { 0x0648d36bL /* 0.392779751 */, 18 }, { 0x064932f6L /* 0.392870865 */, 18 },
 { 0x06499281L /* 0.392961984 */, 18 }, { 0x0649f20eL /* 0.393053108 */, 18 },
 { 0x064a519cL /* 0.393144238 */, 18 }, { 0x064ab12cL /* 0.393235372 */, 18 },
 { 0x064b10beL /* 0.393326513 */, 18 }, { 0x064b7050L /* 0.393417658 */, 18 },
 { 0x064bcfe4L /* 0.393508809 */, 18 }, { 0x064c2f7aL /* 0.393599965 */, 18 },
 { 0x064c8f11L /* 0.393691126 */, 18 }, { 0x064ceea9L /* 0.393782292 */, 18 },

 { 0x064d4e43L /* 0.393873464 */, 18 }, { 0x064daddeL /* 0.393964641 */, 18 },
 { 0x064e0d7aL /* 0.394055823 */, 18 }, { 0x064e6d18L /* 0.394147011 */, 18 },
 { 0x064eccb8L /* 0.394238204 */, 18 }, { 0x064f2c59L /* 0.394329402 */, 18 },
 { 0x064f8bfbL /* 0.394420605 */, 18 }, { 0x064feb9eL /* 0.394511814 */, 18 },
 { 0x06504b44L /* 0.394603028 */, 18 }, { 0x0650aaeaL /* 0.394694247 */, 18 },
 { 0x06510a92L /* 0.394785472 */, 18 }, { 0x06516a3bL /* 0.394876702 */, 18 },
 { 0x0651c9e6L /* 0.394967937 */, 18 }, { 0x06522992L /* 0.395059177 */, 18 },
 { 0x06528940L /* 0.395150423 */, 18 }, { 0x0652e8efL /* 0.395241673 */, 18 },

 { 0x0653489fL /* 0.395332930 */, 18 }, { 0x0653a851L /* 0.395424191 */, 18 },
 { 0x06540804L /* 0.395515458 */, 18 }, { 0x065467b9L /* 0.395606730 */, 18 },
 { 0x0654c76fL /* 0.395698007 */, 18 }, { 0x06552726L /* 0.395789289 */, 18 },
 { 0x065586dfL /* 0.395880577 */, 18 }, { 0x0655e699L /* 0.395971870 */, 18 },
 { 0x06564655L /* 0.396063168 */, 18 }, { 0x0656a612L /* 0.396154472 */, 18 },
 { 0x065705d0L /* 0.396245780 */, 18 }, { 0x06576590L /* 0.396337094 */, 18 },
 { 0x0657c552L /* 0.396428414 */, 18 }, { 0x06582514L /* 0.396519738 */, 18 },
 { 0x065884d9L /* 0.396611068 */, 18 }, { 0x0658e49eL /* 0.396702403 */, 18 },

 { 0x06594465L /* 0.396793743 */, 18 }, { 0x0659a42eL /* 0.396885089 */, 18 },
 { 0x065a03f7L /* 0.396976440 */, 18 }, { 0x065a63c3L /* 0.397067796 */, 18 },
 { 0x065ac38fL /* 0.397159157 */, 18 }, { 0x065b235dL /* 0.397250524 */, 18 },
 { 0x065b832dL /* 0.397341896 */, 18 }, { 0x065be2feL /* 0.397433273 */, 18 },
 { 0x065c42d0L /* 0.397524655 */, 18 }, { 0x065ca2a3L /* 0.397616043 */, 18 },
 { 0x065d0279L /* 0.397707436 */, 18 }, { 0x065d624fL /* 0.397798834 */, 18 },
 { 0x065dc227L /* 0.397890237 */, 18 }, { 0x065e2200L /* 0.397981646 */, 18 },
 { 0x065e81dbL /* 0.398073059 */, 18 }, { 0x065ee1b7L /* 0.398164479 */, 18 },

 { 0x065f4195L /* 0.398255903 */, 18 }, { 0x065fa174L /* 0.398347333 */, 18 },
 { 0x06600154L /* 0.398438767 */, 18 }, { 0x06606136L /* 0.398530207 */, 18 },
 { 0x0660c119L /* 0.398621653 */, 18 }, { 0x066120fdL /* 0.398713103 */, 18 },
 { 0x066180e3L /* 0.398804559 */, 18 }, { 0x0661e0cbL /* 0.398896020 */, 18 },
 { 0x066240b4L /* 0.398987487 */, 18 }, { 0x0662a09eL /* 0.399078958 */, 18 },
 { 0x06630089L /* 0.399170435 */, 18 }, { 0x06636077L /* 0.399261917 */, 18 },
 { 0x0663c065L /* 0.399353404 */, 18 }, { 0x06642055L /* 0.399444897 */, 18 },
 { 0x06648046L /* 0.399536395 */, 18 }, { 0x0664e039L /* 0.399627898 */, 18 },

 { 0x0665402dL /* 0.399719406 */, 18 }, { 0x0665a022L /* 0.399810919 */, 18 },
 { 0x06660019L /* 0.399902438 */, 18 }, { 0x06666011L /* 0.399993962 */, 18 },
 { 0x0666c00bL /* 0.400085491 */, 18 }, { 0x06672006L /* 0.400177026 */, 18 },
 { 0x06678003L /* 0.400268565 */, 18 }, { 0x0667e000L /* 0.400360110 */, 18 },
 { 0x06684000L /* 0.400451660 */, 18 }, { 0x0668a000L /* 0.400543216 */, 18 },
 { 0x06690003L /* 0.400634776 */, 18 }, { 0x06696006L /* 0.400726342 */, 18 },
 { 0x0669c00bL /* 0.400817913 */, 18 }, { 0x066a2011L /* 0.400909489 */, 18 },
 { 0x066a8019L /* 0.401001071 */, 18 }, { 0x066ae022L /* 0.401092657 */, 18 },

 { 0x066b402dL /* 0.401184249 */, 18 }, { 0x066ba039L /* 0.401275847 */, 18 },
 { 0x066c0046L /* 0.401367449 */, 18 }, { 0x066c6055L /* 0.401459057 */, 18 },
 { 0x066cc065L /* 0.401550670 */, 18 }, { 0x066d2076L /* 0.401642288 */, 18 },
 { 0x066d8089L /* 0.401733911 */, 18 }, { 0x066de09eL /* 0.401825540 */, 18 },
 { 0x066e40b3L /* 0.401917173 */, 18 }, { 0x066ea0cbL /* 0.402008812 */, 18 },
 { 0x066f00e3L /* 0.402100457 */, 18 }, { 0x066f60fdL /* 0.402192106 */, 18 },
 { 0x066fc118L /* 0.402283761 */, 18 }, { 0x06702135L /* 0.402375420 */, 18 },
 { 0x06708153L /* 0.402467086 */, 18 }, { 0x0670e173L /* 0.402558756 */, 18 },

 { 0x06714194L /* 0.402650431 */, 18 }, { 0x0671a1b6L /* 0.402742112 */, 18 },
 { 0x067201daL /* 0.402833798 */, 18 }, { 0x067261ffL /* 0.402925489 */, 18 },
 { 0x0672c226L /* 0.403017186 */, 18 }, { 0x0673224eL /* 0.403108887 */, 18 },
 { 0x06738277L /* 0.403200594 */, 18 }, { 0x0673e2a2L /* 0.403292306 */, 18 },
 { 0x067442ceL /* 0.403384024 */, 18 }, { 0x0674a2fcL /* 0.403475746 */, 18 },
 { 0x0675032bL /* 0.403567474 */, 18 }, { 0x0675635bL /* 0.403659207 */, 18 },
 { 0x0675c38dL /* 0.403750945 */, 18 }, { 0x067623c0L /* 0.403842688 */, 18 },
 { 0x067683f4L /* 0.403934437 */, 18 }, { 0x0676e42aL /* 0.404026190 */, 18 },

 { 0x06774462L /* 0.404117949 */, 18 }, { 0x0677a49bL /* 0.404209714 */, 18 },
 { 0x067804d5L /* 0.404301483 */, 18 }, { 0x06786510L /* 0.404393258 */, 18 },
 { 0x0678c54dL /* 0.404485037 */, 18 }, { 0x0679258cL /* 0.404576822 */, 18 },
 { 0x067985cbL /* 0.404668613 */, 18 }, { 0x0679e60cL /* 0.404760408 */, 18 },
 { 0x067a464fL /* 0.404852209 */, 18 }, { 0x067aa693L /* 0.404944014 */, 18 },
 { 0x067b06d8L /* 0.405035825 */, 18 }, { 0x067b671fL /* 0.405127642 */, 18 },
 { 0x067bc767L /* 0.405219463 */, 18 }, { 0x067c27b1L /* 0.405311290 */, 18 },
 { 0x067c87fcL /* 0.405403122 */, 18 }, { 0x067ce848L /* 0.405494959 */, 18 },

 { 0x067d4896L /* 0.405586801 */, 18 }, { 0x067da8e5L /* 0.405678648 */, 18 },
 { 0x067e0935L /* 0.405770501 */, 18 }, { 0x067e6987L /* 0.405862359 */, 18 },
 { 0x067ec9daL /* 0.405954222 */, 18 }, { 0x067f2a2fL /* 0.406046090 */, 18 },
 { 0x067f8a85L /* 0.406137963 */, 18 }, { 0x067feaddL /* 0.406229842 */, 18 },
 { 0x06804b36L /* 0.406321726 */, 18 }, { 0x0680ab90L /* 0.406413615 */, 18 },
 { 0x06810bebL /* 0.406505509 */, 18 }, { 0x06816c49L /* 0.406597408 */, 18 },
 { 0x0681cca7L /* 0.406689313 */, 18 }, { 0x06822d07L /* 0.406781223 */, 18 },
 { 0x06828d68L /* 0.406873138 */, 18 }, { 0x0682edcbL /* 0.406965058 */, 18 },

 { 0x06834e2fL /* 0.407056983 */, 18 }, { 0x0683ae94L /* 0.407148914 */, 18 },
 { 0x06840efbL /* 0.407240850 */, 18 }, { 0x06846f63L /* 0.407332791 */, 18 },
 { 0x0684cfcdL /* 0.407424737 */, 18 }, { 0x06853038L /* 0.407516688 */, 18 },
 { 0x068590a4L /* 0.407608645 */, 18 }, { 0x0685f112L /* 0.407700606 */, 18 },
 { 0x06865181L /* 0.407792573 */, 18 }, { 0x0686b1f2L /* 0.407884545 */, 18 },
 { 0x06871264L /* 0.407976522 */, 18 }, { 0x068772d7L /* 0.408068505 */, 18 },
 { 0x0687d34cL /* 0.408160492 */, 18 }, { 0x068833c2L /* 0.408252485 */, 18 },
 { 0x06889439L /* 0.408344483 */, 18 }, { 0x0688f4b2L /* 0.408436486 */, 18 },

 { 0x0689552cL /* 0.408528495 */, 18 }, { 0x0689b5a8L /* 0.408620508 */, 18 },
 { 0x068a1625L /* 0.408712527 */, 18 }, { 0x068a76a4L /* 0.408804551 */, 18 },
 { 0x068ad724L /* 0.408896580 */, 18 }, { 0x068b37a5L /* 0.408988614 */, 18 },
 { 0x068b9827L /* 0.409080653 */, 18 }, { 0x068bf8acL /* 0.409172698 */, 18 },
 { 0x068c5931L /* 0.409264748 */, 18 }, { 0x068cb9b8L /* 0.409356803 */, 18 },
 { 0x068d1a40L /* 0.409448863 */, 18 }, { 0x068d7acaL /* 0.409540928 */, 18 },
 { 0x068ddb54L /* 0.409632999 */, 18 }, { 0x068e3be1L /* 0.409725074 */, 18 },
 { 0x068e9c6fL /* 0.409817155 */, 18 }, { 0x068efcfeL /* 0.409909241 */, 18 },

 { 0x068f5d8eL /* 0.410001332 */, 18 }, { 0x068fbe20L /* 0.410093428 */, 18 },
 { 0x06901eb4L /* 0.410185530 */, 18 }, { 0x06907f48L /* 0.410277637 */, 18 },
 { 0x0690dfdeL /* 0.410369748 */, 18 }, { 0x06914076L /* 0.410461865 */, 18 },
 { 0x0691a10fL /* 0.410553988 */, 18 }, { 0x069201a9L /* 0.410646115 */, 18 },
 { 0x06926245L /* 0.410738247 */, 18 }, { 0x0692c2e2L /* 0.410830385 */, 18 },
 { 0x06932380L /* 0.410922528 */, 18 }, { 0x06938420L /* 0.411014676 */, 18 },
 { 0x0693e4c1L /* 0.411106829 */, 18 }, { 0x06944563L /* 0.411198987 */, 18 },
 { 0x0694a607L /* 0.411291151 */, 18 }, { 0x069506adL /* 0.411383320 */, 18 },

 { 0x06956753L /* 0.411475493 */, 18 }, { 0x0695c7fcL /* 0.411567672 */, 18 },
 { 0x069628a5L /* 0.411659857 */, 18 }, { 0x06968950L /* 0.411752046 */, 18 },
 { 0x0696e9fcL /* 0.411844240 */, 18 }, { 0x06974aaaL /* 0.411936440 */, 18 },
 { 0x0697ab59L /* 0.412028645 */, 18 }, { 0x06980c09L /* 0.412120855 */, 18 },
 { 0x06986cbbL /* 0.412213070 */, 18 }, { 0x0698cd6eL /* 0.412305290 */, 18 },
 { 0x06992e23L /* 0.412397516 */, 18 }, { 0x06998ed9L /* 0.412489746 */, 18 },
 { 0x0699ef90L /* 0.412581982 */, 18 }, { 0x069a5049L /* 0.412674223 */, 18 },
 { 0x069ab103L /* 0.412766469 */, 18 }, { 0x069b11bfL /* 0.412858720 */, 18 },

 { 0x069b727bL /* 0.412950976 */, 18 }, { 0x069bd33aL /* 0.413043238 */, 18 },
 { 0x069c33f9L /* 0.413135505 */, 18 }, { 0x069c94baL /* 0.413227776 */, 18 },
 { 0x069cf57dL /* 0.413320053 */, 18 }, { 0x069d5641L /* 0.413412335 */, 18 },
 { 0x069db706L /* 0.413504623 */, 18 }, { 0x069e17ccL /* 0.413596915 */, 18 },
 { 0x069e7894L /* 0.413689213 */, 18 }, { 0x069ed95eL /* 0.413781515 */, 18 },
 { 0x069f3a28L /* 0.413873823 */, 18 }, { 0x069f9af4L /* 0.413966136 */, 18 },
 { 0x069ffbc2L /* 0.414058454 */, 18 }, { 0x06a05c91L /* 0.414150778 */, 18 },
 { 0x06a0bd61L /* 0.414243106 */, 18 }, { 0x06a11e32L /* 0.414335440 */, 18 },

 { 0x06a17f05L /* 0.414427779 */, 18 }, { 0x06a1dfdaL /* 0.414520122 */, 18 },
 { 0x06a240b0L /* 0.414612471 */, 18 }, { 0x06a2a187L /* 0.414704826 */, 18 },
 { 0x06a3025fL /* 0.414797185 */, 18 }, { 0x06a36339L /* 0.414889549 */, 18 },
 { 0x06a3c414L /* 0.414981919 */, 18 }, { 0x06a424f1L /* 0.415074294 */, 18 },
 { 0x06a485cfL /* 0.415166674 */, 18 }, { 0x06a4e6aeL /* 0.415259059 */, 18 },
 { 0x06a5478fL /* 0.415351449 */, 18 }, { 0x06a5a871L /* 0.415443844 */, 18 },
 { 0x06a60955L /* 0.415536244 */, 18 }, { 0x06a66a3aL /* 0.415628650 */, 18 },
 { 0x06a6cb20L /* 0.415721061 */, 18 }, { 0x06a72c08L /* 0.415813476 */, 18 },

 { 0x06a78cf1L /* 0.415905897 */, 18 }, { 0x06a7eddbL /* 0.415998324 */, 18 },
 { 0x06a84ec7L /* 0.416090755 */, 18 }, { 0x06a8afb4L /* 0.416183191 */, 18 },
 { 0x06a910a3L /* 0.416275633 */, 18 }, { 0x06a97193L /* 0.416368079 */, 18 },
 { 0x06a9d284L /* 0.416460531 */, 18 }, { 0x06aa3377L /* 0.416552988 */, 18 },
 { 0x06aa946bL /* 0.416645450 */, 18 }, { 0x06aaf561L /* 0.416737917 */, 18 },
 { 0x06ab5657L /* 0.416830389 */, 18 }, { 0x06abb750L /* 0.416922867 */, 18 },
 { 0x06ac1849L /* 0.417015349 */, 18 }, { 0x06ac7944L /* 0.417107837 */, 18 },
 { 0x06acda41L /* 0.417200330 */, 18 }, { 0x06ad3b3eL /* 0.417292828 */, 18 },

 { 0x06ad9c3dL /* 0.417385331 */, 18 }, { 0x06adfd3eL /* 0.417477839 */, 18 },
 { 0x06ae5e40L /* 0.417570352 */, 18 }, { 0x06aebf43L /* 0.417662871 */, 18 },
 { 0x06af2047L /* 0.417755394 */, 18 }, { 0x06af814dL /* 0.417847923 */, 18 },
 { 0x06afe255L /* 0.417940457 */, 18 }, { 0x06b0435eL /* 0.418032996 */, 18 },
 { 0x06b0a468L /* 0.418125540 */, 18 }, { 0x06b10573L /* 0.418218089 */, 18 },
 { 0x06b16680L /* 0.418310643 */, 18 }, { 0x06b1c78eL /* 0.418403203 */, 18 },
 { 0x06b2289eL /* 0.418495767 */, 18 }, { 0x06b289afL /* 0.418588337 */, 18 },
 { 0x06b2eac1L /* 0.418680911 */, 18 }, { 0x06b34bd5L /* 0.418773491 */, 18 },

 { 0x06b3aceaL /* 0.418866076 */, 18 }, { 0x06b40e00L /* 0.418958666 */, 18 },
 { 0x06b46f18L /* 0.419051262 */, 18 }, { 0x06b4d031L /* 0.419143862 */, 18 },
 { 0x06b5314cL /* 0.419236467 */, 18 }, { 0x06b59268L /* 0.419329078 */, 18 },
 { 0x06b5f385L /* 0.419421694 */, 18 }, { 0x06b654a4L /* 0.419514314 */, 18 },
 { 0x06b6b5c4L /* 0.419606940 */, 18 }, { 0x06b716e6L /* 0.419699571 */, 18 },
 { 0x06b77808L /* 0.419792208 */, 18 }, { 0x06b7d92dL /* 0.419884849 */, 18 },
 { 0x06b83a52L /* 0.419977495 */, 18 }, { 0x06b89b79L /* 0.420070147 */, 18 },
 { 0x06b8fca1L /* 0.420162803 */, 18 }, { 0x06b95dcbL /* 0.420255465 */, 18 },

 { 0x06b9bef6L /* 0.420348132 */, 18 }, { 0x06ba2023L /* 0.420440803 */, 18 },
 { 0x06ba8150L /* 0.420533481 */, 18 }, { 0x06bae280L /* 0.420626163 */, 18 },
 { 0x06bb43b0L /* 0.420718850 */, 18 }, { 0x06bba4e2L /* 0.420811542 */, 18 },
 { 0x06bc0615L /* 0.420904240 */, 18 }, { 0x06bc674aL /* 0.420996942 */, 18 },
 { 0x06bcc880L /* 0.421089650 */, 18 }, { 0x06bd29b7L /* 0.421182362 */, 18 },
 { 0x06bd8af0L /* 0.421275080 */, 18 }, { 0x06bdec2aL /* 0.421367803 */, 18 },
 { 0x06be4d66L /* 0.421460531 */, 18 }, { 0x06beaea3L /* 0.421553264 */, 18 },
 { 0x06bf0fe1L /* 0.421646003 */, 18 }, { 0x06bf7120L /* 0.421738746 */, 18 },

 { 0x06bfd261L /* 0.421831494 */, 18 }, { 0x06c033a4L /* 0.421924248 */, 18 },
 { 0x06c094e7L /* 0.422017007 */, 18 }, { 0x06c0f62cL /* 0.422109770 */, 18 },
 { 0x06c15773L /* 0.422202539 */, 18 }, { 0x06c1b8bbL /* 0.422295313 */, 18 },
 { 0x06c21a04L /* 0.422388092 */, 18 }, { 0x06c27b4eL /* 0.422480876 */, 18 },
 { 0x06c2dc9aL /* 0.422573665 */, 18 }, { 0x06c33de8L /* 0.422666460 */, 18 },
 { 0x06c39f36L /* 0.422759259 */, 18 }, { 0x06c40086L /* 0.422852064 */, 18 },
 { 0x06c461d8L /* 0.422944873 */, 18 }, { 0x06c4c32aL /* 0.423037688 */, 18 },
 { 0x06c5247fL /* 0.423130508 */, 18 }, { 0x06c585d4L /* 0.423223333 */, 18 },

 { 0x06c5e72bL /* 0.423316162 */, 18 }, { 0x06c64883L /* 0.423408997 */, 18 },
 { 0x06c6a9ddL /* 0.423501838 */, 18 }, { 0x06c70b38L /* 0.423594683 */, 18 },
 { 0x06c76c94L /* 0.423687533 */, 18 }, { 0x06c7cdf2L /* 0.423780389 */, 18 },
 { 0x06c82f51L /* 0.423873249 */, 18 }, { 0x06c890b1L /* 0.423966115 */, 18 },
 { 0x06c8f213L /* 0.424058985 */, 18 }, { 0x06c95376L /* 0.424151861 */, 18 },
 { 0x06c9b4daL /* 0.424244742 */, 18 }, { 0x06ca1640L /* 0.424337628 */, 18 },
 { 0x06ca77a8L /* 0.424430519 */, 18 }, { 0x06cad910L /* 0.424523415 */, 18 },
 { 0x06cb3a7aL /* 0.424616316 */, 18 }, { 0x06cb9be5L /* 0.424709222 */, 18 },

 { 0x06cbfd52L /* 0.424802133 */, 18 }, { 0x06cc5ec0L /* 0.424895050 */, 18 },
 { 0x06ccc030L /* 0.424987971 */, 18 }, { 0x06cd21a0L /* 0.425080898 */, 18 },
 { 0x06cd8313L /* 0.425173829 */, 18 }, { 0x06cde486L /* 0.425266766 */, 18 },
 { 0x06ce45fbL /* 0.425359708 */, 18 }, { 0x06cea771L /* 0.425452655 */, 18 },
 { 0x06cf08e9L /* 0.425545607 */, 18 }, { 0x06cf6a62L /* 0.425638564 */, 18 },
 { 0x06cfcbdcL /* 0.425731526 */, 18 }, { 0x06d02d58L /* 0.425824493 */, 18 },
 { 0x06d08ed5L /* 0.425917465 */, 18 }, { 0x06d0f053L /* 0.426010443 */, 18 },
 { 0x06d151d3L /* 0.426103425 */, 18 }, { 0x06d1b354L /* 0.426196412 */, 18 },

 { 0x06d214d7L /* 0.426289405 */, 18 }, { 0x06d2765aL /* 0.426382403 */, 18 },
 { 0x06d2d7e0L /* 0.426475405 */, 18 }, { 0x06d33966L /* 0.426568413 */, 18 },
 { 0x06d39aeeL /* 0.426661426 */, 18 }, { 0x06d3fc77L /* 0.426754444 */, 18 },
 { 0x06d45e02L /* 0.426847467 */, 18 }, { 0x06d4bf8eL /* 0.426940495 */, 18 },
 { 0x06d5211cL /* 0.427033528 */, 18 }, { 0x06d582aaL /* 0.427126566 */, 18 },
 { 0x06d5e43aL /* 0.427219609 */, 18 }, { 0x06d645ccL /* 0.427312657 */, 18 },
 { 0x06d6a75fL /* 0.427405711 */, 18 }, { 0x06d708f3L /* 0.427498769 */, 18 },
 { 0x06d76a88L /* 0.427591833 */, 18 }, { 0x06d7cc1fL /* 0.427684901 */, 18 },

 { 0x06d82db8L /* 0.427777975 */, 18 }, { 0x06d88f51L /* 0.427871054 */, 18 },
 { 0x06d8f0ecL /* 0.427964137 */, 18 }, { 0x06d95288L /* 0.428057226 */, 18 },
 { 0x06d9b426L /* 0.428150320 */, 18 }, { 0x06da15c5L /* 0.428243419 */, 18 },
 { 0x06da7766L /* 0.428336523 */, 18 }, { 0x06dad907L /* 0.428429632 */, 18 },
 { 0x06db3aaaL /* 0.428522746 */, 18 }, { 0x06db9c4fL /* 0.428615865 */, 18 },
 { 0x06dbfdf5L /* 0.428708989 */, 18 }, { 0x06dc5f9cL /* 0.428802119 */, 18 },
 { 0x06dcc145L /* 0.428895253 */, 18 }, { 0x06dd22eeL /* 0.428988392 */, 18 },
 { 0x06dd849aL /* 0.429081537 */, 18 }, { 0x06dde646L /* 0.429174686 */, 18 },

 { 0x06de47f4L /* 0.429267841 */, 18 }, { 0x06dea9a4L /* 0.429361001 */, 18 },
 { 0x06df0b54L /* 0.429454165 */, 18 }, { 0x06df6d06L /* 0.429547335 */, 18 },
 { 0x06dfcebaL /* 0.429640510 */, 18 }, { 0x06e0306fL /* 0.429733690 */, 18 },
 { 0x06e09225L /* 0.429826874 */, 18 }, { 0x06e0f3dcL /* 0.429920064 */, 18 },
 { 0x06e15595L /* 0.430013259 */, 18 }, { 0x06e1b74fL /* 0.430106459 */, 18 },
 { 0x06e2190bL /* 0.430199664 */, 18 }, { 0x06e27ac8L /* 0.430292875 */, 18 },
 { 0x06e2dc86L /* 0.430386090 */, 18 }, { 0x06e33e46L /* 0.430479310 */, 18 },
 { 0x06e3a007L /* 0.430572535 */, 18 }, { 0x06e401c9L /* 0.430665765 */, 18 },

 { 0x06e4638dL /* 0.430759001 */, 18 }, { 0x06e4c552L /* 0.430852241 */, 18 },
 { 0x06e52718L /* 0.430945487 */, 18 }, { 0x06e588e0L /* 0.431038737 */, 18 },
 { 0x06e5eaa9L /* 0.431131993 */, 18 }, { 0x06e64c73L /* 0.431225253 */, 18 },
 { 0x06e6ae3fL /* 0.431318519 */, 18 }, { 0x06e7100cL /* 0.431411790 */, 18 },
 { 0x06e771dbL /* 0.431505065 */, 18 }, { 0x06e7d3abL /* 0.431598346 */, 18 },
 { 0x06e8357cL /* 0.431691632 */, 18 }, { 0x06e8974eL /* 0.431784923 */, 18 },
 { 0x06e8f922L /* 0.431878218 */, 18 }, { 0x06e95af8L /* 0.431971519 */, 18 },
 { 0x06e9bcceL /* 0.432064825 */, 18 }, { 0x06ea1ea6L /* 0.432158136 */, 18 },

 { 0x06ea807fL /* 0.432251452 */, 18 }, { 0x06eae25aL /* 0.432344773 */, 18 },
 { 0x06eb4436L /* 0.432438099 */, 18 }, { 0x06eba614L /* 0.432531431 */, 18 },
 { 0x06ec07f2L /* 0.432624767 */, 18 }, { 0x06ec69d2L /* 0.432718108 */, 18 },
 { 0x06eccbb4L /* 0.432811454 */, 18 }, { 0x06ed2d97L /* 0.432904805 */, 18 },
 { 0x06ed8f7bL /* 0.432998162 */, 18 }, { 0x06edf160L /* 0.433091523 */, 18 },
 { 0x06ee5347L /* 0.433184889 */, 18 }, { 0x06eeb52fL /* 0.433278261 */, 18 },
 { 0x06ef1719L /* 0.433371637 */, 18 }, { 0x06ef7904L /* 0.433465019 */, 18 },
 { 0x06efdaf0L /* 0.433558405 */, 18 }, { 0x06f03cdeL /* 0.433651797 */, 18 },

 { 0x06f09eccL /* 0.433745193 */, 18 }, { 0x06f100bdL /* 0.433838595 */, 18 },
 { 0x06f162aeL /* 0.433932001 */, 18 }, { 0x06f1c4a1L /* 0.434025413 */, 18 },
 { 0x06f22696L /* 0.434118830 */, 18 }, { 0x06f2888bL /* 0.434212251 */, 18 },
 { 0x06f2ea82L /* 0.434305678 */, 18 }, { 0x06f34c7bL /* 0.434399110 */, 18 },
 { 0x06f3ae75L /* 0.434492546 */, 18 }, { 0x06f41070L /* 0.434585988 */, 18 },
 { 0x06f4726cL /* 0.434679435 */, 18 }, { 0x06f4d46aL /* 0.434772887 */, 18 },
 { 0x06f53669L /* 0.434866344 */, 18 }, { 0x06f59869L /* 0.434959806 */, 18 },
 { 0x06f5fa6bL /* 0.435053272 */, 18 }, { 0x06f65c6eL /* 0.435146744 */, 18 },

 { 0x06f6be73L /* 0.435240221 */, 18 }, { 0x06f72079L /* 0.435333703 */, 18 },
 { 0x06f78280L /* 0.435427190 */, 18 }, { 0x06f7e489L /* 0.435520682 */, 18 },
 { 0x06f84693L /* 0.435614179 */, 18 }, { 0x06f8a89eL /* 0.435707681 */, 18 },
 { 0x06f90aaaL /* 0.435801188 */, 18 }, { 0x06f96cb8L /* 0.435894700 */, 18 },
 { 0x06f9cec8L /* 0.435988217 */, 18 }, { 0x06fa30d8L /* 0.436081739 */, 18 },
 { 0x06fa92eaL /* 0.436175266 */, 18 }, { 0x06faf4feL /* 0.436268799 */, 18 },
 { 0x06fb5712L /* 0.436362336 */, 18 }, { 0x06fbb928L /* 0.436455878 */, 18 },
 { 0x06fc1b40L /* 0.436549425 */, 18 }, { 0x06fc7d58L /* 0.436642977 */, 18 },

 { 0x06fcdf72L /* 0.436736534 */, 18 }, { 0x06fd418eL /* 0.436830096 */, 18 },
 { 0x06fda3abL /* 0.436923664 */, 18 }, { 0x06fe05c9L /* 0.437017236 */, 18 },
 { 0x06fe67e8L /* 0.437110813 */, 18 }, { 0x06feca09L /* 0.437204395 */, 18 },
 { 0x06ff2c2bL /* 0.437297982 */, 18 }, { 0x06ff8e4fL /* 0.437391575 */, 18 },
 { 0x06fff073L /* 0.437485172 */, 18 }, { 0x0700529aL /* 0.437578774 */, 18 },
 { 0x0700b4c1L /* 0.437672381 */, 18 }, { 0x070116eaL /* 0.437765994 */, 18 },
 { 0x07017914L /* 0.437859611 */, 18 }, { 0x0701db40L /* 0.437953233 */, 18 },
 { 0x07023d6cL /* 0.438046860 */, 18 }, { 0x07029f9bL /* 0.438140493 */, 18 },

 { 0x070301caL /* 0.438234130 */, 18 }, { 0x070363fbL /* 0.438327772 */, 18 },
 { 0x0703c62dL /* 0.438421419 */, 18 }, { 0x07042861L /* 0.438515072 */, 18 },
 { 0x07048a96L /* 0.438608729 */, 18 }, { 0x0704ecccL /* 0.438702391 */, 18 },
 { 0x07054f04L /* 0.438796059 */, 18 }, { 0x0705b13dL /* 0.438889731 */, 18 },
 { 0x07061377L /* 0.438983408 */, 18 }, { 0x070675b3L /* 0.439077090 */, 18 },
 { 0x0706d7f0L /* 0.439170778 */, 18 }, { 0x07073a2eL /* 0.439264470 */, 18 },
 { 0x07079c6eL /* 0.439358167 */, 18 }, { 0x0707feafL /* 0.439451869 */, 18 },
 { 0x070860f1L /* 0.439545577 */, 18 }, { 0x0708c335L /* 0.439639289 */, 18 },

 { 0x0709257aL /* 0.439733006 */, 18 }, { 0x070987c0L /* 0.439826728 */, 18 },
 { 0x0709ea08L /* 0.439920456 */, 18 }, { 0x070a4c51L /* 0.440014188 */, 18 },
 { 0x070aae9bL /* 0.440107925 */, 18 }, { 0x070b10e7L /* 0.440201667 */, 18 },
 { 0x070b7334L /* 0.440295414 */, 18 }, { 0x070bd583L /* 0.440389167 */, 18 },
 { 0x070c37d2L /* 0.440482924 */, 18 }, { 0x070c9a23L /* 0.440576686 */, 18 },
 { 0x070cfc76L /* 0.440670453 */, 18 }, { 0x070d5ecaL /* 0.440764225 */, 18 },
 { 0x070dc11fL /* 0.440858002 */, 18 }, { 0x070e2375L /* 0.440951784 */, 18 },
 { 0x070e85cdL /* 0.441045572 */, 18 }, { 0x070ee826L /* 0.441139364 */, 18 },

 { 0x070f4a80L /* 0.441233161 */, 18 }, { 0x070facdcL /* 0.441326963 */, 18 },
 { 0x07100f39L /* 0.441420770 */, 18 }, { 0x07107198L /* 0.441514582 */, 18 },
 { 0x0710d3f8L /* 0.441608399 */, 18 }, { 0x07113659L /* 0.441702221 */, 18 },
 { 0x071198bbL /* 0.441796048 */, 18 }, { 0x0711fb1fL /* 0.441889880 */, 18 },
 { 0x07125d84L /* 0.441983717 */, 18 }, { 0x0712bfebL /* 0.442077559 */, 18 },
 { 0x07132253L /* 0.442171406 */, 18 }, { 0x071384bcL /* 0.442265257 */, 18 },
 { 0x0713e726L /* 0.442359114 */, 18 }, { 0x07144992L /* 0.442452976 */, 18 },
 { 0x0714abffL /* 0.442546843 */, 18 }, { 0x07150e6eL /* 0.442640715 */, 18 },

 { 0x071570deL /* 0.442734592 */, 18 }, { 0x0715d34fL /* 0.442828473 */, 18 },
 { 0x071635c1L /* 0.442922360 */, 18 }, { 0x07169835L /* 0.443016252 */, 18 },
 { 0x0716faaaL /* 0.443110148 */, 18 }, { 0x07175d21L /* 0.443204050 */, 18 },
 { 0x0717bf99L /* 0.443297957 */, 18 }, { 0x07182212L /* 0.443391868 */, 18 },
 { 0x0718848dL /* 0.443485785 */, 18 }, { 0x0718e709L /* 0.443579706 */, 18 },
 { 0x07194986L /* 0.443673633 */, 18 }, { 0x0719ac04L /* 0.443767564 */, 18 },
 { 0x071a0e84L /* 0.443861501 */, 18 }, { 0x071a7105L /* 0.443955442 */, 18 },
 { 0x071ad388L /* 0.444049389 */, 18 }, { 0x071b360cL /* 0.444143340 */, 18 },

 { 0x071b9891L /* 0.444237296 */, 18 }, { 0x071bfb18L /* 0.444331258 */, 18 },
 { 0x071c5d9fL /* 0.444425224 */, 18 }, { 0x071cc029L /* 0.444519195 */, 18 },
 { 0x071d22b3L /* 0.444613171 */, 18 }, { 0x071d853fL /* 0.444707153 */, 18 },
 { 0x071de7ccL /* 0.444801139 */, 18 }, { 0x071e4a5bL /* 0.444895130 */, 18 },
 { 0x071eacebL /* 0.444989126 */, 18 }, { 0x071f0f7cL /* 0.445083127 */, 18 },
 { 0x071f720eL /* 0.445177133 */, 18 }, { 0x071fd4a2L /* 0.445271144 */, 18 },
 { 0x07203737L /* 0.445365160 */, 18 }, { 0x072099ceL /* 0.445459181 */, 18 },
 { 0x0720fc66L /* 0.445553206 */, 18 }, { 0x07215effL /* 0.445647237 */, 18 },

 { 0x0721c19aL /* 0.445741273 */, 18 }, { 0x07222436L /* 0.445835314 */, 18 },
 { 0x072286d3L /* 0.445929359 */, 18 }, { 0x0722e971L /* 0.446023410 */, 18 },
 { 0x07234c11L /* 0.446117466 */, 18 }, { 0x0723aeb2L /* 0.446211526 */, 18 },
 { 0x07241155L /* 0.446305592 */, 18 }, { 0x072473f9L /* 0.446399662 */, 18 },
 { 0x0724d69eL /* 0.446493738 */, 18 }, { 0x07253944L /* 0.446587818 */, 18 },
 { 0x07259becL /* 0.446681903 */, 18 }, { 0x0725fe95L /* 0.446775994 */, 18 },
 { 0x07266140L /* 0.446870089 */, 18 }, { 0x0726c3ecL /* 0.446964189 */, 18 },
 { 0x07272699L /* 0.447058294 */, 18 }, { 0x07278947L /* 0.447152404 */, 18 },

 { 0x0727ebf7L /* 0.447246519 */, 18 }, { 0x07284ea8L /* 0.447340639 */, 18 },
 { 0x0728b15bL /* 0.447434764 */, 18 }, { 0x0729140fL /* 0.447528894 */, 18 },
 { 0x072976c4L /* 0.447623029 */, 18 }, { 0x0729d97aL /* 0.447717169 */, 18 },
 { 0x072a3c32L /* 0.447811314 */, 18 }, { 0x072a9eebL /* 0.447905463 */, 18 },
 { 0x072b01a6L /* 0.447999618 */, 18 }, { 0x072b6461L /* 0.448093778 */, 18 },
 { 0x072bc71eL /* 0.448187942 */, 18 }, { 0x072c29ddL /* 0.448282112 */, 18 },
 { 0x072c8c9dL /* 0.448376286 */, 18 }, { 0x072cef5eL /* 0.448470466 */, 18 },
 { 0x072d5220L /* 0.448564650 */, 18 }, { 0x072db4e4L /* 0.448658839 */, 18 },

 { 0x072e17a9L /* 0.448753033 */, 18 }, { 0x072e7a6fL /* 0.448847233 */, 18 },
 { 0x072edd37L /* 0.448941437 */, 18 }, { 0x072f4000L /* 0.449035646 */, 18 },
 { 0x072fa2caL /* 0.449129860 */, 18 }, { 0x07300596L /* 0.449224079 */, 18 },
 { 0x07306863L /* 0.449318303 */, 18 }, { 0x0730cb32L /* 0.449412531 */, 18 },
 { 0x07312e01L /* 0.449506765 */, 18 }, { 0x073190d2L /* 0.449601004 */, 18 },
 { 0x0731f3a5L /* 0.449695247 */, 18 }, { 0x07325678L /* 0.449789496 */, 18 },
 { 0x0732b94dL /* 0.449883749 */, 18 }, { 0x07331c23L /* 0.449978008 */, 18 },
 { 0x07337efbL /* 0.450072271 */, 18 }, { 0x0733e1d4L /* 0.450166540 */, 18 },

 { 0x073444aeL /* 0.450260813 */, 18 }, { 0x0734a78aL /* 0.450355091 */, 18 },
 { 0x07350a67L /* 0.450449374 */, 18 }, { 0x07356d45L /* 0.450543662 */, 18 },
 { 0x0735d025L /* 0.450637955 */, 18 }, { 0x07363306L /* 0.450732253 */, 18 },
 { 0x073695e8L /* 0.450826556 */, 18 }, { 0x0736f8cbL /* 0.450920864 */, 18 },
 { 0x07375bb0L /* 0.451015176 */, 18 }, { 0x0737be96L /* 0.451109494 */, 18 },
 { 0x0738217eL /* 0.451203817 */, 18 }, { 0x07388467L /* 0.451298144 */, 18 },
 { 0x0738e751L /* 0.451392477 */, 18 }, { 0x07394a3dL /* 0.451486814 */, 18 },
 { 0x0739ad29L /* 0.451581156 */, 18 }, { 0x073a1017L /* 0.451675503 */, 18 },

 { 0x073a7307L /* 0.451769856 */, 18 }, { 0x073ad5f8L /* 0.451864213 */, 18 },
 { 0x073b38eaL /* 0.451958575 */, 18 }, { 0x073b9bddL /* 0.452052942 */, 18 },
 { 0x073bfed2L /* 0.452147313 */, 18 }, { 0x073c61c8L /* 0.452241690 */, 18 },
 { 0x073cc4bfL /* 0.452336072 */, 18 }, { 0x073d27b8L /* 0.452430458 */, 18 },
 { 0x073d8ab2L /* 0.452524850 */, 18 }, { 0x073dedaeL /* 0.452619246 */, 18 },
 { 0x073e50aaL /* 0.452713648 */, 18 }, { 0x073eb3a8L /* 0.452808054 */, 18 },
 { 0x073f16a8L /* 0.452902465 */, 18 }, { 0x073f79a8L /* 0.452996882 */, 18 },
 { 0x073fdcaaL /* 0.453091303 */, 18 }, { 0x07403fadL /* 0.453185729 */, 18 },

 { 0x0740a2b2L /* 0.453280160 */, 18 }, { 0x074105b8L /* 0.453374595 */, 18 },
 { 0x074168bfL /* 0.453469036 */, 18 }, { 0x0741cbc8L /* 0.453563482 */, 18 },
 { 0x07422ed2L /* 0.453657932 */, 18 }, { 0x074291ddL /* 0.453752388 */, 18 },
 { 0x0742f4e9L /* 0.453846848 */, 18 }, { 0x074357f7L /* 0.453941314 */, 18 },
 { 0x0743bb06L /* 0.454035784 */, 18 }, { 0x07441e17L /* 0.454130259 */, 18 },
 { 0x07448129L /* 0.454224739 */, 18 }, { 0x0744e43cL /* 0.454319224 */, 18 },
 { 0x07454750L /* 0.454413714 */, 18 }, { 0x0745aa66L /* 0.454508209 */, 18 },
 { 0x07460d7dL /* 0.454602708 */, 18 }, { 0x07467095L /* 0.454697213 */, 18 },

 { 0x0746d3afL /* 0.454791723 */, 18 }, { 0x074736caL /* 0.454886237 */, 18 },
 { 0x074799e7L /* 0.454980756 */, 18 }, { 0x0747fd04L /* 0.455075281 */, 18 },
 { 0x07486023L /* 0.455169810 */, 18 }, { 0x0748c344L /* 0.455264344 */, 18 },
 { 0x07492665L /* 0.455358883 */, 18 }, { 0x07498988L /* 0.455453427 */, 18 },
 { 0x0749ecacL /* 0.455547976 */, 18 }, { 0x074a4fd2L /* 0.455642529 */, 18 },
 { 0x074ab2f9L /* 0.455737088 */, 18 }, { 0x074b1621L /* 0.455831652 */, 18 },
 { 0x074b794bL /* 0.455926220 */, 18 }, { 0x074bdc75L /* 0.456020793 */, 18 },
 { 0x074c3fa1L /* 0.456115372 */, 18 }, { 0x074ca2cfL /* 0.456209955 */, 18 },

 { 0x074d05feL /* 0.456304543 */, 18 }, { 0x074d692eL /* 0.456399136 */, 18 },
 { 0x074dcc5fL /* 0.456493733 */, 18 }, { 0x074e2f92L /* 0.456588336 */, 18 },
 { 0x074e92c6L /* 0.456682944 */, 18 }, { 0x074ef5fbL /* 0.456777556 */, 18 },
 { 0x074f5932L /* 0.456872174 */, 18 }, { 0x074fbc6aL /* 0.456966796 */, 18 },
 { 0x07501fa3L /* 0.457061423 */, 18 }, { 0x075082deL /* 0.457156056 */, 18 },
 { 0x0750e61aL /* 0.457250693 */, 18 }, { 0x07514957L /* 0.457345335 */, 18 },
 { 0x0751ac96L /* 0.457439981 */, 18 }, { 0x07520fd6L /* 0.457534633 */, 18 },
 { 0x07527317L /* 0.457629290 */, 18 }, { 0x0752d659L /* 0.457723951 */, 18 },

 { 0x0753399dL /* 0.457818618 */, 18 }, { 0x07539ce2L /* 0.457913289 */, 18 },
 { 0x07540029L /* 0.458007965 */, 18 }, { 0x07546371L /* 0.458102646 */, 18 },
 { 0x0754c6baL /* 0.458197332 */, 18 }, { 0x07552a04L /* 0.458292023 */, 18 },
 { 0x07558d50L /* 0.458386719 */, 18 }, { 0x0755f09dL /* 0.458481420 */, 18 },
 { 0x075653ebL /* 0.458576125 */, 18 }, { 0x0756b73bL /* 0.458670836 */, 18 },
 { 0x07571a8cL /* 0.458765551 */, 18 }, { 0x07577ddeL /* 0.458860271 */, 18 },
 { 0x0757e131L /* 0.458954996 */, 18 }, { 0x07584486L /* 0.459049726 */, 18 },
 { 0x0758a7ddL /* 0.459144461 */, 18 }, { 0x07590b34L /* 0.459239201 */, 18 },

 { 0x07596e8dL /* 0.459333946 */, 18 }, { 0x0759d1e7L /* 0.459428695 */, 18 },
 { 0x075a3542L /* 0.459523450 */, 18 }, { 0x075a989fL /* 0.459618209 */, 18 },
 { 0x075afbfdL /* 0.459712973 */, 18 }, { 0x075b5f5dL /* 0.459807742 */, 18 },
 { 0x075bc2bdL /* 0.459902516 */, 18 }, { 0x075c261fL /* 0.459997295 */, 18 },
 { 0x075c8983L /* 0.460092079 */, 18 }, { 0x075cece7L /* 0.460186867 */, 18 },
 { 0x075d504dL /* 0.460281661 */, 18 }, { 0x075db3b5L /* 0.460376459 */, 18 },
 { 0x075e171dL /* 0.460471262 */, 18 }, { 0x075e7a87L /* 0.460566071 */, 18 },
 { 0x075eddf2L /* 0.460660884 */, 18 }, { 0x075f415fL /* 0.460755701 */, 18 },

 { 0x075fa4ccL /* 0.460850524 */, 18 }, { 0x0760083bL /* 0.460945352 */, 18 },
 { 0x07606bacL /* 0.461040184 */, 18 }, { 0x0760cf1eL /* 0.461135022 */, 18 },
 { 0x07613291L /* 0.461229864 */, 18 }, { 0x07619605L /* 0.461324711 */, 18 },
 { 0x0761f97bL /* 0.461419563 */, 18 }, { 0x07625cf2L /* 0.461514420 */, 18 },
 { 0x0762c06aL /* 0.461609282 */, 18 }, { 0x076323e3L /* 0.461704149 */, 18 },
 { 0x0763875eL /* 0.461799020 */, 18 }, { 0x0763eadbL /* 0.461893897 */, 18 },
 { 0x07644e58L /* 0.461988778 */, 18 }, { 0x0764b1d7L /* 0.462083664 */, 18 },
 { 0x07651557L /* 0.462178555 */, 18 }, { 0x076578d8L /* 0.462273451 */, 18 },

 { 0x0765dc5bL /* 0.462368352 */, 18 }, { 0x07663fdfL /* 0.462463257 */, 18 },
 { 0x0766a364L /* 0.462558168 */, 18 }, { 0x076706ebL /* 0.462653083 */, 18 },
 { 0x07676a73L /* 0.462748003 */, 18 }, { 0x0767cdfcL /* 0.462842928 */, 18 },
 { 0x07683187L /* 0.462937858 */, 18 }, { 0x07689513L /* 0.463032793 */, 18 },
 { 0x0768f8a0L /* 0.463127733 */, 18 }, { 0x07695c2eL /* 0.463222678 */, 18 },
 { 0x0769bfbeL /* 0.463317627 */, 18 }, { 0x076a234fL /* 0.463412581 */, 18 },
 { 0x076a86e2L /* 0.463507540 */, 18 }, { 0x076aea75L /* 0.463602504 */, 18 },
 { 0x076b4e0aL /* 0.463697473 */, 18 }, { 0x076bb1a1L /* 0.463792447 */, 18 },

 { 0x076c1538L /* 0.463887426 */, 18 }, { 0x076c78d1L /* 0.463982409 */, 18 },
 { 0x076cdc6cL /* 0.464077398 */, 18 }, { 0x076d4007L /* 0.464172391 */, 18 },
 { 0x076da3a4L /* 0.464267389 */, 18 }, { 0x076e0742L /* 0.464362392 */, 18 },
 { 0x076e6ae2L /* 0.464457399 */, 18 }, { 0x076ece82L /* 0.464552412 */, 18 },
 { 0x076f3224L /* 0.464647430 */, 18 }, { 0x076f95c8L /* 0.464742452 */, 18 },
 { 0x076ff96cL /* 0.464837479 */, 18 }, { 0x07705d12L /* 0.464932511 */, 18 },
 { 0x0770c0baL /* 0.465027548 */, 18 }, { 0x07712462L /* 0.465122590 */, 18 },
 { 0x0771880cL /* 0.465217637 */, 18 }, { 0x0771ebb7L /* 0.465312688 */, 18 },

 { 0x07724f64L /* 0.465407744 */, 18 }, { 0x0772b312L /* 0.465502806 */, 18 },
 { 0x077316c1L /* 0.465597872 */, 18 }, { 0x07737a71L /* 0.465692943 */, 18 },
 { 0x0773de23L /* 0.465788018 */, 18 }, { 0x077441d6L /* 0.465883099 */, 18 },
 { 0x0774a58aL /* 0.465978184 */, 18 }, { 0x07750940L /* 0.466073275 */, 18 },
 { 0x07756cf7L /* 0.466168370 */, 18 }, { 0x0775d0afL /* 0.466263470 */, 18 },
 { 0x07763468L /* 0.466358575 */, 18 }, { 0x07769823L /* 0.466453684 */, 18 },
 { 0x0776fbdfL /* 0.466548799 */, 18 }, { 0x07775f9dL /* 0.466643918 */, 18 },
 { 0x0777c35cL /* 0.466739043 */, 18 }, { 0x0778271cL /* 0.466834172 */, 18 },

 { 0x07788addL /* 0.466929306 */, 18 }, { 0x0778ee9fL /* 0.467024445 */, 18 },
 { 0x07795263L /* 0.467119588 */, 18 }, { 0x0779b629L /* 0.467214737 */, 18 },
 { 0x077a19efL /* 0.467309890 */, 18 }, { 0x077a7db7L /* 0.467405048 */, 18 },
 { 0x077ae180L /* 0.467500211 */, 18 }, { 0x077b454bL /* 0.467595379 */, 18 },
 { 0x077ba916L /* 0.467690552 */, 18 }, { 0x077c0ce3L /* 0.467785729 */, 18 },
 { 0x077c70b2L /* 0.467880912 */, 18 }, { 0x077cd481L /* 0.467976099 */, 18 },
 { 0x077d3852L /* 0.468071291 */, 18 }, { 0x077d9c24L /* 0.468166488 */, 18 },
 { 0x077dfff8L /* 0.468261690 */, 18 }, { 0x077e63cdL /* 0.468356896 */, 18 },

 { 0x077ec7a3L /* 0.468452108 */, 18 }, { 0x077f2b7aL /* 0.468547324 */, 18 },
 { 0x077f8f53L /* 0.468642545 */, 18 }, { 0x077ff32dL /* 0.468737771 */, 18 },
 { 0x07805708L /* 0.468833002 */, 18 }, { 0x0780bae5L /* 0.468928237 */, 18 },
 { 0x07811ec3L /* 0.469023478 */, 18 }, { 0x078182a2L /* 0.469118723 */, 18 },
 { 0x0781e683L /* 0.469213973 */, 18 }, { 0x07824a64L /* 0.469309228 */, 18 },
 { 0x0782ae47L /* 0.469404488 */, 18 }, { 0x0783122cL /* 0.469499752 */, 18 },
 { 0x07837612L /* 0.469595022 */, 18 }, { 0x0783d9f9L /* 0.469690296 */, 18 },
 { 0x07843de1L /* 0.469785575 */, 18 }, { 0x0784a1caL /* 0.469880859 */, 18 },

 { 0x078505b5L /* 0.469976148 */, 18 }, { 0x078569a2L /* 0.470071442 */, 18 },
 { 0x0785cd8fL /* 0.470166740 */, 18 }, { 0x0786317eL /* 0.470262043 */, 18 },
 { 0x0786956eL /* 0.470357351 */, 18 }, { 0x0786f95fL /* 0.470452664 */, 18 },
 { 0x07875d52L /* 0.470547982 */, 18 }, { 0x0787c146L /* 0.470643305 */, 18 },
 { 0x0788253bL /* 0.470738632 */, 18 }, { 0x07888932L /* 0.470833964 */, 18 },
 { 0x0788ed2aL /* 0.470929301 */, 18 }, { 0x07895123L /* 0.471024643 */, 18 },
 { 0x0789b51dL /* 0.471119990 */, 18 }, { 0x078a1919L /* 0.471215341 */, 18 },
 { 0x078a7d16L /* 0.471310698 */, 18 }, { 0x078ae114L /* 0.471406059 */, 18 },

 { 0x078b4514L /* 0.471501425 */, 18 }, { 0x078ba915L /* 0.471596796 */, 18 },
 { 0x078c0d17L /* 0.471692171 */, 18 }, { 0x078c711aL /* 0.471787552 */, 18 },
 { 0x078cd51fL /* 0.471882937 */, 18 }, { 0x078d3925L /* 0.471978327 */, 18 },
 { 0x078d9d2dL /* 0.472073722 */, 18 }, { 0x078e0135L /* 0.472169122 */, 18 },
 { 0x078e653fL /* 0.472264527 */, 18 }, { 0x078ec94bL /* 0.472359936 */, 18 },
 { 0x078f2d57L /* 0.472455350 */, 18 }, { 0x078f9165L /* 0.472550769 */, 18 },
 { 0x078ff574L /* 0.472646193 */, 18 }, { 0x07905985L /* 0.472741622 */, 18 },
 { 0x0790bd96L /* 0.472837055 */, 18 }, { 0x079121a9L /* 0.472932493 */, 18 },

 { 0x079185beL /* 0.473027937 */, 18 }, { 0x0791e9d3L /* 0.473123384 */, 18 },
 { 0x07924deaL /* 0.473218837 */, 18 }, { 0x0792b202L /* 0.473314295 */, 18 },
 { 0x0793161cL /* 0.473409757 */, 18 }, { 0x07937a37L /* 0.473505224 */, 18 },
 { 0x0793de53L /* 0.473600696 */, 18 }, { 0x07944270L /* 0.473696173 */, 18 },
 { 0x0794a68fL /* 0.473791655 */, 18 }, { 0x07950aafL /* 0.473887141 */, 18 },
 { 0x07956ed0L /* 0.473982632 */, 18 }, { 0x0795d2f2L /* 0.474078128 */, 18 },
 { 0x07963716L /* 0.474173629 */, 18 }, { 0x07969b3bL /* 0.474269135 */, 18 },
 { 0x0796ff62L /* 0.474364645 */, 18 }, { 0x07976389L /* 0.474460161 */, 18 },

 { 0x0797c7b2L /* 0.474555681 */, 18 }, { 0x07982bddL /* 0.474651205 */, 18 },
 { 0x07989008L /* 0.474746735 */, 18 }, { 0x0798f435L /* 0.474842270 */, 18 },
 { 0x07995863L /* 0.474937809 */, 18 }, { 0x0799bc92L /* 0.475033353 */, 18 },
 { 0x079a20c3L /* 0.475128902 */, 18 }, { 0x079a84f5L /* 0.475224456 */, 18 },
 { 0x079ae929L /* 0.475320014 */, 18 }, { 0x079b4d5dL /* 0.475415578 */, 18 },
 { 0x079bb193L /* 0.475511146 */, 18 }, { 0x079c15caL /* 0.475606719 */, 18 },
 { 0x079c7a03L /* 0.475702296 */, 18 }, { 0x079cde3cL /* 0.475797879 */, 18 },
 { 0x079d4277L /* 0.475893466 */, 18 }, { 0x079da6b4L /* 0.475989058 */, 18 },

 { 0x079e0af1L /* 0.476084655 */, 18 }, { 0x079e6f30L /* 0.476180257 */, 18 },
 { 0x079ed370L /* 0.476275863 */, 18 }, { 0x079f37b2L /* 0.476371475 */, 18 },
 { 0x079f9bf5L /* 0.476467091 */, 18 }, { 0x07a00039L /* 0.476562712 */, 18 },
 { 0x07a0647eL /* 0.476658338 */, 18 }, { 0x07a0c8c5L /* 0.476753968 */, 18 },
 { 0x07a12d0cL /* 0.476849603 */, 18 }, { 0x07a19156L /* 0.476945243 */, 18 },
 { 0x07a1f5a0L /* 0.477040888 */, 18 }, { 0x07a259ecL /* 0.477136538 */, 18 },
 { 0x07a2be39L /* 0.477232193 */, 18 }, { 0x07a32287L /* 0.477327852 */, 18 },
 { 0x07a386d7L /* 0.477423516 */, 18 }, { 0x07a3eb28L /* 0.477519185 */, 18 },

 { 0x07a44f7aL /* 0.477614858 */, 18 }, { 0x07a4b3ceL /* 0.477710537 */, 18 },
 { 0x07a51822L /* 0.477806220 */, 18 }, { 0x07a57c78L /* 0.477901908 */, 18 },
 { 0x07a5e0d0L /* 0.477997601 */, 18 }, { 0x07a64528L /* 0.478093299 */, 18 },
 { 0x07a6a982L /* 0.478189001 */, 18 }, { 0x07a70dddL /* 0.478284708 */, 18 },
 { 0x07a7723aL /* 0.478380420 */, 18 }, { 0x07a7d698L /* 0.478476137 */, 18 },
 { 0x07a83af7L /* 0.478571858 */, 18 }, { 0x07a89f57L /* 0.478667585 */, 18 },
 { 0x07a903b9L /* 0.478763316 */, 18 }, { 0x07a9681cL /* 0.478859052 */, 18 },
 { 0x07a9cc80L /* 0.478954793 */, 18 }, { 0x07aa30e5L /* 0.479050538 */, 18 },

 { 0x07aa954cL /* 0.479146288 */, 18 }, { 0x07aaf9b4L /* 0.479242043 */, 18 },
 { 0x07ab5e1eL /* 0.479337803 */, 18 }, { 0x07abc288L /* 0.479433568 */, 18 },
 { 0x07ac26f4L /* 0.479529337 */, 18 }, { 0x07ac8b61L /* 0.479625111 */, 18 },
 { 0x07acefd0L /* 0.479720890 */, 18 }, { 0x07ad543fL /* 0.479816674 */, 18 },
 { 0x07adb8b0L /* 0.479912463 */, 18 }, { 0x07ae1d23L /* 0.480008256 */, 18 },
 { 0x07ae8196L /* 0.480104054 */, 18 }, { 0x07aee60bL /* 0.480199857 */, 18 },
 { 0x07af4a81L /* 0.480295664 */, 18 }, { 0x07afaef9L /* 0.480391477 */, 18 },
 { 0x07b01372L /* 0.480487294 */, 18 }, { 0x07b077ecL /* 0.480583116 */, 18 },

 { 0x07b0dc67L /* 0.480678943 */, 18 }, { 0x07b140e4L /* 0.480774774 */, 18 },
 { 0x07b1a561L /* 0.480870611 */, 18 }, { 0x07b209e1L /* 0.480966452 */, 18 },
 { 0x07b26e61L /* 0.481062298 */, 18 }, { 0x07b2d2e3L /* 0.481158148 */, 18 },
 { 0x07b33766L /* 0.481254004 */, 18 }, { 0x07b39beaL /* 0.481349864 */, 18 },
 { 0x07b4006fL /* 0.481445729 */, 18 }, { 0x07b464f6L /* 0.481541598 */, 18 },
 { 0x07b4c97eL /* 0.481637473 */, 18 }, { 0x07b52e08L /* 0.481733352 */, 18 },
 { 0x07b59292L /* 0.481829236 */, 18 }, { 0x07b5f71eL /* 0.481925125 */, 18 },
 { 0x07b65bacL /* 0.482021019 */, 18 }, { 0x07b6c03aL /* 0.482116917 */, 18 },

 { 0x07b724caL /* 0.482212820 */, 18 }, { 0x07b7895bL /* 0.482308728 */, 18 },
 { 0x07b7ededL /* 0.482404640 */, 18 }, { 0x07b85281L /* 0.482500558 */, 18 },
 { 0x07b8b716L /* 0.482596480 */, 18 }, { 0x07b91bacL /* 0.482692407 */, 18 },
 { 0x07b98044L /* 0.482788339 */, 18 }, { 0x07b9e4dcL /* 0.482884275 */, 18 },
 { 0x07ba4976L /* 0.482980216 */, 18 }, { 0x07baae12L /* 0.483076162 */, 18 },
 { 0x07bb12aeL /* 0.483172113 */, 18 }, { 0x07bb774cL /* 0.483268069 */, 18 },
 { 0x07bbdbebL /* 0.483364029 */, 18 }, { 0x07bc408cL /* 0.483459994 */, 18 },
 { 0x07bca52dL /* 0.483555964 */, 18 }, { 0x07bd09d0L /* 0.483651939 */, 18 },

 { 0x07bd6e75L /* 0.483747918 */, 18 }, { 0x07bdd31aL /* 0.483843902 */, 18 },
 { 0x07be37c1L /* 0.483939891 */, 18 }, { 0x07be9c69L /* 0.484035885 */, 18 },
 { 0x07bf0113L /* 0.484131883 */, 18 }, { 0x07bf65bdL /* 0.484227886 */, 18 },
 { 0x07bfca69L /* 0.484323894 */, 18 }, { 0x07c02f16L /* 0.484419907 */, 18 },
 { 0x07c093c5L /* 0.484515924 */, 18 }, { 0x07c0f875L /* 0.484611946 */, 18 },
 { 0x07c15d26L /* 0.484707973 */, 18 }, { 0x07c1c1d8L /* 0.484804005 */, 18 },
 { 0x07c2268bL /* 0.484900041 */, 18 }, { 0x07c28b40L /* 0.484996083 */, 18 },
 { 0x07c2eff6L /* 0.485092128 */, 18 }, { 0x07c354aeL /* 0.485188179 */, 18 },

 { 0x07c3b967L /* 0.485284235 */, 18 }, { 0x07c41e21L /* 0.485380295 */, 18 },
 { 0x07c482dcL /* 0.485476360 */, 18 }, { 0x07c4e798L /* 0.485572430 */, 18 },
 { 0x07c54c56L /* 0.485668504 */, 18 }, { 0x07c5b115L /* 0.485764583 */, 18 },
 { 0x07c615d6L /* 0.485860667 */, 18 }, { 0x07c67a97L /* 0.485956756 */, 18 },
 { 0x07c6df5aL /* 0.486052849 */, 18 }, { 0x07c7441eL /* 0.486148948 */, 18 },
 { 0x07c7a8e4L /* 0.486245051 */, 18 }, { 0x07c80daaL /* 0.486341158 */, 18 },
 { 0x07c87272L /* 0.486437271 */, 18 }, { 0x07c8d73cL /* 0.486533388 */, 18 },
 { 0x07c93c06L /* 0.486629510 */, 18 }, { 0x07c9a0d2L /* 0.486725637 */, 18 },

 { 0x07ca059fL /* 0.486821768 */, 18 }, { 0x07ca6a6dL /* 0.486917905 */, 18 },
 { 0x07cacf3dL /* 0.487014045 */, 18 }, { 0x07cb340eL /* 0.487110191 */, 18 },
 { 0x07cb98e0L /* 0.487206342 */, 18 }, { 0x07cbfdb4L /* 0.487302497 */, 18 },
 { 0x07cc6288L /* 0.487398657 */, 18 }, { 0x07ccc75eL /* 0.487494821 */, 18 },
 { 0x07cd2c36L /* 0.487590991 */, 18 }, { 0x07cd910eL /* 0.487687165 */, 18 },
 { 0x07cdf5e8L /* 0.487783344 */, 18 }, { 0x07ce5ac3L /* 0.487879528 */, 18 },
 { 0x07cebfa0L /* 0.487975716 */, 18 }, { 0x07cf247dL /* 0.488071909 */, 18 },
 { 0x07cf895cL /* 0.488168107 */, 18 }, { 0x07cfee3cL /* 0.488264310 */, 18 },

 { 0x07d0531eL /* 0.488360517 */, 18 }, { 0x07d0b801L /* 0.488456729 */, 18 },
 { 0x07d11ce5L /* 0.488552946 */, 18 }, { 0x07d181caL /* 0.488649167 */, 18 },
 { 0x07d1e6b0L /* 0.488745394 */, 18 }, { 0x07d24b98L /* 0.488841625 */, 18 },
 { 0x07d2b081L /* 0.488937860 */, 18 }, { 0x07d3156cL /* 0.489034101 */, 18 },
 { 0x07d37a57L /* 0.489130346 */, 18 }, { 0x07d3df44L /* 0.489226596 */, 18 },
 { 0x07d44432L /* 0.489322851 */, 18 }, { 0x07d4a922L /* 0.489419110 */, 18 },
 { 0x07d50e13L /* 0.489515375 */, 18 }, { 0x07d57305L /* 0.489611643 */, 18 },
 { 0x07d5d7f8L /* 0.489707917 */, 18 }, { 0x07d63cecL /* 0.489804195 */, 18 },

 { 0x07d6a1e2L /* 0.489900479 */, 18 }, { 0x07d706d9L /* 0.489996766 */, 18 },
 { 0x07d76bd2L /* 0.490093059 */, 18 }, { 0x07d7d0cbL /* 0.490189356 */, 18 },
 { 0x07d835c6L /* 0.490285658 */, 18 }, { 0x07d89ac2L /* 0.490381965 */, 18 },
 { 0x07d8ffc0L /* 0.490478277 */, 18 }, { 0x07d964beL /* 0.490574593 */, 18 },
 { 0x07d9c9beL /* 0.490670914 */, 18 }, { 0x07da2ebfL /* 0.490767239 */, 18 },
 { 0x07da93c2L /* 0.490863570 */, 18 }, { 0x07daf8c6L /* 0.490959905 */, 18 },
 { 0x07db5dcbL /* 0.491056245 */, 18 }, { 0x07dbc2d1L /* 0.491152589 */, 18 },
 { 0x07dc27d9L /* 0.491248939 */, 18 }, { 0x07dc8ce1L /* 0.491345293 */, 18 },

 { 0x07dcf1ecL /* 0.491441651 */, 18 }, { 0x07dd56f7L /* 0.491538015 */, 18 },
 { 0x07ddbc04L /* 0.491634383 */, 18 }, { 0x07de2111L /* 0.491730756 */, 18 },
 { 0x07de8621L /* 0.491827134 */, 18 }, { 0x07deeb31L /* 0.491923516 */, 18 },
 { 0x07df5043L /* 0.492019903 */, 18 }, { 0x07dfb556L /* 0.492116295 */, 18 },
 { 0x07e01a6aL /* 0.492212691 */, 18 }, { 0x07e07f80L /* 0.492309093 */, 18 },
 { 0x07e0e496L /* 0.492405499 */, 18 }, { 0x07e149aeL /* 0.492501909 */, 18 },
 { 0x07e1aec8L /* 0.492598325 */, 18 }, { 0x07e213e2L /* 0.492694745 */, 18 },
 { 0x07e278feL /* 0.492791170 */, 18 }, { 0x07e2de1bL /* 0.492887599 */, 18 },

 { 0x07e3433aL /* 0.492984033 */, 18 }, { 0x07e3a859L /* 0.493080472 */, 18 },
 { 0x07e40d7aL /* 0.493176916 */, 18 }, { 0x07e4729cL /* 0.493273365 */, 18 },
 { 0x07e4d7c0L /* 0.493369818 */, 18 }, { 0x07e53ce4L /* 0.493466275 */, 18 },
 { 0x07e5a20aL /* 0.493562738 */, 18 }, { 0x07e60732L /* 0.493659205 */, 18 },
 { 0x07e66c5aL /* 0.493755677 */, 18 }, { 0x07e6d184L /* 0.493852154 */, 18 },
 { 0x07e736afL /* 0.493948635 */, 18 }, { 0x07e79bdbL /* 0.494045122 */, 18 },
 { 0x07e80109L /* 0.494141612 */, 18 }, { 0x07e86638L /* 0.494238108 */, 18 },
 { 0x07e8cb68L /* 0.494334608 */, 18 }, { 0x07e93099L /* 0.494431113 */, 18 },

 { 0x07e995ccL /* 0.494527623 */, 18 }, { 0x07e9fb00L /* 0.494624137 */, 18 },
 { 0x07ea6035L /* 0.494720656 */, 18 }, { 0x07eac56bL /* 0.494817180 */, 18 },
 { 0x07eb2aa3L /* 0.494913709 */, 18 }, { 0x07eb8fdcL /* 0.495010242 */, 18 },
 { 0x07ebf516L /* 0.495106780 */, 18 }, { 0x07ec5a51L /* 0.495203322 */, 18 },
 { 0x07ecbf8eL /* 0.495299870 */, 18 }, { 0x07ed24ccL /* 0.495396422 */, 18 },
 { 0x07ed8a0bL /* 0.495492978 */, 18 }, { 0x07edef4cL /* 0.495589540 */, 18 },
 { 0x07ee548eL /* 0.495686106 */, 18 }, { 0x07eeb9d1L /* 0.495782677 */, 18 },
 { 0x07ef1f15L /* 0.495879252 */, 18 }, { 0x07ef845bL /* 0.495975833 */, 18 },

 { 0x07efe9a1L /* 0.496072418 */, 18 }, { 0x07f04ee9L /* 0.496169007 */, 18 },
 { 0x07f0b433L /* 0.496265602 */, 18 }, { 0x07f1197dL /* 0.496362201 */, 18 },
 { 0x07f17ec9L /* 0.496458804 */, 18 }, { 0x07f1e416L /* 0.496555413 */, 18 },
 { 0x07f24965L /* 0.496652026 */, 18 }, { 0x07f2aeb5L /* 0.496748644 */, 18 },
 { 0x07f31405L /* 0.496845266 */, 18 }, { 0x07f37958L /* 0.496941894 */, 18 },
 { 0x07f3deabL /* 0.497038526 */, 18 }, { 0x07f44400L /* 0.497135162 */, 18 },
 { 0x07f4a956L /* 0.497231804 */, 18 }, { 0x07f50eadL /* 0.497328450 */, 18 },
 { 0x07f57405L /* 0.497425100 */, 18 }, { 0x07f5d95fL /* 0.497521756 */, 18 },

 { 0x07f63ebaL /* 0.497618416 */, 18 }, { 0x07f6a416L /* 0.497715081 */, 18 },
 { 0x07f70974L /* 0.497811750 */, 18 }, { 0x07f76ed3L /* 0.497908425 */, 18 },
 { 0x07f7d433L /* 0.498005103 */, 18 }, { 0x07f83994L /* 0.498101787 */, 18 },
 { 0x07f89ef7L /* 0.498198475 */, 18 }, { 0x07f9045aL /* 0.498295168 */, 18 },
 { 0x07f969c0L /* 0.498391866 */, 18 }, { 0x07f9cf26L /* 0.498488568 */, 18 },
 { 0x07fa348eL /* 0.498585275 */, 18 }, { 0x07fa99f6L /* 0.498681987 */, 18 },
 { 0x07faff60L /* 0.498778704 */, 18 }, { 0x07fb64ccL /* 0.498875425 */, 18 },
 { 0x07fbca38L /* 0.498972150 */, 18 }, { 0x07fc2fa6L /* 0.499068881 */, 18 },

 { 0x07fc9516L /* 0.499165616 */, 18 }, { 0x07fcfa86L /* 0.499262356 */, 18 },
 { 0x07fd5ff8L /* 0.499359101 */, 18 }, { 0x07fdc56bL /* 0.499455850 */, 18 },
 { 0x07fe2adfL /* 0.499552604 */, 18 }, { 0x07fe9054L /* 0.499649362 */, 18 },
 { 0x07fef5cbL /* 0.499746126 */, 18 }, { 0x07ff5b43L /* 0.499842894 */, 18 },
 { 0x07ffc0bcL /* 0.499939666 */, 18 }, { 0x0400131bL /* 0.250018222 */, 19 },
 { 0x040045d9L /* 0.250066613 */, 19 }, { 0x04007897L /* 0.250115006 */, 19 },
 { 0x0400ab57L /* 0.250163402 */, 19 }, { 0x0400de16L /* 0.250211800 */, 19 },
 { 0x040110d7L /* 0.250260200 */, 19 }, { 0x04014398L /* 0.250308603 */, 19 },

 { 0x04017659L /* 0.250357008 */, 19 }, { 0x0401a91cL /* 0.250405415 */, 19 },
 { 0x0401dbdfL /* 0.250453825 */, 19 }, { 0x04020ea2L /* 0.250502237 */, 19 },
 { 0x04024166L /* 0.250550652 */, 19 }, { 0x0402742bL /* 0.250599068 */, 19 },
 { 0x0402a6f0L /* 0.250647488 */, 19 }, { 0x0402d9b6L /* 0.250695909 */, 19 },
 { 0x04030c7dL /* 0.250744333 */, 19 }, { 0x04033f44L /* 0.250792759 */, 19 },
 { 0x0403720cL /* 0.250841187 */, 19 }, { 0x0403a4d5L /* 0.250889618 */, 19 },
 { 0x0403d79eL /* 0.250938051 */, 19 }, { 0x04040a68L /* 0.250986487 */, 19 },
 { 0x04043d32L /* 0.251034924 */, 19 }, { 0x04046ffdL /* 0.251083365 */, 19 },

 { 0x0404a2c9L /* 0.251131807 */, 19 }, { 0x0404d595L /* 0.251180252 */, 19 },
 { 0x04050862L /* 0.251228699 */, 19 }, { 0x04053b30L /* 0.251277148 */, 19 },
 { 0x04056dfeL /* 0.251325600 */, 19 }, { 0x0405a0cdL /* 0.251374054 */, 19 },
 { 0x0405d39cL /* 0.251422511 */, 19 }, { 0x0406066cL /* 0.251470970 */, 19 },
 { 0x0406393dL /* 0.251519431 */, 19 }, { 0x04066c0eL /* 0.251567894 */, 19 },
 { 0x04069ee0L /* 0.251616360 */, 19 }, { 0x0406d1b3L /* 0.251664828 */, 19 },
 { 0x04070486L /* 0.251713299 */, 19 }, { 0x0407375aL /* 0.251761772 */, 19 },
 { 0x04076a2eL /* 0.251810247 */, 19 }, { 0x04079d03L /* 0.251858724 */, 19 },

 { 0x0407cfd9L /* 0.251907204 */, 19 }, { 0x040802afL /* 0.251955686 */, 19 },
 { 0x04083586L /* 0.252004171 */, 19 }, { 0x0408685eL /* 0.252052658 */, 19 },
 { 0x04089b36L /* 0.252101147 */, 19 }, { 0x0408ce0fL /* 0.252149638 */, 19 },
 { 0x040900e8L /* 0.252198132 */, 19 }, { 0x040933c2L /* 0.252246628 */, 19 },
 { 0x0409669dL /* 0.252295127 */, 19 }, { 0x04099978L /* 0.252343627 */, 19 },
 { 0x0409cc54L /* 0.252392131 */, 19 }, { 0x0409ff31L /* 0.252440636 */, 19 },
 { 0x040a320eL /* 0.252489144 */, 19 }, { 0x040a64ecL /* 0.252537654 */, 19 },
 { 0x040a97cbL /* 0.252586166 */, 19 }, { 0x040acaaaL /* 0.252634681 */, 19 },

 { 0x040afd89L /* 0.252683198 */, 19 }, { 0x040b306aL /* 0.252731718 */, 19 },
 { 0x040b634bL /* 0.252780240 */, 19 }, { 0x040b962cL /* 0.252828764 */, 19 },
 { 0x040bc90eL /* 0.252877290 */, 19 }, { 0x040bfbf1L /* 0.252925819 */, 19 },
 { 0x040c2ed5L /* 0.252974350 */, 19 }, { 0x040c61b9L /* 0.253022883 */, 19 },
 { 0x040c949eL /* 0.253071419 */, 19 }, { 0x040cc783L /* 0.253119957 */, 19 },
 { 0x040cfa69L /* 0.253168498 */, 19 }, { 0x040d2d4fL /* 0.253217040 */, 19 },
 { 0x040d6037L /* 0.253265585 */, 19 }, { 0x040d931eL /* 0.253314133 */, 19 },
 { 0x040dc607L /* 0.253362682 */, 19 }, { 0x040df8f0L /* 0.253411234 */, 19 },

 { 0x040e2bdaL /* 0.253459789 */, 19 }, { 0x040e5ec4L /* 0.253508345 */, 19 },
 { 0x040e91afL /* 0.253556904 */, 19 }, { 0x040ec49bL /* 0.253605466 */, 19 },
 { 0x040ef787L /* 0.253654029 */, 19 }, { 0x040f2a74L /* 0.253702595 */, 19 },
 { 0x040f5d61L /* 0.253751164 */, 19 }, { 0x040f904fL /* 0.253799734 */, 19 },
 { 0x040fc33eL /* 0.253848307 */, 19 }, { 0x040ff62dL /* 0.253896883 */, 19 },
 { 0x0410291dL /* 0.253945460 */, 19 }, { 0x04105c0eL /* 0.253994040 */, 19 },
 { 0x04108effL /* 0.254042622 */, 19 }, { 0x0410c1f1L /* 0.254091207 */, 19 },
 { 0x0410f4e3L /* 0.254139794 */, 19 }, { 0x041127d6L /* 0.254188383 */, 19 },

 { 0x04115acaL /* 0.254236974 */, 19 }, { 0x04118dbeL /* 0.254285568 */, 19 },
 { 0x0411c0b3L /* 0.254334165 */, 19 }, { 0x0411f3a9L /* 0.254382763 */, 19 },
 { 0x0412269fL /* 0.254431364 */, 19 }, { 0x04125996L /* 0.254479967 */, 19 },
 { 0x04128c8dL /* 0.254528572 */, 19 }, { 0x0412bf85L /* 0.254577180 */, 19 },
 { 0x0412f27eL /* 0.254625790 */, 19 }, { 0x04132577L /* 0.254674403 */, 19 },
 { 0x04135871L /* 0.254723017 */, 19 }, { 0x04138b6cL /* 0.254771635 */, 19 },
 { 0x0413be67L /* 0.254820254 */, 19 }, { 0x0413f163L /* 0.254868876 */, 19 },
 { 0x0414245fL /* 0.254917500 */, 19 }, { 0x0414575cL /* 0.254966126 */, 19 },

 { 0x04148a5aL /* 0.255014755 */, 19 }, { 0x0414bd58L /* 0.255063386 */, 19 },
 { 0x0414f057L /* 0.255112019 */, 19 }, { 0x04152356L /* 0.255160655 */, 19 },
 { 0x04155657L /* 0.255209292 */, 19 }, { 0x04158957L /* 0.255257933 */, 19 },
 { 0x0415bc59L /* 0.255306575 */, 19 }, { 0x0415ef5bL /* 0.255355220 */, 19 },
 { 0x0416225dL /* 0.255403867 */, 19 }, { 0x04165561L /* 0.255452517 */, 19 },
 { 0x04168864L /* 0.255501169 */, 19 }, { 0x0416bb69L /* 0.255549823 */, 19 },
 { 0x0416ee6eL /* 0.255598479 */, 19 }, { 0x04172174L /* 0.255647138 */, 19 },
 { 0x0417547aL /* 0.255695799 */, 19 }, { 0x04178781L /* 0.255744463 */, 19 },

 { 0x0417ba89L /* 0.255793128 */, 19 }, { 0x0417ed91L /* 0.255841796 */, 19 },
 { 0x0418209aL /* 0.255890467 */, 19 }, { 0x041853a3L /* 0.255939139 */, 19 },
 { 0x041886adL /* 0.255987814 */, 19 }, { 0x0418b9b8L /* 0.256036492 */, 19 },
 { 0x0418ecc3L /* 0.256085171 */, 19 }, { 0x04191fcfL /* 0.256133853 */, 19 },
 { 0x041952dcL /* 0.256182537 */, 19 }, { 0x041985e9L /* 0.256231224 */, 19 },
 { 0x0419b8f7L /* 0.256279913 */, 19 }, { 0x0419ec05L /* 0.256328604 */, 19 },
 { 0x041a1f15L /* 0.256377297 */, 19 }, { 0x041a5224L /* 0.256425993 */, 19 },
 { 0x041a8534L /* 0.256474691 */, 19 }, { 0x041ab845L /* 0.256523392 */, 19 },

 { 0x041aeb57L /* 0.256572095 */, 19 }, { 0x041b1e69L /* 0.256620800 */, 19 },
 { 0x041b517cL /* 0.256669507 */, 19 }, { 0x041b848fL /* 0.256718217 */, 19 },
 { 0x041bb7a3L /* 0.256766929 */, 19 }, { 0x041beab8L /* 0.256815643 */, 19 },
 { 0x041c1dcdL /* 0.256864359 */, 19 }, { 0x041c50e3L /* 0.256913078 */, 19 },
 { 0x041c83faL /* 0.256961800 */, 19 }, { 0x041cb711L /* 0.257010523 */, 19 },
 { 0x041cea28L /* 0.257059249 */, 19 }, { 0x041d1d41L /* 0.257107977 */, 19 },
 { 0x041d505aL /* 0.257156708 */, 19 }, { 0x041d8373L /* 0.257205440 */, 19 },
 { 0x041db68eL /* 0.257254175 */, 19 }, { 0x041de9a8L /* 0.257302913 */, 19 },

 { 0x041e1cc4L /* 0.257351652 */, 19 }, { 0x041e4fe0L /* 0.257400394 */, 19 },
 { 0x041e82fdL /* 0.257449139 */, 19 }, { 0x041eb61aL /* 0.257497885 */, 19 },
 { 0x041ee938L /* 0.257546634 */, 19 }, { 0x041f1c57L /* 0.257595386 */, 19 },
 { 0x041f4f76L /* 0.257644139 */, 19 }, { 0x041f8296L /* 0.257692895 */, 19 },
 { 0x041fb5b6L /* 0.257741653 */, 19 }, { 0x041fe8d7L /* 0.257790414 */, 19 },
 { 0x04201bf9L /* 0.257839176 */, 19 }, { 0x04204f1bL /* 0.257887941 */, 19 },
 { 0x0420823eL /* 0.257936709 */, 19 }, { 0x0420b561L /* 0.257985478 */, 19 },
 { 0x0420e885L /* 0.258034250 */, 19 }, { 0x04211baaL /* 0.258083025 */, 19 },

 { 0x04214ed0L /* 0.258131801 */, 19 }, { 0x042181f6L /* 0.258180580 */, 19 },
 { 0x0421b51cL /* 0.258229361 */, 19 }, { 0x0421e843L /* 0.258278145 */, 19 },
 { 0x04221b6bL /* 0.258326931 */, 19 }, { 0x04224e94L /* 0.258375719 */, 19 },
 { 0x042281bdL /* 0.258424509 */, 19 }, { 0x0422b4e6L /* 0.258473302 */, 19 },
 { 0x0422e811L /* 0.258522097 */, 19 }, { 0x04231b3cL /* 0.258570894 */, 19 },
 { 0x04234e67L /* 0.258619694 */, 19 }, { 0x04238193L /* 0.258668496 */, 19 },
 { 0x0423b4c0L /* 0.258717300 */, 19 }, { 0x0423e7eeL /* 0.258766106 */, 19 },
 { 0x04241b1cL /* 0.258814915 */, 19 }, { 0x04244e4aL /* 0.258863726 */, 19 },

 { 0x04248179L /* 0.258912540 */, 19 }, { 0x0424b4a9L /* 0.258961356 */, 19 },
 { 0x0424e7daL /* 0.259010174 */, 19 }, { 0x04251b0bL /* 0.259058994 */, 19 },
 { 0x04254e3dL /* 0.259107817 */, 19 }, { 0x0425816fL /* 0.259156642 */, 19 },
 { 0x0425b4a2L /* 0.259205469 */, 19 }, { 0x0425e7d6L /* 0.259254298 */, 19 },
 { 0x04261b0aL /* 0.259303130 */, 19 }, { 0x04264e3fL /* 0.259351964 */, 19 },
 { 0x04268174L /* 0.259400801 */, 19 }, { 0x0426b4aaL /* 0.259449639 */, 19 },
 { 0x0426e7e1L /* 0.259498480 */, 19 }, { 0x04271b18L /* 0.259547324 */, 19 },
 { 0x04274e50L /* 0.259596169 */, 19 }, { 0x04278188L /* 0.259645017 */, 19 },

 { 0x0427b4c2L /* 0.259693868 */, 19 }, { 0x0427e7fbL /* 0.259742720 */, 19 },
 { 0x04281b36L /* 0.259791575 */, 19 }, { 0x04284e71L /* 0.259840432 */, 19 },
 { 0x042881acL /* 0.259889291 */, 19 }, { 0x0428b4e8L /* 0.259938153 */, 19 },
 { 0x0428e825L /* 0.259987017 */, 19 }, { 0x04291b63L /* 0.260035883 */, 19 },
 { 0x04294ea1L /* 0.260084752 */, 19 }, { 0x042981dfL /* 0.260133623 */, 19 },
 { 0x0429b51fL /* 0.260182496 */, 19 }, { 0x0429e85fL /* 0.260231372 */, 19 },
 { 0x042a1b9fL /* 0.260280249 */, 19 }, { 0x042a4ee0L /* 0.260329129 */, 19 },
 { 0x042a8222L /* 0.260378012 */, 19 }, { 0x042ab564L /* 0.260426896 */, 19 },

 { 0x042ae8a7L /* 0.260475783 */, 19 }, { 0x042b1bebL /* 0.260524673 */, 19 },
 { 0x042b4f2fL /* 0.260573564 */, 19 }, { 0x042b8274L /* 0.260622458 */, 19 },
 { 0x042bb5baL /* 0.260671354 */, 19 }, { 0x042be900L /* 0.260720252 */, 19 },
 { 0x042c1c46L /* 0.260769153 */, 19 }, { 0x042c4f8eL /* 0.260818056 */, 19 },
 { 0x042c82d6L /* 0.260866961 */, 19 }, { 0x042cb61eL /* 0.260915869 */, 19 },
 { 0x042ce967L /* 0.260964779 */, 19 }, { 0x042d1cb1L /* 0.261013691 */, 19 },
 { 0x042d4ffbL /* 0.261062606 */, 19 }, { 0x042d8346L /* 0.261111522 */, 19 },
 { 0x042db692L /* 0.261160441 */, 19 }, { 0x042de9deL /* 0.261209363 */, 19 },

 { 0x042e1d2bL /* 0.261258286 */, 19 }, { 0x042e5078L /* 0.261307212 */, 19 },
 { 0x042e83c6L /* 0.261356140 */, 19 }, { 0x042eb715L /* 0.261405071 */, 19 },
 { 0x042eea64L /* 0.261454004 */, 19 }, { 0x042f1db4L /* 0.261502939 */, 19 },
 { 0x042f5105L /* 0.261551876 */, 19 }, { 0x042f8456L /* 0.261600816 */, 19 },
 { 0x042fb7a8L /* 0.261649758 */, 19 }, { 0x042feafaL /* 0.261698702 */, 19 },
 { 0x04301e4dL /* 0.261747649 */, 19 }, { 0x043051a1L /* 0.261796597 */, 19 },
 { 0x043084f5L /* 0.261845548 */, 19 }, { 0x0430b84aL /* 0.261894502 */, 19 },
 { 0x0430eb9fL /* 0.261943458 */, 19 }, { 0x04311ef5L /* 0.261992416 */, 19 },

 { 0x0431524cL /* 0.262041376 */, 19 }, { 0x043185a3L /* 0.262090338 */, 19 },
 { 0x0431b8fbL /* 0.262139303 */, 19 }, { 0x0431ec54L /* 0.262188270 */, 19 },
 { 0x04321fadL /* 0.262237240 */, 19 }, { 0x04325306L /* 0.262286211 */, 19 },
 { 0x04328661L /* 0.262335185 */, 19 }, { 0x0432b9bcL /* 0.262384162 */, 19 },
 { 0x0432ed17L /* 0.262433140 */, 19 }, { 0x04332074L /* 0.262482121 */, 19 },
 { 0x043353d0L /* 0.262531104 */, 19 }, { 0x0433872eL /* 0.262580089 */, 19 },
 { 0x0433ba8cL /* 0.262629077 */, 19 }, { 0x0433edeaL /* 0.262678067 */, 19 },
 { 0x0434214aL /* 0.262727059 */, 19 }, { 0x043454aaL /* 0.262776054 */, 19 },

 { 0x0434880aL /* 0.262825051 */, 19 }, { 0x0434bb6bL /* 0.262874050 */, 19 },
 { 0x0434eecdL /* 0.262923051 */, 19 }, { 0x0435222fL /* 0.262972055 */, 19 },
 { 0x04355592L /* 0.263021061 */, 19 }, { 0x043588f6L /* 0.263070069 */, 19 },
 { 0x0435bc5aL /* 0.263119079 */, 19 }, { 0x0435efbfL /* 0.263168092 */, 19 },
 { 0x04362324L /* 0.263217107 */, 19 }, { 0x0436568aL /* 0.263266125 */, 19 },
 { 0x043689f1L /* 0.263315144 */, 19 }, { 0x0436bd58L /* 0.263364166 */, 19 },
 { 0x0436f0c0L /* 0.263413191 */, 19 }, { 0x04372428L /* 0.263462217 */, 19 },
 { 0x04375791L /* 0.263511246 */, 19 }, { 0x04378afbL /* 0.263560277 */, 19 },

 { 0x0437be65L /* 0.263609310 */, 19 }, { 0x0437f1d0L /* 0.263658346 */, 19 },
 { 0x0438253cL /* 0.263707384 */, 19 }, { 0x043858a8L /* 0.263756424 */, 19 },
 { 0x04388c14L /* 0.263805466 */, 19 }, { 0x0438bf82L /* 0.263854511 */, 19 },
 { 0x0438f2f0L /* 0.263903558 */, 19 }, { 0x0439265eL /* 0.263952607 */, 19 },
 { 0x043959cdL /* 0.264001659 */, 19 }, { 0x04398d3dL /* 0.264050713 */, 19 },
 { 0x0439c0aeL /* 0.264099769 */, 19 }, { 0x0439f41fL /* 0.264148827 */, 19 },
 { 0x043a2790L /* 0.264197888 */, 19 }, { 0x043a5b02L /* 0.264246951 */, 19 },
 { 0x043a8e75L /* 0.264296016 */, 19 }, { 0x043ac1e9L /* 0.264345084 */, 19 },

 { 0x043af55dL /* 0.264394153 */, 19 }, { 0x043b28d2L /* 0.264443225 */, 19 },
 { 0x043b5c47L /* 0.264492300 */, 19 }, { 0x043b8fbdL /* 0.264541376 */, 19 },
 { 0x043bc333L /* 0.264590455 */, 19 }, { 0x043bf6aaL /* 0.264639536 */, 19 },
 { 0x043c2a22L /* 0.264688620 */, 19 }, { 0x043c5d9aL /* 0.264737706 */, 19 },
 { 0x043c9113L /* 0.264786794 */, 19 }, { 0x043cc48dL /* 0.264835884 */, 19 },
 { 0x043cf807L /* 0.264884976 */, 19 }, { 0x043d2b82L /* 0.264934071 */, 19 },
 { 0x043d5efdL /* 0.264983168 */, 19 }, { 0x043d9279L /* 0.265032268 */, 19 },
 { 0x043dc5f6L /* 0.265081369 */, 19 }, { 0x043df973L /* 0.265130473 */, 19 },

 { 0x043e2cf1L /* 0.265179580 */, 19 }, { 0x043e6070L /* 0.265228688 */, 19 },
 { 0x043e93efL /* 0.265277799 */, 19 }, { 0x043ec76eL /* 0.265326912 */, 19 },
 { 0x043efaefL /* 0.265376027 */, 19 }, { 0x043f2e6fL /* 0.265425145 */, 19 },
 { 0x043f61f1L /* 0.265474264 */, 19 }, { 0x043f9573L /* 0.265523387 */, 19 },
 { 0x043fc8f6L /* 0.265572511 */, 19 }, { 0x043ffc79L /* 0.265621638 */, 19 },
 { 0x04402ffdL /* 0.265670766 */, 19 }, { 0x04406382L /* 0.265719898 */, 19 },
 { 0x04409707L /* 0.265769031 */, 19 }, { 0x0440ca8dL /* 0.265818167 */, 19 },
 { 0x0440fe13L /* 0.265867305 */, 19 }, { 0x0441319aL /* 0.265916445 */, 19 },

 { 0x04416522L /* 0.265965588 */, 19 }, { 0x044198aaL /* 0.266014732 */, 19 },
 { 0x0441cc33L /* 0.266063880 */, 19 }, { 0x0441ffbcL /* 0.266113029 */, 19 },
 { 0x04423346L /* 0.266162181 */, 19 }, { 0x044266d1L /* 0.266211334 */, 19 },
 { 0x04429a5cL /* 0.266260491 */, 19 }, { 0x0442cde8L /* 0.266309649 */, 19 },
 { 0x04430174L /* 0.266358810 */, 19 }, { 0x04433501L /* 0.266407973 */, 19 },
 { 0x0443688fL /* 0.266457138 */, 19 }, { 0x04439c1dL /* 0.266506305 */, 19 },
 { 0x0443cfacL /* 0.266555475 */, 19 }, { 0x0444033cL /* 0.266604647 */, 19 },
 { 0x044436ccL /* 0.266653822 */, 19 }, { 0x04446a5dL /* 0.266702998 */, 19 },

 { 0x04449deeL /* 0.266752177 */, 19 }, { 0x0444d180L /* 0.266801358 */, 19 },
 { 0x04450513L /* 0.266850541 */, 19 }, { 0x044538a6L /* 0.266899727 */, 19 },
 { 0x04456c39L /* 0.266948915 */, 19 }, { 0x04459fceL /* 0.266998105 */, 19 },
 { 0x0445d363L /* 0.267047298 */, 19 }, { 0x044606f8L /* 0.267096492 */, 19 },
 { 0x04463a8fL /* 0.267145689 */, 19 }, { 0x04466e25L /* 0.267194888 */, 19 },
 { 0x0446a1bdL /* 0.267244090 */, 19 }, { 0x0446d555L /* 0.267293294 */, 19 },
 { 0x044708eeL /* 0.267342500 */, 19 }, { 0x04473c87L /* 0.267391708 */, 19 },
 { 0x04477021L /* 0.267440919 */, 19 }, { 0x0447a3bbL /* 0.267490131 */, 19 },

 { 0x0447d756L /* 0.267539347 */, 19 }, { 0x04480af2L /* 0.267588564 */, 19 },
 { 0x04483e8eL /* 0.267637783 */, 19 }, { 0x0448722bL /* 0.267687005 */, 19 },
 { 0x0448a5c9L /* 0.267736229 */, 19 }, { 0x0448d967L /* 0.267785456 */, 19 },
 { 0x04490d05L /* 0.267834685 */, 19 }, { 0x044940a5L /* 0.267883915 */, 19 },
 { 0x04497445L /* 0.267933149 */, 19 }, { 0x0449a7e5L /* 0.267982384 */, 19 },
 { 0x0449db86L /* 0.268031622 */, 19 }, { 0x044a0f28L /* 0.268080862 */, 19 },
 { 0x044a42caL /* 0.268130104 */, 19 }, { 0x044a766dL /* 0.268179349 */, 19 },
 { 0x044aaa11L /* 0.268228595 */, 19 }, { 0x044addb5L /* 0.268277844 */, 19 },

 { 0x044b115aL /* 0.268327096 */, 19 }, { 0x044b44ffL /* 0.268376349 */, 19 },
 { 0x044b78a5L /* 0.268425605 */, 19 }, { 0x044bac4cL /* 0.268474863 */, 19 },
 { 0x044bdff3L /* 0.268524123 */, 19 }, { 0x044c139bL /* 0.268573386 */, 19 },
 { 0x044c4743L /* 0.268622651 */, 19 }, { 0x044c7aecL /* 0.268671918 */, 19 },
 { 0x044cae96L /* 0.268721187 */, 19 }, { 0x044ce240L /* 0.268770459 */, 19 },
 { 0x044d15ebL /* 0.268819733 */, 19 }, { 0x044d4997L /* 0.268869009 */, 19 },
 { 0x044d7d43L /* 0.268918287 */, 19 }, { 0x044db0efL /* 0.268967568 */, 19 },
 { 0x044de49dL /* 0.269016851 */, 19 }, { 0x044e184bL /* 0.269066136 */, 19 },

 { 0x044e4bf9L /* 0.269115423 */, 19 }, { 0x044e7fa8L /* 0.269164713 */, 19 },
 { 0x044eb358L /* 0.269214005 */, 19 }, { 0x044ee708L /* 0.269263299 */, 19 },
 { 0x044f1ab9L /* 0.269312595 */, 19 }, { 0x044f4e6bL /* 0.269361894 */, 19 },
 { 0x044f821dL /* 0.269411195 */, 19 }, { 0x044fb5cfL /* 0.269460498 */, 19 },
 { 0x044fe983L /* 0.269509804 */, 19 }, { 0x04501d37L /* 0.269559111 */, 19 },
 { 0x045050ebL /* 0.269608421 */, 19 }, { 0x045084a0L /* 0.269657734 */, 19 },
 { 0x0450b856L /* 0.269707048 */, 19 }, { 0x0450ec0dL /* 0.269756365 */, 19 },
 { 0x04511fc4L /* 0.269805684 */, 19 }, { 0x0451537bL /* 0.269855005 */, 19 },

 { 0x04518733L /* 0.269904329 */, 19 }, { 0x0451baecL /* 0.269953654 */, 19 },
 { 0x0451eea5L /* 0.270002982 */, 19 }, { 0x0452225fL /* 0.270052313 */, 19 },
 { 0x0452561aL /* 0.270101645 */, 19 }, { 0x045289d5L /* 0.270150980 */, 19 },
 { 0x0452bd91L /* 0.270200317 */, 19 }, { 0x0452f14dL /* 0.270249656 */, 19 },
 { 0x0453250aL /* 0.270298998 */, 19 }, { 0x045358c8L /* 0.270348341 */, 19 },
 { 0x04538c86L /* 0.270397687 */, 19 }, { 0x0453c045L /* 0.270447036 */, 19 },
 { 0x0453f405L /* 0.270496386 */, 19 }, { 0x045427c5L /* 0.270545739 */, 19 },
 { 0x04545b85L /* 0.270595094 */, 19 }, { 0x04548f46L /* 0.270644451 */, 19 },

 { 0x0454c308L /* 0.270693811 */, 19 }, { 0x0454f6cbL /* 0.270743173 */, 19 },
 { 0x04552a8eL /* 0.270792537 */, 19 }, { 0x04555e51L /* 0.270841903 */, 19 },
 { 0x04559216L /* 0.270891271 */, 19 }, { 0x0455c5dbL /* 0.270940642 */, 19 },
 { 0x0455f9a0L /* 0.270990015 */, 19 }, { 0x04562d66L /* 0.271039390 */, 19 },
 { 0x0456612dL /* 0.271088768 */, 19 }, { 0x045694f4L /* 0.271138148 */, 19 },
 { 0x0456c8bcL /* 0.271187530 */, 19 }, { 0x0456fc84L /* 0.271236914 */, 19 },
 { 0x0457304eL /* 0.271286301 */, 19 }, { 0x04576417L /* 0.271335689 */, 19 },
 { 0x045797e2L /* 0.271385080 */, 19 }, { 0x0457cbacL /* 0.271434474 */, 19 },

 { 0x0457ff78L /* 0.271483869 */, 19 }, { 0x04583344L /* 0.271533267 */, 19 },
 { 0x04586711L /* 0.271582667 */, 19 }, { 0x04589adeL /* 0.271632069 */, 19 },
 { 0x0458ceacL /* 0.271681474 */, 19 }, { 0x0459027bL /* 0.271730880 */, 19 },
 { 0x0459364aL /* 0.271780289 */, 19 }, { 0x04596a19L /* 0.271829701 */, 19 },
 { 0x04599deaL /* 0.271879114 */, 19 }, { 0x0459d1bbL /* 0.271928530 */, 19 },
 { 0x045a058cL /* 0.271977948 */, 19 }, { 0x045a395eL /* 0.272027368 */, 19 },
 { 0x045a6d31L /* 0.272076790 */, 19 }, { 0x045aa104L /* 0.272126215 */, 19 },
 { 0x045ad4d8L /* 0.272175642 */, 19 }, { 0x045b08adL /* 0.272225071 */, 19 },

 { 0x045b3c82L /* 0.272274503 */, 19 }, { 0x045b7058L /* 0.272323936 */, 19 },
 { 0x045ba42eL /* 0.272373372 */, 19 }, { 0x045bd805L /* 0.272422810 */, 19 },
 { 0x045c0bddL /* 0.272472251 */, 19 }, { 0x045c3fb5L /* 0.272521693 */, 19 },
 { 0x045c738eL /* 0.272571138 */, 19 }, { 0x045ca767L /* 0.272620585 */, 19 },
 { 0x045cdb41L /* 0.272670035 */, 19 }, { 0x045d0f1bL /* 0.272719486 */, 19 },
 { 0x045d42f7L /* 0.272768940 */, 19 }, { 0x045d76d2L /* 0.272818396 */, 19 },
 { 0x045daaafL /* 0.272867855 */, 19 }, { 0x045dde8cL /* 0.272917315 */, 19 },
 { 0x045e1269L /* 0.272966778 */, 19 }, { 0x045e4647L /* 0.273016243 */, 19 },

 { 0x045e7a26L /* 0.273065710 */, 19 }, { 0x045eae06L /* 0.273115180 */, 19 },
 { 0x045ee1e6L /* 0.273164652 */, 19 }, { 0x045f15c6L /* 0.273214126 */, 19 },
 { 0x045f49a7L /* 0.273263602 */, 19 }, { 0x045f7d89L /* 0.273313081 */, 19 },
 { 0x045fb16cL /* 0.273362561 */, 19 }, { 0x045fe54fL /* 0.273412044 */, 19 },
 { 0x04601932L /* 0.273461530 */, 19 }, { 0x04604d16L /* 0.273511017 */, 19 },
 { 0x046080fbL /* 0.273560507 */, 19 }, { 0x0460b4e1L /* 0.273609999 */, 19 },
 { 0x0460e8c7L /* 0.273659493 */, 19 }, { 0x04611cadL /* 0.273708989 */, 19 },
 { 0x04615094L /* 0.273758488 */, 19 }, { 0x0461847cL /* 0.273807989 */, 19 },

 { 0x0461b864L /* 0.273857492 */, 19 }, { 0x0461ec4dL /* 0.273906997 */, 19 },
 { 0x04622037L /* 0.273956505 */, 19 }, { 0x04625421L /* 0.274006015 */, 19 },
 { 0x0462880cL /* 0.274055527 */, 19 }, { 0x0462bbf7L /* 0.274105041 */, 19 },
 { 0x0462efe3L /* 0.274154558 */, 19 }, { 0x046323d0L /* 0.274204076 */, 19 },
 { 0x046357bdL /* 0.274253597 */, 19 }, { 0x04638babL /* 0.274303121 */, 19 },
 { 0x0463bf99L /* 0.274352646 */, 19 }, { 0x0463f388L /* 0.274402174 */, 19 },
 { 0x04642778L /* 0.274451704 */, 19 }, { 0x04645b68L /* 0.274501236 */, 19 },
 { 0x04648f59L /* 0.274550771 */, 19 }, { 0x0464c34aL /* 0.274600307 */, 19 },

 { 0x0464f73cL /* 0.274649846 */, 19 }, { 0x04652b2fL /* 0.274699387 */, 19 },
 { 0x04655f22L /* 0.274748931 */, 19 }, { 0x04659316L /* 0.274798476 */, 19 },
 { 0x0465c70aL /* 0.274848024 */, 19 }, { 0x0465faffL /* 0.274897574 */, 19 },
 { 0x04662ef5L /* 0.274947126 */, 19 }, { 0x046662ebL /* 0.274996681 */, 19 },
 { 0x046696e2L /* 0.275046238 */, 19 }, { 0x0466cad9L /* 0.275095797 */, 19 },
 { 0x0466fed1L /* 0.275145358 */, 19 }, { 0x046732caL /* 0.275194921 */, 19 },
 { 0x046766c3L /* 0.275244487 */, 19 }, { 0x04679abdL /* 0.275294055 */, 19 },
 { 0x0467ceb7L /* 0.275343625 */, 19 }, { 0x046802b2L /* 0.275393198 */, 19 },

 { 0x046836aeL /* 0.275442772 */, 19 }, { 0x04686aaaL /* 0.275492349 */, 19 },
 { 0x04689ea7L /* 0.275541928 */, 19 }, { 0x0468d2a4L /* 0.275591509 */, 19 },
 { 0x046906a2L /* 0.275641093 */, 19 }, { 0x04693aa1L /* 0.275690679 */, 19 },
 { 0x04696ea0L /* 0.275740267 */, 19 }, { 0x0469a2a0L /* 0.275789857 */, 19 },
 { 0x0469d6a0L /* 0.275839449 */, 19 }, { 0x046a0aa1L /* 0.275889044 */, 19 },
 { 0x046a3ea3L /* 0.275938641 */, 19 }, { 0x046a72a5L /* 0.275988240 */, 19 },
 { 0x046aa6a8L /* 0.276037842 */, 19 }, { 0x046adaabL /* 0.276087445 */, 19 },
 { 0x046b0eafL /* 0.276137051 */, 19 }, { 0x046b42b3L /* 0.276186659 */, 19 },

 { 0x046b76b9L /* 0.276236269 */, 19 }, { 0x046baabeL /* 0.276285882 */, 19 },
 { 0x046bdec5L /* 0.276335497 */, 19 }, { 0x046c12ccL /* 0.276385113 */, 19 },
 { 0x046c46d3L /* 0.276434733 */, 19 }, { 0x046c7adbL /* 0.276484354 */, 19 },
 { 0x046caee4L /* 0.276533978 */, 19 }, { 0x046ce2eeL /* 0.276583604 */, 19 },
 { 0x046d16f7L /* 0.276633232 */, 19 }, { 0x046d4b02L /* 0.276682862 */, 19 },
 { 0x046d7f0dL /* 0.276732495 */, 19 }, { 0x046db319L /* 0.276782129 */, 19 },
 { 0x046de725L /* 0.276831766 */, 19 }, { 0x046e1b32L /* 0.276881406 */, 19 },
 { 0x046e4f40L /* 0.276931047 */, 19 }, { 0x046e834eL /* 0.276980691 */, 19 },

 { 0x046eb75cL /* 0.277030337 */, 19 }, { 0x046eeb6cL /* 0.277079985 */, 19 },
 { 0x046f1f7cL /* 0.277129635 */, 19 }, { 0x046f538cL /* 0.277179288 */, 19 },
 { 0x046f879dL /* 0.277228942 */, 19 }, { 0x046fbbafL /* 0.277278600 */, 19 },
 { 0x046fefc1L /* 0.277328259 */, 19 }, { 0x047023d4L /* 0.277377920 */, 19 },
 { 0x047057e8L /* 0.277427584 */, 19 }, { 0x04708bfcL /* 0.277477250 */, 19 },
 { 0x0470c011L /* 0.277526918 */, 19 }, { 0x0470f426L /* 0.277576588 */, 19 },
 { 0x0471283cL /* 0.277626261 */, 19 }, { 0x04715c52L /* 0.277675936 */, 19 },
 { 0x04719069L /* 0.277725613 */, 19 }, { 0x0471c481L /* 0.277775292 */, 19 },

 { 0x0471f899L /* 0.277824973 */, 19 }, { 0x04722cb2L /* 0.277874657 */, 19 },
 { 0x047260ccL /* 0.277924343 */, 19 }, { 0x047294e6L /* 0.277974031 */, 19 },
 { 0x0472c900L /* 0.278023722 */, 19 }, { 0x0472fd1bL /* 0.278073414 */, 19 },
 { 0x04733137L /* 0.278123109 */, 19 }, { 0x04736554L /* 0.278172806 */, 19 },
 { 0x04739971L /* 0.278222505 */, 19 }, { 0x0473cd8eL /* 0.278272207 */, 19 },
 { 0x047401adL /* 0.278321910 */, 19 }, { 0x047435cbL /* 0.278371616 */, 19 },
 { 0x047469ebL /* 0.278421324 */, 19 }, { 0x04749e0bL /* 0.278471035 */, 19 },
 { 0x0474d22cL /* 0.278520747 */, 19 }, { 0x0475064dL /* 0.278570462 */, 19 },

 { 0x04753a6fL /* 0.278620179 */, 19 }, { 0x04756e91L /* 0.278669898 */, 19 },
 { 0x0475a2b4L /* 0.278719619 */, 19 }, { 0x0475d6d7L /* 0.278769343 */, 19 },
 { 0x04760afcL /* 0.278819069 */, 19 }, { 0x04763f20L /* 0.278868797 */, 19 },
 { 0x04767346L /* 0.278918527 */, 19 }, { 0x0476a76cL /* 0.278968260 */, 19 },
 { 0x0476db92L /* 0.279017995 */, 19 }, { 0x04770fbaL /* 0.279067731 */, 19 },
 { 0x047743e1L /* 0.279117471 */, 19 }, { 0x0477780aL /* 0.279167212 */, 19 },
 { 0x0477ac33L /* 0.279216956 */, 19 }, { 0x0477e05cL /* 0.279266701 */, 19 },
 { 0x04781486L /* 0.279316449 */, 19 }, { 0x047848b1L /* 0.279366200 */, 19 },

 { 0x04787cdcL /* 0.279415952 */, 19 }, { 0x0478b108L /* 0.279465707 */, 19 },
 { 0x0478e535L /* 0.279515464 */, 19 }, { 0x04791962L /* 0.279565223 */, 19 },
 { 0x04794d8fL /* 0.279614984 */, 19 }, { 0x047981beL /* 0.279664748 */, 19 },
 { 0x0479b5edL /* 0.279714513 */, 19 }, { 0x0479ea1cL /* 0.279764281 */, 19 },
 { 0x047a1e4cL /* 0.279814051 */, 19 }, { 0x047a527dL /* 0.279863824 */, 19 },
 { 0x047a86aeL /* 0.279913598 */, 19 }, { 0x047abae0L /* 0.279963375 */, 19 },
 { 0x047aef12L /* 0.280013154 */, 19 }, { 0x047b2346L /* 0.280062935 */, 19 },
 { 0x047b5779L /* 0.280112719 */, 19 }, { 0x047b8badL /* 0.280162504 */, 19 },

 { 0x047bbfe2L /* 0.280212292 */, 19 }, { 0x047bf418L /* 0.280262082 */, 19 },
 { 0x047c284eL /* 0.280311875 */, 19 }, { 0x047c5c84L /* 0.280361669 */, 19 },
 { 0x047c90bbL /* 0.280411466 */, 19 }, { 0x047cc4f3L /* 0.280461265 */, 19 },
 { 0x047cf92cL /* 0.280511066 */, 19 }, { 0x047d2d65L /* 0.280560869 */, 19 },
 { 0x047d619eL /* 0.280610675 */, 19 }, { 0x047d95d8L /* 0.280660483 */, 19 },
 { 0x047dca13L /* 0.280710292 */, 19 }, { 0x047dfe4eL /* 0.280760105 */, 19 },
 { 0x047e328aL /* 0.280809919 */, 19 }, { 0x047e66c7L /* 0.280859736 */, 19 },
 { 0x047e9b04L /* 0.280909554 */, 19 }, { 0x047ecf42L /* 0.280959375 */, 19 },

 { 0x047f0380L /* 0.281009199 */, 19 }, { 0x047f37bfL /* 0.281059024 */, 19 },
 { 0x047f6bffL /* 0.281108852 */, 19 }, { 0x047fa03fL /* 0.281158682 */, 19 },
 { 0x047fd47fL /* 0.281208514 */, 19 }, { 0x048008c1L /* 0.281258348 */, 19 },
 { 0x04803d02L /* 0.281308184 */, 19 }, { 0x04807145L /* 0.281358023 */, 19 },
 { 0x0480a588L /* 0.281407864 */, 19 }, { 0x0480d9ccL /* 0.281457707 */, 19 },
 { 0x04810e10L /* 0.281507552 */, 19 }, { 0x04814255L /* 0.281557400 */, 19 },
 { 0x0481769aL /* 0.281607250 */, 19 }, { 0x0481aae0L /* 0.281657101 */, 19 },
 { 0x0481df27L /* 0.281706956 */, 19 }, { 0x0482136eL /* 0.281756812 */, 19 },

 { 0x048247b6L /* 0.281806670 */, 19 }, { 0x04827bfeL /* 0.281856531 */, 19 },
 { 0x0482b047L /* 0.281906394 */, 19 }, { 0x0482e491L /* 0.281956259 */, 19 },
 { 0x048318dbL /* 0.282006127 */, 19 }, { 0x04834d26L /* 0.282055996 */, 19 },
 { 0x04838171L /* 0.282105868 */, 19 }, { 0x0483b5bdL /* 0.282155742 */, 19 },
 { 0x0483ea0aL /* 0.282205618 */, 19 }, { 0x04841e57L /* 0.282255496 */, 19 },
 { 0x048452a4L /* 0.282305377 */, 19 }, { 0x048486f3L /* 0.282355260 */, 19 },
 { 0x0484bb42L /* 0.282405145 */, 19 }, { 0x0484ef91L /* 0.282455032 */, 19 },
 { 0x048523e1L /* 0.282504921 */, 19 }, { 0x04855832L /* 0.282554813 */, 19 },

 { 0x04858c83L /* 0.282604707 */, 19 }, { 0x0485c0d5L /* 0.282654603 */, 19 },
 { 0x0485f527L /* 0.282704501 */, 19 }, { 0x0486297aL /* 0.282754401 */, 19 },
 { 0x04865dceL /* 0.282804304 */, 19 }, { 0x04869222L /* 0.282854209 */, 19 },
 { 0x0486c677L /* 0.282904116 */, 19 }, { 0x0486faccL /* 0.282954025 */, 19 },
 { 0x04872f22L /* 0.283003936 */, 19 }, { 0x04876379L /* 0.283053850 */, 19 },
 { 0x048797d0L /* 0.283103766 */, 19 }, { 0x0487cc28L /* 0.283153684 */, 19 },
 { 0x04880080L /* 0.283203604 */, 19 }, { 0x048834d9L /* 0.283253527 */, 19 },
 { 0x04886933L /* 0.283303451 */, 19 }, { 0x04889d8dL /* 0.283353378 */, 19 },

 { 0x0488d1e8L /* 0.283403307 */, 19 }, { 0x04890643L /* 0.283453238 */, 19 },
 { 0x04893a9fL /* 0.283503172 */, 19 }, { 0x04896efbL /* 0.283553107 */, 19 },
 { 0x0489a358L /* 0.283603045 */, 19 }, { 0x0489d7b6L /* 0.283652985 */, 19 },
 { 0x048a0c14L /* 0.283702927 */, 19 }, { 0x048a4073L /* 0.283752872 */, 19 },
 { 0x048a74d3L /* 0.283802818 */, 19 }, { 0x048aa933L /* 0.283852767 */, 19 },
 { 0x048add93L /* 0.283902718 */, 19 }, { 0x048b11f5L /* 0.283952671 */, 19 },
 { 0x048b4656L /* 0.284002627 */, 19 }, { 0x048b7ab9L /* 0.284052584 */, 19 },
 { 0x048baf1cL /* 0.284102544 */, 19 }, { 0x048be37fL /* 0.284152506 */, 19 },

 { 0x048c17e3L /* 0.284202470 */, 19 }, { 0x048c4c48L /* 0.284252436 */, 19 },
 { 0x048c80adL /* 0.284302405 */, 19 }, { 0x048cb513L /* 0.284352376 */, 19 },
 { 0x048ce97aL /* 0.284402349 */, 19 }, { 0x048d1de1L /* 0.284452324 */, 19 },
 { 0x048d5249L /* 0.284502301 */, 19 }, { 0x048d86b1L /* 0.284552281 */, 19 },
 { 0x048dbb1aL /* 0.284602263 */, 19 }, { 0x048def83L /* 0.284652246 */, 19 },
 { 0x048e23edL /* 0.284702233 */, 19 }, { 0x048e5858L /* 0.284752221 */, 19 },
 { 0x048e8cc3L /* 0.284802211 */, 19 }, { 0x048ec12fL /* 0.284852204 */, 19 },
 { 0x048ef59bL /* 0.284902199 */, 19 }, { 0x048f2a08L /* 0.284952196 */, 19 },

 { 0x048f5e76L /* 0.285002195 */, 19 }, { 0x048f92e4L /* 0.285052197 */, 19 },
 { 0x048fc753L /* 0.285102201 */, 19 }, { 0x048ffbc2L /* 0.285152206 */, 19 },
 { 0x04903032L /* 0.285202214 */, 19 }, { 0x049064a3L /* 0.285252225 */, 19 },
 { 0x04909914L /* 0.285302237 */, 19 }, { 0x0490cd86L /* 0.285352252 */, 19 },
 { 0x049101f8L /* 0.285402269 */, 19 }, { 0x0491366bL /* 0.285452288 */, 19 },
 { 0x04916adeL /* 0.285502309 */, 19 }, { 0x04919f52L /* 0.285552332 */, 19 },
 { 0x0491d3c7L /* 0.285602358 */, 19 }, { 0x0492083cL /* 0.285652386 */, 19 },
 { 0x04923cb2L /* 0.285702416 */, 19 }, { 0x04927128L /* 0.285752448 */, 19 },

 { 0x0492a59fL /* 0.285802482 */, 19 }, { 0x0492da17L /* 0.285852519 */, 19 },
 { 0x04930e8fL /* 0.285902557 */, 19 }, { 0x04934308L /* 0.285952598 */, 19 },
 { 0x04937781L /* 0.286002641 */, 19 }, { 0x0493abfbL /* 0.286052687 */, 19 },
 { 0x0493e076L /* 0.286102734 */, 19 }, { 0x049414f1L /* 0.286152784 */, 19 },
 { 0x0494496cL /* 0.286202836 */, 19 }, { 0x04947de9L /* 0.286252890 */, 19 },
 { 0x0494b266L /* 0.286302946 */, 19 }, { 0x0494e6e3L /* 0.286353005 */, 19 },
 { 0x04951b61L /* 0.286403065 */, 19 }, { 0x04954fe0L /* 0.286453128 */, 19 },
 { 0x0495845fL /* 0.286503193 */, 19 }, { 0x0495b8dfL /* 0.286553260 */, 19 },

 { 0x0495ed5fL /* 0.286603329 */, 19 }, { 0x049621e0L /* 0.286653401 */, 19 },
 { 0x04965662L /* 0.286703475 */, 19 }, { 0x04968ae4L /* 0.286753551 */, 19 },
 { 0x0496bf67L /* 0.286803629 */, 19 }, { 0x0496f3eaL /* 0.286853709 */, 19 },
 { 0x0497286eL /* 0.286903792 */, 19 }, { 0x04975cf2L /* 0.286953876 */, 19 },
 { 0x04979177L /* 0.287003963 */, 19 }, { 0x0497c5fdL /* 0.287054052 */, 19 },
 { 0x0497fa83L /* 0.287104143 */, 19 }, { 0x04982f0aL /* 0.287154237 */, 19 },
 { 0x04986392L /* 0.287204332 */, 19 }, { 0x0498981aL /* 0.287254430 */, 19 },
 { 0x0498cca2L /* 0.287304530 */, 19 }, { 0x0499012cL /* 0.287354632 */, 19 },

 { 0x049935b5L /* 0.287404737 */, 19 }, { 0x04996a40L /* 0.287454843 */, 19 },
 { 0x04999ecbL /* 0.287504952 */, 19 }, { 0x0499d356L /* 0.287555063 */, 19 },
 { 0x049a07e2L /* 0.287605176 */, 19 }, { 0x049a3c6fL /* 0.287655291 */, 19 },
 { 0x049a70fcL /* 0.287705409 */, 19 }, { 0x049aa58aL /* 0.287755528 */, 19 },
 { 0x049ada19L /* 0.287805650 */, 19 }, { 0x049b0ea8L /* 0.287855774 */, 19 },
 { 0x049b4337L /* 0.287905900 */, 19 }, { 0x049b77c8L /* 0.287956028 */, 19 },
 { 0x049bac58L /* 0.288006159 */, 19 }, { 0x049be0eaL /* 0.288056292 */, 19 },
 { 0x049c157cL /* 0.288106427 */, 19 }, { 0x049c4a0eL /* 0.288156564 */, 19 },

 { 0x049c7ea1L /* 0.288206703 */, 19 }, { 0x049cb335L /* 0.288256844 */, 19 },
 { 0x049ce7caL /* 0.288306988 */, 19 }, { 0x049d1c5eL /* 0.288357134 */, 19 },
 { 0x049d50f4L /* 0.288407282 */, 19 }, { 0x049d858aL /* 0.288457432 */, 19 },
 { 0x049dba21L /* 0.288507584 */, 19 }, { 0x049deeb8L /* 0.288557739 */, 19 },
 { 0x049e2350L /* 0.288607895 */, 19 }, { 0x049e57e8L /* 0.288658054 */, 19 },
 { 0x049e8c81L /* 0.288708215 */, 19 }, { 0x049ec11bL /* 0.288758379 */, 19 },
 { 0x049ef5b5L /* 0.288808544 */, 19 }, { 0x049f2a50L /* 0.288858712 */, 19 },
 { 0x049f5eebL /* 0.288908881 */, 19 }, { 0x049f9387L /* 0.288959053 */, 19 },

 { 0x049fc824L /* 0.289009227 */, 19 }, { 0x049ffcc1L /* 0.289059404 */, 19 },
 { 0x04a0315eL /* 0.289109582 */, 19 }, { 0x04a065fdL /* 0.289159763 */, 19 },
 { 0x04a09a9bL /* 0.289209946 */, 19 }, { 0x04a0cf3bL /* 0.289260131 */, 19 },
 { 0x04a103dbL /* 0.289310318 */, 19 }, { 0x04a1387bL /* 0.289360507 */, 19 },
 { 0x04a16d1dL /* 0.289410699 */, 19 }, { 0x04a1a1beL /* 0.289460893 */, 19 },
 { 0x04a1d661L /* 0.289511088 */, 19 }, { 0x04a20b04L /* 0.289561287 */, 19 },
 { 0x04a23fa7L /* 0.289611487 */, 19 }, { 0x04a2744bL /* 0.289661689 */, 19 },
 { 0x04a2a8f0L /* 0.289711894 */, 19 }, { 0x04a2dd95L /* 0.289762101 */, 19 },

 { 0x04a3123bL /* 0.289812309 */, 19 }, { 0x04a346e2L /* 0.289862521 */, 19 },
 { 0x04a37b89L /* 0.289912734 */, 19 }, { 0x04a3b030L /* 0.289962949 */, 19 },
 { 0x04a3e4d8L /* 0.290013167 */, 19 }, { 0x04a41981L /* 0.290063387 */, 19 },
 { 0x04a44e2bL /* 0.290113609 */, 19 }, { 0x04a482d5L /* 0.290163833 */, 19 },
 { 0x04a4b77fL /* 0.290214059 */, 19 }, { 0x04a4ec2aL /* 0.290264288 */, 19 },
 { 0x04a520d6L /* 0.290314519 */, 19 }, { 0x04a55582L /* 0.290364751 */, 19 },
 { 0x04a58a2fL /* 0.290414986 */, 19 }, { 0x04a5beddL /* 0.290465224 */, 19 },
 { 0x04a5f38bL /* 0.290515463 */, 19 }, { 0x04a62839L /* 0.290565705 */, 19 },

 { 0x04a65ce8L /* 0.290615948 */, 19 }, { 0x04a69198L /* 0.290666194 */, 19 },
 { 0x04a6c648L /* 0.290716442 */, 19 }, { 0x04a6faf9L /* 0.290766692 */, 19 },
 { 0x04a72fabL /* 0.290816945 */, 19 }, { 0x04a7645dL /* 0.290867199 */, 19 },
 { 0x04a79910L /* 0.290917456 */, 19 }, { 0x04a7cdc3L /* 0.290967715 */, 19 },
 { 0x04a80277L /* 0.291017976 */, 19 }, { 0x04a8372bL /* 0.291068239 */, 19 },
 { 0x04a86be0L /* 0.291118505 */, 19 }, { 0x04a8a096L /* 0.291168772 */, 19 },
 { 0x04a8d54cL /* 0.291219042 */, 19 }, { 0x04a90a03L /* 0.291269314 */, 19 },
 { 0x04a93ebaL /* 0.291319588 */, 19 }, { 0x04a97372L /* 0.291369865 */, 19 },

 { 0x04a9a82bL /* 0.291420143 */, 19 }, { 0x04a9dce4L /* 0.291470424 */, 19 },
 { 0x04aa119dL /* 0.291520706 */, 19 }, { 0x04aa4658L /* 0.291570991 */, 19 },
 { 0x04aa7b13L /* 0.291621278 */, 19 }, { 0x04aaafceL /* 0.291671568 */, 19 },
 { 0x04aae48aL /* 0.291721859 */, 19 }, { 0x04ab1947L /* 0.291772153 */, 19 },
 { 0x04ab4e04L /* 0.291822449 */, 19 }, { 0x04ab82c2L /* 0.291872747 */, 19 },
 { 0x04abb780L /* 0.291923047 */, 19 }, { 0x04abec3fL /* 0.291973349 */, 19 },
 { 0x04ac20feL /* 0.292023653 */, 19 }, { 0x04ac55beL /* 0.292073960 */, 19 },
 { 0x04ac8a7fL /* 0.292124269 */, 19 }, { 0x04acbf40L /* 0.292174580 */, 19 },

 { 0x04acf402L /* 0.292224893 */, 19 }, { 0x04ad28c5L /* 0.292275208 */, 19 },
 { 0x04ad5d88L /* 0.292325526 */, 19 }, { 0x04ad924bL /* 0.292375845 */, 19 },
 { 0x04adc70fL /* 0.292426167 */, 19 }, { 0x04adfbd4L /* 0.292476491 */, 19 },
 { 0x04ae3099L /* 0.292526817 */, 19 }, { 0x04ae655fL /* 0.292577145 */, 19 },
 { 0x04ae9a26L /* 0.292627476 */, 19 }, { 0x04aeceedL /* 0.292677808 */, 19 },
 { 0x04af03b4L /* 0.292728143 */, 19 }, { 0x04af387dL /* 0.292778480 */, 19 },
 { 0x04af6d45L /* 0.292828819 */, 19 }, { 0x04afa20fL /* 0.292879160 */, 19 },
 { 0x04afd6d9L /* 0.292929504 */, 19 }, { 0x04b00ba3L /* 0.292979849 */, 19 },

 { 0x04b0406eL /* 0.293030197 */, 19 }, { 0x04b0753aL /* 0.293080547 */, 19 },
 { 0x04b0aa06L /* 0.293130899 */, 19 }, { 0x04b0ded3L /* 0.293181253 */, 19 },
 { 0x04b113a1L /* 0.293231610 */, 19 }, { 0x04b1486fL /* 0.293281968 */, 19 },
 { 0x04b17d3dL /* 0.293332329 */, 19 }, { 0x04b1b20cL /* 0.293382692 */, 19 },
 { 0x04b1e6dcL /* 0.293433057 */, 19 }, { 0x04b21badL /* 0.293483424 */, 19 },
 { 0x04b2507dL /* 0.293533794 */, 19 }, { 0x04b2854fL /* 0.293584165 */, 19 },
 { 0x04b2ba21L /* 0.293634539 */, 19 }, { 0x04b2eef4L /* 0.293684915 */, 19 },
 { 0x04b323c7L /* 0.293735293 */, 19 }, { 0x04b3589bL /* 0.293785673 */, 19 },

 { 0x04b38d6fL /* 0.293836055 */, 19 }, { 0x04b3c244L /* 0.293886440 */, 19 },
 { 0x04b3f71aL /* 0.293936826 */, 19 }, { 0x04b42bf0L /* 0.293987215 */, 19 },
 { 0x04b460c7L /* 0.294037606 */, 19 }, { 0x04b4959eL /* 0.294087999 */, 19 },
 { 0x04b4ca76L /* 0.294138395 */, 19 }, { 0x04b4ff4eL /* 0.294188792 */, 19 },
 { 0x04b53427L /* 0.294239192 */, 19 }, { 0x04b56901L /* 0.294289593 */, 19 },
 { 0x04b59ddbL /* 0.294339997 */, 19 }, { 0x04b5d2b6L /* 0.294390403 */, 19 },
 { 0x04b60791L /* 0.294440812 */, 19 }, { 0x04b63c6dL /* 0.294491222 */, 19 },
 { 0x04b6714aL /* 0.294541635 */, 19 }, { 0x04b6a627L /* 0.294592049 */, 19 },

 { 0x04b6db05L /* 0.294642466 */, 19 }, { 0x04b70fe3L /* 0.294692885 */, 19 },
 { 0x04b744c2L /* 0.294743306 */, 19 }, { 0x04b779a1L /* 0.294793730 */, 19 },
 { 0x04b7ae81L /* 0.294844155 */, 19 }, { 0x04b7e362L /* 0.294894583 */, 19 },
 { 0x04b81843L /* 0.294945013 */, 19 }, { 0x04b84d24L /* 0.294995445 */, 19 },
 { 0x04b88207L /* 0.295045879 */, 19 }, { 0x04b8b6eaL /* 0.295096315 */, 19 },
 { 0x04b8ebcdL /* 0.295146753 */, 19 }, { 0x04b920b1L /* 0.295197194 */, 19 },
 { 0x04b95596L /* 0.295247637 */, 19 }, { 0x04b98a7bL /* 0.295298082 */, 19 },
 { 0x04b9bf61L /* 0.295348529 */, 19 }, { 0x04b9f447L /* 0.295398978 */, 19 },

 { 0x04ba292eL /* 0.295449429 */, 19 }, { 0x04ba5e16L /* 0.295499883 */, 19 },
 { 0x04ba92feL /* 0.295550338 */, 19 }, { 0x04bac7e6L /* 0.295600796 */, 19 },
 { 0x04bafcd0L /* 0.295651256 */, 19 }, { 0x04bb31b9L /* 0.295701718 */, 19 },
 { 0x04bb66a4L /* 0.295752183 */, 19 }, { 0x04bb9b8fL /* 0.295802649 */, 19 },
 { 0x04bbd07aL /* 0.295853118 */, 19 }, { 0x04bc0566L /* 0.295903588 */, 19 },
 { 0x04bc3a53L /* 0.295954061 */, 19 }, { 0x04bc6f40L /* 0.296004536 */, 19 },
 { 0x04bca42eL /* 0.296055013 */, 19 }, { 0x04bcd91dL /* 0.296105493 */, 19 },
 { 0x04bd0e0cL /* 0.296155974 */, 19 }, { 0x04bd42fbL /* 0.296206458 */, 19 },

 { 0x04bd77ecL /* 0.296256944 */, 19 }, { 0x04bdacdcL /* 0.296307432 */, 19 },
 { 0x04bde1ceL /* 0.296357922 */, 19 }, { 0x04be16c0L /* 0.296408414 */, 19 },
 { 0x04be4bb2L /* 0.296458908 */, 19 }, { 0x04be80a5L /* 0.296509405 */, 19 },
 { 0x04beb599L /* 0.296559904 */, 19 }, { 0x04beea8dL /* 0.296610404 */, 19 },
 { 0x04bf1f82L /* 0.296660907 */, 19 }, { 0x04bf5477L /* 0.296711413 */, 19 },
 { 0x04bf896dL /* 0.296761920 */, 19 }, { 0x04bfbe64L /* 0.296812429 */, 19 },
 { 0x04bff35bL /* 0.296862941 */, 19 }, { 0x04c02852L /* 0.296913455 */, 19 },
 { 0x04c05d4bL /* 0.296963971 */, 19 }, { 0x04c09243L /* 0.297014489 */, 19 },

 { 0x04c0c73dL /* 0.297065009 */, 19 }, { 0x04c0fc37L /* 0.297115531 */, 19 },
 { 0x04c13131L /* 0.297166056 */, 19 }, { 0x04c1662dL /* 0.297216582 */, 19 },
 { 0x04c19b28L /* 0.297267111 */, 19 }, { 0x04c1d025L /* 0.297317642 */, 19 },
 { 0x04c20521L /* 0.297368175 */, 19 }, { 0x04c23a1fL /* 0.297418710 */, 19 },
 { 0x04c26f1dL /* 0.297469248 */, 19 }, { 0x04c2a41bL /* 0.297519787 */, 19 },
 { 0x04c2d91bL /* 0.297570329 */, 19 }, { 0x04c30e1aL /* 0.297620873 */, 19 },
 { 0x04c3431bL /* 0.297671418 */, 19 }, { 0x04c3781cL /* 0.297721967 */, 19 },
 { 0x04c3ad1dL /* 0.297772517 */, 19 }, { 0x04c3e21fL /* 0.297823069 */, 19 },

 { 0x04c41722L /* 0.297873624 */, 19 }, { 0x04c44c25L /* 0.297924180 */, 19 },
 { 0x04c48129L /* 0.297974739 */, 19 }, { 0x04c4b62dL /* 0.298025300 */, 19 },
 { 0x04c4eb32L /* 0.298075863 */, 19 }, { 0x04c52038L /* 0.298126429 */, 19 },
 { 0x04c5553eL /* 0.298176996 */, 19 }, { 0x04c58a44L /* 0.298227565 */, 19 },
 { 0x04c5bf4cL /* 0.298278137 */, 19 }, { 0x04c5f453L /* 0.298328711 */, 19 },
 { 0x04c6295cL /* 0.298379287 */, 19 }, { 0x04c65e65L /* 0.298429865 */, 19 },
 { 0x04c6936eL /* 0.298480445 */, 19 }, { 0x04c6c878L /* 0.298531028 */, 19 },
 { 0x04c6fd83L /* 0.298581612 */, 19 }, { 0x04c7328eL /* 0.298632199 */, 19 },

 { 0x04c7679aL /* 0.298682788 */, 19 }, { 0x04c79ca7L /* 0.298733379 */, 19 },
 { 0x04c7d1b4L /* 0.298783972 */, 19 }, { 0x04c806c1L /* 0.298834567 */, 19 },
 { 0x04c83bcfL /* 0.298885165 */, 19 }, { 0x04c870deL /* 0.298935764 */, 19 },
 { 0x04c8a5edL /* 0.298986366 */, 19 }, { 0x04c8dafdL /* 0.299036970 */, 19 },
 { 0x04c9100dL /* 0.299087576 */, 19 }, { 0x04c9451eL /* 0.299138184 */, 19 },
 { 0x04c97a30L /* 0.299188794 */, 19 }, { 0x04c9af42L /* 0.299239406 */, 19 },
 { 0x04c9e455L /* 0.299290021 */, 19 }, { 0x04ca1968L /* 0.299340638 */, 19 },
 { 0x04ca4e7cL /* 0.299391256 */, 19 }, { 0x04ca8391L /* 0.299441877 */, 19 },

 { 0x04cab8a6L /* 0.299492500 */, 19 }, { 0x04caedbbL /* 0.299543126 */, 19 },
 { 0x04cb22d1L /* 0.299593753 */, 19 }, { 0x04cb57e8L /* 0.299644382 */, 19 },
 { 0x04cb8d00L /* 0.299695014 */, 19 }, { 0x04cbc217L /* 0.299745648 */, 19 },
 { 0x04cbf730L /* 0.299796284 */, 19 }, { 0x04cc2c49L /* 0.299846922 */, 19 },
 { 0x04cc6163L /* 0.299897562 */, 19 }, { 0x04cc967dL /* 0.299948204 */, 19 },
 { 0x04cccb98L /* 0.299998849 */, 19 }, { 0x04cd00b3L /* 0.300049495 */, 19 },
 { 0x04cd35cfL /* 0.300100144 */, 19 }, { 0x04cd6aebL /* 0.300150795 */, 19 },
 { 0x04cda008L /* 0.300201448 */, 19 }, { 0x04cdd526L /* 0.300252103 */, 19 },

 { 0x04ce0a44L /* 0.300302761 */, 19 }, { 0x04ce3f63L /* 0.300353420 */, 19 },
 { 0x04ce7482L /* 0.300404082 */, 19 }, { 0x04cea9a2L /* 0.300454745 */, 19 },
 { 0x04cedec3L /* 0.300505411 */, 19 }, { 0x04cf13e4L /* 0.300556079 */, 19 },
 { 0x04cf4906L /* 0.300606749 */, 19 }, { 0x04cf7e28L /* 0.300657421 */, 19 },
 { 0x04cfb34bL /* 0.300708096 */, 19 }, { 0x04cfe86eL /* 0.300758772 */, 19 },
 { 0x04d01d92L /* 0.300809451 */, 19 }, { 0x04d052b6L /* 0.300860132 */, 19 },
 { 0x04d087dbL /* 0.300910815 */, 19 }, { 0x04d0bd01L /* 0.300961500 */, 19 },
 { 0x04d0f227L /* 0.301012187 */, 19 }, { 0x04d1274eL /* 0.301062876 */, 19 },

 { 0x04d15c76L /* 0.301113568 */, 19 }, { 0x04d1919eL /* 0.301164261 */, 19 },
 { 0x04d1c6c6L /* 0.301214957 */, 19 }, { 0x04d1fbefL /* 0.301265655 */, 19 },
 { 0x04d23119L /* 0.301316355 */, 19 }, { 0x04d26643L /* 0.301367057 */, 19 },
 { 0x04d29b6eL /* 0.301417761 */, 19 }, { 0x04d2d099L /* 0.301468468 */, 19 },
 { 0x04d305c5L /* 0.301519176 */, 19 }, { 0x04d33af2L /* 0.301569887 */, 19 },
 { 0x04d3701fL /* 0.301620599 */, 19 }, { 0x04d3a54dL /* 0.301671314 */, 19 },
 { 0x04d3da7bL /* 0.301722031 */, 19 }, { 0x04d40faaL /* 0.301772751 */, 19 },
 { 0x04d444d9L /* 0.301823472 */, 19 }, { 0x04d47a09L /* 0.301874195 */, 19 },

 { 0x04d4af3aL /* 0.301924921 */, 19 }, { 0x04d4e46bL /* 0.301975649 */, 19 },
 { 0x04d5199cL /* 0.302026378 */, 19 }, { 0x04d54ecfL /* 0.302077110 */, 19 },
 { 0x04d58401L /* 0.302127845 */, 19 }, { 0x04d5b935L /* 0.302178581 */, 19 },
 { 0x04d5ee69L /* 0.302229319 */, 19 }, { 0x04d6239dL /* 0.302280060 */, 19 },
 { 0x04d658d2L /* 0.302330802 */, 19 }, { 0x04d68e08L /* 0.302381547 */, 19 },
 { 0x04d6c33eL /* 0.302432294 */, 19 }, { 0x04d6f875L /* 0.302483043 */, 19 },
 { 0x04d72dadL /* 0.302533794 */, 19 }, { 0x04d762e5L /* 0.302584547 */, 19 },
 { 0x04d7981dL /* 0.302635303 */, 19 }, { 0x04d7cd56L /* 0.302686060 */, 19 },

 { 0x04d80290L /* 0.302736820 */, 19 }, { 0x04d837caL /* 0.302787581 */, 19 },
 { 0x04d86d05L /* 0.302838345 */, 19 }, { 0x04d8a240L /* 0.302889111 */, 19 },
 { 0x04d8d77cL /* 0.302939879 */, 19 }, { 0x04d90cb9L /* 0.302990650 */, 19 },
 { 0x04d941f6L /* 0.303041422 */, 19 }, { 0x04d97734L /* 0.303092197 */, 19 },
 { 0x04d9ac72L /* 0.303142973 */, 19 }, { 0x04d9e1b1L /* 0.303193752 */, 19 },
 { 0x04da16f0L /* 0.303244533 */, 19 }, { 0x04da4c30L /* 0.303295316 */, 19 },
 { 0x04da8171L /* 0.303346101 */, 19 }, { 0x04dab6b2L /* 0.303396889 */, 19 },
 { 0x04daebf4L /* 0.303447678 */, 19 }, { 0x04db2136L /* 0.303498469 */, 19 },

 { 0x04db5679L /* 0.303549263 */, 19 }, { 0x04db8bbcL /* 0.303600059 */, 19 },
 { 0x04dbc100L /* 0.303650857 */, 19 }, { 0x04dbf644L /* 0.303701657 */, 19 },
 { 0x04dc2b8aL /* 0.303752459 */, 19 }, { 0x04dc60cfL /* 0.303803263 */, 19 },
 { 0x04dc9616L /* 0.303854070 */, 19 }, { 0x04dccb5cL /* 0.303904878 */, 19 },
 { 0x04dd00a4L /* 0.303955689 */, 19 }, { 0x04dd35ecL /* 0.304006502 */, 19 },
 { 0x04dd6b34L /* 0.304057317 */, 19 }, { 0x04dda07dL /* 0.304108134 */, 19 },
 { 0x04ddd5c7L /* 0.304158953 */, 19 }, { 0x04de0b11L /* 0.304209774 */, 19 },
 { 0x04de405cL /* 0.304260597 */, 19 }, { 0x04de75a7L /* 0.304311423 */, 19 },

 { 0x04deaaf3L /* 0.304362251 */, 19 }, { 0x04dee040L /* 0.304413080 */, 19 },
 { 0x04df158dL /* 0.304463912 */, 19 }, { 0x04df4adbL /* 0.304514746 */, 19 },
 { 0x04df8029L /* 0.304565582 */, 19 }, { 0x04dfb578L /* 0.304616421 */, 19 },
 { 0x04dfeac7L /* 0.304667261 */, 19 }, { 0x04e02017L /* 0.304718103 */, 19 },
 { 0x04e05567L /* 0.304768948 */, 19 }, { 0x04e08ab8L /* 0.304819795 */, 19 },
 { 0x04e0c00aL /* 0.304870644 */, 19 }, { 0x04e0f55cL /* 0.304921495 */, 19 },
 { 0x04e12aafL /* 0.304972348 */, 19 }, { 0x04e16002L /* 0.305023203 */, 19 },
 { 0x04e19556L /* 0.305074060 */, 19 }, { 0x04e1caabL /* 0.305124920 */, 19 },

 { 0x04e20000L /* 0.305175781 */, 19 }, { 0x04e23555L /* 0.305226645 */, 19 },
 { 0x04e26aacL /* 0.305277511 */, 19 }, { 0x04e2a002L /* 0.305328379 */, 19 },
 { 0x04e2d55aL /* 0.305379249 */, 19 }, { 0x04e30ab2L /* 0.305430121 */, 19 },
 { 0x04e3400aL /* 0.305480995 */, 19 }, { 0x04e37563L /* 0.305531872 */, 19 },
 { 0x04e3aabdL /* 0.305582750 */, 19 }, { 0x04e3e017L /* 0.305633631 */, 19 },
 { 0x04e41572L /* 0.305684513 */, 19 }, { 0x04e44acdL /* 0.305735398 */, 19 },
 { 0x04e48029L /* 0.305786285 */, 19 }, { 0x04e4b585L /* 0.305837174 */, 19 },
 { 0x04e4eae2L /* 0.305888066 */, 19 }, { 0x04e52040L /* 0.305938959 */, 19 },

 { 0x04e5559eL /* 0.305989854 */, 19 }, { 0x04e58afdL /* 0.306040752 */, 19 },
 { 0x04e5c05cL /* 0.306091652 */, 19 }, { 0x04e5f5bcL /* 0.306142554 */, 19 },
 { 0x04e62b1cL /* 0.306193457 */, 19 }, { 0x04e6607dL /* 0.306244364 */, 19 },
 { 0x04e695dfL /* 0.306295272 */, 19 }, { 0x04e6cb41L /* 0.306346182 */, 19 },
 { 0x04e700a3L /* 0.306397094 */, 19 }, { 0x04e73607L /* 0.306448009 */, 19 },
 { 0x04e76b6bL /* 0.306498925 */, 19 }, { 0x04e7a0cfL /* 0.306549844 */, 19 },
 { 0x04e7d634L /* 0.306600765 */, 19 }, { 0x04e80b99L /* 0.306651688 */, 19 },
 { 0x04e84100L /* 0.306702613 */, 19 }, { 0x04e87666L /* 0.306753540 */, 19 },

 { 0x04e8abcdL /* 0.306804470 */, 19 }, { 0x04e8e135L /* 0.306855401 */, 19 },
 { 0x04e9169eL /* 0.306906334 */, 19 }, { 0x04e94c07L /* 0.306957270 */, 19 },
 { 0x04e98170L /* 0.307008208 */, 19 }, { 0x04e9b6daL /* 0.307059148 */, 19 },
 { 0x04e9ec45L /* 0.307110090 */, 19 }, { 0x04ea21b0L /* 0.307161034 */, 19 },
 { 0x04ea571cL /* 0.307211980 */, 19 }, { 0x04ea8c88L /* 0.307262928 */, 19 },
 { 0x04eac1f5L /* 0.307313879 */, 19 }, { 0x04eaf762L /* 0.307364831 */, 19 },
 { 0x04eb2cd0L /* 0.307415786 */, 19 }, { 0x04eb623fL /* 0.307466743 */, 19 },
 { 0x04eb97aeL /* 0.307517702 */, 19 }, { 0x04ebcd1eL /* 0.307568663 */, 19 },

 { 0x04ec028eL /* 0.307619626 */, 19 }, { 0x04ec37ffL /* 0.307670591 */, 19 },
 { 0x04ec6d71L /* 0.307721558 */, 19 }, { 0x04eca2e3L /* 0.307772528 */, 19 },
 { 0x04ecd855L /* 0.307823499 */, 19 }, { 0x04ed0dc8L /* 0.307874473 */, 19 },
 { 0x04ed433cL /* 0.307925449 */, 19 }, { 0x04ed78b0L /* 0.307976426 */, 19 },
 { 0x04edae25L /* 0.308027406 */, 19 }, { 0x04ede39aL /* 0.308078389 */, 19 },
 { 0x04ee1910L /* 0.308129373 */, 19 }, { 0x04ee4e87L /* 0.308180359 */, 19 },
 { 0x04ee83feL /* 0.308231347 */, 19 }, { 0x04eeb976L /* 0.308282338 */, 19 },
 { 0x04eeeeeeL /* 0.308333331 */, 19 }, { 0x04ef2467L /* 0.308384325 */, 19 },

 { 0x04ef59e0L /* 0.308435322 */, 19 }, { 0x04ef8f5aL /* 0.308486321 */, 19 },
 { 0x04efc4d5L /* 0.308537322 */, 19 }, { 0x04effa50L /* 0.308588325 */, 19 },
 { 0x04f02fcbL /* 0.308639331 */, 19 }, { 0x04f06547L /* 0.308690338 */, 19 },
 { 0x04f09ac4L /* 0.308741348 */, 19 }, { 0x04f0d041L /* 0.308792359 */, 19 },
 { 0x04f105bfL /* 0.308843373 */, 19 }, { 0x04f13b3eL /* 0.308894389 */, 19 },
 { 0x04f170bdL /* 0.308945407 */, 19 }, { 0x04f1a63cL /* 0.308996427 */, 19 },
 { 0x04f1dbbdL /* 0.309047449 */, 19 }, { 0x04f2113dL /* 0.309098473 */, 19 },
 { 0x04f246bfL /* 0.309149499 */, 19 }, { 0x04f27c40L /* 0.309200528 */, 19 },

 { 0x04f2b1c3L /* 0.309251558 */, 19 }, { 0x04f2e746L /* 0.309302591 */, 19 },
 { 0x04f31cc9L /* 0.309353626 */, 19 }, { 0x04f3524dL /* 0.309404663 */, 19 },
 { 0x04f387d2L /* 0.309455702 */, 19 }, { 0x04f3bd57L /* 0.309506743 */, 19 },
 { 0x04f3f2ddL /* 0.309557786 */, 19 }, { 0x04f42864L /* 0.309608831 */, 19 },
 { 0x04f45deaL /* 0.309659879 */, 19 }, { 0x04f49372L /* 0.309710928 */, 19 },
 { 0x04f4c8faL /* 0.309761980 */, 19 }, { 0x04f4fe83L /* 0.309813033 */, 19 },
 { 0x04f5340cL /* 0.309864089 */, 19 }, { 0x04f56996L /* 0.309915147 */, 19 },
 { 0x04f59f20L /* 0.309966207 */, 19 }, { 0x04f5d4abL /* 0.310017269 */, 19 },

 { 0x04f60a36L /* 0.310068333 */, 19 }, { 0x04f63fc2L /* 0.310119400 */, 19 },
 { 0x04f6754fL /* 0.310170468 */, 19 }, { 0x04f6aadcL /* 0.310221539 */, 19 },
 { 0x04f6e06aL /* 0.310272611 */, 19 }, { 0x04f715f8L /* 0.310323686 */, 19 },
 { 0x04f74b87L /* 0.310374763 */, 19 }, { 0x04f78116L /* 0.310425842 */, 19 },
 { 0x04f7b6a6L /* 0.310476923 */, 19 }, { 0x04f7ec37L /* 0.310528006 */, 19 },
 { 0x04f821c8L /* 0.310579091 */, 19 }, { 0x04f85759L /* 0.310630179 */, 19 },
 { 0x04f88cecL /* 0.310681268 */, 19 }, { 0x04f8c27eL /* 0.310732360 */, 19 },
 { 0x04f8f812L /* 0.310783453 */, 19 }, { 0x04f92da6L /* 0.310834549 */, 19 },

 { 0x04f9633aL /* 0.310885647 */, 19 }, { 0x04f998cfL /* 0.310936747 */, 19 },
 { 0x04f9ce65L /* 0.310987849 */, 19 }, { 0x04fa03fbL /* 0.311038953 */, 19 },
 { 0x04fa3992L /* 0.311090059 */, 19 }, { 0x04fa6f29L /* 0.311141168 */, 19 },
 { 0x04faa4c1L /* 0.311192278 */, 19 }, { 0x04fada59L /* 0.311243390 */, 19 },
 { 0x04fb0ff2L /* 0.311294505 */, 19 }, { 0x04fb458cL /* 0.311345622 */, 19 },
 { 0x04fb7b26L /* 0.311396741 */, 19 }, { 0x04fbb0c1L /* 0.311447862 */, 19 },
 { 0x04fbe65cL /* 0.311498985 */, 19 }, { 0x04fc1bf8L /* 0.311550110 */, 19 },
 { 0x04fc5194L /* 0.311601237 */, 19 }, { 0x04fc8731L /* 0.311652366 */, 19 },

 { 0x04fcbcceL /* 0.311703498 */, 19 }, { 0x04fcf26cL /* 0.311754631 */, 19 },
 { 0x04fd280bL /* 0.311805767 */, 19 }, { 0x04fd5daaL /* 0.311856905 */, 19 },
 { 0x04fd934aL /* 0.311908044 */, 19 }, { 0x04fdc8eaL /* 0.311959186 */, 19 },
 { 0x04fdfe8bL /* 0.312010330 */, 19 }, { 0x04fe342cL /* 0.312061476 */, 19 },
 { 0x04fe69ceL /* 0.312112625 */, 19 }, { 0x04fe9f71L /* 0.312163775 */, 19 },
 { 0x04fed514L /* 0.312214927 */, 19 }, { 0x04ff0ab8L /* 0.312266082 */, 19 },
 { 0x04ff405cL /* 0.312317238 */, 19 }, { 0x04ff7601L /* 0.312368397 */, 19 },
 { 0x04ffaba6L /* 0.312419558 */, 19 }, { 0x04ffe14cL /* 0.312470720 */, 19 },

 { 0x050016f3L /* 0.312521885 */, 19 }, { 0x05004c9aL /* 0.312573052 */, 19 },
 { 0x05008241L /* 0.312624222 */, 19 }, { 0x0500b7e9L /* 0.312675393 */, 19 },
 { 0x0500ed92L /* 0.312726566 */, 19 }, { 0x0501233bL /* 0.312777742 */, 19 },
 { 0x050158e5L /* 0.312828919 */, 19 }, { 0x05018e90L /* 0.312880099 */, 19 },
 { 0x0501c43bL /* 0.312931280 */, 19 }, { 0x0501f9e6L /* 0.312982464 */, 19 },
 { 0x05022f92L /* 0.313033650 */, 19 }, { 0x0502653fL /* 0.313084838 */, 19 },
 { 0x05029aecL /* 0.313136028 */, 19 }, { 0x0502d09aL /* 0.313187220 */, 19 },
 { 0x05030648L /* 0.313238414 */, 19 }, { 0x05033bf7L /* 0.313289611 */, 19 },

 { 0x050371a7L /* 0.313340809 */, 19 }, { 0x0503a757L /* 0.313392010 */, 19 },
 { 0x0503dd07L /* 0.313443212 */, 19 }, { 0x050412b9L /* 0.313494417 */, 19 },
 { 0x0504486aL /* 0.313545624 */, 19 }, { 0x05047e1dL /* 0.313596833 */, 19 },
 { 0x0504b3cfL /* 0.313648044 */, 19 }, { 0x0504e983L /* 0.313699257 */, 19 },
 { 0x05051f37L /* 0.313750472 */, 19 }, { 0x050554ebL /* 0.313801689 */, 19 },
 { 0x05058aa0L /* 0.313852909 */, 19 }, { 0x0505c056L /* 0.313904130 */, 19 },
 { 0x0505f60cL /* 0.313955354 */, 19 }, { 0x05062bc3L /* 0.314006579 */, 19 },
 { 0x0506617aL /* 0.314057807 */, 19 }, { 0x05069732L /* 0.314109037 */, 19 },

 { 0x0506ccebL /* 0.314160269 */, 19 }, { 0x050702a4L /* 0.314211502 */, 19 },
 { 0x0507385dL /* 0.314262739 */, 19 }, { 0x05076e17L /* 0.314313977 */, 19 },
 { 0x0507a3d2L /* 0.314365217 */, 19 }, { 0x0507d98dL /* 0.314416459 */, 19 },
 { 0x05080f49L /* 0.314467704 */, 19 }, { 0x05084506L /* 0.314518950 */, 19 },
 { 0x05087ac2L /* 0.314570199 */, 19 }, { 0x0508b080L /* 0.314621449 */, 19 },
 { 0x0508e63eL /* 0.314672702 */, 19 }, { 0x05091bfdL /* 0.314723957 */, 19 },
 { 0x050951bcL /* 0.314775214 */, 19 }, { 0x0509877cL /* 0.314826473 */, 19 },
 { 0x0509bd3cL /* 0.314877734 */, 19 }, { 0x0509f2fdL /* 0.314928997 */, 19 },

 { 0x050a28beL /* 0.314980262 */, 19 }, { 0x050a5e80L /* 0.315031530 */, 19 },
 { 0x050a9443L /* 0.315082799 */, 19 }, { 0x050aca06L /* 0.315134071 */, 19 },
 { 0x050affc9L /* 0.315185344 */, 19 }, { 0x050b358eL /* 0.315236620 */, 19 },
 { 0x050b6b52L /* 0.315287898 */, 19 }, { 0x050ba118L /* 0.315339178 */, 19 },
 { 0x050bd6deL /* 0.315390460 */, 19 }, { 0x050c0ca4L /* 0.315441744 */, 19 },
 { 0x050c426bL /* 0.315493030 */, 19 }, { 0x050c7833L /* 0.315544318 */, 19 },
 { 0x050cadfbL /* 0.315595608 */, 19 }, { 0x050ce3c4L /* 0.315646901 */, 19 },
 { 0x050d198dL /* 0.315698195 */, 19 }

};


/* fractional powers of two used during requantization */
static
mad_fixed_t const root_table[7] = {
  0x09837f05L /* 2^(-3/4) == 0.59460355750136 */,
  0x0b504f33L /* 2^(-2/4) == 0.70710678118655 */,
  0x0d744fcdL /* 2^(-1/4) == 0.84089641525371 */,
  0x10000000L /* 2^( 0/4) == 1.00000000000000 */,
  0x1306fe0aL /* 2^(+1/4) == 1.18920711500272 */,
  0x16a09e66L /* 2^(+2/4) == 1.41421356237310 */,
  0x1ae89f99L /* 2^(+3/4) == 1.68179283050743 */
};

/*
 * coefficients for aliasing reduction
 * derived from Table B.9 of ISO/IEC 11172-3
 *
 *  c[]  = { -0.6, -0.535, -0.33, -0.185, -0.095, -0.041, -0.0142, -0.0037 }
 * cs[i] =    1 / sqrt(1 + c[i]^2)
 * ca[i] = c[i] / sqrt(1 + c[i]^2)
 */
static
mad_fixed_t const cs[8] = {
  +0x0db84a81L /* +0.857492926 */, +0x0e1b9d7fL /* +0.881741997 */,
  +0x0f31adcfL /* +0.949628649 */, +0x0fbba815L /* +0.983314592 */,
  +0x0feda417L /* +0.995517816 */, +0x0ffc8fc8L /* +0.999160558 */,
  +0x0fff964cL /* +0.999899195 */, +0x0ffff8d3L /* +0.999993155 */
};

static
mad_fixed_t const ca[8] = {
  -0x083b5fe7L /* -0.514495755 */, -0x078c36d2L /* -0.471731969 */,
  -0x05039814L /* -0.313377454 */, -0x02e91dd1L /* -0.181913200 */,
  -0x0183603aL /* -0.094574193 */, -0x00a7cb87L /* -0.040965583 */,
  -0x003a2847L /* -0.014198569 */, -0x000f27b4L /* -0.003699975 */
};

/*
 * IMDCT coefficients for short blocks
 * derived from section 2.4.3.4.10.2 of ISO/IEC 11172-3
 *
 * imdct_s[i][k] = cos((PI / 24) * (2 * i + 7) * (2 * k + 1))
 */
static
mad_fixed_t const imdct_s[12][6] = {
 {  0x09bd7ca0L /*  0.608761429 */, -0x0ec835e8L /* -0.923879533 */,   /* 0 */
    -0x0216a2a2L /* -0.130526192 */,  0x0fdcf549L /*  0.991444861 */,
    -0x061f78aaL /* -0.382683432 */, -0x0cb19346L /* -0.793353340 */ },

  {  0x061f78aaL /*  0.382683432 */, -0x0ec835e8L /* -0.923879533 */,   /* 1 */
     0x0ec835e8L /*  0.923879533 */, -0x061f78aaL /* -0.382683432 */,
    -0x061f78aaL /* -0.382683432 */,  0x0ec835e8L /*  0.923879533 */ },

  {  0x0216a2a2L /*  0.130526192 */, -0x061f78aaL /* -0.382683432 */,   /* 2 */
     0x09bd7ca0L /*  0.608761429 */, -0x0cb19346L /* -0.793353340 */,
     0x0ec835e8L /*  0.923879533 */, -0x0fdcf549L /* -0.991444861 */ },

  { -0x0216a2a2L /* -0.130526192 */,  0x061f78aaL /*  0.382683432 */,   /* 3 */
    -0x09bd7ca0L /* -0.608761429 */,  0x0cb19346L /*  0.793353340 */,
    -0x0ec835e8L /* -0.923879533 */,  0x0fdcf549L /*  0.991444861 */ },

  { -0x061f78aaL /* -0.382683432 */,  0x0ec835e8L /*  0.923879533 */,   /* 4 */
    -0x0ec835e8L /* -0.923879533 */,  0x061f78aaL /*  0.382683432 */,
     0x061f78aaL /*  0.382683432 */, -0x0ec835e8L /* -0.923879533 */ },

  { -0x09bd7ca0L /* -0.608761429 */,  0x0ec835e8L /*  0.923879533 */,   /* 5 */
     0x0216a2a2L /*  0.130526192 */, -0x0fdcf549L /* -0.991444861 */,
     0x061f78aaL /*  0.382683432 */,  0x0cb19346L /*  0.793353340 */ },

  { -0x0cb19346L /* -0.793353340 */,  0x061f78aaL /*  0.382683432 */,   /* 6 */
     0x0fdcf549L /*  0.991444861 */,  0x0216a2a2L /*  0.130526192 */,
    -0x0ec835e8L /* -0.923879533 */, -0x09bd7ca0L /* -0.608761429 */ },

  { -0x0ec835e8L /* -0.923879533 */, -0x061f78aaL /* -0.382683432 */,   /* 7 */
     0x061f78aaL /*  0.382683432 */,  0x0ec835e8L /*  0.923879533 */,
     0x0ec835e8L /*  0.923879533 */,  0x061f78aaL /*  0.382683432 */ },

  { -0x0fdcf549L /* -0.991444861 */, -0x0ec835e8L /* -0.923879533 */,   /* 8 */
    -0x0cb19346L /* -0.793353340 */, -0x09bd7ca0L /* -0.608761429 */,
    -0x061f78aaL /* -0.382683432 */, -0x0216a2a2L /* -0.130526192 */ },

  { -0x0fdcf549L /* -0.991444861 */, -0x0ec835e8L /* -0.923879533 */,   /* 9 */
    -0x0cb19346L /* -0.793353340 */, -0x09bd7ca0L /* -0.608761429 */,
    -0x061f78aaL /* -0.382683432 */, -0x0216a2a2L /* -0.130526192 */ },

  { -0x0ec835e8L /* -0.923879533 */, -0x061f78aaL /* -0.382683432 */,  /* 10 */
     0x061f78aaL /*  0.382683432 */,  0x0ec835e8L /*  0.923879533 */,
     0x0ec835e8L /*  0.923879533 */,  0x061f78aaL /*  0.382683432 */ },

  { -0x0cb19346L /* -0.793353340 */,  0x061f78aaL /*  0.382683432 */,  /* 11 */
     0x0fdcf549L /*  0.991444861 */,  0x0216a2a2L /*  0.130526192 */,
    -0x0ec835e8L /* -0.923879533 */, -0x09bd7ca0L /* -0.608761429 */ }

};

# if !defined(ASO_IMDCT)
/*
 * windowing coefficients for long blocks
 * derived from section 2.4.3.4.10.3 of ISO/IEC 11172-3
 *
 * window_l[i] = sin((PI / 36) * (i + 1/2))
 */
static
mad_fixed_t const window_l[36] = {
  0x00b2aa3eL /* 0.043619387 */, 0x0216a2a2L /* 0.130526192 */,
  0x03768962L /* 0.216439614 */, 0x04cfb0e2L /* 0.300705800 */,
  0x061f78aaL /* 0.382683432 */, 0x07635284L /* 0.461748613 */,
  0x0898c779L /* 0.537299608 */, 0x09bd7ca0L /* 0.608761429 */,
  0x0acf37adL /* 0.675590208 */, 0x0bcbe352L /* 0.737277337 */,
  0x0cb19346L /* 0.793353340 */, 0x0d7e8807L /* 0.843391446 */,

  0x0e313245L /* 0.887010833 */, 0x0ec835e8L /* 0.923879533 */,
  0x0f426cb5L /* 0.953716951 */, 0x0f9ee890L /* 0.976296007 */,
  0x0fdcf549L /* 0.991444861 */, 0x0ffc19fdL /* 0.999048222 */,
  0x0ffc19fdL /* 0.999048222 */, 0x0fdcf549L /* 0.991444861 */,
  0x0f9ee890L /* 0.976296007 */, 0x0f426cb5L /* 0.953716951 */,
  0x0ec835e8L /* 0.923879533 */, 0x0e313245L /* 0.887010833 */,

  0x0d7e8807L /* 0.843391446 */, 0x0cb19346L /* 0.793353340 */,
  0x0bcbe352L /* 0.737277337 */, 0x0acf37adL /* 0.675590208 */,
  0x09bd7ca0L /* 0.608761429 */, 0x0898c779L /* 0.537299608 */,
  0x07635284L /* 0.461748613 */, 0x061f78aaL /* 0.382683432 */,
  0x04cfb0e2L /* 0.300705800 */, 0x03768962L /* 0.216439614 */,
  0x0216a2a2L /* 0.130526192 */, 0x00b2aa3eL /* 0.043619387 */,
};
# endif  /* ASO_IMDCT */

/*
 * windowing coefficients for short blocks
 * derived from section 2.4.3.4.10.3 of ISO/IEC 11172-3
 *
 * window_s[i] = sin((PI / 12) * (i + 1/2))
 */
static
mad_fixed_t const window_s[12] = {
  0x0216a2a2L /* 0.130526192 */, 0x061f78aaL /* 0.382683432 */,
  0x09bd7ca0L /* 0.608761429 */, 0x0cb19346L /* 0.793353340 */,
  0x0ec835e8L /* 0.923879533 */, 0x0fdcf549L /* 0.991444861 */,
  0x0fdcf549L /* 0.991444861 */, 0x0ec835e8L /* 0.923879533 */,
  0x0cb19346L /* 0.793353340 */, 0x09bd7ca0L /* 0.608761429 */,
  0x061f78aaL /* 0.382683432 */, 0x0216a2a2L /* 0.130526192 */,
};

/*
 * coefficients for intensity stereo processing
 * derived from section 2.4.3.4.9.3 of ISO/IEC 11172-3
 *
 * is_ratio[i] = tan(i * (PI / 12))
 * is_table[i] = is_ratio[i] / (1 + is_ratio[i])
 */
static
mad_fixed_t const is_table[7] = {
  0x00000000L /* 0.000000000 */,
  0x0361962fL /* 0.211324865 */,
  0x05db3d74L /* 0.366025404 */,
  0x08000000L /* 0.500000000 */,
  0x0a24c28cL /* 0.633974596 */,
  0x0c9e69d1L /* 0.788675135 */,
  0x10000000L /* 1.000000000 */
};

/*
 * coefficients for LSF intensity stereo processing
 * derived from section 2.4.3.2 of ISO/IEC 13838-3
 *
 * lsf_is_table[0][i] = (1 / sqrt(sqrt(2)))^(i + 1)
 * lsf_is_table[1][i] = (1 / sqrt(2))^(i + 1)
 */
static
mad_fixed_t const lsf_is_table[2][3] = {
  {
    0x0d744fcdL /* 0.840896415 */,
    0x0b504f33L /* 0.707106781 */,
    0x09837f05L /* 0.594603558 */
  }, {
    0x0b504f33L /* 0.707106781 */,
    0x08000000L /* 0.500000000 */,
    0x05a8279aL /* 0.353553391 */
  }
};

/*
 * NAME:	III_sideinfo()
 * DESCRIPTION:	decode frame side information from a bitstream
 */
static
int III_sideinfo(struct mad_bitptr *ptr, unsigned int nch,
		 int lsf, struct sideinfo *si,
		 unsigned int *md_bitlen, unsigned int *priv_bitlen)
{
  unsigned int ngr, gr, ch, i;
  int result = 0;

  *md_bitlen   = 0;
  *priv_bitlen = lsf ? ((nch == 1) ? 1 : 2) : ((nch == 1) ? 5 : 3);

  si->main_data_begin = mad_bit_read(ptr, lsf ? 8 : 9);
  si->private_bits    = mad_bit_read(ptr, *priv_bitlen);

  if (lsf)
    ngr = 1;
  else {
    ngr = 2;

    for (ch = 0; ch < nch; ++ch)
      si->scfsi[ch] = mad_bit_read(ptr, 4);
  }

  for (gr = 0; gr < ngr; ++gr) {
    struct granule *granule = &si->gr[gr];

    for (ch = 0; ch < nch; ++ch) {
      struct channel *channel = &granule->ch[ch];

      channel->part2_3_length    = mad_bit_read(ptr, 12);
      channel->big_values        = mad_bit_read(ptr, 9);
      channel->global_gain       = mad_bit_read(ptr, 8);
      channel->scalefac_compress = mad_bit_read(ptr, lsf ? 9 : 4);

      *md_bitlen += channel->part2_3_length;

      if (channel->big_values > 288 && result == 0)
	result = MAD_ERR_BADBIGVALUES;

      channel->flags = 0;

      if (mad_bit_read(ptr, 1)) {
	channel->flags |= window_switching_flag;

	channel->block_type = mad_bit_read(ptr, 2);
	if (channel->block_type == 0 && result == 0)
	  result = MAD_ERR_BADBLOCKTYPE;

	channel->region0_count = 7;
	channel->region1_count = 36;

	if (mad_bit_read(ptr, 1))
	  channel->flags |= mixed_block_flag;
	else if (channel->block_type == 2)
	  channel->region0_count = 8;

	for (i = 0; i < 2; ++i)
	  channel->table_select[i] = mad_bit_read(ptr, 5);

# ifdef DEBUG
	channel->table_select[2] = 4;  /* not used */
# endif

	for (i = 0; i < 3; ++i)
	  channel->subblock_gain[i] = mad_bit_read(ptr, 3);
      }
      else {
	channel->block_type = 0;

	for (i = 0; i < 3; ++i)
	  channel->table_select[i] = mad_bit_read(ptr, 5);

	channel->region0_count = mad_bit_read(ptr, 4);
	channel->region1_count = mad_bit_read(ptr, 3);
      }

      /* [preflag,] scalefac_scale, count1table_select */
      channel->flags |= mad_bit_read(ptr, lsf ? 2 : 3);
    }
  }

  return result;
}

/*
 * NAME:	III_lsf_scalefactors()
 * DESCRIPTION:	decode channel scalefactors for LSF from a bitstream
 */
static
unsigned int III_lsf_scalefactors(struct mad_bitptr *ptr, int mode_extension,
				  unsigned int ch, struct channel *channel)
{
  struct mad_bitptr start;
  unsigned int scalefac_compress, index, slen[4], part, n, sfb, w;
  unsigned short const *nsfb;
  unsigned short scalefac[36];

  start = *ptr;

  scalefac_compress = channel->scalefac_compress;
  index = (channel->block_type == 2) ?
    ((channel->flags & mixed_block_flag) ? 2 : 1) : 0;

  if (!((mode_extension & 0x1) && ch == 1)) {
    if (scalefac_compress < 400) {
      slen[0] = (scalefac_compress >> 4) / 5;
      slen[1] = (scalefac_compress >> 4) % 5;
      slen[2] = (scalefac_compress % 16) >> 2;
      slen[3] =  scalefac_compress %  4;

      nsfb = nsfb_table[0][index];
    }
    else if (scalefac_compress < 500) {
      scalefac_compress -= 400;

      slen[0] = (scalefac_compress >> 2) / 5;
      slen[1] = (scalefac_compress >> 2) % 5;
      slen[2] =  scalefac_compress %  4;
      slen[3] = 0;

      nsfb = nsfb_table[1][index];
    }
    else {
      scalefac_compress -= 500;

      slen[0] = scalefac_compress / 3;
      slen[1] = scalefac_compress % 3;
      slen[2] = 0;
      slen[3] = 0;

      channel->flags |= preflag;

      nsfb = nsfb_table[2][index];
    }
  }
  else {  /* (mode_extension & 0x1) && ch == 1 */
    scalefac_compress >>= 1;

    if (scalefac_compress < 180) {
      slen[0] =  scalefac_compress / 36;
      slen[1] = (scalefac_compress % 36) / 6;
      slen[2] = (scalefac_compress % 36) % 6;
      slen[3] = 0;

      nsfb = nsfb_table[3][index];
    }
    else if (scalefac_compress < 244) {
      scalefac_compress -= 180;

      slen[0] = (scalefac_compress % 64) >> 4;
      slen[1] = (scalefac_compress % 16) >> 2;
      slen[2] =  scalefac_compress %  4;
      slen[3] = 0;

      nsfb = nsfb_table[4][index];
    }
    else {
      scalefac_compress -= 244;

      slen[0] = scalefac_compress / 3;
      slen[1] = scalefac_compress % 3;
      slen[2] = 0;
      slen[3] = 0;

      nsfb = nsfb_table[5][index];
    }
  }

  n = 0;
  for (part = 0; part < 4; ++part) {
    for (sfb = 0; sfb < nsfb[part]; ++sfb)
      scalefac[n++] = mad_bit_read(ptr, slen[part]);
  }

  n = 0;
  if (channel->block_type == 2) {
    if (channel->flags & mixed_block_flag) {
      for (sfb = 0; sfb < 6; ++sfb)
	channel->scalefac_l[sfb] = scalefac[n++];

      for (sfb = 3; sfb < 12; ++sfb) {
	for (w = 0; w < 3; ++w)
	  channel->scalefac_s[sfb][w] = scalefac[n++];
      }
    }
    else {  /* !(channel->flags & mixed_block_flag) */
      for (sfb = 0; sfb < 12; ++sfb) {
	for (w = 0; w < 3; ++w)
	  channel->scalefac_s[sfb][w] = scalefac[n++];
      }
    }

    for (w = 0; w < 3; ++w)
      channel->scalefac_s[12][w] = 0;
  }
  else {  /* channel->block_type != 2 */
    for (sfb = 0; sfb < 21; ++sfb)
      channel->scalefac_l[sfb] = scalefac[n++];

    channel->scalefac_l[21] = 0;
  }

  return mad_bit_length(&start, ptr);
}

/*
 * NAME:	III_scalefactors()
 * DESCRIPTION:	decode channel scalefactors of one granule from a bitstream
 */
static
unsigned int III_scalefactors(struct mad_bitptr *ptr, struct channel *channel,
			      struct channel const *gr0ch, unsigned int scfsi)
{
  struct mad_bitptr start;
  unsigned int sfb, w;
  unsigned int slen1, slen2;

  start = *ptr;

  slen1 = sflen[channel->scalefac_compress].slen1;
  slen2 = sflen[channel->scalefac_compress].slen2;

  if (channel->block_type == 2) {
    if (channel->flags & mixed_block_flag) {
      for (sfb = 0; sfb < 8; ++sfb)
	channel->scalefac_l[sfb] = mad_bit_read(ptr, slen1);

      for (sfb = 3; sfb < 6; ++sfb) {
	for (w = 0; w < 3; ++w)
	  channel->scalefac_s[sfb][w] = mad_bit_read(ptr, slen1);
      }
    }
    else {  /* !(channel->flags & mixed_block_flag) */
      for (sfb = 0; sfb < 6; ++sfb) {
	for (w = 0; w < 3; ++w)
	  channel->scalefac_s[sfb][w] = mad_bit_read(ptr, slen1);
      }
    }

    for (sfb = 6; sfb < 12; ++sfb) {
      for (w = 0; w < 3; ++w)
	channel->scalefac_s[sfb][w] = mad_bit_read(ptr, slen2);
    }

    for (w = 0; w < 3; ++w)
      channel->scalefac_s[12][w] = 0;
  }
  else {  /* channel->block_type != 2 */
    if (scfsi & 0x8) {
      for (sfb = 0; sfb < 6; ++sfb)
	channel->scalefac_l[sfb] = gr0ch->scalefac_l[sfb];
    }
    else {
      for (sfb = 0; sfb < 6; ++sfb)
	channel->scalefac_l[sfb] = mad_bit_read(ptr, slen1);
    }

    if (scfsi & 0x4) {
      for (sfb = 6; sfb < 11; ++sfb)
	channel->scalefac_l[sfb] = gr0ch->scalefac_l[sfb];
    }
    else {
      for (sfb = 6; sfb < 11; ++sfb)
	channel->scalefac_l[sfb] = mad_bit_read(ptr, slen1);
    }

    if (scfsi & 0x2) {
      for (sfb = 11; sfb < 16; ++sfb)
	channel->scalefac_l[sfb] = gr0ch->scalefac_l[sfb];
    }
    else {
      for (sfb = 11; sfb < 16; ++sfb)
	channel->scalefac_l[sfb] = mad_bit_read(ptr, slen2);
    }

    if (scfsi & 0x1) {
      for (sfb = 16; sfb < 21; ++sfb)
	channel->scalefac_l[sfb] = gr0ch->scalefac_l[sfb];
    }
    else {
      for (sfb = 16; sfb < 21; ++sfb)
	channel->scalefac_l[sfb] = mad_bit_read(ptr, slen2);
    }

    channel->scalefac_l[21] = 0;
  }

  return mad_bit_length(&start, ptr);
}

/*
 * NAME:	III_requantize()
 * DESCRIPTION:	requantize one sample
 */
static
mad_fixed_t III_requantize(signed int value, signed int exp, unsigned int scf,
			   struct channel const *channel)
{
  mad_fixed_t requantized;
  signed int frac;
  struct fixedfloat const *power;

  /*
   * long blocks:
   * xr[i] = sign(is[i]) * abs(is[i])^(4/3) *
   *         2^((1/4) * (global_gain - 210)) *
   *         2^-(scalefac_multiplier *
   *             (scalefac_l[sfb] + preflag * pretab[sfb]))
   *
   * short blocks:
   * xr[i] = sign(is[i]) * abs(is[i])^(4/3) *
   *         2^((1/4) * (global_gain - 210 - 8 * subblock_gain[w])) *
   *         2^-(scalefac_multiplier * scalefac_s[sfb][w])
   *
   * where:
   * scalefac_multiplier = (scalefac_scale + 1) / 2
   */

  scf <<= (channel->flags & scalefac_scale) ? 2 : 1;
  exp  += (signed) channel->global_gain - 210 - (signed) scf;

  frac = exp % 4;
  exp /= 4;

  if (value < 0) {
    power = &rq_table[-value];
    requantized = -power->mantissa;
  }
  else {
    power = &rq_table[value];
    requantized = power->mantissa;
  }

  exp += power->exponent;

  if (exp < 0) {
    if (-exp >= sizeof(mad_fixed_t) * CHAR_BIT) {
      /* underflow */
      requantized = 0;
    }
    else
      requantized >>= -exp;
  }
  else {
    if (exp >= 5) {
      /* overflow */
# ifdef DEBUG
      fprintf(stderr, "requantize overflow (%f * 2^%d)\n",
	      mad_f_todouble(requantized), exp);
# endif
      requantized = (requantized < 0) ? MAD_F_MIN : MAD_F_MAX;
    }
    else
      requantized <<= exp;
  }

  return frac ? mad_f_mul(requantized, root_table[3 + frac]) : requantized;
}

/*
 * NAME:	III_requantize_l()
 * DESCRIPTION:	requantize one sample from a long block
 */
static
mad_fixed_t III_requantize_l(signed int value, unsigned int sfb,
			     struct channel const *channel)
{
  unsigned int scf;

  /* channel->block_type != 2 ||
     (l < 36 && (channel->flags & mixed_block_flag)) */

  scf = channel->scalefac_l[sfb];
  if (channel->flags & preflag)
    scf += pretab[sfb];

  return III_requantize(value, 0, scf, channel);
}

/*
 * NAME:	III_requantize_s()
 * DESCRIPTION:	requantize one sample from a short block
 */
static
mad_fixed_t III_requantize_s(signed int value, unsigned int sfb,
			     unsigned int window,
			     struct channel const *channel)
{
  signed int exp;
  unsigned int scf;

  /* channel->block_type == 2 &&
     (l >= 36 || !(channel->flags & mixed_block_flag)) */

  exp = (signed) channel->subblock_gain[window] * 8;
  scf = channel->scalefac_s[sfb][window];

  return III_requantize(value, -exp, scf, channel);
}

/* we must take care that sz >= bits and sz < sizeof(long) lest bits == 0 */
# define MASK(cache, sz, bits)	\
    (((cache) >> ((sz) - (bits))) & ((1 << (bits)) - 1))
# define MASK1BIT(cache, sz)  \
    ((cache) & (1 << ((sz) - 1)))

/*
 * NAME:	III_huffdecode()
 * DESCRIPTION:	decode Huffman code words of one channel of one granule
 */
static
int III_huffdecode(struct mad_bitptr *ptr, signed short is[576],
		   struct channel *channel, unsigned int sfreqi,
		   unsigned int part2_length)
{
  struct mad_bitptr peek;
  signed int bits_left, cachesz;
  unsigned int l;
  unsigned long bitcache;

  bits_left = (signed) channel->part2_3_length - (signed) part2_length;
  if (bits_left < 0)
    return MAD_ERR_BADPART3LEN;

  peek = *ptr;
  mad_bit_skip(ptr, bits_left);

  /* align bit reads to byte boundaries */
  cachesz  = mad_bit_bitsleft(&peek);
  cachesz += (7 + (24 - cachesz)) & ~7;

  bitcache   = mad_bit_read(&peek, cachesz);
  bits_left -= cachesz;

  l = 0;

  /* big_values */
  {
    unsigned int region, rcount, pcount;
    unsigned short const *fwidth;
    struct hufftable const *entry;
    union huffpair const *table;
    unsigned int linbits, startbits, big_values;

    if (channel->block_type == 2) {
      fwidth = &sfwidth_s[sfreqi][3];

      pcount = 1 + (36 / 2);
      rcount = 1;
    }
    else {
      fwidth = &sfwidth_l[sfreqi][0];

      pcount = 1 + (*fwidth++ / 2);
      rcount = channel->region0_count + 1;
    }

    entry     = &mad_huff_pair_table[channel->table_select[region = 0]];
    table     = entry->table;
    linbits   = entry->linbits;
    startbits = entry->startbits;

    if (table == 0)
      return MAD_ERR_BADHUFFTABLE;

    big_values = channel->big_values;

    while (big_values-- && cachesz + bits_left > 0) {
      union huffpair const *pair;
      unsigned int clumpsz;
      signed int x, y;

      /* change table if region boundary */

      if (--pcount == 0) {
	if (--rcount == 0) {
	  if (region == 0 && !(channel->flags & window_switching_flag)) {
	    pcount = *fwidth++ / 2;
	    rcount = channel->region1_count + 1;
	  }
	  else
	    pcount = 576 / 2;  /* all remaining */

	  entry     = &mad_huff_pair_table[channel->table_select[++region]];
	  table     = entry->table;
	  linbits   = entry->linbits;
	  startbits = entry->startbits;

	  if (table == 0)
	    return MAD_ERR_BADHUFFTABLE;
	}
	else {
	  pcount = *fwidth++ / 2;
	  if (channel->block_type == 2)
	    pcount *= 3;
	}
      }

      /* hcod (0..19) */

      if (cachesz < 19) {
	unsigned int bits;

	bits       = (12 + (19 - cachesz)) & ~7;
	bitcache   = (bitcache << bits) | mad_bit_read(&peek, bits);
	cachesz   += bits;
	bits_left -= bits;
      }

      clumpsz = startbits;
      pair    = &table[MASK(bitcache, cachesz, clumpsz)];

      while (!pair->final) {
	cachesz -= clumpsz;

	clumpsz = pair->ptr.bits;
	pair    = &table[pair->ptr.offset + MASK(bitcache, cachesz, clumpsz)];
      }

      cachesz -= pair->value.hlen;

      /* x (0..14) */

      if (cachesz < 14) {
	bitcache   = (bitcache << 16) | mad_bit_read(&peek, 16);
	cachesz   += 16;
	bits_left -= 16;
      }

      x = pair->value.x;
      if (x == 15 && linbits) {
	x += MASK(bitcache, cachesz, linbits);
	cachesz -= linbits;
      }

      if (x && MASK1BIT(bitcache, cachesz--))
	x = -x;

      is[l++] = x;

      /* y (0..14) */

      if (cachesz < 14) {
	bitcache   = (bitcache << 16) | mad_bit_read(&peek, 16);
	cachesz   += 16;
	bits_left -= 16;
      }

      y = pair->value.y;
      if (y == 15 && linbits) {
	y += MASK(bitcache, cachesz, linbits);
	cachesz -= linbits;
      }

      if (y && MASK1BIT(bitcache, cachesz--))
	y = -y;

      is[l++] = y;
    }
  }

  if (cachesz + bits_left < 0) {
# if 1 && defined(DEBUG)
    fprintf(stderr, "huffman big_values overrun (%d bits)\n",
	    -(cachesz + bits_left));
# endif

    is[l - 2] = 0;
    is[l - 1] = 0;

    channel->count1 = 0;
  }
  else {  /* count1 */
    union huffquad const *table;
    unsigned int count1 = 0;

    table = mad_huff_quad_table[channel->flags & count1table_select];

    while (cachesz + bits_left > 0 && l <= 572) {
      union huffquad const *quad;
      signed int v, w, x, y;

      /* hcod (1..6) */

      if (cachesz < 10) {
	bitcache   = (bitcache << 16) | mad_bit_read(&peek, 16);
	cachesz   += 16;
	bits_left -= 16;
      }

      quad = &table[MASK(bitcache, cachesz, 4)];

      /* quad tables guaranteed to have at most one extra lookup */
      if (!quad->final) {
	cachesz -= 4;

	quad = &table[quad->ptr.offset +
		      MASK(bitcache, cachesz, quad->ptr.bits)];
      }

      cachesz -= quad->value.hlen;

      /* v (0..1) */

      v = quad->value.v;
      if (v && MASK1BIT(bitcache, cachesz--))
	v = -v;

      is[l++] = v;

      /* w (0..1) */

      w = quad->value.w;
      if (w && MASK1BIT(bitcache, cachesz--))
	w = -w;

      is[l++] = w;

      /* x (0..1) */

      x = quad->value.x;
      if (x && MASK1BIT(bitcache, cachesz--))
	x = -x;

      is[l++] = x;

      /* y (0..1) */

      y = quad->value.y;
      if (y && MASK1BIT(bitcache, cachesz--))
	y = -y;

      is[l++] = y;

      ++count1;
    }

    if (cachesz + bits_left < 0) {
# if 0 && defined(DEBUG)
      fprintf(stderr, "huffman count1 overrun (%d bits)\n",
	      -(cachesz + bits_left));
# endif

      /* technically the bitstream is misformatted, but apparently
	 some encoders are just a bit sloppy with stuffing bits */

      is[l - 4] = 0;
      is[l - 3] = 0;
      is[l - 2] = 0;
      is[l - 1] = 0;

      --count1;
    }

    channel->count1 = count1;
  }

# if 0 && defined(DEBUG)
  if (bits_left < 0)
    fprintf(stderr, "read %d bits too many\n", -bits_left);
  else if (cachesz + bits_left > 0)
    fprintf(stderr, "%d stuffing bits\n", cachesz + bits_left);
# endif

  /* rzero */
  while (l < 576) {
    is[l++] = 0;
    is[l++] = 0;
  }

  return 0;
}

# undef MASK
# undef MASK1BIT

/*
 * NAME:	III_reorder()
 * DESCRIPTION:	reorder frequency lines of a short block into subband order
 */
static
void III_reorder(mad_fixed_t xr[576], struct channel const *channel,
		 unsigned short const sfwidth[13])
{
  mad_fixed_t tmp[32][3][6];
  unsigned int sb, l, sfb, f, w, sbw[3], sw[3];

  if (channel->flags & mixed_block_flag)
    sb = 2, sfb = 3;
  else
    sb = 0, sfb = 0;

  for (w = 0; w < 3; ++w) {
    sbw[w] = sb;
    sw[w]  = 0;
  }

  f = sfwidth[sfb];
  w = 0;

  for (l = 18 * sb; l < 576; ++l) {
    tmp[sbw[w]][w][sw[w]++] = xr[l];

    if (sw[w] == 6) {
      sw[w] = 0;
      ++sbw[w];
    }

    if (--f == 0) {
      if (w++ == 2) {
	w = 0;
	++sfb;
      }

      f = sfwidth[sfb];
    }
  }

  memcpy(&xr[18 * sb], &tmp[sb], (576 - 18 * sb) * sizeof(mad_fixed_t));
}

/*
 * NAME:	III_stereo()
 * DESCRIPTION:	perform joint stereo processing on a granule
 */
static
int III_stereo(mad_fixed_t xr[2][576], struct granule *granule,
	       struct mad_frame *frame, unsigned int sfreqi)
{
  int i_stereo, ms_stereo;
  unsigned int bound;

  i_stereo  = frame->mode_ext & 0x1;
  ms_stereo = frame->mode_ext & 0x2;

  bound = 576;

  /* intensity stereo */

  if (i_stereo) {
    struct channel const *right_ch = &granule->ch[1];
    unsigned int zero_part, i, sfb_l, sfb_s, f, w, n;
    mad_fixed_t const *lsf_scale;

    frame->flags |= MAD_FLAG_I_STEREO;

    zero_part = 576 - right_ch->big_values * 2 - right_ch->count1 * 4;

# if 0 && defined(DEBUG)
    fprintf(stderr, "intensity_stereo zero_part:%u, block_type:%u\n",
	    zero_part, right_ch->block_type);
# endif

    sfb_l = (right_ch->block_type == 2) ?
      ((frame->flags & MAD_FLAG_LSF_EXT) ? 6 : 8) : 22;
    sfb_s = 13;

    while (bound > 0) {
      if (right_ch->block_type != 2 ||
	  (bound <= 36 && (right_ch->flags & mixed_block_flag))) {
	/* long blocks */
	f = sfwidth_l[sfreqi][sfb_l - 1];

	if (zero_part < f)
	  break;
	else {
	  zero_part -= f;
	  bound     -= f;
	  --sfb_l;
	}
      }
      else {
	/* short blocks */
	f = sfwidth_s[sfreqi][sfb_s - 1] * 3;

	if (zero_part < f)
	  break;
	else {
	  zero_part -= f;
	  bound     -= f;
	  --sfb_s;
	}
      }
    }

    f = 0;
    w = 0;
    n = 0;

    /* intensity_scale */
    lsf_scale = lsf_is_table[right_ch->scalefac_compress & 0x1];

    for (i = bound; i < 576; ++i) {
      unsigned int is_pos = 7;

      if (f-- == 0) {
	if (w-- == 0) {
	  if (right_ch->block_type != 2 ||
	      (i < 36 && (right_ch->flags & mixed_block_flag))) {
	    w = 0;
	    f = sfwidth_l[sfreqi][sfb_l] - 1;
	    is_pos = right_ch->scalefac_l[sfb_l++];
	  }
	  else {
	    w = 2;
	    f = n = sfwidth_s[sfreqi][sfb_s] - 1;
	    is_pos = right_ch->scalefac_s[sfb_s++][0];
	  }
	}
	else {
	  f = n;
	  is_pos = right_ch->scalefac_s[sfb_s - 1][2 - w];
	}
      }

      if (is_pos < 7) {
	mad_fixed_t left;

	left = xr[0][i];

	if (frame->flags & MAD_FLAG_LSF_EXT) {
	  if (is_pos == 0)
	    xr[1][i] = left;
	  else if (is_pos & 0x1) {
	    xr[0][i] = mad_f_mul(left, lsf_scale[(is_pos - 1) / 2]);
	    xr[1][i] = left;
	  }
	  else
	    xr[1][i] = mad_f_mul(left, lsf_scale[(is_pos - 1) / 2]);
	}
	else {
	  xr[0][i] = mad_f_mul(left, is_table[is_pos]);
	  xr[1][i] = mad_f_mul(left, is_table[6 - is_pos]);
	}
      }
# if defined(OPT_ISKLUGE)
      else if (ms_stereo) {
	mad_fixed_t m, s;

	/*
	 * There seems to be some discrepancy with respect to simultaneous
	 * MS and intensity stereo encoding. The ISO/IEC 11172-3 standard
	 * clearly states that "...the scalefactor band in which the last
	 * non-zero (right channel) frequency line occurs is the last
	 * scalefactor band to which the MS_stereo equations apply" when
	 * intensity stereo is also enabled. "Above this bound intensity
	 * stereo may be applied if enabled in the header."
	 *
	 * However, the ISO/IEC 13818-3 standard states, "As in ISO/IEC
	 * 11172-3 scalefactor bands with an illegal intensity position
	 * have to be decoded according to the MS equations as defined in
	 * ISO/IEC 11172-3 if MS stereo is enabled, or both channels
	 * independently if MS stereo is not enabled."
	 *
	 * The ISO reference source and many other decoders are apparently
	 * interpreting this with mixed results. It's possible many
	 * encoders have a questionable implementation as well, but it
	 * happens to go unnoticed with the same decoders.
	 *
	 * It's not clear what the right answer is. For now, the intensity
	 * stereo bands can also optionally be processed with MS for
	 * is_pos values outside the intensity stereo range, which seems
	 * to be what at least one other decoder implementation is doing.
	 * This does not necessarily produce optimal results, but it is
	 * better than no processing for many bitstreams.
	 */

	m = xr[0][i];
	s = xr[1][i];

	xr[0][i] = mad_f_mul(m + s, root_table[1]); /* l = (m + s) / sqrt(2) */
	xr[1][i] = mad_f_mul(m - s, root_table[1]); /* r = (m - s) / sqrt(2) */
      }
# endif  /* OPT_ISKLUGE */
    }
  }

  /* middle/side stereo */

  if (ms_stereo) {
    unsigned int i;

    frame->flags |= MAD_FLAG_MS_STEREO;

    if (granule->ch[0].block_type != granule->ch[1].block_type)
      return MAD_ERR_BADSTEREO;

    i = bound;
    while (i--) {
      mad_fixed_t m, s;

      m = xr[0][i];
      s = xr[1][i];

      xr[0][i] = mad_f_mul(m + s, root_table[1]);  /* l = (m + s) / sqrt(2) */
      xr[1][i] = mad_f_mul(m - s, root_table[1]);  /* r = (m - s) / sqrt(2) */
    }
  }

  return 0;
}

/*
 * NAME:	III_aliasreduce()
 * DESCRIPTION:	perform frequency line alias reduction
 */
static inline
void III_aliasreduce(mad_fixed_t xr[576])
{
  unsigned int l, i;

  for (l = 18; l < 576; l += 18) {
    for (i = 0; i < 8; ++i) {
      mad_fixed_t a, b;

      a = xr[l - 1 - i];
      b = xr[l     + i];

      xr[l - 1 - i] = mad_f_mul(a, cs[i]) - mad_f_mul(b, ca[i]);
      xr[l     + i] = mad_f_mul(b, cs[i]) + mad_f_mul(a, ca[i]);
    }
  }
}

# if defined(ASO_IMDCT)
void III_imdct_l(mad_fixed_t const [18], mad_fixed_t [36], unsigned int);
# else
/*
 * NAME:	imdct36
 * DESCRIPTION:	perform X[18]->x[36] IMDCT
 */
static inline
void imdct36(mad_fixed_t const X[18], mad_fixed_t x[36])
{
  mad_fixed_t t0,  t1,  t2,  t3,  t4,  t5;
  mad_fixed_t t6,  t7,  t8,  t9,  t10, t11;
  mad_fixed_t t12, t13, t14, t15, t16, t17;

  t6 =
    mad_f_mul(X[4],  0x0ec835e8L) +
    mad_f_mul(X[13], 0x061f78aaL);

  t0 = t6 +
    (t7 = mad_f_mul((t16 = X[1] - X[10]), -0x061f78aaL)) +
    (t8 = mad_f_mul((t17 = X[7] + X[16]), -0x0ec835e8L));

  x[7] = t0 +
    mad_f_mul((t10 = X[0] - X[11] - X[12]),  0x0216a2a2L) +
    mad_f_mul((t11 = X[2] - X[9]  - X[14]),  0x09bd7ca0L) +
    mad_f_mul((t12 = X[3] - X[8]  - X[15]), -0x0cb19346L) +
    mad_f_mul((t13 = X[5] - X[6]  - X[17]), -0x0fdcf549L);
  x[10] = -x[7];

  x[19] = x[34] = -t6 + -t7 + -t8 +
    mad_f_mul(t10, -0x0cb19346L) +
    mad_f_mul(t11,  0x0fdcf549L) +
    mad_f_mul(t12,  0x0216a2a2L) +
    mad_f_mul(t13, -0x09bd7ca0L);

  t14 = X[0] - X[3] + X[8] - X[11] - X[12] + X[15];
  t15 = X[2] + X[5] - X[6] - X[9]  - X[14] - X[17];

  x[22] = x[31] = t0 +
    mad_f_mul(t14, -0x0ec835e8L) +
    mad_f_mul(t15,  0x061f78aaL);

  t1 = t6 +
    mad_f_mul(X[1],  -0x09bd7ca0L) +
    mad_f_mul(X[7],   0x0216a2a2L) +
    mad_f_mul(X[10], -0x0fdcf549L) +
    mad_f_mul(X[16],  0x0cb19346L);

  x[6] = t1 +
    mad_f_mul(X[0],   0x03768962L) +
    mad_f_mul(X[2],   0x0e313245L) +
    mad_f_mul(X[3],  -0x0ffc19fdL) +
    mad_f_mul(X[5],  -0x0acf37adL) +
    mad_f_mul(X[6],   0x04cfb0e2L) +
    mad_f_mul(X[8],  -0x0898c779L) +
    mad_f_mul(X[9],   0x0d7e8807L) +
    mad_f_mul(X[11],  0x0f426cb5L) +
    mad_f_mul(X[12], -0x0bcbe352L) +
    mad_f_mul(X[14],  0x00b2aa3eL) +
    mad_f_mul(X[15], -0x07635284L) +
    mad_f_mul(X[17], -0x0f9ee890L);
  x[11] = -x[6];

  x[23] = x[30] = t1 +
    mad_f_mul(X[0],  -0x0f426cb5L) +
    mad_f_mul(X[2],  -0x00b2aa3eL) +
    mad_f_mul(X[3],   0x0898c779L) +
    mad_f_mul(X[5],   0x0f9ee890L) +
    mad_f_mul(X[6],   0x0acf37adL) +
    mad_f_mul(X[8],  -0x07635284L) +
    mad_f_mul(X[9],  -0x0e313245L) +
    mad_f_mul(X[11], -0x0bcbe352L) +
    mad_f_mul(X[12], -0x03768962L) +
    mad_f_mul(X[14],  0x0d7e8807L) +
    mad_f_mul(X[15],  0x0ffc19fdL) +
    mad_f_mul(X[17],  0x04cfb0e2L);

  x[18] = x[35] = -t1 +
    mad_f_mul(X[0],  -0x0bcbe352L) +
    mad_f_mul(X[2],   0x0d7e8807L) +
    mad_f_mul(X[3],  -0x07635284L) +
    mad_f_mul(X[5],   0x04cfb0e2L) +
    mad_f_mul(X[6],   0x0f9ee890L) +
    mad_f_mul(X[8],  -0x0ffc19fdL) +
    mad_f_mul(X[9],  -0x00b2aa3eL) +
    mad_f_mul(X[11],  0x03768962L) +
    mad_f_mul(X[12], -0x0f426cb5L) +
    mad_f_mul(X[14],  0x0e313245L) +
    mad_f_mul(X[15],  0x0898c779L) +
    mad_f_mul(X[17], -0x0acf37adL);

  t9 =
    mad_f_mul(X[4],   0x061f78aaL) +
    mad_f_mul(X[13], -0x0ec835e8L);

  t2 = t9 +
    mad_f_mul(X[1],  -0x0cb19346L)  +
    mad_f_mul(X[7],   0x0fdcf549L)  +
    mad_f_mul(X[10],  0x0216a2a2L)  +
    mad_f_mul(X[16], -0x09bd7ca0L);

  x[5] = t2 +
    mad_f_mul(X[0],   0x04cfb0e2L) +
    mad_f_mul(X[2],   0x0ffc19fdL) +
    mad_f_mul(X[3],  -0x0d7e8807L) +
    mad_f_mul(X[5],   0x03768962L) +
    mad_f_mul(X[6],  -0x0bcbe352L) +
    mad_f_mul(X[8],  -0x0e313245L) +
    mad_f_mul(X[9],   0x07635284L) +
    mad_f_mul(X[11], -0x0acf37adL) +
    mad_f_mul(X[12],  0x0f9ee890L) +
    mad_f_mul(X[14],  0x0898c779L) +
    mad_f_mul(X[15],  0x00b2aa3eL) +
    mad_f_mul(X[17],  0x0f426cb5L);
  x[12] = -x[5];

  x[0] = t2 +
    mad_f_mul(X[0],   0x0acf37adL) +
    mad_f_mul(X[2],  -0x0898c779L) +
    mad_f_mul(X[3],   0x0e313245L) +
    mad_f_mul(X[5],  -0x0f426cb5L) +
    mad_f_mul(X[6],  -0x03768962L) +
    mad_f_mul(X[8],   0x00b2aa3eL) +
    mad_f_mul(X[9],  -0x0ffc19fdL) +
    mad_f_mul(X[11],  0x0f9ee890L) +
    mad_f_mul(X[12], -0x04cfb0e2L) +
    mad_f_mul(X[14],  0x07635284L) +
    mad_f_mul(X[15],  0x0d7e8807L) +
    mad_f_mul(X[17], -0x0bcbe352L);
  x[17] = -x[0];

  x[24] = x[29] = t2 +
    mad_f_mul(X[0],  -0x0f9ee890L) +
    mad_f_mul(X[2],  -0x07635284L) +
    mad_f_mul(X[3],  -0x00b2aa3eL) +
    mad_f_mul(X[5],   0x0bcbe352L) +
    mad_f_mul(X[6],   0x0f426cb5L) +
    mad_f_mul(X[8],   0x0d7e8807L) +
    mad_f_mul(X[9],   0x0898c779L) +
    mad_f_mul(X[11], -0x04cfb0e2L) +
    mad_f_mul(X[12], -0x0acf37adL) +
    mad_f_mul(X[14], -0x0ffc19fdL) +
    mad_f_mul(X[15], -0x0e313245L) +
    mad_f_mul(X[17], -0x03768962L);

  t3 = t9 +
    mad_f_mul(X[1],  -0x0216a2a2L) +
    mad_f_mul(X[7],  -0x09bd7ca0L) +
    mad_f_mul(X[10],  0x0cb19346L) +
    mad_f_mul(X[16],  0x0fdcf549L);

  x[8] = t3 +
    mad_f_mul(X[0],   0x00b2aa3eL) +
    mad_f_mul(X[2],   0x03768962L) +
    mad_f_mul(X[3],  -0x04cfb0e2L) +
    mad_f_mul(X[5],  -0x07635284L) +
    mad_f_mul(X[6],   0x0898c779L) +
    mad_f_mul(X[8],   0x0acf37adL) +
    mad_f_mul(X[9],  -0x0bcbe352L) +
    mad_f_mul(X[11], -0x0d7e8807L) +
    mad_f_mul(X[12],  0x0e313245L) +
    mad_f_mul(X[14],  0x0f426cb5L) +
    mad_f_mul(X[15], -0x0f9ee890L) +
    mad_f_mul(X[17], -0x0ffc19fdL);
  x[9] = -x[8];

  x[21] = x[32] = t3 +
    mad_f_mul(X[0],  -0x0e313245L) +
    mad_f_mul(X[2],   0x0bcbe352L) +
    mad_f_mul(X[3],   0x0f9ee890L) +
    mad_f_mul(X[5],  -0x0898c779L) +
    mad_f_mul(X[6],  -0x0ffc19fdL) +
    mad_f_mul(X[8],   0x04cfb0e2L) +
    mad_f_mul(X[9],   0x0f426cb5L) +
    mad_f_mul(X[11], -0x00b2aa3eL) +
    mad_f_mul(X[12], -0x0d7e8807L) +
    mad_f_mul(X[14], -0x03768962L) +
    mad_f_mul(X[15],  0x0acf37adL) +
    mad_f_mul(X[17],  0x07635284L);

  x[20] = x[33] = -t3 +
    mad_f_mul(X[0],  -0x0d7e8807L) +
    mad_f_mul(X[2],   0x0f426cb5L) +
    mad_f_mul(X[3],   0x0acf37adL) +
    mad_f_mul(X[5],  -0x0ffc19fdL) +
    mad_f_mul(X[6],  -0x07635284L) +
    mad_f_mul(X[8],   0x0f9ee890L) +
    mad_f_mul(X[9],   0x03768962L) +
    mad_f_mul(X[11], -0x0e313245L) +
    mad_f_mul(X[12],  0x00b2aa3eL) +
    mad_f_mul(X[14],  0x0bcbe352L) +
    mad_f_mul(X[15], -0x04cfb0e2L) +
    mad_f_mul(X[17], -0x0898c779L);

  t4 = -t9 +
    mad_f_mul(t16, -0x0ec835e8L) +
    mad_f_mul(t17,  0x061f78aaL);

  x[4] = t4 +
    mad_f_mul(t14, 0x061f78aaL) +
    mad_f_mul(t15, 0x0ec835e8L);
  x[13] = -x[4];

  x[1] = t4 +
    mad_f_mul(t10,  0x09bd7ca0L) +
    mad_f_mul(t11, -0x0216a2a2L) +
    mad_f_mul(t12,  0x0fdcf549L) +
    mad_f_mul(t13, -0x0cb19346L);
  x[16] = -x[1];

  x[25] = x[28] = t4 +
    mad_f_mul(t10, -0x0fdcf549L) +
    mad_f_mul(t11, -0x0cb19346L) +
    mad_f_mul(t12, -0x09bd7ca0L) +
    mad_f_mul(t13, -0x0216a2a2L);

  t5 = -t6 +
    mad_f_mul(X[1],  -0x0fdcf549L) +
    mad_f_mul(X[7],  -0x0cb19346L) +
    mad_f_mul(X[10], -0x09bd7ca0L) +
    mad_f_mul(X[16], -0x0216a2a2L);

  x[2] = t5 +
    mad_f_mul(X[0],   0x0898c779L) +
    mad_f_mul(X[2],   0x04cfb0e2L) +
    mad_f_mul(X[3],   0x0bcbe352L) +
    mad_f_mul(X[5],   0x00b2aa3eL) +
    mad_f_mul(X[6],   0x0e313245L) +
    mad_f_mul(X[8],  -0x03768962L) +
    mad_f_mul(X[9],   0x0f9ee890L) +
    mad_f_mul(X[11], -0x07635284L) +
    mad_f_mul(X[12],  0x0ffc19fdL) +
    mad_f_mul(X[14], -0x0acf37adL) +
    mad_f_mul(X[15],  0x0f426cb5L) +
    mad_f_mul(X[17], -0x0d7e8807L);
  x[15] = -x[2];

  x[3] = t5 +
    mad_f_mul(X[0],   0x07635284L) +
    mad_f_mul(X[2],   0x0acf37adL) +
    mad_f_mul(X[3],   0x03768962L) +
    mad_f_mul(X[5],   0x0d7e8807L) +
    mad_f_mul(X[6],  -0x00b2aa3eL) +
    mad_f_mul(X[8],   0x0f426cb5L) +
    mad_f_mul(X[9],  -0x04cfb0e2L) +
    mad_f_mul(X[11],  0x0ffc19fdL) +
    mad_f_mul(X[12], -0x0898c779L) +
    mad_f_mul(X[14],  0x0f9ee890L) +
    mad_f_mul(X[15], -0x0bcbe352L) +
    mad_f_mul(X[17],  0x0e313245L);
  x[14] = -x[3];

  x[26] = x[27] = t5 +
    mad_f_mul(X[0],  -0x0ffc19fdL) +
    mad_f_mul(X[2],  -0x0f9ee890L) +
    mad_f_mul(X[3],  -0x0f426cb5L) +
    mad_f_mul(X[5],  -0x0e313245L) +
    mad_f_mul(X[6],  -0x0d7e8807L) +
    mad_f_mul(X[8],  -0x0bcbe352L) +
    mad_f_mul(X[9],  -0x0acf37adL) +
    mad_f_mul(X[11], -0x0898c779L) +
    mad_f_mul(X[12], -0x07635284L) +
    mad_f_mul(X[14], -0x04cfb0e2L) +
    mad_f_mul(X[15], -0x03768962L) +
    mad_f_mul(X[17], -0x00b2aa3eL);
}

/*
 * NAME:	III_imdct_l()
 * DESCRIPTION:	perform IMDCT and windowing for long blocks
 */
static
void III_imdct_l(mad_fixed_t const X[18], mad_fixed_t z[36],
		 unsigned int block_type)
{
  unsigned int i;

  /* IMDCT */

  imdct36(X, z);

  /* windowing */

  switch (block_type) {
  case 0:  /* normal window */
    for (i =  0; i < 36; ++i) z[i] = mad_f_mul(z[i], window_l[i]);
    break;

  case 1:  /* start block */
    for (i =  0; i < 18; ++i) z[i] = mad_f_mul(z[i], window_l[i]);
    /*  (i = 18; i < 24; ++i) z[i] unchanged */
    for (i = 24; i < 30; ++i) z[i] = mad_f_mul(z[i], window_s[i - 18]);
    for (i = 30; i < 36; ++i) z[i] = 0;
    break;

  case 3:  /* stop block */
    for (i =  0; i <  6; ++i) z[i] = 0;
    for (i =  6; i < 12; ++i) z[i] = mad_f_mul(z[i], window_s[i - 6]);
    /*  (i = 12; i < 18; ++i) z[i] unchanged */
    for (i = 18; i < 36; ++i) z[i] = mad_f_mul(z[i], window_l[i]);
    break;
  }
}
# endif  /* ASO_IMDCT */

/*
 * NAME:	III_imdct_s()
 * DESCRIPTION:	perform IMDCT and windowing for short blocks
 */
static
void III_imdct_s(mad_fixed_t const X[18], mad_fixed_t z[36])
{
  unsigned int w, i, k;
  mad_fixed_t y[3][12];

  /* IMDCT and windowing */

  for (w = 0; w < 3; ++w) {
    for (i = 0; i < 12; ++i) {
      register mad_fixed_t sum;

      sum = 0;

      for (k = 0; k < 6; ++k)
	sum += mad_f_mul(X[6 * w + k], imdct_s[i][k]);

      y[w][i] = mad_f_mul(sum, window_s[i]);
    }
  }

  /* overlapping and concatenation */

  for (i =  0; i <  6; ++i) z[i] = 0;
  for (i =  6; i < 12; ++i) z[i] = y[0][i - 6];
  for (i = 12; i < 18; ++i) z[i] = y[0][i - 6] + y[1][i - 12];
  for (i = 18; i < 24; ++i) z[i] =               y[1][i - 12] + y[2][i - 18];
  for (i = 24; i < 30; ++i) z[i] =                              y[2][i - 18];
  for (i = 30; i < 36; ++i) z[i] = 0;
}

/*
 * NAME:	III_overlap()
 * DESCRIPTION:	perform overlap-add of windowed IMDCT outputs
 */
static
void III_overlap(mad_fixed_t const output[36], mad_fixed_t overlap[18],
		 mad_fixed_t sample[18][32], unsigned int sb)
{
  unsigned int i;

  for (i = 0; i < 18; ++i) {
    sample[i][sb] = output[i] + overlap[i];
    overlap[i]    = output[i + 18];
  }
}

/*
 * NAME:	III_overlap_z()
 * DESCRIPTION:	perform "overlap-add" of zero IMDCT outputs
 */
static inline
void III_overlap_z(mad_fixed_t overlap[18],
		   mad_fixed_t sample[18][32], unsigned int sb)
{
  unsigned int i;

  for (i = 0; i < 18; ++i) {
    sample[i][sb] = overlap[i];
    overlap[i]    = 0;
  }
}

/*
 * NAME:	III_freqinver()
 * DESCRIPTION:	perform subband frequency inversion for odd sample lines
 */
static
void III_freqinver(mad_fixed_t sample[18][32], unsigned int sb)
{
  unsigned int i;

  for (i = 1; i < 18; i += 2)
    sample[i][sb] = -sample[i][sb];
}

/*
 * NAME:	III_decode()
 * DESCRIPTION:	decode frame main_data
 */
static
int III_decode(struct mad_bitptr *ptr, struct mad_frame *frame,
		struct sideinfo *si, unsigned int nch)
{
  unsigned int sfreqi, ngr, gr, sfbcut;

  /* 48000 => 0, 44100 => 1, 32000 => 2,
     24000 => 3, 22050 => 4, 16000 => 5 */
  sfreqi = ((frame->sfreq >>  7) & 0x000f) +
           ((frame->sfreq >> 15) & 0x0001) - 8;

  /* scalefactors, Huffman decoding, requantization */

  if (frame->flags & MAD_FLAG_LSF_EXT)
    ngr = 1, sfbcut = 6;
  else
    ngr = 2, sfbcut = 8;

  for (gr = 0; gr < ngr; ++gr) {
    struct granule *granule = &si->gr[gr];
    static mad_fixed_t xr[2][576];
    unsigned int ch;
    int error;

    for (ch = 0; ch < nch; ++ch) {
      struct channel *channel = &granule->ch[ch];
      static signed short is[576];
      unsigned int part2_length, sfb, i, f, w;

      if (frame->flags & MAD_FLAG_LSF_EXT)
	part2_length = III_lsf_scalefactors(ptr, frame->mode_ext, ch, channel);
      else {
	part2_length = III_scalefactors(ptr, channel, &si->gr[0].ch[ch],
					gr == 0 ? 0 : si->scfsi[ch]);
      }

      error = III_huffdecode(ptr, is, channel, sfreqi, part2_length);
      if (error)
	return error;

      i = 0;

      /* subbands 0-1 */

      if (channel->block_type != 2 || (channel->flags & mixed_block_flag)) {
	/* long blocks sfb 0-7 (0-5 LSF) */
	for (sfb = 0; sfb < sfbcut; ++sfb) {
	  for (f = sfwidth_l[sfreqi][sfb]; f--; ++i)
	    xr[ch][i] = is[i] ? III_requantize_l(is[i], sfb, channel) : 0;
	}
      }
      else {
	/* short blocks sfb 0-2 */
	for (sfb = 0; sfb < 3; ++sfb) {
	  for (w = 0; w < 3; ++w) {
	    for (f = sfwidth_s[sfreqi][sfb]; f--; ++i)
	      xr[ch][i] = is[i] ? III_requantize_s(is[i], sfb, w, channel) : 0;
	  }
	}
      }

      /* subbands 2-31 */

      if (channel->block_type != 2) {
	/* long blocks sfb 8-21 (6-21 LSF) */
	for (sfb = sfbcut; sfb < 22; ++sfb) {
	  for (f = sfwidth_l[sfreqi][sfb]; f--; ++i)
	    xr[ch][i] = is[i] ? III_requantize_l(is[i], sfb, channel) : 0;
	}
      }
      else {  /* channel->block_type == 2 */
	/* short blocks sfb 3-12 */
	for (sfb = 3; sfb < 13; ++sfb) {
	  for (w = 0; w < 3; ++w) {
	    for (f = sfwidth_s[sfreqi][sfb]; f--; ++i)
	      xr[ch][i] = is[i] ? III_requantize_s(is[i], sfb, w, channel) : 0;
	  }
	}
      }
    }

    /* joint stereo processing */

    if (frame->mode == MAD_MODE_JOINT_STEREO) {
      error = III_stereo(xr, granule, frame, sfreqi);
      if (error)
	return error;
    }

    /* reordering, alias reduction, IMDCT, overlap-add, frequency inversion */

    for (ch = 0; ch < nch; ++ch) {
      struct channel const *channel = &granule->ch[ch];
      mad_fixed_t (*sample)[32] = &frame->sbsample[ch][18 * gr];
      unsigned int sb, l, i, sblimit;
      mad_fixed_t output[36];

      if (channel->block_type == 2)
	III_reorder(xr[ch], channel, sfwidth_s[sfreqi]);
      else
	III_aliasreduce(xr[ch]);

      l = 0;

      /* subbands 0-1 */

      if (channel->block_type != 2 || (channel->flags & mixed_block_flag)) {
	/* long blocks */
	for (sb = 0; sb < 2; ++sb, l += 18) {
	  III_imdct_l(&xr[ch][l], output, channel->block_type);
	  III_overlap(output, (*frame->overlap)[ch][sb], sample, sb);
	}
      }
      else {
	/* short blocks */
	for (sb = 0; sb < 2; ++sb, l += 18) {
	  III_imdct_s(&xr[ch][l], output);
	  III_overlap(output, (*frame->overlap)[ch][sb], sample, sb);
	}
      }

      III_freqinver(sample, 1);

      /* (nonzero) subbands 2-31 */

      i = 576;
      while (i > 36 && xr[ch][i - 1] == 0)
	--i;

      sblimit = 32 - (576 - i) / 18;

      if (channel->block_type != 2) {
	/* long blocks */
	for (sb = 2; sb < sblimit; ++sb, l += 18) {
	  III_imdct_l(&xr[ch][l], output, channel->block_type);
	  III_overlap(output, (*frame->overlap)[ch][sb], sample, sb);

	  if (sb & 1)
	    III_freqinver(sample, sb);
	}
      }
      else {
	/* short blocks */
	for (sb = 2; sb < sblimit; ++sb, l += 18) {
	  III_imdct_s(&xr[ch][l], output);
	  III_overlap(output, (*frame->overlap)[ch][sb], sample, sb);

	  if (sb & 1)
	    III_freqinver(sample, sb);
	}
      }

      /* (zero) remaining subbands */

      for (sb = sblimit; sb < 32; ++sb) {
	III_overlap_z((*frame->overlap)[ch][sb], sample, sb);

	if (sb & 1)
	  III_freqinver(sample, sb);
      }
    }
  }

  return 0;
}

/*
 * NAME:	layer->III()
 * DESCRIPTION:	decode a single Layer III frame
 */
int mad_layer_III(struct mad_stream *stream, struct mad_frame *frame,
		  unsigned short const crc[2])
{
  unsigned int nch, private_bitlen;
  unsigned int main_data_bitlen, main_data_length, sideinfo_length;
  unsigned int frame_space, frame_used, frame_free;
  struct mad_bitptr ptr;
  struct sideinfo si;
  int error, result = 0;

  /* allocate Layer III dynamic structures */

  if (stream->main_data == 0) {
    stream->main_data = malloc(1935);
    if (stream->main_data == 0) {
      stream->error = MAD_ERR_NOMEM;
      return -1;
    }
  }

  if (frame->overlap == 0) {
    frame->overlap = calloc(2 * 32 * 18, sizeof(mad_fixed_t));
    if (frame->overlap == 0) {
      stream->error = MAD_ERR_NOMEM;
      return -1;
    }
  }

  nch = MAD_NCHANNELS(frame);
  sideinfo_length = (frame->flags & MAD_FLAG_LSF_EXT) ?
    (nch == 1 ? 9 : 17) : (nch == 1 ? 17 : 32);

  /* check frame sanity */

  if (stream->next_frame - mad_bit_nextbyte(&stream->ptr) <
      (signed int) sideinfo_length) {
    stream->error = MAD_ERR_BADFRAMELEN;
    stream->md_len = 0;
    return -1;
  }

  /* check CRC word */

  if ((frame->flags & MAD_FLAG_PROTECTION) &&
      mad_bit_crc(stream->ptr, sideinfo_length * CHAR_BIT, crc[0]) != crc[1]) {
    stream->error = MAD_ERR_BADCRC;
    result = -1;
  }

  /* decode frame side information */

  error = III_sideinfo(&stream->ptr, nch, frame->flags & MAD_FLAG_LSF_EXT,
		       &si, &main_data_bitlen, &private_bitlen);
  if (error && result == 0) {
    stream->error = error;
    result = -1;
  }

  frame->flags   |= private_bitlen;
  frame->private |= si.private_bits;

  /* find main_data */

  frame_space      = stream->next_frame - mad_bit_nextbyte(&stream->ptr);
  main_data_length = (main_data_bitlen + (CHAR_BIT - 1)) / CHAR_BIT;

  if (si.main_data_begin == 0) {
    ptr = stream->ptr;
    stream->md_len = 0;

    frame_used = main_data_length;
  }
  else {
    if (si.main_data_begin > stream->md_len) {
      if (result == 0) {
	stream->error = MAD_ERR_BADDATAPTR;
	result = -1;
      }

      frame_used = 0;
    }
    else {
      mad_bit_init(&ptr,
		   &(*stream->main_data)[stream->md_len] - si.main_data_begin);

      if (main_data_length > si.main_data_begin) {
	if (main_data_length - si.main_data_begin > frame_space) {
	  if (result == 0) {
	    stream->error = MAD_ERR_BADDATALEN;
	    result = -1;
	  }

	  frame_used = 0;
	}
	else {
	  memcpy(&(*stream->main_data)[stream->md_len],
		 mad_bit_nextbyte(&stream->ptr),
		 main_data_length - si.main_data_begin);
	  stream->md_len += main_data_length - si.main_data_begin;

	  frame_used = main_data_length - si.main_data_begin;
	}
      }
      else
	frame_used = 0;
    }
  }

  frame_free = frame_space - frame_used;

  /* improve free bitrate discovery */
  if (frame->bitrate == 0 && frame_used)
    mad_bit_skip(&stream->ptr, frame_used * CHAR_BIT);

# if 0 && defined(DEBUG)
  fprintf(stderr, "main_data_begin:%u, main_data_length:%u, "
	  "frame_space:%u, md_len:%u      \n",
	  si.main_data_begin, main_data_length, frame_space, stream->md_len);
# endif

  /* decode main_data */

  if (result == 0) {
    error = III_decode(&ptr, frame, &si, nch);
    if (error) {
      stream->error = error;
      result = -1;
    }
  }

  /* preload main_data buffer with up to 511 bytes for next frame(s) */

  if (frame_free >= 511) {
    memcpy(*stream->main_data, stream->next_frame - 511, 511);
    stream->md_len = 511;
  }
  else {
    if (main_data_length < si.main_data_begin) {
      unsigned int extra;

      extra = si.main_data_begin - main_data_length;
      if (extra + frame_free > 511)
	extra = 511 - frame_free;

      memmove(*stream->main_data,
	      *stream->main_data + stream->md_len - extra,
	      extra);

      stream->md_len = extra;
    }
    else
      stream->md_len = 0;

    memcpy(*stream->main_data + stream->md_len,
	   stream->next_frame - frame_free, frame_free);
    stream->md_len += frame_free;
  }

  /* designate ancillary bits */
  /* FIXME */
  stream->anc_bitlen = 0;

  return result;
}
